/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "genobj.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupresimple.h"
#include "spnonsp.h"

//#include "p_histo.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

#define MAXNRCHAN 20
#define MaxRuns 20000
#define MAXNRFRAMEFORMAXDET 1080000


#define TESTLEVEL
#define TESTLEVEL2
#define TEST_RUNS

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif




// -- global variables
int nrchanwav, totnrwav, nrpzm ;
int wavfileoffset, labfileoffset ;
int mixfileindex ;
int *pzmfileindex ;
int outpathindex ;

FILE *chanfp, *spnonspchancfgfp, *spnonspmixcfgfp ; 
CTool      *Tool ;
CAkuPreSimple    **AkuPre ;
CAkuPreSimple    *AkuPreMixFile  ;
CAkuPreSimple    **AkuPrePZMFile ;

CSpeechNonSpeech **SpNonSp  ;
CSpeechNonSpeech *SpNonSpMixFile ;

FILE **featoutfp ;
FILE **laboutfp ;

int NrSpLabels, NrSiLabels ; 
char **SpLabel, **SiLabel ;
//char *SpOutLabel, *SiOutLabel ;
CListOfLabel **Lab ;
FILE **labfp ;

long *nr_runs ;
long **start ;
long **stop ;

char **ChanId ;
int  *NrSkipBytes ;

SPreFeature **prefp ;
SPreFeature *prefpmixfile ;
SPreFeature **prefppzmfile ;

float meanenergy ;
float meanenergyohnemix ;
float *energymin  ;
float *energymintmp  ;
float pzm_mean ;
// float *entropymin = new float[nrchanwav+1] ;
float *sumloudmax  ;
//float *sumloudmin  ;
float *modloudmax ;
//float *modloudmin ;
//float *zcratemin  ;
int usedframes ;
int nrframe ;


int CountFrames(  char **argument ) {

  // local variables
  int count ;
  
#ifdef TESTLEVEL
  fprintf( stderr, "Determine length of signal-files ...\n" ) ;
#endif
  // start akupre of first wavfile
  AkuPre[0]->clearAkuPre() ;
  AkuPre[0]->startAkuPre( argument[wavfileoffset], NrSkipBytes[0] ) ;

  // count frames
  count = 0 ;
  while( AkuPre[0]->getNextPre() != 0 ) {
    count++ ;
  } ;

  // restart first wavfile
  AkuPre[0]->clearAkuPre() ;
  AkuPre[0]->startAkuPre( argument[wavfileoffset], NrSkipBytes[0] ) ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif
  
  return( count ) ;
  
} ;




// global functions
int GetChanId( char* chanfilename, char* chanid ) {

  // get ChanIds 
  char *tmpchar = new char[strlen(chanfilename)+1] ; ;
  char *tmpchar2;
  strcpy( tmpchar, chanfilename ) ;

  // parse filenames of wav-files
  tmpchar2 = strrchr( tmpchar, '.' ) ;
  tmpchar2--;
  strncpy( chanid, tmpchar2, 1 ) ;
  chanid[1] = '\0'  ;

  return(1);
  
} ;


void ReadCommandLine( int nrarg, char **argument) {
  int i ; 

  nrpzm = strtol( argument[nrarg-1], NULL, 10 );
  nrchanwav = (nrarg-5-nrpzm)/2 ;
  wavfileoffset = 3 ;
  labfileoffset = wavfileoffset + nrchanwav ;
  totnrwav = nrchanwav+1 ; // includes mixfile 
  mixfileindex = nrarg-2-nrpzm ;
  pzmfileindex = new int[nrpzm] ;
  for( i=0 ; i<nrpzm ; i++ )
    pzmfileindex[i] = nrarg-i-2 ;
  outpathindex = 2;
} ;

int OpenCfgFiles( char **argument ) {
  chanfp = fopen( argument[1], "r+" );
  if( chanfp == 0 ) {
    fprintf( stderr, "can't open akupre-cfg-file: %s\n", argument[1] );
    return(-1) ;
  }
  //spnonspchancfgfp = fopen( argument[2], "r+" );
  //if( spnonspchancfgfp == 0 ) {
  //  fprintf( stderr, "can't open spnonsp-cfg-file: %s\n", argument[2]);
  //  return(-1) ;
  //}
  //spnonspmixcfgfp = fopen( argument[3], "r+" );
  //if( spnonspmixcfgfp == 0 ) {
  //  fprintf( stderr, "can't open spnonsp-cfg-file: %s\n", argument[3]);
  //  return(-1) ;
  //}

  return(1) ;

} ;


void InitAkuPre( char **argument) {
  int i ;
#ifdef TESTLEVEL
  fprintf( stdout, "Initializing AkuPre of all %d channels and of MixFile '%s' ... \n", nrchanwav, argument[mixfileindex] ) ;
#endif
  // construct Preprocessing
  Tool = new CTool( chanfp );
  AkuPre = new CAkuPreSimple*[nrchanwav] ;
  AkuPreMixFile = new CAkuPreSimple( chanfp ) ;
  AkuPrePZMFile = new CAkuPreSimple*[nrpzm] ;
  for( i=0 ; i<nrpzm ; i++ ) {
    AkuPrePZMFile[i] = new CAkuPreSimple( chanfp ) ;
  } ;
  for( i=0 ; i<nrchanwav; i++ ) {
    AkuPre[i] = new CAkuPreSimple( chanfp ) ;
  }
  fprintf( stdout, "... Done\n" ) ;
}

void  InitSpnonsp( int usecommonforall,int  usepzmenergy, int usemodloudtimeszcrate) {
  int i ;

  // construct SpNonSp
  SpNonSp = new CSpeechNonSpeech*[nrchanwav] ;
  SpNonSpMixFile = new CSpeechNonSpeech( (sizeof(sPreFeature)/sizeof(float))-2, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspmixcfgfp ) ;  // Achtung hard coded length of feature file
  
  // init channel specific akupre and spnonsp
  // use only one SpeechNonSpeech-Detector for all channels!!!
  if( usecommonforall )
    if( usepzmenergy && usemodloudtimeszcrate )
      SpNonSp[0] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
    else if( !usepzmenergy && !usemodloudtimeszcrate )
      SpNonSp[0] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float)-2, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
    else
      SpNonSp[0] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float)-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
  
  for( i=0 ; i<nrchanwav; i++ ) {
    if( usecommonforall ) {
      // use one detector for all channels:
      if( i!= 0 )
	SpNonSp[i] = SpNonSp[0] ;
    } else {
      // use indiv Detector for each channel: 
      if( usepzmenergy && usemodloudtimeszcrate )
	SpNonSp[i] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
      else if( !usepzmenergy && !usemodloudtimeszcrate )
	SpNonSp[i] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float)-2, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
      else
	SpNonSp[i] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float)-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
} ;

void OpenOutFiles( int chanfeatout, int mixfeatout, int pzmfeatout, int chanlabout, int mixlabout, int pzmlabout, char *labending, char *featending, char **argument ) {
  int i,chan ;

#ifdef TESTLEVEL
  fprintf( stdout, "Open Outfiles ...\n" ) ;
#endif

  char labtmp[MaxLineLength] ;
  char feattmp[MaxLineLength] ;
  char tmptmp[MaxLineLength] ;
  featoutfp = new FILE*[nrchanwav+1+nrpzm] ;
  laboutfp = new FILE*[nrchanwav+1+nrpzm] ;  
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    strcpy( tmptmp, argument[(chan)+wavfileoffset] ) ;
    sprintf( labtmp, "%s/%s.%s", argument[outpathindex], tmptmp, labending );
    strcpy( tmptmp, argument[(chan)+wavfileoffset] ) ;
    sprintf( feattmp, "%s/%s.%s", argument[outpathindex], tmptmp, featending );
    if( chanfeatout ) 
      featoutfp[chan] = fopen( feattmp, "w" ) ;
    if( chanlabout )
      laboutfp[chan] = fopen( labtmp, "w" ) ;
  } ;

  strcpy( tmptmp, argument[mixfileindex] ) ;
  sprintf( labtmp, "%s/%s.%s", argument[outpathindex], tmptmp, labending );
  strcpy( tmptmp, argument[mixfileindex] ) ;
  sprintf( feattmp, "%s/%s.%s", argument[outpathindex], tmptmp, featending );
  if( mixlabout ) 
    laboutfp[nrchanwav] = fopen( labtmp, "w" ) ;
  if( mixfeatout ) 
    featoutfp[nrchanwav] = fopen( feattmp, "w" ) ;

  for( i=0 ; i<nrpzm ; i++ ) {
    strcpy( tmptmp, argument[pzmfileindex[i]] ) ;
    sprintf( labtmp, "%s/%s.%s", argument[outpathindex], tmptmp, labending );
    strcpy( tmptmp, argument[pzmfileindex[i]] ) ;
    sprintf( feattmp, "%s/%s.%s", argument[outpathindex], tmptmp, featending );
    if( pzmlabout ) 
      laboutfp[nrchanwav+i+1] = fopen( labtmp, "w" ) ;
    if( pzmfeatout ) 
      featoutfp[nrchanwav+i+1] = fopen( feattmp, "w" ) ;
  }
#ifdef TESTLEVEL
  fprintf( stdout, "... Done\n" ) ;
#endif
}
 

void InitLabelInfo( EOutPutMode inputmode, CTool *tool, char *keyword, char* speechlabel, char* nonspeechlabel, char **argument ) {
  int chan,i ;

#ifdef TESTLEVEL
  fprintf( stdout, "InitLabelInfo ...\n" ) ;
#endif

  // initialize label-info for all channels
  Lab = new CListOfLabel*[nrchanwav] ; 
  // just quick and dirty, class CListOfLabel should be redesigned!
  if( inputmode == XML )
    // XML Input, use XML as KeyWord
    for( chan=0 ; chan<nrchanwav ; chan++ )
      Lab[chan] = new CListOfLabel( keyword );
  else 
    for( chan=0 ; chan<nrchanwav ; chan++ )
      // MUC or Kiel Input! use Keyword as Keyword
      Lab[chan] = new CListOfLabel( tool->LabelKeyword.Par );

#ifdef TESTLEVEL
  fprintf( stdout, "Keyword read\n" ) ;
#endif

  // set Labels to empty string if requested!
  if( strcmp( speechlabel, "NONE" ) == 0 ) 
    strcpy( speechlabel, "" ) ;
  if( strcmp( nonspeechlabel, "NONE" ) == 0 ) 
    strcpy( nonspeechlabel, "" ) ;
  
  // create array of Silence-Labels
  NrSiLabels = CParInt( "Tools.NrSiLabel", chanfp, 1 ).Par ;
  SiLabel = new char*[NrSiLabels] ;
  char tmp[MAXKEYWORDLEN] ;
  for( i=0 ; i<NrSiLabels ; i++ ) {
    SiLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SilentName.%d", i ) ;
    strcpy( SiLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SiLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SilentLabel in cfg-file '%s'\n", i, argument[1] ) ;
      exit(-1) ;
    } ;
  } ;
  
  // create array of Speech-Labels
  // NrSpLabels = SpNonSp[0]->getNrOfMixSp() ; 
  NrSpLabels = CParInt( "Tools.NrSpLabel", chanfp, 1 ).Par ;
  SpLabel = new char*[NrSpLabels] ;
  for( i=0 ; i<NrSpLabels ; i++ ) {
    SpLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SpeechName.%d", i ) ;
    strcpy( SpLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SpLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SpeechLabel in cfg-file '%s'\n", i, argument[1] ) ;
      exit(-1) ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... Done\n" ) ;
#endif
}



void  OpenInputLabelFiles( char **argument ) {
  int chan ;
#ifdef TESTLEVEL
  fprintf( stdout, "Open labfiles for all channels ...\n" ) ;
#endif
  // open lab-files
  labfp = new FILE*[nrchanwav] ;
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    labfp[chan]= fopen( argument[labfileoffset+chan], "r");
    if( labfp[chan] == 0 ) {   
      ERROR_MSG_S("can't open label-file:", argument[labfileoffset+chan] );
      exit(-1) ; 
    }
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
}



void GetRuns( char **argument, int controloutput ) {
  int chan, run  ;
  int count ;
#ifdef TESTLEVEL
  fprintf( stdout, "Define 'runs' for all channels ...\n" ) ;
#endif
  nr_runs = new long[nrchanwav] ; 
  
  // define array for starts and stops of runs
  //fprintf( stderr, "maxRuns: %d\n", MaxRuns ) ;
  start = new long*[nrchanwav] ;
  stop = new long*[nrchanwav] ;
  
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    start[chan] = new long[MaxRuns] ;
    stop[chan] = new long[MaxRuns] ;
    
    // default values for start and stop of each run
    for( run=0 ; run<MaxRuns ; run++ ) 
      start[chan][run] = stop[chan][run] = -1 ;
    
    // read multiple starts and stops from label-file
    count = 0 ;
    if( Lab[chan]->readLabels( labfp[chan] ) == _ERR_ ) { 
      ERROR_MSG_S("can't read label-file:", argument[labfileoffset+chan] );
      fclose( labfp[chan] );
      continue; 
    }
  
    // loop over all runs to determine start and stop-frame
    while( count < MaxRuns && Lab[chan]->getNextStartStop( (count>0 ? stop[chan][count-1] : 0 ), start[chan][count], stop[chan][count], NrSiLabels, SiLabel ) != _ERR_ ) { 
      if( controloutput ) 
	fprintf( stdout, "chan %d: %ld: start: %ld stop: %ld\n", chan, count, start[chan][count], stop[chan][count] );
      count++ ;
    } ;
    fclose( labfp[chan] );
    nr_runs[chan] = count ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
}


void CorrectRuns() {
  int chan ;
  int run ;
  int correction ; // in frames 



  // go through runs of each channel and correct boundaries according to SkipInfo of each chennel:
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    
    if( NrSkipBytes[chan] % (WINDOW_DELTA*2) >= WINDOW_DELTA )
      correction = (NrSkipBytes[chan]/(WINDOW_DELTA*2)) + 1 ;
    else
      correction = (NrSkipBytes[chan]/(WINDOW_DELTA*2)) ;
      
#ifdef TESTLEVEL
    fprintf( stderr, "Correction of Labels of Chan %s by %d frames\n", ChanId[chan], correction ) ;
#endif

    for( run=0 ; run<nr_runs[chan] ; run++ ) {

      start[chan][run] -= correction ;

    }
  }
}


void GetChanIds( char **argument) {
  int i, chan ;
  ChanId = new char*[nrchanwav+nrpzm] ;
  for( i=0 ; i<nrchanwav+nrpzm ; i++ )
    ChanId[i] = new char[2] ;

  for( chan=0 ; chan<nrchanwav ; chan++ )
    GetChanId( argument[(chan)+wavfileoffset], ChanId[chan] ) ;
  for( i=0 ; i<nrpzm ; i++ )
    GetChanId( argument[pzmfileindex[i]], ChanId[nrchanwav+i] ) ;
}


void GetSkipInfo() {
  int i ;
  NrSkipBytes = new int[nrchanwav+nrpzm] ;
  for( i=0 ; i<nrchanwav+nrpzm ; i++ ) {
    switch( ChanId[i][0] ) {
    case '0':
    case '1':
    case '2': 
    case '3': 
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      NrSkipBytes[i] = (15-strtol(ChanId[i],NULL,10))*128 ;
#ifdef TESTLEVEL
      fprintf( stderr, "chan: %d, Id: %s, skew: %d\n", i, ChanId[i], NrSkipBytes[i] ) ;
#endif
      break;
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
      NrSkipBytes[i] = ('F'-ChanId[i][0])*128 ;
#ifdef TESTLEVEL
      fprintf( stderr, "chan: %d, Id: %s, skew: %d\n", i, ChanId[i],NrSkipBytes[i] ) ;
#endif
      break;
    } ;
  } ;
} ;


int StartAkuPreAll( char **argument ) {
  int i, chan ;

#ifdef TESTLEVEL
  fprintf( stdout, "Start AkuPre of all %d channels ... \n", nrchanwav ) ;
#endif
  for( chan=0 ; chan<nrchanwav ; chan++) {
#ifdef TESTLEVEL
    fprintf( stdout, "... %d\n", chan  ) ;
#endif    
    // start AkuPre
    if( AkuPre[chan]->startAkuPre( argument[(chan)+wavfileoffset], NrSkipBytes[i] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argument[(chan)+wavfileoffset] ) ;
      return(-1) ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
 
#ifdef TESTLEVEL
  fprintf( stdout, "Start AkuPre of Mixfile ...\n" ) ;
#endif
  if(AkuPreMixFile->startAkuPre( argument[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argument[mixfileindex] ) ;
    return(-1) ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

#ifdef TESTLEVEL
  fprintf( stdout, "Start AkuPre of all PZMfiles ...\n" ) ;
#endif
  for( i=0 ; i<nrpzm ; i++ ) {
#ifdef TESTLEVEL
    fprintf( stdout, "%s\n", argument[pzmfileindex[i]] ) ;
#endif
    if(AkuPrePZMFile[i]->startAkuPre( argument[pzmfileindex[i]], NrSkipBytes[nrchanwav+i] ) == _ERR_ ) {
      fprintf( stderr, "Can't open pzmfile '%s'\n", argument[pzmfileindex[i]] ) ;
      return(-1) ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
  
  return(1) ;
}

void InitFeatPointer() {

  prefp = new SPreFeature*[nrchanwav];
  prefppzmfile = new SPreFeature*[nrpzm]  ;
} ;




void GetMinMax( float normspecfeatchanperc, int MinMaxBeg=30000, int MinMaxEnd=48000, CParFloat** normchanpercindiv=0 ) {
  int i, chan, frame, running ;

  CListOfObj ModLoudList, SumLoudList ;

  energymin = new float[nrchanwav+1+nrpzm] ;
  // float *entropymin = new float[nrchanwav+1] ;
  sumloudmax = new float[nrchanwav+1+nrpzm] ;
  modloudmax = new float[nrchanwav+1+nrpzm] ;

  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( i=0 ; i<nrchanwav+1+nrpzm ; i++ ) {
    energymin[i] = 1.0e30 ;
    // entropymin[i] = 1.0e30 ;
    sumloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
  } ;

#ifdef TESTLEVEL
  fprintf( stdout, "Determine minima and maxima of special features ...\n" ) ;
#endif

  usedframes = MAXNRFRAMEFORMAXDET ;

  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    frame = 0 ;
    running = 1 ;

    if( normspecfeatchanperc != 0.0 ) {
      // reset sortlists!
      if( !ModLoudList.isEmpty() )
	ModLoudList.removeAll() ;
      if( !SumLoudList.isEmpty() )
	SumLoudList.removeAll() ;
      //if( !ZCRateList.isEmpty() )
      //ZCRateList.removeAll() ;
    } ;
   
    // check if any of the sigfiles faces EOF yet
    while( running ) { 
      if( (prefp[chan] = AkuPre[chan]->getNextPre()) == 0 ) {
	running = 0 ;
	break ;
      } ;
      
      if( normspecfeatchanperc != 0.0 ) {
	CFloatObj* modloud = new CFloatObj(prefp[chan]->ModLoud) ;
	CFloatObj* sumloud = new CFloatObj(prefp[chan]->SumLoud) ;
	//CFloatObj* zcrate  = new CFloatObj(prefp[chan]->ZCRate) ;

	ModLoudList+=( (CObj*)modloud ) ;
	SumLoudList+=( (CObj*)sumloud ) ;
	//ZCRateList+=( (CObj*)zcrate ) ;
      } ;
      
      //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( frame > MinMaxBeg && frame < MinMaxEnd ) {
	if( prefp[chan]->TotalEnergy < energymin[chan] )
	  energymin[chan] =  prefp[chan]->TotalEnergy ;
	//if( prefp[chan]->ModLoudTimesZCRate < entropymin[chan] )
	//entropymin[chan] = prefp[chan]->ModLoudTimesZCRate ;
      }
      if( prefp[chan]->SumLoud > sumloudmax[chan] )
	sumloudmax[chan] = prefp[chan]->SumLoud ;
      if( prefp[chan]->ModLoud > modloudmax[chan] )
	modloudmax[chan] = prefp[chan]->ModLoud ;
	
      if( ++frame > MAXNRFRAMEFORMAXDET )
	break ;
    } ;

#ifdef TEST_LEVEL
    fprintf( stderr, "Minimum energy found in each channel:\n" ) ;
    for( chan=0 ; chan<nrchanwav; chan++ ) {
      fprintf( stderr, "chan[%d]: %f\n", chan, energymin[chan] ) ;
    }
#endif
      
    // memorize length of shortest wavfile!
    if( frame < usedframes )
      usedframes = frame ;
    
    // overwrite max-values with percentiles, if requested!
    CListIterator modloudit ;
    CListIterator sumloudit;
    if( normspecfeatchanperc != 0.0 ) {
      ModLoudList.quickSort() ;
      SumLoudList.quickSort() ;
      //modloudit = ModLoudList.head() ;
      //sumloudit = SumLoudList.head() ;
      modloudit = ModLoudList.tail() ;
      sumloudit = SumLoudList.tail() ;

      if( normchanpercindiv != 0 ) {
	int percframe = (int)( usedframes*((float)(normchanpercindiv[chan]->Par))/100 );
#ifdef TESTLEVEL
	fprintf( stderr, "PercFrame: %d\n", percframe ) ;
#endif
	for( i=0 ; i<percframe ; i++ ) {
	  --modloudit ;
	  --sumloudit ;
	}
      } else {
	for( i=0 ; i<(int)(usedframes*normspecfeatchanperc/100) ; i++ ) {
	  --modloudit ;
	  --sumloudit ;
	}
      }
      
      sumloudmax[chan] = ((CFloatObj*)(*sumloudit))->getData() ; 
      modloudmax[chan] = ((CFloatObj*)(*modloudit))->getData() ; 
      
      ModLoudList.removeAll() ;
      SumLoudList.removeAll() ;
    }
  } ;


  // process mixfile
  for( frame=0 ; frame<usedframes ; frame++ ) {

    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 )
      break ;
    //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
    if( frame > MinMaxBeg && frame < MinMaxEnd ) {
      if( prefpmixfile->TotalEnergy < energymin[nrchanwav] )
	energymin[nrchanwav] =  prefpmixfile->TotalEnergy ;
      //if( prefpmixfile->ModLoudTimesZCRate < entropymin[nrchanwav] )
      //entropymin[nrchanwav] = prefpmixfile->ModLoudTimesZCRate ;
    }    
    if( prefpmixfile->SumLoud > sumloudmax[nrchanwav] )
      sumloudmax[nrchanwav] = prefpmixfile->SumLoud ;
    if( prefpmixfile->ModLoud > modloudmax[nrchanwav] )
      modloudmax[nrchanwav] = prefpmixfile->ModLoud ;
  } ;

  if( frame<usedframes )
    usedframes = frame ;

  // process pzmfiles
  for( i=0 ; i< nrpzm ; i++ ) {
    for( frame=0 ; frame<usedframes ; frame++ ) {
      prefppzmfile[i] = AkuPrePZMFile[i]->getNextPre() ;
      if( prefppzmfile[i] == 0 )
	break ;
      //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( frame > MinMaxBeg && frame < MinMaxEnd ) {
	if( prefppzmfile[i]->TotalEnergy < energymin[nrchanwav+1+i] )
	  energymin[nrchanwav+1+i] =  prefppzmfile[i]->TotalEnergy ;
	//if( prefpmixfile->ModLoudTimesZCRate < entropymin[nrchanwav] )
	//entropymin[nrchanwav] = prefpmixfile->ModLoudTimesZCRate ;
      }    
      if( prefppzmfile[i]->SumLoud > sumloudmax[nrchanwav+1+i] )
	sumloudmax[nrchanwav+1+i] = prefppzmfile[i]->SumLoud ;
      if( prefppzmfile[i]->ModLoud > modloudmax[nrchanwav+1+i] )
	modloudmax[nrchanwav+1+i] = prefppzmfile[i]->ModLoud ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  usedframes = frame ; // memorize how many frames were used for max-detection!!! (use for percentiles!)


#ifdef TESTLEVEL
  // output of min energies. max sumloud, max modloud, max zcrate: 
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    fprintf( stdout, "channel %d: Min energy: %f\n", chan, energymin[chan] ) ;
    fprintf( stdout, "channel %d: Max sumloud: %f\n", chan, sumloudmax[chan] ) ;
    fprintf( stdout, "channel %d: Max modloud: %f\n", chan, modloudmax[chan] ) ;
  } ;
  fprintf( stdout, "mixfile: Min energy: %f\n", energymin[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Max sumloud: %f\n",  sumloudmax[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Max modloud: %f\n", modloudmax[nrchanwav] ) ;

  for(i=0 ; i<nrpzm ; i++ ) {
    fprintf( stdout, "pzm-file: Min energy: %f\n", energymin[nrchanwav+1+i] ) ;
    fprintf( stdout, "pzm-file: Max sumloud: %f\n",  sumloudmax[nrchanwav+1+i] ) ;
    fprintf( stdout, "pzm-file: Max modloud: %f\n", modloudmax[nrchanwav+1+i] ) ;
  } ;
#endif
} ;


void GetMinMaxMult( float normspecfeatchanperc, CParFloat** normchanpercindiv=0 ) {
  int i, j, chan, frame, running, nrrun, atbeg, atend, inrun ;

  int MinMaxBeg = nrframe/5 ;
  int MinMaxRunLen = nrframe/20 ;
  int MinMaxNrRuns = 4 ;
  int MinMaxSkipBetween = nrframe/5 ;
 
  int *RunStart = new int[MinMaxNrRuns] ;
  int *RunEnd = new int[MinMaxNrRuns] ;

  for( i=0 ; i< MinMaxNrRuns ; i++ ) {
    RunStart[i] = MinMaxBeg + i*MinMaxSkipBetween ;
    RunEnd[i] = MinMaxBeg + i*MinMaxSkipBetween + MinMaxRunLen -1 ;
#ifdef TEST_RUNS
    fprintf( stderr, "MinMax: RunStart[%d]: %d, RunEnd[%d]: %d\n",i, RunStart[i], i, RunEnd[i] ) ;
#endif
  } ;


  CListOfObj ModLoudList, SumLoudList ;

  energymin = new float[nrchanwav+1+nrpzm] ;
  energymintmp = new float[nrchanwav+1+nrpzm] ;
  // float *entropymin = new float[nrchanwav+1] ;
  sumloudmax = new float[nrchanwav+1+nrpzm] ;
  modloudmax = new float[nrchanwav+1+nrpzm] ;

  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( i=0 ; i<nrchanwav+1+nrpzm ; i++ ) {
    energymin[i] = -1.0e30 ;
    energymintmp[i] = 1.0e30 ;
    // entropymin[i] = 1.0e30 ;
    sumloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
  } ;

#ifdef TESTLEVEL
  fprintf( stdout, "Determine minima and maxima of special features ...\n" ) ;
#endif

  usedframes = MAXNRFRAMEFORMAXDET ;

  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    frame = 0 ;
    running = 1 ;

    if( normspecfeatchanperc != 0.0 ) {
      // reset sortlists!
      if( !ModLoudList.isEmpty() )
	ModLoudList.removeAll() ;
      if( !SumLoudList.isEmpty() )
	SumLoudList.removeAll() ;
      //if( !ZCRateList.isEmpty() )
      //ZCRateList.removeAll() ;
    } ;
   
    // check if any of the sigfiles faces EOF yet
    while( running ) { 
      if( (prefp[chan] = AkuPre[chan]->getNextPre()) == 0 ) {
	running = 0 ;
	break ;
      } ;
      
      if( normspecfeatchanperc != 0.0 ) {
	CFloatObj* modloud = new CFloatObj(prefp[chan]->ModLoud) ;
	CFloatObj* sumloud = new CFloatObj(prefp[chan]->SumLoud) ;
	//CFloatObj* zcrate  = new CFloatObj(prefp[chan]->ZCRate) ;

	ModLoudList+=( (CObj*)modloud ) ;
	SumLoudList+=( (CObj*)sumloud ) ;
	//ZCRateList+=( (CObj*)zcrate ) ;
      } ;
    
      // check if at beg, within or at end of Run for detecting Min Energy:
      atbeg = 0 ;
      atend = 0 ;
      nrrun = 0 ;
      inrun = 0 ;

      for( i=0 ; i<MinMaxNrRuns ; i++ ) {
	if( frame==RunStart[i] ) {
	  atbeg = 1 ;
	  nrrun = i ;
	  break ;
	} else if( frame == RunEnd[i] ) {
	  atend = 1 ;
	  nrrun = i ;
	  break ;
	} else if( frame >RunStart[i] && frame<RunEnd[i] ) {
	  inrun = 1 ;
	  nrrun = i ;
	  break ;
	} 
      }

      if( atbeg ) {
	// reset tmp counter:
	energymintmp[chan] = 1.0e30 ;
      } else if ( atend ) {
	// search for max minimum of all runs
	if( energymintmp[chan] > energymin[chan] )
	  energymin[chan] = energymintmp[chan] ;
      } else if( inrun ==1 ) {
	// within run:
	if( prefp[chan]->TotalEnergy < energymintmp[chan] )
	  energymintmp[chan] =  prefp[chan]->TotalEnergy ;
      }
  
      if( prefp[chan]->SumLoud > sumloudmax[chan] )
	sumloudmax[chan] = prefp[chan]->SumLoud ;
      if( prefp[chan]->ModLoud > modloudmax[chan] )
	modloudmax[chan] = prefp[chan]->ModLoud ;
	
      if( ++frame > MAXNRFRAMEFORMAXDET )
	break ;
    } ;

#ifdef TEST_LEVEL
    fprintf( stderr, "Minimum energy found in each channel:\n" ) ;
    for( chan=0 ; chan<nrchanwav; chan++ ) {
      fprintf( stderr, "chan[%d]: %f\n", chan, energymin[chan] ) ;
    }
#endif
      
    // memorize length of shortest wavfile!
    if( frame < usedframes )
      usedframes = frame ;
    
    // overwrite max-values with percentiles, if requested!
    CListIterator modloudit ;
    CListIterator sumloudit;
    if( normspecfeatchanperc != 0.0 ) {
      ModLoudList.quickSort() ;
      SumLoudList.quickSort() ;
      //modloudit = ModLoudList.head() ;
      //sumloudit = SumLoudList.head() ;
      modloudit = ModLoudList.tail() ;
      sumloudit = SumLoudList.tail() ;

      if( normchanpercindiv != 0 ) {
	int percframe = (int)( usedframes*((float)(normchanpercindiv[chan]->Par))/100 );
#ifdef TESTLEVEL
	fprintf( stderr, "PercFrame: %d\n", percframe ) ;
#endif
	for( i=0 ; i<percframe ; i++ ) {
	  --modloudit ;
	  --sumloudit ;
	}
      } else {
	for( i=0 ; i<(int)(usedframes*normspecfeatchanperc/100) ; i++ ) {
	  --modloudit ;
	  --sumloudit ;
	}
      }
      
      sumloudmax[chan] = ((CFloatObj*)(*sumloudit))->getData() ; 
      modloudmax[chan] = ((CFloatObj*)(*modloudit))->getData() ; 
      
      ModLoudList.removeAll() ;
      SumLoudList.removeAll() ;
    }
  } ;


  // process mixfile
  for( frame=0 ; frame<usedframes ; frame++ ) {

    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 )
      break ;

    atbeg = 0 ;
    atend = 0 ;
    nrrun = 0 ;
    inrun = 0 ;

    for( i=0 ; i<MinMaxNrRuns ; i++ ) {
      if( frame==RunStart[i] ) {
	atbeg = 1 ;
	nrrun = i ;
	break ;
      } else if( frame == RunEnd[i] ) {
	atend = 1 ;
	nrrun = i ;
	break ;
      } else if( frame >RunStart[i] && frame<RunEnd[i] ) {
	inrun = 1 ;
	nrrun = i ;
	break ;
      } 
    }

    if( atbeg ) {
      // reset tmp counter:
      energymintmp[nrchanwav] = 1.0e30 ;
    } else if ( atend ) {
      // search for max minimum of all runs
      if( energymintmp[nrchanwav] > energymin[nrchanwav] )
	energymin[nrchanwav] = energymintmp[nrchanwav] ;
    } else if( inrun ) {
      // within run:
      if( prefpmixfile->TotalEnergy < energymintmp[nrchanwav] )
	energymintmp[nrchanwav] =  prefpmixfile->TotalEnergy ;
    }
 
    if( prefpmixfile->SumLoud > sumloudmax[nrchanwav] )
      sumloudmax[nrchanwav] = prefpmixfile->SumLoud ;
    if( prefpmixfile->ModLoud > modloudmax[nrchanwav] )
      modloudmax[nrchanwav] = prefpmixfile->ModLoud ;
  } ;

  if( frame<usedframes )
    usedframes = frame ;

  // process pzmfiles
  for( i=0 ; i< nrpzm ; i++ ) {
    for( frame=0 ; frame<usedframes ; frame++ ) {
      prefppzmfile[i] = AkuPrePZMFile[i]->getNextPre() ;
      if( prefppzmfile[i] == 0 )
	break ;

      // look only at certains portions of the signal file for finding minima
      atbeg = 0 ;
      atend = 0 ;
      nrrun = 0 ;
      inrun = 0 ;

      for( j=0 ; j<MinMaxNrRuns ; j++ ) {
	if( frame==RunStart[j] ) {
	  atbeg = 1 ;
	  nrrun = i ;
	  break ;
	} else if( frame == RunEnd[j] ) {
	  atend = 1 ;
	  nrrun = i ;
	  break ;
	} else if( frame >RunStart[j] && frame<RunEnd[j] ) {
	  inrun = 1 ;
	  nrrun = i ;
	  break ;
	} 
      }
      
      if( atbeg ) {
	// reset tmp counter:
	energymintmp[nrchanwav+1+i] = 1.0e30 ;
      } else if ( atend ) {
	// search for max minimum of all runs
	if( energymintmp[nrchanwav+1+i] > energymin[nrchanwav+1+i] )
	  energymin[nrchanwav+1+i] = energymintmp[nrchanwav+1+i] ;
      } else if( inrun ) {
	// within run:
	if( prefppzmfile[i]->TotalEnergy < energymintmp[nrchanwav+1+i] )
	  energymintmp[nrchanwav+1+i] =  prefppzmfile[i]->TotalEnergy ;
      }
      
      if( prefppzmfile[i]->SumLoud > sumloudmax[nrchanwav+1+i] )
	sumloudmax[nrchanwav+1+i] = prefppzmfile[i]->SumLoud ;
      if( prefppzmfile[i]->ModLoud > modloudmax[nrchanwav+1+i] )
	modloudmax[nrchanwav+1+i] = prefppzmfile[i]->ModLoud ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  usedframes = frame ; // memorize how many frames were used for max-detection!!! (use for percentiles!)


#ifdef TESTLEVEL
  // output of min energies. max sumloud, max modloud, max zcrate: 
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    fprintf( stderr, "channel %d: Min energy: %f\n", chan, energymin[chan] ) ;
    fprintf( stderr, "channel %d: Max sumloud: %f\n", chan, sumloudmax[chan] ) ;
    fprintf( stderr, "channel %d: Max modloud: %f\n", chan, modloudmax[chan] ) ;
  } ;
  fprintf( stderr, "mixfile: Min energy: %f\n", energymin[nrchanwav] ) ;
  fprintf( stderr, "mixfile: Max sumloud: %f\n",  sumloudmax[nrchanwav] ) ;
  fprintf( stderr, "mixfile: Max modloud: %f\n", modloudmax[nrchanwav] ) ;

  for(i=0 ; i<nrpzm ; i++ ) {
    fprintf( stderr, "pzm-file: Min energy: %f\n", energymin[nrchanwav+1+i] ) ;
    fprintf( stderr, "pzm-file: Max sumloud: %f\n",  sumloudmax[nrchanwav+1+i] ) ;
    fprintf( stderr, "pzm-file: Max modloud: %f\n", modloudmax[nrchanwav+1+i] ) ;
  } ;
#endif

  delete[] RunStart;
  delete[] RunEnd ;
} ;





int GetLengthOfLabFiles() {
  int chan ;
#ifdef TESTLEVEL
  fprintf( stdout, "\nDetermine length of lab-files ...\n" ) ;
#endif
  // how long are files expected to be?
  int minlabfilelen = Lab[0]->getEndFrameOfLastEntry() ;
  for( chan=1 ; chan<nrchanwav ; chan++ ) {
    if( Lab[chan]->getEndFrameOfLastEntry() < minlabfilelen )
      minlabfilelen = Lab[chan]->getEndFrameOfLastEntry() ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... length of shortest labfile is %d frames\n", minlabfilelen ) ;
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  return( minlabfilelen ) ;
}

int RestartAkupre( char **argument) {
  int chan, i ;
#ifdef TESTLEVEL
  fprintf( stdout, "... Clear and restart AkuPre of all %d channels ... \n", nrchanwav ) ;
#endif
  // reset acoustic preprocessing for channels
  for( chan=0 ; chan<nrchanwav ; chan++) {
#ifdef TESTLEVEL
    fprintf( stdout, "... %d\n", chan  ) ;
#endif
    // clear AkuPre
    AkuPre[chan]->clearAkuPre() ;
    if( AkuPre[chan]->startAkuPre( argument[(chan)+wavfileoffset], NrSkipBytes[i] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argument[(chan)+wavfileoffset] ) ;
      return(-1) ;
    } ;
  } ;
  // clear AkuPre (MixFile)
  AkuPreMixFile->clearAkuPre() ;
  if( AkuPreMixFile->startAkuPre( argument[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argument[mixfileindex] ) ;
    return(-1) ;
  } ;
  for( i=0 ; i<nrpzm ; i++ ) {
    // clear AkuPre (PZM-File)
    AkuPrePZMFile[i]->clearAkuPre() ;
    if( AkuPrePZMFile[i]->startAkuPre( argument[pzmfileindex[i]], NrSkipBytes[nrchanwav+i] ) == _ERR_ ) {
      fprintf( stderr, "Can't open mixfile '%s'\n", argument[pzmfileindex[i]] ) ;
      return(-1) ;
    } ;
  } ;
  return(1) ;
} 


void PrintChanFeatFileHeaders() {
  int chan ;
  // print headers in pfiles and labfiles
  // channels
  for( chan=0 ;chan<nrchanwav ; chan++ ) {
    fprintf( featoutfp[chan], "-pfile_header version 0 size 32768\n" ) ;
    fprintf( featoutfp[chan], "-num_sentences 1\n" ) ;
    fprintf( featoutfp[chan], "-num_frames %d\n", nrframe ) ;
    fprintf( featoutfp[chan], "-first_feature_column 2\n", nrframe ) ;
    fprintf( featoutfp[chan], "-num_features %d\n", sizeof(sPreFeature)/sizeof(float) ) ;
    fprintf( featoutfp[chan], "-first_label_column %d\n", sizeof(sPreFeature)/sizeof(float)+2 ) ;
    fprintf( featoutfp[chan], "-num_labels %d\n", 0 ) ;
    fprintf( featoutfp[chan], "-format %s\n", "2d26f" ) ;
    fprintf( featoutfp[chan], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(26+2), 0, 2, nrframe, 26+2  ) ;
    fseek( featoutfp[chan], 32768, 0 ) ;
  } ;
}

void PrintPzmFeatFileHeader( ) {
  int i ;
  // print headers in pfiles and labfiles
  // channels
  for( i=0 ;i<nrpzm ; i++ ) {
    fprintf( featoutfp[nrchanwav+1+i], "-pfile_header version 0 size 32768\n" ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-num_sentences 1\n" ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-num_frames %d\n", nrframe ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-first_feature_column 2\n", nrframe ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-num_features %d\n", sizeof(sPreFeature)/sizeof(float)-2 ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-first_label_column %d\n", sizeof(sPreFeature)/sizeof(float) ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-num_labels %d\n", 0 ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-format %s\n", "2d24f" ) ;
    fprintf( featoutfp[nrchanwav+1+i], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(24+2), 0, 2, nrframe, 24+2  ) ;
    fseek( featoutfp[nrchanwav+1+i], 32768, 0 ) ;
  } ;
}

void PrintMixFeatFileHeader() {
  //mixfile:
  fprintf( featoutfp[nrchanwav], "-pfile_header version 0 size 32768\n" ) ;
  fprintf( featoutfp[nrchanwav], "-num_sentences 1\n" ) ;
  fprintf( featoutfp[nrchanwav], "-num_frames %d\n", nrframe ) ;
  fprintf( featoutfp[nrchanwav], "-first_feature_column 2\n" ) ;
  fprintf( featoutfp[nrchanwav], "-num_features %d\n", sizeof(sPreFeature)/sizeof(float)-2 ) ;
  fprintf( featoutfp[nrchanwav], "-first_label_column %d\n", sizeof(sPreFeature)/sizeof(float) ) ;
  fprintf( featoutfp[nrchanwav], "-num_labels %d\n", 0 ) ;
  fprintf( featoutfp[nrchanwav], "-format %s\n", "2d24f" ) ;
  fprintf( featoutfp[nrchanwav], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(24+2), 0, 2, nrframe, 24+2  ) ;
  fseek( featoutfp[nrchanwav], 32768, 0 ) ;
}
 


void PrintChanLabFileHeaders() {
  int chan ;
  // channels
  for( chan=0 ;chan<nrchanwav ; chan++ ) {
    fprintf( laboutfp[chan], "-pfile_header version 0 size 32768\n" ) ;
    fprintf( laboutfp[chan], "-num_sentences 1\n" ) ;
    fprintf( laboutfp[chan], "-num_frames %d\n", nrframe ) ;
    fprintf( laboutfp[chan], "-first_feature_column 2\n" ) ;
    fprintf( laboutfp[chan], "-num_features %d\n", 0 ) ;
    fprintf( laboutfp[chan], "-first_label_column %d\n", 2 ) ;
    fprintf( laboutfp[chan], "-num_labels %d\n", 1 ) ;
    fprintf( laboutfp[chan], "-format %s\n", "2dd" ) ;
    fprintf( laboutfp[chan], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(1+2), 0, 2, nrframe, 1+2  ) ;
    fseek( laboutfp[chan], 32768, 0 ) ;
  } ;
}

void PrintPzmLabFileHeader() {
  int i ;
  // pzm files
  for( i=0 ;i<nrpzm ; i++ ) {
    fprintf( laboutfp[nrchanwav+1+i], "-pfile_header version 0 size 32768\n" ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-num_sentences 1\n" ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-num_frames %d\n", nrframe ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-first_feature_column 2\n" ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-num_features %d\n", 0 ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-first_label_column %d\n", 2 ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-num_labels %d\n", 1 ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-format %s\n", "2dd" ) ;
    fprintf( laboutfp[nrchanwav+1+i], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(1+2), 0, 2, nrframe, 1+2  ) ;
    fseek( laboutfp[nrchanwav+1+i], 32768, 0 ) ;
  } ;
}

void PrintMixLabFileHeader() {
  //mixfile:
  fprintf( laboutfp[nrchanwav], "-pfile_header version 0 size 32768\n" ) ;
  fprintf( laboutfp[nrchanwav], "-num_sentences 1\n" ) ;
  fprintf( laboutfp[nrchanwav], "-num_frames %d\n", nrframe ) ;
  fprintf( laboutfp[nrchanwav], "-first_feature_column 2\n" ) ;
  fprintf( laboutfp[nrchanwav], "-num_features %d\n", 0 ) ;
  fprintf( laboutfp[nrchanwav], "-first_label_column %d\n", 2 ) ;
  fprintf( laboutfp[nrchanwav], "-num_labels %d\n", 1 ) ;
  fprintf( laboutfp[nrchanwav], "-format %s\n", "2dd" ) ;
  fprintf( laboutfp[nrchanwav], "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(1+2), 0, 2, nrframe, 1+2  ) ;
  fseek( laboutfp[nrchanwav], 32768, 0 ) ;
}


void NormSpecialFeatures( int normspecfeatchan, float normspecfeatchanperc, int normspecfeatmix, int normspecfeatpzm, int usepzmenergy, int usemodloudtimeszcrate, char **argument ) {

  int chan, i ;


  // get feature vecs, prepare feature vecs by normalizing
  meanenergy = 0.0 ;
  for( chan=0 ; chan<nrchanwav ; chan++) {
    prefp[chan] = AkuPre[chan]->getNextPre() ;
    if( prefp[chan] == 0 ) {
      fprintf( stderr, "Unexpected EOF reached for file '%s'\n", argument[(chan)+wavfileoffset] ) ;
      break ;
    } ;
    
    // use prduct of modloud and zcrate as madiitinal feature
    prefp[chan]->ModLoudTimesZCRate = (prefp[chan]->ZCRate) * ((prefp[chan]->ModLoud )/modloudmax[chan]) ;
    if( normspecfeatchan || normspecfeatchanperc != 0.0 ) {
      // normalize channel features: 
      prefp[chan]->TotalEnergy -= energymin[chan] ;
      prefp[chan]->ModLoud /= modloudmax[chan] ;
      prefp[chan]->SumLoud /= sumloudmax[chan] ;
      //prefp[chan]->ZCRate /= zcratemax[chan] ;    
    } ;
    // calc mean energy over all channel energies
    meanenergy += prefp[chan]->TotalEnergy ;
  } ;
  
  // normalize mixfile features:
  prefpmixfile = AkuPreMixFile->getNextPre() ;
  if( prefpmixfile == 0 ) {
    fprintf( stderr, "Unexpected EOF reached for file '%s'\n", argument[mixfileindex] ) ;
    //break ;
  } ;
  // use product of modloud and zcrate as additional feature
  prefpmixfile->ModLoudTimesZCRate = (prefpmixfile->ZCRate) * ((prefpmixfile->ModLoud)/(modloudmax[nrchanwav])) ;
  if( normspecfeatmix ) {
    prefpmixfile->TotalEnergy -= energymin[nrchanwav] ;
    prefpmixfile->SumLoud /= sumloudmax[nrchanwav] ;
    //prefpmixfile->ZCRate /= zcratemax[nrchanwav] ; 
    prefpmixfile->ModLoud /= modloudmax[nrchanwav] ;
  } ;
  
  for( int i=0 ; i<nrpzm ; i++ ) {
    // normalize PZM-special-features (anyway)
    prefppzmfile[i] = AkuPrePZMFile[i]->getNextPre() ;
    if( usepzmenergy || normspecfeatpzm ) {
      prefppzmfile[i]->TotalEnergy -= energymin[nrchanwav+1+i] ;
      prefppzmfile[i]->SumLoud /= sumloudmax[nrchanwav+1+i] ;
      //prefpmixfile[i]->ZCRate /= zcratemax[nrchanwav+1+i] ; 
      prefppzmfile[i]->ModLoud /= modloudmax[nrchanwav+1+i] ;
    } ;
  };
  
  // calc mean energy:
  meanenergyohnemix = meanenergy ;
  meanenergyohnemix /= nrchanwav ;
  meanenergy += prefpmixfile->TotalEnergy ;
  meanenergy /= (nrchanwav+1) ;
  
  // normalize energies with mean over channels
  if( normspecfeatchan || normspecfeatchanperc != 0.0 ) {
    for( chan=0 ; chan<nrchanwav ; chan++) {
      prefp[chan]->TotalEnergy -= meanenergyohnemix ;
      //(prefp[chan]->ModLoudTimesZCRate -= entropymin[chan]) *= pow( 10.0,(prefp[chan]->TotalEnergy)/10) ;
      //prefp[chan]->ModLoudTimesZCRate *= (prefp[chan]->ModLoud) ;
    } ;
  } ;
  if( normspecfeatmix ) {
    // normalize with mean over channels
    prefpmixfile->TotalEnergy -= meanenergyohnemix ;
    //(prefpmixfile->ModLoudTimesZCRate -= entropymin[nrchanwav]) *= pow(10.0,(prefpmixfile->TotalEnergy/10)) ;
    //prefpmixfile->ModLoudTimesZCRate *= (prefpmixfile->ModLoud) ;
  } 
  if( normspecfeatpzm ) {
    for( i=0 ; i<nrpzm ; i++ ) {
      // normalize with mean over channels
      prefppzmfile[i]->TotalEnergy -= meanenergyohnemix ;
    }
  } 
  
  // normalize PZM-energy with mean across channels 
  pzm_mean = 0.0 ;
  if( usepzmenergy ) {
    for( i=0 ; i<nrpzm ; i++ ) {
      //prefppzmfile[i]->TotalEnergy -= meanenergyohnemix ;
      pzm_mean += prefppzmfile[i]->TotalEnergy ;
    } ;
    pzm_mean /= nrpzm ;
    for( chan=0 ; chan<nrchanwav ; chan++ ) 
      //prefp[chan]->ModLoudTimesZCRate = prefp[chan]->TotalEnergy - prefppzmfile->TotalEnergy ;
      if( usemodloudtimeszcrate )
	prefp[chan]->PZMEnergy = prefp[chan]->TotalEnergy - pzm_mean ;
      else 
	prefp[chan]->ModLoudTimesZCRate = prefp[chan]->TotalEnergy - pzm_mean ;
  } 
}




void WriteFeaturesLabels( int chanfeatfileout, int mixfeatfileout, int pzmfeatfileout, int chanlabfileout, int mixlabfileout, int pzmlabfileout, int sent, int frame ) {

  int chan, run, i ;
  // first decide if there is speech or nonspeech on indiv channels -> if there is speech in any channel,
  // it is assumed that there is also speech in the mixfile!!!
  int speechmix = 0 ;
  int lapel=0 ;
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    int speechchan = 0 ;
    // is 'speech' in the current channel?
    for( run=0 ; run<nr_runs[chan] ; run++ ) {
      if( frame >= start[chan][run] && frame <= stop[chan][run] ) {
	// speech in this chan found
	speechchan = 1 ;
	if( chan == 0 )
	  lapel = 1 ;
	break ; // matching run found, leave loop over all runs!
      } ;
    } ;
    
    // FeatureFiles Wanted?
    if( chanfeatfileout ) {
      fwrite( &sent,  sizeof(int), 1, featoutfp[chan] ) ;
      fwrite( &frame, sizeof(int), 1, featoutfp[chan] ) ;
      fwrite( (float*)prefp[chan], sizeof(float), sizeof(sPreFeature)/sizeof(float), featoutfp[chan] ) ;
    } ;
    
    // LabelFiles Wanted?
    if( chanlabfileout ) {
      fwrite( &sent,  sizeof(int), 1, laboutfp[chan] ) ;
      fwrite( &frame, sizeof(int), 1, laboutfp[chan] ) ;
      fwrite( &speechchan, sizeof(int), 1, laboutfp[chan] ) ; 
    } 
    speechmix |= speechchan ;
    
  } ; // END loop over all channels to decide whether speech or nonspeech!
  
  // write features and labels for mixfile
  // FeatureFiles Wanted?
  if( mixfeatfileout ) {
    fwrite( &sent,  sizeof(int), 1, featoutfp[nrchanwav] ) ;
    fwrite( &frame, sizeof(int), 1, featoutfp[nrchanwav] ) ;
    fwrite( (float*)prefpmixfile, sizeof(float), sizeof(sPreFeature)/sizeof(float)-2, featoutfp[nrchanwav] ) ;
  } ;
  // LabelFiles Wanted?
  if(  mixlabfileout ) {
    fwrite( &sent,  sizeof(int), 1, laboutfp[nrchanwav] ) ;
    fwrite( &frame, sizeof(int), 1, laboutfp[nrchanwav] ) ;
    fwrite( &speechmix, sizeof(int), 1, laboutfp[nrchanwav] ) ; 
  } 
  
  // write features and labels for pzmfiles
  // FeatureFiles Wanted?
  if( pzmfeatfileout ) {
    for( i=0 ; i<nrpzm ; i++ ) {
      fwrite( &sent,  sizeof(int), 1, featoutfp[nrchanwav+1+i] ) ;
      fwrite( &frame, sizeof(int), 1, featoutfp[nrchanwav+1+i] ) ;
      fwrite( (float*)prefppzmfile[i], sizeof(float), sizeof(sPreFeature)/sizeof(float)-2, featoutfp[nrchanwav+1+i] ) ;
    } ;
  } ;
  // LabelFiles Wanted?
  if( pzmlabfileout ) {
    for( i=0 ; i<nrpzm ; i++ ) {
      fwrite( &sent,  sizeof(int), 1, laboutfp[nrchanwav+1+i] ) ;
      fwrite( &frame, sizeof(int), 1, laboutfp[nrchanwav+1+i] ) ;
      fwrite( &speechmix, sizeof(int), 1, laboutfp[nrchanwav+1+i] ) ;
    }  
  } 
}


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i, run, chan, frame, tmpframe, sent=0 ;
  // open files
  if( argc < 8) 
    {   fprintf(stderr,"usage: spdtrain <akupre-cfg-file> <outpath> <wavfile1> <wavfile2> ... <wavfileN> <labfile1> <labfile2> ... <labfileN> <mixfile> <pzmfile1> ... <nrpzm>\n");
    return _ERR_;
    }

  ReadCommandLine( argc, argv ) ;
  if( OpenCfgFiles( argv ) != 1 ) {
    fprintf( stderr, "Can't open all cfg-files, aborting!\n" );
    exit(-1) ;
  }

  GetChanIds( argv ) ;

  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParEnum InputPar( "Tools.Label.InputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode InputMode = (EOutPutMode)InputPar.Par ;  
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  //CParBool  SpdAutoUpdateSilent( "Tools.SpdAutoUpdate.Silent", chanfp, 1 ) ;
  //CParBool  SpdAutoUpdateSpeech( "Tools.SpdAutoUpdate.Speech", chanfp, 1 ) ;
  //CParBool  SpDetUseCommonForAllChan( "Tools.SpdUseCommonForAllChan", chanfp, 1 ) ;
  CParBool NormSpecialFeatChan( "Tools.NormSpecialFeatChan", chanfp, 1 ) ;
  CParFloat NormSpecialFeatChanPerc( "Tools.NormSpecialFeatChanPerc", chanfp, 0.0 ) ; // use percentile of max instead of max!
  CParFloat** NormSpecialFeatChanPercIndiv = new CParFloat*[nrchanwav] ;
  char *tmpperc= new char[strlen("Tools.NormSpecialFeatChanPercIndiv.Chan")+3];
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    sprintf( tmpperc, "%s%s\0",  "Tools.NormSpecialFeatChanPercIndiv.Chan", ChanId[chan] ) ;
    NormSpecialFeatChanPercIndiv[chan] = new CParFloat( tmpperc, chanfp, 0.1 ) ;
  }
  CParBool NormSpecialFeatMix( "Tools.NormSpecialFeatMix", chanfp, 0 ) ;
  CParBool NormSpecialFeatPzm( "Tools.NormSpecialFeatPzm", chanfp, 0 ) ;
  CParBool UsePZMEnergy( "Tools.UsePZMEnergy", chanfp, 1 ) ;
  CParBool UseModLoudTimesZCRate( "Tools.UseModLoudTimesZCRate", chanfp, 1 ) ;
  CParBool  ControlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;
  CParInt SmoothWindowLen( "Tools.Result.SmoothWindowLen", chanfp, 41 ) ;
  CParBool OverlapWanted( "Tools.Result.OverlapWanted", chanfp, 0 ) ;
  CParBool MixFeatFileOut( "Tools.MixFeatFileOut", chanfp, 0 ) ;
  CParBool ChanFeatFileOut( "Tools.ChanFeatFileOut", chanfp, 0 ) ;
  CParBool PzmFeatFileOut( "Tools.PzmFeatFileOut", chanfp, 0 ) ;
  CParBool MixLabFileOut( "Tools.MixLabFileOut", chanfp, 0 ) ;
  CParBool ChanLabFileOut( "Tools.ChanLabFileOut", chanfp, 0 ) ;
  CParBool PzmLabFileOut( "Tools.PzmLabFileOut", chanfp, 0 ) ;
  CParString FeatEnding( "Tools.FeatEnding", chanfp, "pfile" ) ;
  CParString LabEnding( "Tools.LabEnding", chanfp, "pflab" ) ;
  CParInt MinMaxBegFrame( "Tools.MinMaxBegFrame", chanfp, 60000 ) ;
  CParInt MinMaxRunLen( "Tools.MinMaxRunLen", chanfp, 6000 ) ;
  CParInt MinMaxNrRun( "Tools.MinMaxNrRun", chanfp, 5 ) ;

  CParBool CorrectLabelInfo( "Tools.CorrectLabelInfo", chanfp, 0 ) ;
  CParBool CorrectForSkew( "Tools.CorrectForSkew", chanfp, 1 ) ;
  
  // read cfg files for starting akupre of all channels and pzmfiles and mixfile
  InitAkuPre( argv ) ;

  //InitSpnonsp(SpDetUseCommonForAllChan.Par, UsePZMEnergy.Par, UseModLoudTimesZCRate.Par) ;


  OpenOutFiles( ChanFeatFileOut.Par, MixFeatFileOut.Par, PzmFeatFileOut.Par, ChanLabFileOut.Par, MixLabFileOut.Par, PzmLabFileOut.Par, LabEnding.Par, FeatEnding.Par, argv ) ;
  
  InitLabelInfo( InputMode, Tool, InputPar.getValue( XML ), SpeechLabel.Par, NonspeechLabel.Par, argv ) ;
    
  OpenInputLabelFiles( argv ) ;

  GetRuns( argv, ControlOutput.Par ) ;


  if( CorrectForSkew.Par )
    GetSkipInfo() ;
  else {
    NrSkipBytes = new int[nrchanwav+nrpzm] ;
    for( i=0 ; i<nrchanwav+nrpzm ; i++ )
      NrSkipBytes[i] = 0 ;
  }

  if( CorrectLabelInfo.Par ) 
    CorrectRuns() ;

  if( StartAkuPreAll( argv ) != 1 ) {
    fprintf( stderr, "Could not start akupre of all files! Aborting ...\n" ) ;
    exit(-1) ;
  } ;

  InitFeatPointer() ;

  //nrframe = GetLengthOfLabFiles() ;
  nrframe= CountFrames( argv ) ;

  GetMinMaxMult(NormSpecialFeatChanPerc.Par ) ;

  // get min of labfilelen and sigfilelen
  //if( nrframe < usedframes )
  //usedframes = nrframe ;
  //else
  //nrframe = usedframes ;


#ifdef TESTLEVEL
  if( NormSpecialFeatChan.Par || NormSpecialFeatChanPerc.Par != 0.0 )
    fprintf( stdout, "Normalize Special features of individual channels ...\n" ) ;
  if( NormSpecialFeatMix.Par )
    fprintf( stdout, "Normalize Special features of mixfile ...\n" ) ;
  if( NormSpecialFeatPzm.Par )
    fprintf( stdout, "Normalize Special features of pzmfiles ...\n" ) ;
#endif


 if( RestartAkupre( argv ) != 1 ) {
   fprintf( stderr, "Can't restart akupre of all sigfiles! Aborting ... \n" );
   exit(-1) ;
 } 

 if( ChanFeatFileOut.Par ) 
   PrintChanFeatFileHeaders() ;
 
 if( PzmFeatFileOut.Par )
   PrintPzmFeatFileHeader() ;
   
 if( MixFeatFileOut.Par ) 
   PrintMixFeatFileHeader() ;
 
 if( ChanLabFileOut.Par ) 
   PrintChanLabFileHeaders() ;
   
 if( PzmFeatFileOut.Par )
   PrintPzmLabFileHeader() ;
   
 if( MixFeatFileOut.Par ) 
   PrintMixLabFileHeader() ;


 // loop over all frames, normalize feature and train system:
 for( frame=0 ; frame<nrframe ; frame++ ) {
   
   NormSpecialFeatures( NormSpecialFeatChan.Par, NormSpecialFeatChanPerc.Par, NormSpecialFeatMix.Par, NormSpecialFeatPzm.Par, UsePZMEnergy.Par, UseModLoudTimesZCRate.Par, argv ) ;

   // write features and or labels:
   WriteFeaturesLabels( ChanFeatFileOut.Par, MixFeatFileOut.Par, PzmFeatFileOut.Par, ChanLabFileOut.Par, MixLabFileOut.Par, PzmLabFileOut.Par, sent, frame ) ;

 } ; // END loop over all frames

#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
      
#ifdef TESTLEVEL
  fprintf( stdout, "Free memory ...\n", nrchanwav, argv[argc-1] ) ;
#endif

  // free memory:
  for( i=0 ; i<nrchanwav ; i++ ) {
    AkuPre[i]->clearAkuPre();
    delete AkuPre[i] ;
    //if( !SpDetUseCommonForAllChan.Par ) 
    //delete SpNonSp[i] ;
  } ;
  //if( !SpDetUseCommonForAllChan.Par ) 
  //delete SpNonSp[0] ;

  //delete[] SpNonSp ;
  delete[] AkuPre ;

  AkuPreMixFile->clearAkuPre() ;
  delete   AkuPreMixFile ;
  
  for( i=0 ;i<nrpzm ; i++ ) {
    AkuPrePZMFile[i]->clearAkuPre();
    delete AkuPrePZMFile[i] ;
  } ;

  delete[] energymin ;
  delete[] prefp ;
  delete[] prefppzmfile ;
  
  delete[] nr_runs ;
  delete[] start ;
  delete[] stop ;

  delete[] energymin ;
  delete[] energymintmp ;

  delete[] sumloudmax  ;
  delete[] modloudmax ;
  

#ifdef TESTLEVEL
  fprintf( stdout, "Done\n", nrchanwav, argv[argc-1] ) ;
#endif
  return _OK_;
}
