/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "genobj.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupresimple.h"
#include "spnonsp.h"

//#include "p_histo.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

#define MAXNRCHAN 20
#define MaxRuns 20000
#define MAXNRFRAMEFORMAXDET 1080000


#define TESTLEVEL
#define TESTLEVEL2

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif




// -- global variables

CTool      *Tool ;

FILE *laboutfp, *chanfp ;

int NrSpLabels, NrSiLabels ; 
char **SpLabel, **SiLabel ;
//char *SpOutLabel, *SiOutLabel ;
CListOfLabel *Lab ;
FILE *labfp ;

long nr_runs ;
long *start ;
long *stop ;

int usedframes ;
int nrframe ;






int OpenCfgFile( char **argument ) {
  chanfp = fopen( argument[1], "r" );
  if( chanfp == 0 ) {
    fprintf( stderr, "can't open akupre-cfg-file: %s\n", argument[1] );
    return(-1) ;
  }

  return(1) ;

} ;



void OpenOutFile( char *labending, char **argument ) {
  int i ;

#ifdef TESTLEVEL
  fprintf( stdout, "Open Outfile ...\n" ) ;
#endif

  char labtmp[MaxLineLength] ;
  char tmptmp[MaxLineLength] ;

  strcpy( tmptmp, argument[2] ) ;
  sprintf( labtmp, "%s.%s", tmptmp, labending );
  laboutfp = fopen( labtmp, "w" ) ;

#ifdef TESTLEVEL
  fprintf( stdout, "... Done\n" ) ;
#endif
}
 

void InitLabelInfo( EOutPutMode inputmode, CTool *tool, char *keyword, char* speechlabel, char* nonspeechlabel, char **argument ) {
  int chan,i ;

#ifdef TESTLEVEL
  fprintf( stdout, "InitLabelInfo ...\n" ) ;
#endif

  // initialize label-info
  // just quick and dirty, class CListOfLabel should be redesigned!
  if( inputmode == XML )
    // XML Input, use XML as KeyWord
    Lab = new CListOfLabel( keyword );
  else 
    // MUC or Kiel Input! use Keyword as Keyword
    Lab = new CListOfLabel( tool->LabelKeyword.Par );

#ifdef TESTLEVEL
  fprintf( stdout, "Keyword read\n" ) ;
#endif

  // set Labels to empty string if requested!
  if( strcmp( speechlabel, "NONE" ) == 0 ) 
    strcpy( speechlabel, "" ) ;
  if( strcmp( nonspeechlabel, "NONE" ) == 0 ) 
    strcpy( nonspeechlabel, "" ) ;
  
  // create array of Silence-Labels
  NrSiLabels = CParInt( "Tools.NrSiLabel", chanfp, 1 ).Par ;
  SiLabel = new char*[NrSiLabels] ;
  char tmp[MAXKEYWORDLEN] ;
  for( i=0 ; i<NrSiLabels ; i++ ) {
    SiLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SilentName.%d", i ) ;
    strcpy( SiLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SiLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SilentLabel in cfg-file '%s'\n", i, argument[1] ) ;
      exit(-1) ;
    } ;
  } ;
  
  // create array of Speech-Labels
  // NrSpLabels = SpNonSp[0]->getNrOfMixSp() ; 
  NrSpLabels = CParInt( "Tools.NrSpLabel", chanfp, 1 ).Par ;
  SpLabel = new char*[NrSpLabels] ;
  for( i=0 ; i<NrSpLabels ; i++ ) {
    SpLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SpeechName.%d", i ) ;
    strcpy( SpLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SpLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SpeechLabel in cfg-file '%s'\n", i, argument[1] ) ;
      exit(-1) ;
    } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... Done\n" ) ;
#endif
}



void  OpenInputLabelFile( char **argument ) {
#ifdef TESTLEVEL
  fprintf( stdout, "Open labfile ...\n" ) ;
#endif
  // open lab-file
  labfp = fopen( argument[2], "r");
  if( labfp == 0 ) {   
    ERROR_MSG_S("can't open label-file:", argument[2] );
    exit(-1) ; 
  }
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
}



void GetRuns( char **argument, int controloutput ) {
  int run  ;
  int count ;
#ifdef TESTLEVEL
  fprintf( stdout, "Define 'runs' ...\n" ) ;
#endif
  
  // define array for starts and stops of runs
  //fprintf( stderr, "maxRuns: %d\n", MaxRuns ) ;
  
  start = new long[MaxRuns] ;
  stop = new long[MaxRuns] ;
    
  // default values for start and stop of each run
  for( run=0 ; run<MaxRuns ; run++ ) 
    start[run] = stop[run] = -1 ;
    
  // read multiple starts and stops from label-file
  count = 0 ;
  if( Lab->readLabels( labfp ) == _ERR_ ) { 
    ERROR_MSG_S("can't read label-file:", argument[2] );
    fclose( labfp );
  }
  
  // loop over all runs to determine start and stop-frame
  while( count < MaxRuns && Lab->getNextStartStop( (count>0 ? stop[count-1] : 0 ), start[count], stop[count], NrSiLabels, SiLabel ) != _ERR_ ) { 
    if( controloutput ) 
      fprintf( stdout, "start: %ld stop: %ld\n", start[count], stop[count] );
    count++ ;
  } ;
  fclose( labfp );
  nr_runs = count ;
    
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
}









int GetLengthOfLabFiles() {
#ifdef TESTLEVEL
  fprintf( stdout, "\nDetermine length of lab-file ...\n" ) ;
#endif
  // how long are files expected to be?
  int minlabfilelen = Lab->getEndFrameOfLastEntry() ;

  return( minlabfilelen ) ;
}




void PrintChanLabFileHeader() {
  fprintf( laboutfp, "-pfile_header version 0 size 32768\n" ) ;
  fprintf( laboutfp, "-num_sentences 1\n" ) ;
  fprintf( laboutfp, "-num_frames %d\n", nrframe ) ;
  fprintf( laboutfp, "-first_feature_column 2\n" ) ;
  fprintf( laboutfp, "-num_features %d\n", 0 ) ;
  fprintf( laboutfp, "-first_label_column %d\n", 2 ) ;
  fprintf( laboutfp, "-num_labels %d\n", 1 ) ;
  fprintf( laboutfp, "-format %s\n", "2dd" ) ;
  fprintf( laboutfp, "-data size %d offset %d ndim %d nrow %d ncol %d\n",  nrframe*(1+2), 0, 2, nrframe, 1+2  ) ;
  fseek( laboutfp, 32768, 0 ) ;
}



void WriteFeaturesLabels( int sent, int frame ) {

  int run, i ;
  // first decide if there is speech or nonspeech on indiv channels -> if there is speech in any channel,
  // it is assumed that there is also speech in the mixfile!!!
  int speechchan = 0 ;
  // is 'speech' in the current channel?
  for( run=0 ; run<nr_runs ; run++ ) {
    if( frame >= start[run] && frame <= stop[run] ) {
      // speech in this chan found
      speechchan = 1 ;
      break ; // matching run found, leave loop over all runs!
    } ;
  } 
   
  fwrite( &sent,  sizeof(int), 1, laboutfp ) ;
  fwrite( &frame, sizeof(int), 1, laboutfp ) ;
  fwrite( &speechchan, sizeof(int), 1, laboutfp ) ;  
  
  
}


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i, run, chan, frame, tmpframe, sent=0 ;
  // open files
  if( argc != 3) 
    {   fprintf(stderr,"usage: %s <akupre-cfg-file> <labfile1> \n", argv[0] );
    return _ERR_;
    }

  if( OpenCfgFile( argv ) != 1 ) {
    fprintf( stderr, "Can't open all cfg-files, aborting!\n" );
    exit(-1) ;
  }


  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParEnum InputPar( "Tools.Label.InputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode InputMode = (EOutPutMode)InputPar.Par ;  
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  CParBool  ControlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;
  CParString LabEnding( "Tools.LabEnding", chanfp, "pflab" ) ;

  OpenOutFile( LabEnding.Par,  argv ) ;
  
  InitLabelInfo( InputMode, Tool, InputPar.getValue( XML ), SpeechLabel.Par, NonspeechLabel.Par, argv ) ;
    
  OpenInputLabelFile( argv ) ;

  GetRuns( argv, ControlOutput.Par ) ;

  nrframe = GetLengthOfLabFiles() ;

  PrintChanLabFileHeader() ;
   

  // loop over all frames, normalize feature and train system:
  for( frame=0 ; frame<nrframe ; frame++ ) {
    
    // write features and or labels:
    WriteFeaturesLabels( sent, frame ) ;

  } ; // END loop over all frames

#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
      
#ifdef TESTLEVEL
  fprintf( stdout, "Free memory ...\n" ) ;
#endif
 
  delete[] start ;
  delete[] stop ;

  
#ifdef TESTLEVEL
  fprintf( stdout, "Done\n" ) ;
#endif
  return _OK_;
}
