/******************************************************************************
 * file:        akup_muc.h  (for akupre.cc, TU-Muenchen Vorverarbeitung)
 * info:        class declarations for CAkuPre, acoustic preprocessing 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 *
 * Die Implementierung realisiert ein innerhalb des Projekts VERBMOBIL  
 * entwickeltes Signalanalyseverfahren, da"s an der TU-Muenchen erg"anzt und 
 * neu implementiert wurde.
 *
 * Die Software ist ausschlie"slich f"ur die Verwendung innerhalb des Projektes
 * VERBMOBIL bestimmt. Eine Weitergabe der Programme an Partner au"serhalb des 
 * Projektes VERBMOBIL ist nur mit Zustimmung des Lehrstuhls MMK der TU
 * Muenchen zul"assig. 
 * 
 * Das Projekt VERBMOBIL wird vom Bundesminister f"ur Forschung und Technologie
 * unter dem F"orderkennzeichen 01IV 102K gef"ordert.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

/******************************************************************************

CLASS 'CAkuPre':

Beschreibung siehe: _readme

 ******************************************************************************/


#if !defined( _AKUPRE_H_ )
#define _AKUPRE_H_

#include <stdio.h>

#include "pref_muc.h"

#include "dbg_mess.h"
#include "siganl.h"
#include "chancomp.h"
#include "spnonsp.h"
#include "paramet.h"

enum ERdFlag    { Undef, Wanted, NotWanted };

// class definition for CAkuPre
class CAkuPre
{
private:
    // readable parameter for akupre
    CParEnum           ReadMode;
    CParString         DeviceStr;
    CParInt            SampleShift;
    CParFloat          Preemphase;
  
    CParBool           SpNonSpOverlap; // use augmented spnonsp-detector (overlap-state added)

    // object members
    CSigAnalyse        SigAnl;
    CSpeechNonSpeech   *SpNonSp;
    CPreFeature        PreFeature;
    CChanComp          ChanComp;
    CGainControl       GainControl;
    CDeltaFeature      Delta;
    // fix parameters
    const unsigned int  FrqSamp;
    const  int          BufLen;
    const  int          DeltaF;
    const  int          DeltaB;
    const  int          PastHold;
    const  int          MinBufFill;
    // speech non-speech state dependent
    ESpeechFlag         LastFlag;
    unsigned int        UndefCount;
    // fifo control
    int                 BufFill;  
    unsigned long       WrTotal;
    unsigned long       RdTotal;
    SFeatBufEntry      *BasePtr;
    SFeatBufEntry      *WPtr;
    SFeatBufEntry      *RPtr;
    enum EState       { Stopped, Started, Running, Preproc };
    EState              RdState;
    EState              WrState;
    // private function
    SFeatBufEntry  *writeEntry();
    SFeature       *readEntry();
    SFeature       *readDummy();
    void            resetBuf();
    ERdFlag         getRd();
public:
   CAkuPre( FILE *channelfile );
  ~CAkuPre();
   int            startAkuPre( const char *filename );
   void           stopAkuPre();
   void           clearAkuPre();
   SFeature      *getNext();

   SPreFeature   *getNextPre();
   int            getFill()     const { return BufFill;  }
   long           getRdNum()    const { return RPtr->Last->Num; }
   ESpeechFlag    getRdFlag()   const { return RPtr->Last->Flag; }
  double getRdProbSi() const { return RPtr->Last->ProbSi ; } ; // tpfau, 12/21/2000
  double getRdProbSp() const { return RPtr->Last->ProbSp ; } ; // tpfau, 12/21/2000

   float         *getRdMax()    const { return GainControl.getActMax(); }
   unsigned long  getRdTotal()  const { return RdTotal; }
   unsigned long  getWrTotal()  const { return WrTotal; }
   int            testRdIsRun() const { return RdState != Stopped; }

    int &SpdUpdateFlag()      { return SpNonSp->getUpdateFlag(); }
    int &SpdUpdateShortTime() { return SpNonSp->getUpdateShortTime(); }
    int &SpdUpdateWrite()     { return SpNonSp->getUpdateWrite(); }

    int writePar( FILE *fp )  { return SpNonSp->writePar(fp) | ChanComp.writePar(fp) | GainControl.writePar(fp); } 
    int appendPar( FILE *fp ) ;
    int appendAkuPrePar( FILE *fp ) ;
    void printFeature( SFeature *fp ) const { PreFeature.printFeature( fp ); }
    void updateSpNonSp( SPreFeature *pf, ESpeechFlag ind ) { SpNonSp->update( PreFeature.getSpDetFeature( pf ), ind  ); }
    void updateSpNonSpAuto( SPreFeature *pf, ESpeechFlag ind ) { SpNonSp->update( PreFeature.getSpDetFeature( pf ), ind  ); }
    int updateSpNonSpOneMix( SPreFeature *pf, ESpeechFlag ind, int mixnr ) { return( SpNonSp->updateOneMix( PreFeature.getSpDetFeature( pf ), ind, mixnr ) ); }
  int getSpNonSpNrOfMixSi() { return( SpNonSp->getNrOfMixSi() ) ; } ;
  int getSpNonSpNrOfMixSp() { return( SpNonSp->getNrOfMixSp() ) ; } ;
  int getSpNonSpNrOfMixOv() { return( (SpNonSpOverlap.Par ? ((CSpeechNonSpeechOverlap*)SpNonSp)->getNrOfMixOv() : 0 ) ) ; } ;

  int getSpNonSpUseOverlap() { return( SpNonSpOverlap.Par ) ; } ;


};


#endif

