/******************************************************************************
 * file:        akup_muc.h  (for akupre.cc, TU-Muenchen Vorverarbeitung)
 * info:        class declarations for CAkuPre, acoustic preprocessing 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 *
 * Die Implementierung realisiert ein innerhalb des Projekts VERBMOBIL  
 * entwickeltes Signalanalyseverfahren, da"s an der TU-Muenchen erg"anzt und 
 * neu implementiert wurde.
 *
 * Die Software ist ausschlie"slich f"ur die Verwendung innerhalb des Projektes
 * VERBMOBIL bestimmt. Eine Weitergabe der Programme an Partner au"serhalb des 
 * Projektes VERBMOBIL ist nur mit Zustimmung des Lehrstuhls MMK der TU
 * Muenchen zul"assig. 
 * 
 * Das Projekt VERBMOBIL wird vom Bundesminister f"ur Forschung und Technologie
 * unter dem F"orderkennzeichen 01IV 102K gef"ordert.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

/******************************************************************************

CLASS 'CAkuPre':

Beschreibung siehe: _readme

 ******************************************************************************/


#if !defined( _AKUPRE_H_ )
#define _AKUPRE_H_

#include <stdio.h>

#include "pref_muc.h"

#include "dbg_mess.h"
#include "siganl.h"
//#include "chancomp.h"
#include "paramet.h"

enum ERdFlag    { Undef, Wanted, NotWanted };

// class definition for CAkuPre
class CAkuPreSimple
{
private:
    // readable parameter for akupre
    CParString         DeviceStr;
    CParInt            SampleShift;
    CParFloat          Preemphase;
  
    CParBool           SpNonSpOverlap; // use augmented spnonsp-detector (overlap-state added)

    // object members
    CSigAnalyse        SigAnl;
    CPreFeature        PreFeature;
    // fix parameters
    const unsigned int  FrqSamp;

  int calcPreFeature() ; // calculates new feature vector
  SPreFeature *prefp ; // pointer to current feature vector!

public:
  CAkuPreSimple( FILE *channelfile );
  ~CAkuPreSimple();
  int            startAkuPre( const char *filename, int skipbytes=0 );
  void           stopAkuPre();
  void           clearAkuPre();
  //SFeature      *getNext();
  SPreFeature   *getNextPre();
  
  //int writePar( FILE *fp )  { return SpNonSp->writePar(fp) | ChanComp.writePar(fp) | GainControl.writePar(fp); } 
  int appendPar( FILE *fp ) ;
  int appendAkuPrePar( FILE *fp ) ;
  void printFeature( SFeature *fp ) const { PreFeature.printFeature( fp ); }
  
  float getPreemphase() { return( Preemphase.Par) ; } ;
  int getSampleShift() { return( SampleShift.Par) ; } ;
  
  int skipFrames( int nrframes ) ;

};


#endif

