/******************************************************************************
 * file:        akupre.h (selector for diff. header files)  
 * info:        definition of CAkuPre (Acoustic Preprocessing)  
 * written by:  M.Beham, 12.01.95 
 * last change: M.Beham, 03.03.95
 * 
 ******************************************************************************/
/******************************************************************************
 Beschreibung:
 Der akustische Preprozessor ist in zwei verschiedenen Varianten realisiert 
 worden. F"ur jede Variante mu"s eine spezielle Headerdatei in Programmen, die
 diese Vorverarbeitung verwenden wollen, eingebunden werden (#include). Diese
 Headerdatei bietet die M"oglichkeit, mit Hilfe einer Symboldefinition 
 (#define) die passende Headerdatei auszuwaehlen. Nach der Definition eines 
 entsprechenden Symbols braucht nur noch diese Datei 'akupre.h' eingebunden 
 werden. 

 Symbol:    Header:     Variante der Vorverarbeitung:
  FEAT_MUC  akup_muc.h  Frequenzgruppenzusammenfassung nach der Barkskala zu 
                        20 spektralen Kan"alen, Delta und DeltaDelta f"ur alle
                        Kan"ale und zus"atzlich die Merkmale: Energie, Null-
                        durchgangsrate, Lautheit und modifizierte Lautheit. Die
                        Deltaberechnung erfolgt in einem Zeitfenster "uber -2 
                        und und +2 Frames nach der Rabiner-Formel.
  FEAT_DRS  akup_drs.h  Cepstrale gl"attung und Reduktion des Spektrums durch 
                        eine Mel-skalierte spektrale Faltung zu 30 Kan"alen. 
                        Die Delta und DeltaDelta Berechnung erfolgt durch Diff.
                        von t und t-30ms zu 15 Komponenten und den Merkmalen:
                        Energie, mittlerer spek. Pegel und NDR. (nach der Vor-
                        verarbeitung von Siemens, "uberarbeitet von der TU-
                        Dresden)


 ******************************************************************************/

// include right akup_xxx.h
#if       defined(FEAT_MUC)
#include "akup_muc.h"
#elif     defined(FEAT_DRS)
#include "akup_drs.h"
#else
#error    no match for akup_xxx.h include
#endif

