/******************************************************************************
 * file:        akupre.cc  
 * info:        implementation of CAkuPre (Acoustic Preprocessing)  
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <string.h>
#include <math.h>

// #define DEBUG_LEVEL (3)
// #define AP_TEST 

#include "dbg_mess.h"

// include right akup_xxx.h
#if       defined(FEAT_MUC)
#include "akup_muc_simple.h"
#elif     defined(FEAT_DRS)
#include "akup_drs_simple.h"
#else
#error    no match for akup_xxx.h include
#endif



//--------------------------- public 'CAkuPre' -----------------------------


// constructor: CAkuPre
CAkuPreSimple::CAkuPreSimple( FILE *channelfile )
 :  DeviceStr( "AkuPre.Device", channelfile, "???" ),
    SampleShift( "AkuPre.SampleShift", channelfile, 0 ),
    Preemphase( "AkuPre.Preemphase", channelfile, 0.0 ),
#ifdef FEAT_DRS
    NoisePower( "AkuPre.NoiseEngDB", channelfile, 0.0 ),
    QuantPower( "AkuPre.QuantEngDB", channelfile, 0.0 ),
#endif
    SigAnl( WINDOW_DELTA, WINDOW_LENGTH, FFT_LENGTH, DeviceStr.Par, SampleShift.Par, Preemphase.Par ),
#ifdef FEAT_DRS
    PreFeature( FFT_LENGTH/2, (double)SAMPLE_FRQ, NoisePower.Par, QuantPower.Par ),
#endif
#ifdef FEAT_MUC
    PreFeature( FFT_LENGTH/2, (double)SAMPLE_FRQ ),
#endif
    FrqSamp( SAMPLE_FRQ )
{
  //int i;
    DEBUG_1_MSG("'CAkuPre' constructor called" );
    //fprintf( stderr, "FFT-Len(komplex) %d\n", FFT_LENGTH ) ;
    //fprintf( stderr, "Window-Length    %d\n", WINDOW_LENGTH ) ;
    //fprintf( stderr, "Window-Delta     %d\n", WINDOW_DELTA ) ;
    //fprintf( stderr, "Sample-Freq      %d\n", SAMPLE_FRQ ) ;
	
    // set norm-factor for hamming window
    // double shiftFactor = ldexp( 1.0, SampleShift.Par );
    SigAnl.normWindow( PreFeature.getNormFactor() /* / shiftFactor */ );

    prefp = new SPreFeature ;

}

// destructor: CAkuPre
CAkuPreSimple::~CAkuPreSimple( )
{
    DEBUG_1_MSG("'CAkuPre' destructor called" );
 
    delete prefp ;
}

// start AkuPre
int  CAkuPreSimple::startAkuPre( const char *filename, int skipbytes ) {
    DEBUG_2_MSG_S("'CAkuPre' startAkuPre called with:", filename );

    // start signal analyse 
    IF_ERROR_RETURN( SigAnl.startAnl( filename, skipbytes ) != _OK_, "'CAkuPre' startAkuPre failed" );
    // check sample-frq.
    IF_ERROR_RETURN_S( SigAnl.getFrqSamp() != FrqSamp, "'CAkuPre' startAkuPre: sampling frq. of", filename);

    // skip this nr of frames! (calculate features but don't use them, not very efficient!)
    //for( register i=0 ; i<skipframes ; i++ ) {
      // write one spec to buffer
    //if( calcPreFeature() == _ERR_ )
    //return( _ERR_ )  ;
    //} ;

    return _OK_;
}

// stop AkuPre
void CAkuPreSimple::stopAkuPre() {
    DEBUG_2_MSG("'CAkuPre' stopAkuPre called" );

    // stop signal analyse
    SigAnl.stopAnl();
}

// clear AkuPre
void CAkuPreSimple::clearAkuPre() {
    DEBUG_2_MSG("'CAkuPre' clearAkuPre called" );

    // stop signal analyse
    SigAnl.clearAnl();
  
}

int CAkuPreSimple::skipFrames( int nrframes ) {
  DEBUG_2_MSG("'CAkuPre' skipFrames called" );

  // skip this nr of frames! (calculate features but don't use them, not very efficient!)
  for( register i=0 ; i<nrframes ; i++ ) {
    // write one spec to buffer
    if( calcPreFeature() == _ERR_ )
      return( _ERR_ )  ;
  } ;

  return( _OK_ ) ;

} ;


// get next pre. proc. feature vector
SPreFeature *CAkuPreSimple::getNextPre( ) {
  DEBUG_4_MSG("'CAkuPre' getNextPre called" );
  
  
  // write one spec to buffer
  if( calcPreFeature() == _ERR_ )
    return( NULL )  ;
  else
    return( prefp ) ; // Achtung pointer auf private member!, be careful!
}



//////////////////////////////////////////////////////////////////////////////
//
// appending new cfg-file at the end of the old one!
//       old cfg-file given by filepointer fp
//
/////////////////////////////////////////////////////////////////////////////
int CAkuPreSimple::appendPar( FILE *fp ) {
    int ret = 0 ;
   
    // append fix parameters
    ret |= appendAkuPrePar( fp ) ;
    fprintf( fp, "\n" ) ;

    return( ret == 0 ? ret : -1 ) ;
   
} ; // END appendPar
    
    



//--------------------------- private 'CAkuPre' -------------------------------



// calculate normalized spec, 
int CAkuPreSimple::calcPreFeature()
{
    DEBUG_4_MSG("'CAkuPre' calcPreFeature called" );

    float  zcrate;
    float  entropy;
    float *powspec = SigAnl.getNextSpec( &zcrate );

    if( powspec == 0 )
      return _ERR_ ;

    // calculate log of spez loudnesses of the critical bands (store it to prefp):
    // log is designed to do channel comp, which is not done in the simple version here
    PreFeature.convPowSpec( powspec, zcrate, prefp );

    // therefore: redo log by building exp log of critical bands.
    // then normalize them to the sum over all bands (store it to prefp)
    PreFeature.postPreSpec( prefp );

    return _OK_ ;
}


//////////////////////////////////////////////////////////////////////////////
//
// append params for AkuPre
//
/////////////////////////////////////////////////////////////////////////////
int CAkuPreSimple::appendAkuPrePar( FILE *fp ) {

    int ret = 0 ;
    fprintf( fp, "// 1. AkuPre.  general parameters for preprocessing\n" ) ;
    ret |= DeviceStr.appendPar( fp ) ;
    ret |= SampleShift.appendPar( fp ) ;
    ret |= Preemphase.appendPar( fp ) ;

    return( ret == 0 ? ret : -1 ) ;
} ; // END appendApMainPar



//--------------------------- end of 'CAkuPre' --------------------------------

//--------------------------- test --------------------------------------------

#ifdef AP_TEST

#include <stdlib.h>

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 3, "usage: <chanfile> <samplefile>" );  
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);

    CAkuPre ap( chanfp );
    SFeature  *vf;

    if( ap.startAkuPre( argv[2] ) == _ERR_ ) return(-1);

    for( int i=0; ap.testRdIsRun(); i++) 
    {   if((vf = ap.getNext( )) != 0)
        {   printf( "%d. \n", ap.getRdNum() );
            ap.printFeature( vf );
        }
    }
    ap.clearAkuPre();

    return(0);
}

#endif  // TEST_LEVELcd source
