#include <stdio.h>
#include <string.h>
#include <math.h>

#include "dbg_mess.h"
#include "akupresimple.h"


//#define log


CSigWinBuffer *SigWin1=0 ;             // array of pointers to sigwinbuffer of first signal files
CSigWinBuffer *SigWin2=0 ;             // array of pointers to sigwinbuffer of second signal files
float *Correl ;                        // buffer to store crosscorrel (cc) info
int *CorrelShift ;                     // buffer to store crosscorrel info
int CrossCorrelWinLen ;                // WindowLen for Crosscorrelations (in samples)
int MaxShift ;                         // maximum shift between windows to calc crosscorrelations (in samples)
int WindowShift ;                      // shift between neighboring windows (in samples)
int MaxFileLen = 1000000 ;             // longest treatable file (nr of frames!) ;
float CrossCorrelMin ;                 // use only cc-shifts correponding to cc-values greater than this for histo

float CalcCrossCorrelation( float *window1, float *window2, int &shiftindex ) {

  float crosscorrel = 0.0 ;
  float autocorrel1 = 0.0 ;
  float autocorrel2 = 0.0 ;

  float crosscorrelmax = -1.0 ;
  int shiftindexformax ;

  for( int shift=-MaxShift ; shift<MaxShift ; shift++ ) {
    
    crosscorrel = 0.0 ;
    autocorrel1 = 0.0 ;
    autocorrel2 = 0.0 ;

    for( int i=MaxShift ; i<(CrossCorrelWinLen)-MaxShift ; i++ ) {
      crosscorrel += (window1[i]*window2[i-shift]) ;
      autocorrel1 += window1[i]*window1[i] ;
      autocorrel2 += window2[i-shift]*window2[i-shift] ;
    } ;
    
    crosscorrel /= sqrt( autocorrel1*autocorrel2) ;

    crosscorrel = abs(crosscorrel);
   
    if( crosscorrel > crosscorrelmax ) {
      crosscorrelmax = crosscorrel ;
      shiftindexformax = shift ;
    }
  }

  shiftindex = shiftindexformax;

  return( crosscorrelmax ) ; 
} ;

int main( int argc, char *argv[] ) {

  // check params:
  if( argc < 6 ) {
    fprintf( stderr, "%s <chan1.wav> <chan2.wav> <winlen> <windowshift> <maxcorrelshift> [crosscorrelmin]\n", argv[0] ) ;
    exit(-1) ;
  };

  CrossCorrelWinLen = strtol( argv[3], NULL, 10 ) ;
  WindowShift = strtol( argv[4], NULL, 10 ) ;
  MaxShift = strtol( argv[5], NULL, 10 ) ;

  if( argc == 7 ) 
    CrossCorrelMin = strtod( argv[6], NULL ) ;
  else
    CrossCorrelMin = 0.0 ;

  Correl = new float[1000000];
  CorrelShift = new int[1000000];


  // initialize sigwinbuffers: (10ms = 160 samples shift, sampleshift=0, preemphase=0.0 )
  SigWin1 = new CSigWinBuffer( WindowShift, CrossCorrelWinLen, "FileNIST", 0, 0.0 ) ;
  SigWin2 = new CSigWinBuffer( WindowShift, CrossCorrelWinLen, "FileNIST", 0, 0.0 ) ;
 
  SigWin1->startBuf( argv[1] ) ;
  SigWin2->startBuf( argv[2] ) ;

  float *win1 = new float[CrossCorrelWinLen] ;
  float *win2 = new float[CrossCorrelWinLen] ;

  int histolen = 2*MaxShift+1 ;
  int *histo = new int[histolen] ;
  for( int i=0 ; i<histolen ; i ++ )
    histo[i]=0;

  //int frameshift = (int)(((float)((crosscorrelwinlen/2)-(WINDOW_DELTA/2)) / (float)WINDOW_DELTA)+0.5) ;


  int count=0 ;
  while( SigWin1->getWin( win1 ) == _OK_ && SigWin2->getWin( win2 ) == _OK_ ) {
    
    CorrelShift[count] = -100000;

    Correl[count] = CalcCrossCorrelation( win1, win2, CorrelShift[count] ) ;

    if( Correl[count] > CrossCorrelMin )
      histo[CorrelShift[count]+MaxShift]++;

#ifdef log
      fprintf( stderr, "%d %f %d\n", count, Correl[count], CorrelShift[count] ) ;
#endif

    count++ ;
  } ;

  for( int i=0 ; i<histolen ; i ++ )
    fprintf( stdout, "%d\t%d\n", i-MaxShift, histo[i] ) ;
  
  
  return(1) ;

} ;
