#include <stdio.h>
#include <string.h>
#include <math.h>

#include "dbg_mess.h"
#include "akupresimple.h"

#include <math.h>

#include "sigwin.h"

//#define Fast 0

//#define log
//#define TESTLEVEL

CSigWinBuffer **SigWinPzm=0 ;             // array of pointers to sigwinbuffers of pzm signals
CSigWinBuffer *SigWin=0 ;                    // pointers to sigwinbuffer of first signal file
float *MaxCorrel ;                     // buffer to store crosscorrel (cc) info (max correl to any pzm signal for each time)
int *MaxCorrelShift ;                     // buffer to store crosscorrel info
int **Shift ;
int *DeltaShift ;                       // difference between shift of surrent and previous max!
int *PzmNr ;                           // which pzm had highest crosscorrel!
int CrossCorrelWinLen ;                // WindowLen for Crosscorrelations (in samples)
int MaxShift ;                         // maximum shift between windows to calc crosscorrelations (in samples)
int WindowShift ;                      // shift between neighboring windows (in samples)
int MaxFileLen = 1000000 ;             // longest treatable file (nr of frames!) ;
float CrossCorrelMin ;                 // use only cc-shifts correponding to cc-values greater than this for histo
int NrPzm ;
int MaxNrFrame ;
int NrPad ;
char *outpath=0 ;


float CalcCrossCorrelation( float *window1, float *window2, int &shiftindex ) {

  int stepsize = 1 ; // define bigger stepsize when faster correl is wanted: 
                     // CAVE: loss in accuracy if stepsize > 1

  float crosscorrel = 0.0 ;
  float autocorrel1 = 0.0 ;
  float autocorrel2 = 0.0 ;

  float crosscorrelmax = -1.0 ;
  int shiftindexformax ;

  for( int shift=-MaxShift ; shift<MaxShift ; shift++ ) {
    
    crosscorrel = 0.0 ;
    autocorrel1 = 0.0 ;
    autocorrel2 = 0.0 ;

    for( int i=MaxShift ; i<(CrossCorrelWinLen)-MaxShift ; i+=stepsize ) {
      crosscorrel += (window1[i]*window2[i-shift]) ;
      autocorrel1 += window1[i]*window1[i] ;
      autocorrel2 += window2[i-shift]*window2[i-shift] ;
    } ;
    
    crosscorrel /= sqrt( autocorrel1*autocorrel2) ;

    crosscorrel = abs(crosscorrel);
   
    if( crosscorrel > crosscorrelmax ) {
      crosscorrelmax = crosscorrel ;
      shiftindexformax = shift ;
    }
  }

  shiftindex = shiftindexformax;

  return( crosscorrelmax ) ; 
} ;


float CalcCrossCorrelationApprox( float *window1, float *window2, int &shiftindex ) {

  int stepsize = 4 ; // define bigger stepsize when faster correl is wanted: 
                     // CAVE: loss in accuracy if stepsize > 1
  int shiftstep = 2 ;// define bigger stepsize when faster correl is wanted: 
                     // CAVE: loss in accuracy if stepsize > 1

  float crosscorrel = 0.0 ;
  float autocorrel1 = 0.0 ;
  float autocorrel2 = 0.0 ;

  float crosscorrelmax = -1.0 ;
  int shiftindexformax ;

  // calc autocorrel of complete window (this is an estimate of the autocorrel for the 
  // window used for calculating the crosscorrel!
  autocorrel1 = 0.0 ;
  autocorrel2 = 0.0 ;
  for( register i=MaxShift ;  i<CrossCorrelWinLen-MaxShift ; i+=stepsize ) {
    autocorrel1 += window1[i]*window1[i] ;
    autocorrel2 += window2[i]*window2[i] ;
  }    
  float norm = sqrt(autocorrel1 * autocorrel2) ;

  for( int shift=-MaxShift ; shift<MaxShift ; shift+=shiftstep ) {
    crosscorrel = 0.0 ;
    for( int i=MaxShift ; i<(CrossCorrelWinLen)-MaxShift ; i+=stepsize ) {
      crosscorrel += (window1[i]*window2[i-shift]) ;
    } ;
    
    crosscorrel /= norm ;
    crosscorrel = abs(crosscorrel);
   
    if( crosscorrel > crosscorrelmax ) {
      crosscorrelmax = crosscorrel ;
      shiftindexformax = shift ;
    }
  }

  shiftindex = shiftindexformax;

  return( crosscorrelmax ) ; 
} ;

int getChanId( char* chanfilename, char* chanid ) {

  // get ChanIds 
  char *tmpchar = new char[strlen(chanfilename)+1] ; ;
  char *tmpchar2;
  strcpy( tmpchar, chanfilename ) ;

  // parse filenames of wav-files
  tmpchar2 = strrchr( tmpchar, '.' ) ;
  tmpchar2--;
  strncpy( chanid, tmpchar2, 1 ) ;
  chanid[1] = '\0'  ;

  delete[] tmpchar;
  return(1);
  
} ;

 



int main( int argc, char *argv[] ) {

  int i,j;
  int sent=0;
  int N = 5 ; // use N frames to the left and right to calc average deltashift!

  // check params:
  if( argc < 6 ) {
    fprintf( stderr, "%s <chan1.wav> <outpath> <winlen> <windowshift> <maxcorrelshift> <crosscorrelmin> <pzm1.wav> ... <pzmN.wav> <N>\n", argv[0] ) ;
    exit(-1) ;
  };

  CrossCorrelWinLen = strtol( argv[3], NULL, 10 ) ;
  WindowShift = strtol( argv[4], NULL, 10 ) ;
  MaxShift = strtol( argv[5], NULL, 10 ) ;
  CrossCorrelMin = strtod( argv[6], NULL ) ;
  outpath = argv[2]; 
  NrPzm = strtol( argv[argc-1], NULL, 10 ) ;

  NrPad = CrossCorrelWinLen/WindowShift ; // pad that much frame at beg and end of pfile!!!
  //NrPad = CrossCorrelWinLen/WindowShift ; // pad that much frame at beg and end of pfile!!!

  MaxNrFrame = 1000000 ;
  MaxCorrel = new float[MaxNrFrame];
  MaxCorrelShift = new int[MaxNrFrame];
  PzmNr = new int[MaxNrFrame];
  Shift = new int*[NrPzm] ; // array for storing shift info for all pzms
  for( i=0 ; i<NrPzm ; i++ )
    Shift[i] = new int[MaxNrFrame] ;
  DeltaShift = new int[MaxNrFrame];

  // get channel-ids from filenames
  char **ChanId = new char*[NrPzm+1] ;
  for( i=0 ; i<NrPzm+1 ; i++ )
    ChanId[i] = new char[2] ;

  getChanId( argv[1], ChanId[0] ) ;
  for( i=0 ; i<NrPzm ; i++ )
    getChanId( argv[argc-NrPzm-1+i], ChanId[i+1] ) ;

  int MaxByteSkip = 0 ;
  //int MaxByteSkip = 15 * 128 ;
  int* ByteSkip = new int[NrPzm+1];
  for( i=0 ; i<NrPzm+1 ; i++ ) {
    switch( ChanId[i][0] ) {
    case '0':
    case '1':
    case '2': 
    case '3': 
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      ByteSkip[i] = (15-strtol(ChanId[i],NULL,10))*128 ;
#ifdef TESTLEVEL
      fprintf( stderr, "chan: %d, Id: %s, skew: %d\n", i, ChanId[i], ByteSkip[i] ) ;
#endif
      if( ByteSkip[i] > MaxByteSkip )
	MaxByteSkip = ByteSkip[i] ;
      break;
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
      ByteSkip[i] = ('F'-ChanId[i][0])*128 ;
#ifdef TESTLEVEL
      fprintf( stderr, "chan: %d, Id: %s, skew: %d\n", i, ChanId[i], ByteSkip[i] ) ;
#endif
      if( ByteSkip[i] > MaxByteSkip )
	MaxByteSkip = ByteSkip[i] ;
      break;
    } ;
  }  

  FILE *outfp ;
  char *tmpchar = new char[strlen(argv[1])+1];
  char *outfilename = new char[strlen(outpath)+strlen(argv[1])+10] ;
  strcpy( tmpchar, argv[1] ) ;
  sprintf( outfilename, "%s/%s.correl\0", outpath, strtok( tmpchar, "." ) ) ;
  if( ( outfp = fopen( outfilename, "w" )) == 0 ) {
    fprintf( stderr, "Can't open outfile '%s'\n", outfilename ) ;
    exit(-1) ;
  } ;

  // initialize sigwinbuffers: (10ms = 160 samples shift, sampleshift=0, preemphase=0.0 )
  
  SigWin = new CSigWinBuffer( WindowShift, CrossCorrelWinLen, "FileNIST", 0, 0.0 ) ;
  SigWinPzm = new CSigWinBuffer*[NrPzm] ;
  for( i=0 ; i<NrPzm ; i++ ) {
    SigWinPzm[i] = new CSigWinBuffer( WindowShift, CrossCorrelWinLen, "FileNIST", 0, 0.0 ) ;
    SigWinPzm[i]->startBuf( argv[argc - (NrPzm+1) + i ], ByteSkip[i+1] ) ;
  } ;
 
  SigWin->startBuf( argv[1], ByteSkip[0] ) ;
  //SigWin2->startBuf( argv[2] ) ;

  // pointer to data window:
  float *win = new float[CrossCorrelWinLen] ;
  float **winpzm = new float*[NrPzm] ;
  for( i=0 ; i<NrPzm ; i++ ) {
    winpzm[i] = new float[CrossCorrelWinLen] ;
  }

  int histolen = 2*MaxShift+1 ;
  int *histo = new int[histolen] ;
  for( int i=0 ; i<histolen ; i ++ )
    histo[i]=0;

  //int frameshift = (int)(((float)((crosscorrelwinlen/2)-(WINDOW_DELTA/2)) / (float)WINDOW_DELTA)+0.5) ;
  int count=0 ;
  float correltmp, correltmp2 ;
  int correlshifttmp, tmpshift ;
  while( SigWin->getWin( win ) == _OK_  ) {

    // check if all pzm files did not yet see eof
    int eof = 0 ;
    for( i=0 ; i<NrPzm ; i++ ) {
      if( SigWinPzm[i]->getWin( winpzm[i] ) != _OK_ ) {
	eof =1 ;
	break ;
      } ;
    } 
    // leave loop when at least one file faces eof
    if( eof == 1 )
      break ;
	
    MaxCorrelShift[count] = -100000;
    MaxCorrel[count] = -100000;


    for( i=0 ; i<NrPzm ; i++ ) {
      if( ( correltmp = abs( CalcCrossCorrelationApprox( win, winpzm[i], Shift[i][count] ) ) ) >= MaxCorrel[count] ) {
	MaxCorrel[count] = correltmp ;
	MaxCorrelShift[count] = Shift[i][count] ;
	PzmNr[count] = i ;
      } 
      //correltmp2 = CalcCrossCorrelationApprox( win, winpzm[i], tmpshift ) ;
      //fprintf( stdout, "[%d] Correct: %f (%d), Aprrox: %f (%d), DIFF: %f\n", count, correltmp, Shift[i][count], correltmp2, tmpshift, correltmp-correltmp2 ) ;
    }

	
    if( MaxCorrel[count] > CrossCorrelMin )
      histo[MaxCorrelShift[count]+MaxShift]++;

    DeltaShift[count] = ( count > 0 ? (Shift[PzmNr[count]][count]-Shift[PzmNr[count]][count-1]) : 0 ) ;

    //#ifdef log
    //fprintf( stderr, "%d %d correl: %f d-shift: %d (maxshift: %d, Pzm-Nr: %d)\n", sent, count, MaxCorrel[count], DeltaShift[count], MaxCorrelShift[count], PzmNr[count] ) ;
    //#endif

    count++ ;
  } ;

  // print pfile-header if needed:
  if( 1 ) {
    fprintf( outfp, "-pfile_header version 0 size 32768\n" ) ;
    fprintf( outfp, "-num_sentences 1\n" ) ;
    fprintf( outfp, "-num_frames %d\n", count+NrPad-1+((MaxByteSkip-1)/640)*2 ) ;
    fprintf( outfp, "-first_feature_column %d\n", 2 ) ;
    fprintf( outfp, "-num_features %d\n", 2 ) ;
    fprintf( outfp, "-first_label_column %d\n", 4 ) ;
    fprintf( outfp, "-num_labels %d\n", 0 ) ;
    fprintf( outfp, "-format %s\n", "2dff" ) ;
    fprintf( outfp, "-data size %d offset %d ndim %d nrow %d ncol %d\n",  (count+NrPad-1+((MaxByteSkip-1)/640)*2)*(2+2), 0, 2, (count+NrPad-1+((MaxByteSkip-1)/640)*2), 2+2  ) ;
    fseek( outfp, 32768, 0 ) ;
  } ;

  float deltashift ;

  deltashift = 0.0 ; 
  for( j=0 ; j<=N ; j++ ) 
    deltashift += abs(DeltaShift[j]) ;
  deltashift /= (N+1) ;

  // Padding at beg of file: use NrPad/2 frames
  for( i=0 ; i<NrPad/2 ; i++ ) {
    fwrite( &sent,  sizeof(int), 1, outfp ) ;
    fwrite( &i, sizeof(int), 1, outfp ) ;
    fwrite( &(MaxCorrel[0]), sizeof(float), 1, outfp ) ;
    fwrite( &(deltashift), sizeof(float), 1, outfp ) ;
#ifdef log
    fprintf( stderr, "%d %d %f %f\n", sent, i, MaxCorrel[0], deltashift ) ;
#endif

  } ;


  // process regular frames
  int frame ;
  for( i=0 ; i<count ; i++ ) {
    if( 1  ) {
      frame = i+NrPad/2 ;
      fwrite( &sent,  sizeof(int), 1, outfp ) ;
      fwrite( &frame, sizeof(int), 1, outfp ) ;
      fwrite( &(MaxCorrel[i]), sizeof(float), 1, outfp ) ;
      
      // use averagedeltashift over n (odd) frames:
      deltashift = 0.0 ;
      for( j=i-N ; j<=i+N ; j++ ) {
	if( j<0 )
	  deltashift += abs(DeltaShift[0]) ;
	else if( j>= count )
	  deltashift += abs(DeltaShift[count-1]);
	else
	  deltashift += abs(DeltaShift[j]);
      }
      deltashift /= (2*N+1) ;
      //deltashift = DeltaShift[i] ;
      fwrite( &(deltashift), sizeof(float), 1, outfp ) ;
      //fwrite( &(DeltaShift[i]), sizeof(int), 1, outfp ) ;
#ifdef log
      fprintf( stderr, "%d %d %f %f\n", sent, frame, MaxCorrel[i], deltashift ) ;
#endif
    } ;
  } ;

  deltashift = 0.0 ; 
  for( j=count-N-1 ; j<=count-1 ; j++ ) 
    deltashift += abs(DeltaShift[j]) ;
  deltashift /= (N+1) ;


  // Padding at end!!!
  for( i=0 ; i<(NrPad-(NrPad/2)-1)+((MaxByteSkip-1)/640)*2 ; i++ ) {
    frame = NrPad/2 + count + i ;
    fwrite( &sent,  sizeof(int), 1, outfp ) ;
    fwrite( &frame, sizeof(int), 1, outfp ) ;
    fwrite( &(MaxCorrel[count-1]), sizeof(float), 1, outfp ) ;
    fwrite( &(deltashift), sizeof(float), 1, outfp ) ;
#ifdef log
    fprintf( stderr, "%d %d %f %f\n", sent, frame, MaxCorrel[count-1], deltashift ) ;
#endif

  } ;


  //  for( int i=0 ; i<histolen ; i ++ )
  //fprintf( stdout, "%d\t%d\n", i-MaxShift, histo[i] ) ;
  
  
  return(1) ;

  delete SigWin ;
  for( i=0 ; i< NrPzm ; i++ ) {
    delete SigWinPzm[i] ;
    delete[] Shift[i] ;
    delete[] winpzm[i] ;
  }
  delete[] SigWinPzm ;
  delete[] Shift ;
  delete[] winpzm ;


  delete[] win  ;

  delete[] MaxCorrel ;
  delete[] MaxCorrelShift ;
  delete[] PzmNr ;

  for( i=0 ; i<NrPzm+1 ; i++ )
    delete[] ChanId[i] ;
  delete[] ChanId ;

  delete[] ByteSkip;


  delete[] tmpchar;
  delete[] outfilename ;

  delete[] histo;

  fclose( outfp ) ;

} ;
