/******************************************************************************
 * file:        chancomp.cc  
 * info:        implementation of CChanComp  
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "chancomp.h"



//--------------------------- public 'CChanComp' ------------------------------


// constructor: CChanComp
CChanComp::CChanComp( unsigned int speclen, FILE *file, double dt )
:   NrChannels( speclen ),
    TrainFlag(   "ChanComp.TrainLongTime",  file, FALSE ),
    UpdateFlag(  "ChanComp.UpdateShortTime", file, FALSE ),
    UpdateWrite( "ChanComp.UpdateWrite",   file, FALSE ),
    UpdateUse(   "ChanComp.UpdateUse",     ESPEECHFLAG, file, S ),
    UpdateTau(   "ChanComp.UpdateTau",     file, 1.0   ),
    bFactor( exp( -dt / UpdateTau.Par ) ),
    aFactor( 1.0 - bFactor ),
    cFactor( aFactor * bFactor ),
    CurMean(     "ChanComp.Current.Mean", speclen, file, 0.0 ),
    CurVar(      "ChanComp.Current.Var",  speclen, file, 0.0 ),
    TrainMean(   "ChanComp.Train.Mean", speclen, file, 0.0 ),
    TrainVar(    "ChanComp.Train.Var",  speclen, file, 0.0 ),
    NrVectors( 0 ),  
    ChanFile( file )
{
    DEBUG_1_MSG("'CChanComp' constructor called" );
}

// destructor: CChanComp
CChanComp::~CChanComp( )
{
    DEBUG_1_MSG("'CChanComp' destructor called" );

    // write meanfile
    if( UpdateWrite.Par )
        writePar( ChanFile );
}

// write parameter
int CChanComp::writePar( FILE *fp )
{
    if( CurMean.writePar(fp) == _ERR_ || 
        CurVar.writePar(fp)  == _ERR_ || 
        TrainMean.writePar(fp) == _ERR_ || 
        TrainVar.writePar(fp)  == _ERR_   )
        return _ERR_;
    else
        return _OK_;
}


// append updatable parameters
int CChanComp::appendUpdPar( FILE *fp )
{
    if( CurMean.appendPar(fp) == _ERR_ || 
        CurVar.appendPar(fp)  == _ERR_ || 
        TrainMean.appendPar(fp) == _ERR_ || 
        TrainVar.appendPar(fp)  == _ERR_   )
        return _ERR_;
    else
        return _OK_;
}

// append fix parameters
int CChanComp::appendFixPar( FILE *fp )
{
    if( TrainFlag.appendPar(fp) == _ERR_ || 
        UpdateFlag.appendPar(fp)  == _ERR_ || 
        UpdateWrite.appendPar(fp) == _ERR_ || 
        UpdateUse.appendPar(fp)  == _ERR_ ||
	UpdateTau.appendPar(fp)  == _ERR_   )
        return _ERR_;
    else
        return _OK_;
}


// estimate channel and sub mean channel
void  CChanComp::subMean( float *xvec, ESpeechFlag flag )
{
    DEBUG_4_MSG("'CChanComp' subMean called" );
    register int i;
    register float alpha, mean;

    // update current mean and var
    if( UpdateFlag.Par && flag >= UpdateUse.Par )
    {   for( i=0; i<NrChannels; i++ )
        {   CurVar.Vec[i]  = bFactor * CurVar.Vec[i]  + 
               cFactor * (xvec[i]-CurMean.Vec[i]) * (xvec[i]-CurMean.Vec[i]);
            CurMean.Vec[i] = bFactor * CurMean.Vec[i] + aFactor * xvec[i];
        }
    }

    // train longtime mean and var
    if( TrainFlag.Par && flag >= UpdateUse.Par )
    {   // long time update
        alpha = 1.0/(float)(++NrVectors);
        float beta  = 1.0 - alpha;
        float gamma = alpha * beta;
        for( i=0; i<NrChannels; i++ )
        {   TrainVar.Vec[i]  = beta * TrainVar.Vec[i]  + 
                   gamma * (xvec[i]-TrainMean.Vec[i]) * (xvec[i]-TrainMean.Vec[i]);
            TrainMean.Vec[i] = beta * TrainMean.Vec[i] + alpha * xvec[i];
        }
    }

    // calc mean for sub and subtract
    for( i=0; i<NrChannels; i++ )
    {   alpha   = aFactor * ( CurVar.Vec[i] / TrainVar.Vec[i] );
        mean    = ( alpha * TrainMean.Vec[i] + CurMean.Vec[i] ) / (1.0 + alpha);
        xvec[i] -= mean;
    }
    
}


//--------------------------- end of 'CChanComp' ------------------------------

//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL


#endif  // TEST_LEVEL
