/******************************************************************************
 * file:        cctrain.cc  (main)
 * info:        training of channel compensation means and vars 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>

#include "dbg_mess.h"
#include "akup_muc_db.h"

#include "fileconst.h"

//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
    // open files
    if( argc != 3) 
    {   fprintf(stderr,"usage: chctrain <channel-file> <catalog-file>\n");
        return _ERR_;
    }
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);
    FILE *catfp = fopen( argv[2], "r" );
    IF_ERROR_RETURN_S( catfp == 0, "can't open catalog-file:", argv[2] );
    FILE *fftfp ;
    FILE *zcrfp ;
    int j ;
    float zcrate ;
    float *fftptr = new float[FFT_LENGTH/2] ;
    // construct
    CAkuPreDB    AkuPre( chanfp );

    // loop over catalog
    char  str[MaxLineLength], str2[MaxLineLength], str3[MaxLineLength], *fftfilename, *zcrfilename;
    while( fgets( str, MaxLineLength, catfp) != 0 ) {   
	
	// read name of fft-file
	if( (fftfilename = strtok(str, " \t")) == 0) {
	    fprintf( stderr, "Sorry, can't get fftfilename from catalog-file '%s'\n", argv[2] ) ;
	    continue;
	} ;
	// read name of zcrate-file
	if( (zcrfilename = strtok( NULL, " \n")) == 0 ) {
	    fprintf( stderr, "Sorry, can't get zcrfilename from catalog-file '%s'\n", argv[2] ) ;
	    continue;
	} ;

	// open fftfile
	if( (fftfp = fopen( fftfilename, "r" )) == 0 ) {
	    fprintf( stderr, "Sorry, can't open fft-file '%s'\n", fftfilename ) ;
	    continue ;
	} ;

	// open zcratefile
	if( (zcrfp = fopen( zcrfilename, "r" )) == 0 ) {
	    fprintf( stderr, "Sorry, can't open zcrate-file '%s'\n", zcrfilename ) ;
	    fclose( fftfp ) ;
	    continue ;
	} ;
	    

	fprintf( stderr, "Now processing files '%s' and '%s'\n", fftfilename, zcrfilename ) ;

	// Schleife ueber die frames des fft-files
	while( fgets( str2, MaxLineLength, fftfp ) != 0 ) {
	    
	    // read fft
	    fftptr[0] = (float)strtod( strtok( str2, " " ), NULL ) ;
	    for( j=1 ; j<FFT_LENGTH/2-1 ; j++ )
		fftptr[j] = (float)strtod( strtok( NULL, " " ), NULL ) ;
	    fftptr[FFT_LENGTH/2-1] = (float)strtod( strtok( NULL, "\n" ), NULL ) ;
	    
	    // read zcrate
	    fgets( str3, MaxLineLength, zcrfp ) ;
	    zcrate = (float)strtod( strtok( str3, " \n" ), NULL ) ;
    
	    // give fft and zcrate to AkuPre
	    AkuPre.calcVec( fftptr, zcrate ) ;

	} ;

	fclose( fftfp ) ;
	fclose( zcrfp ) ;

    } ;

    // write params
    FILE *testfp = fopen( "ap_chctrained.cfg", "w" ) ;
    AkuPre.appendPar( testfp ) ;
    fclose( testfp ) ;
    

    if( chanfp != 0 )
	fclose( chanfp ) ;
    if( catfp != 0 ) 
	fclose( catfp ) ;

    delete[] fftptr ;


    return(0) ;
}
