
/******************************************************************************
 * file:        dbg_mess.h  
 * info:        macros for debuging and error messages 
 * written by:  M.Beham, 12.01.95 (org. from fft.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#ifndef _DBG_MESS_H_
#define  _DBG_MESS_H_

#include <stdlib.h>
#include <stdio.h>


// only message (msg) output on stderr

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 1 || DEBUG_LEVEL == -1 )
#define DEBUG_1_MSG(msg)  fprintf( stderr, ">     %-12s %3d, %s\n", __FILE__, __LINE__, msg )
#else
#define DEBUG_1_MSG(msg) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 2 || DEBUG_LEVEL == -2 )
#define DEBUG_2_MSG(msg)  fprintf( stderr, ">>    %-12s %3d, %s\n", __FILE__, __LINE__, msg )
#else
#define DEBUG_2_MSG(msg) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 3 || DEBUG_LEVEL == -3 )
#define DEBUG_3_MSG(msg)  fprintf( stderr, ">>>   %-12s %3d, %s\n", __FILE__, __LINE__, msg )
#else
#define DEBUG_3_MSG(msg) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 4 || DEBUG_LEVEL == -4 )
#define DEBUG_4_MSG(msg)  fprintf( stderr, ">>>>  %-12s %3d, %s\n", __FILE__, __LINE__, msg )
#else
#define DEBUG_4_MSG(msg) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 5 || DEBUG_LEVEL == -5 )
#define DEBUG_5_MSG(msg)  fprintf( stderr, ">>>>> %-12s %3d, %s\n", __FILE__, __LINE__, msg )
#else
#define DEBUG_5_MSG(msg) 
#endif

// message (msg) and string (str) output on stderr

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 1 || DEBUG_LEVEL == -1 )
#define DEBUG_1_MSG_S(msg,str)  fprintf( stderr, ">     %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )
#else
#define DEBUG_1_MSG_S(msg,str) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 2 || DEBUG_LEVEL == -2 )
#define DEBUG_2_MSG_S(msg,str)  fprintf( stderr, ">>    %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )
#else
#define DEBUG_2_MSG_S(msg,str) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 3 || DEBUG_LEVEL == -3 )
#define DEBUG_3_MSG_S(msg,str)  fprintf( stderr, ">>>   %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )
#else
#define DEBUG_3_MSG_S(msg,str) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 4 || DEBUG_LEVEL == -4 )
#define DEBUG_4_MSG_S(msg,str)  fprintf( stderr, ">>>>  %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )
#else
#define DEBUG_4_MSG_S(msg,str) 
#endif

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 5 || DEBUG_LEVEL == -5 )
#define DEBUG_5_MSG_S(msg,str)  fprintf( stderr, ">>>>> %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )
#else
#define DEBUG_5_MSG_S(msg,str) 
#endif


// ERROR values and masseges

#define  _ERR_   -1
#define  _OK_    0
#define  FALSE   0
#define  TRUE    1
#define  False   0
#define  True    1

#define WARNING_MSG(msg)  fprintf( stderr, "!Warn %-12s %3d, %s\n", __FILE__, __LINE__, msg ) 

#define ERROR_MSG(msg)    fprintf( stderr, "ERROR %-12s %3d, %s\n", __FILE__, __LINE__, msg )

#define FATAL_MSG(msg)    fprintf( stderr, "FATAL %-12s %3d, %s\n", __FILE__, __LINE__, msg )

#define WARNING_PERR(msg) { fprintf( stderr, "!Warn %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(""); } 

#define ERROR_PERR(msg)   { fprintf( stderr, "ERROR %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(""); }

#define FATAL_PERR(msg)   { fprintf( stderr, "FATAL %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(""); }

#define WARNING_MSG_S(msg,str)  fprintf( stderr, "!Warn %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )

#define ERROR_MSG_S(msg,str)    fprintf( stderr, "ERROR %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )

#define FATAL_MSG_S(msg,str)    fprintf( stderr, "FATAL %-12s %3d, %s %s\n", __FILE__, __LINE__, msg, str )

#define WARNING_PERR_S(msg,str) { fprintf( stderr, "!Warn %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(str); } 

#define ERROR_PERR_S(msg,str)   { fprintf( stderr, "ERROR %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(str); }

#define FATAL_PERR_S(msg,str)   { fprintf( stderr, "FATAL %-12s %3d, %s; ", __FILE__, __LINE__, msg ); perror(str); }



#define IF_FATAL_EXIT(cond, msg)  if( cond ) { FATAL_MSG(msg); exit(EXIT_FAILURE); }

#define IF_ERROR_RETURN(cond,msg) if( cond ) { ERROR_MSG(msg); return(_ERR_); }

#define IF_WARNING(cond,msg)      if( cond ) { WARNING_MSG(msg); }

#define IF_FATAL_EXIT_S(cond, msg, str)  if( cond ) { FATAL_MSG_S(msg,str); exit(EXIT_FAILURE); }

#define IF_ERROR_RETURN_S(cond,msg, str) if( cond ) { ERROR_MSG_S(msg,str); return(_ERR_); }

#define IF_WARNING_S(cond,msg, str)      if( cond ) { WARNING_MSG_S(msg,str); }


#define IF_FATAL_EXIT_PERR(cond, msg)  if( cond ) { FATAL_PERR(msg); exit(EXIT_FAILURE); }

#define IF_ERROR_RETURN_PERR(cond,msg) if( cond ) { ERROR_PERR(msg); return(_ERR_); }

#define IF_WARNING_PERR(cond,msg)      if( cond ) { WARNING_PERR(msg); }

#define IF_FATAL_EXIT_PERR_S(cond, msg, str)  if( cond ) { FATAL_PERR_S(msg,str); exit(EXIT_FAILURE); }

#define IF_ERROR_RETURN_PERR_S(cond,msg, str) if( cond ) { ERROR_PERR_S(msg,str); return(_ERR_); }

#define IF_WARNING_PERR_S(cond,msg, str)      if( cond ) { WARNING_PERR_S(msg,str); }


// complex WARNING and ERROR (no \n)

#define WARNING_P(msg) fprintf( stderr, "!Warn %-12s %3d, %s ", __FILE__, __LINE__, msg ) 

#define ERROR_P(msg)   fprintf( stderr, "ERROR %-12s %3d, %s ", __FILE__, __LINE__, msg )


#endif   // _DBG_MESS_H_


