#include <stdio.h>
#include <string.h>

#include "p_histo.h"

#include "paramet.h"

#include "dbg_mess.h"
#include "akupresimple.h"



//#define   MAXCATLINELEN  1024
//#define   MaxRuns 20000


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] ) {

  // open files
  int i, j, count, wavindex, labindex ;
  if( argc < 3) {   
    fprintf(stderr, "Usage: %s <cfg-file> <wavfile1> <wavfile2> ... <wavfileN>\n");
    fprintf(stderr, "Need at least one wavfile to process!\n" ) ;
    return _ERR_;
  }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open cfg-file:", argv[1]);
 
  int nrchan = (argc-2) ;
  CAkuPreSimple    **AkuPre = new CAkuPreSimple*[nrchan] ;
  Chisto **EnergyHisto = new Chisto*[nrchan] ;

  // initialize all channels
  for( i=0 ; i<nrchan; i++ ) {
    AkuPre[i] = new CAkuPreSimple( chanfp ) ;
    EnergyHisto[i] = new Chisto( 0.0, 90.0 , 90 ) ;
    rewind( chanfp ) ;
  } ;

  // start AkuPre of each channel
  for( i=0 ; i<nrchan ; i++ )
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[i+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open%d-th  sigfile '%s'\n", i, argv[i+2] ) ;
      exit(-1) ;
    } ;
      
  // loop over all frames and collect energy info
  SPreFeature **prefp = new SPreFeature*[nrchan] ;	  
  count = 0 ;  // counts frames now!
  
  float *energymin = new float[nrchan] ;
  for( int i=0 ; i<nrchan ; i++ ) 
    energymin[i] = 1.0e30 ;
  
  // determine speech energy min in each channel at the same time!
  while(1) { 
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 1 ;
     
    for( i=0 ; i<nrchan ; i++ ) {
      if( (prefp[i] = AkuPre[i]->getNextPre()) == 0 ) {
	stillrunning = 0 ;
	break ;
      } ;
      
      EnergyHisto[i]->update( prefp[i]->TotalEnergy ) ;

      //if( prefp[i]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( prefp[i]->TotalEnergy < energymin[i] )
	energymin[i] =  prefp[i]->TotalEnergy ;
    } ;
    if( stillrunning == 0 )
      break ;
    
    count++ ;
  } ;
  

  // write histos
  char *tmp = new char[255] ;
  for( i=0 ; i<nrchan ; i++ ) {
    
    sprintf( tmp, "channel%d.histo", i ) ;
    EnergyHisto[i]->write( tmp ) ; 

  } ;

  // free memory
  for( i=0 ; i<nrchan ; i++ ) {
    delete AkuPre[i] ;
    delete EnergyHisto[i] ;
  } ;

  delete[] prefp ;
  delete[] AkuPre ;
  delete[] EnergyHisto ;

  return _OK_;
}
