/******************************************************************************
 * file:        feature.h  (structures of feature-vectors)
 * info:        declaration of all feature-vectors for AkuPre (TU Dresden)
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _FEATURE_H_ )
#define _FEATURE_H_

#include "feature.h"

/* constants for all signal analysis modes */
/* signal analysis parameters */
#define   WINDOW_LENGTH 400
#define   WINDOW_DELTA  160
#define   FFT_LENGTH    512
#define   SAMPLE_FRQ    16000
/* length of spectrum in feature vector */
#define   PRESPECLEN    30
#define   DELTASPECLEN  15
/* delta-features */
#define   DELTA_FORW   3
#define   DELTA_BACK   3

/* preprocessing feature vector */
typedef struct 
{
    float  TotalEnergy;
    float  ZCRate;
    float  MidSpec;
    float  Spec[PRESPECLEN];
} SPreFeature;

/* delta features */
typedef struct
{
    float  TotalEnergy;
    float  ZCRate;
    float  MidSpec;
    float  Spec[DELTASPECLEN];
} SDeltaFeature;

/* normalized features */
typedef struct
{
    float  TotalEnergy;
    float  ZCRate;
    float  MidSpec;
} SNormFeature;

/* final feature vector */
typedef struct
{
    SPreFeature     Pre;
    SDeltaFeature   Delta;
    SDeltaFeature   DDelta;
    SNormFeature    Norm;
} SFeature;

/* entry in ring buffer */
typedef struct sFeatBufEntry
{
    SFeature        Vec;
    ESpeechFlag     Flag;
    long            Num;
    short           NrUndef;
    short           InUndef;
    struct sFeatBufEntry  *Next;
    struct sFeatBufEntry  *Last;
    struct sFeatBufEntry  *DeltaForw;
    struct sFeatBufEntry  *DeltaBack;
} SFeatBufEntry;

/* feature vector for speech non-speech detector */
#define   SPDETSPECRED    5
#define   SPDETSPECLEN    (PRESPECLEN / SPDETSPECRED)
#define   SPDETDELTATAU   0.05
typedef struct 
{
    float   TotalEnergy;
    float   ZCRate;
    float   MidSpec;
    float   SmSpec[SPDETSPECLEN];
} SSpDetFeatX;

typedef struct
{
    SSpDetFeatX   Vec;
    SSpDetFeatX   DVec;
} SSpDetFeature;
    


#endif

