 /******************************************************************************
 * file:        feat_muc.h  (structures of feature-vectors)
 * info:        declaration of all feature-vectors for AkuPre (TU Muenchen)
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _FEATURE_H_ )
#define _FEATURE_H_

#include "feature.h"

/*  constants for all signal analysis modes */
/*  signal analysis parameters */
#define   WINDOW_LENGTH 256
#define   WINDOW_DELTA  160
#define   FFT_LENGTH    256
#define   SAMPLE_FRQ    16000
/*  length of spectrum in feature vector */
#define   PRESPECLEN    25
#define   DELTASPECLEN  25
/*  delta-features */
#define   DELTA_FORW   2
#define   DELTA_BACK   2

/*  special constants */
/*  Constants For Beh Convolution (CSpecConv) */
#define   FIRSTBARK   1.0
#define   DELTABARK   0.8
#define   MINLIMIT    1.0E-6
#define   MOD_POS_VON  1
#define   MOD_POS_BIS  16
#define   MOD_NEG_VON  21
#define   MOD_NEG_BIS  24
#define   ENERGYLEVEL  (-90)  // -20*(Quant-1)*log10(2) [dB]

/*  vector from pre-processing */
typedef struct sPreFeature
{
    float  TotalEnergy;
    float  ZCRate;
    float  SumLoud;
    float  ModLoud;
    float  Spec[PRESPECLEN];
} SPreFeature ;

/*  deltas of preprocesing vector */
typedef struct sDeltaFeature
{
    float  TotalEnergy;
    float  ZCRate;
    float  SumLoud;
    float  ModLoud;
    float  Spec[DELTASPECLEN];
} SDeltaFeature;

/*  gain controlled features (normalized) */
typedef struct sNormFeature
{
    float  TotalEnergy;
    float  ZCRate;
    float  SumLoud;
    float  ModLoud;
} SNormFeature;

/*  final feature vector */
typedef struct sFeature
{
    SPreFeature     Pre;
    SDeltaFeature   Delta;
    SDeltaFeature   DDelta;
    SNormFeature    Norm;
    SNormFeature    DNorm;
    SNormFeature    DDNorm;
} SFeature;


/*  feature fifo data-types  */
/*  entry in feature fifo buffer */
typedef struct sFeatBufEntry
{
    SFeature        Vec;
    ESpeechFlag     Flag;
    long            Num;
    short           NrUndef;
    short           InUndef;
    struct sFeatBufEntry  *Next;
    struct sFeatBufEntry  *Last;
    struct sFeatBufEntry  *DeltaForw;
    struct sFeatBufEntry  *DeltaBack;
} SFeatBufEntry;

#endif

