/******************************************************************************
 * file:        hamming.cc  
 * info:        source for CHammingWindow
 * written by:  M.Beham, 12.01.95 (org. from sts.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Info:    Hamming-Window
    w[i] = 0.54 - 0.46 cos( i*2Pi / Len )

*/


#include <math.h>

// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "hamming.h"

//------------------------------- public ---------------------------------------

// constructor: CHammingWindow
CHammingWindow::CHammingWindow( unsigned int len, double normfactor )
 : Len(len), NormFactor( normfactor ) 
{
    DEBUG_1_MSG("'CHammingWindow' constructor called" );
    
    // win-table calloc and calc
    WinDataPtr = new float[Len];
    IF_FATAL_EXIT( WinDataPtr == 0, "'CHammingWindow' out of memory");

    double xnorm = (2.0 * M_PI) / (double)(Len-1);  // or better! (Len)
    for( unsigned int i =0; i < Len; i++ )
        WinDataPtr[i] = NormFactor * (0.54 - 0.46*cos( (double)i*xnorm ));

}

// destructor: CHammingWindow
CHammingWindow::~CHammingWindow()
{
    DEBUG_1_MSG("'CHammingWindow' destructor called" );

    delete [] WinDataPtr;
}

// Win-Mult
void CHammingWindow::setNorm( double norm )
{
    DEBUG_2_MSG("'CHammingWindow' setNorm called" );
    NormFactor *= norm;

    register unsigned int i;
    for( i = 0; i < Len; i++ )
        WinDataPtr[i] *= norm;
}


// Win-Mult
void CHammingWindow::mult( float *sig )
{
    DEBUG_4_MSG("'CHammingWindow' mult called" );
    register unsigned int i;

    for( i = 0; i < Len; i++ )
        sig[i] *= WinDataPtr[i];

}

// subtraction of signal - mean value in window
void CHammingWindow::dcSub( float *sigptr )
{
    DEBUG_4_MSG("'CHammingWindow' dcSub called" );
    register unsigned int i;

    for( WinAverage=0.0, i=0; i<Len; i++ )
        WinAverage += sigptr[i];

    WinAverage /= (float)Len;

    for( i=0; i<Len; i++ )
        sigptr[i] -= WinAverage;
}

// count zero crossing rate in signal window
float CHammingWindow::countZC( float *sigptr )
{
    DEBUG_4_MSG("'CHammingWindow' countZC called" );
    
    float                 hysterese = 0.05;
    register int          Upper = ( sigptr[0] >= 0.0 );
    register unsigned int i;
    register unsigned int zcr = 0;

    for( i=1; i<Len; i++ )
        if( Upper ? sigptr[i] < -hysterese : sigptr[i] > hysterese )
        {    zcr++; Upper = !Upper; }

    return (float)zcr;
}



//------------------------------ end of hamming.cc -----------------------------

// test

#ifdef TEST_LEVEL

int main()
{   unsigned int len;
    printf( "\nHamming lengt: " ); scanf( "%d", &len );

    CHammingWindow Ham(len);
    float *vec = new float[len];

    for( int i = 0; i < len; i++ ) vec[i]=2.0;
    Ham.mult(vec);
 
    printf("\n (win * 2): \n");
    for( i = 0; i < len; i++ ) printf( "%5.2f", vec[i] );

    return(0);
}


#endif  // TEST_LEVEL
