#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fileconst.h" 

int maxnrframes = 1000000 ;

int nrdet ;     // nr of speech frames detected
int nrref ;     // nr of speech frames in ref
int nrdetcor ;  // nr of correctly detected speech frames


int main( int argc, char *argv[] ) {

  int beg, end, lastframe, frame ;
  char line[MaxLineLength] ;
  int *refarray, *hyparray ;
  char* tmp = new char[MaxLineLength] ;
  float precision = -1.0 ;
  float recall = -1.0 ;

  int startframe=0 ;
  int endframe = maxnrframes ;

  // check params:
  if( argc < 5 ) {
    fprintf( stderr, "%s <ref-file.lst> <hyp-file.lst> <refchan> <hypchan> [s=startframe] [e=endframe]\n", argv[0] ) ;
    exit(-1) ;
  }

  if( argc== 6 || argc==7 ) {
    strcpy( tmp, argv[5] ) ;
    if( strncmp( tmp, "s=", 2 ) == 0 ) {
      startframe = strtol( tmp+2, NULL, 10 ) ;
    } else if( strncmp( tmp, "e=", 2 ) == 0 ) {  
      endframe = strtol( tmp+2, NULL, 10 ) ;
    } ;
  }
  if( argc==7 ) {
    strcpy( tmp, argv[6] ) ;
    if( strncmp( tmp, "s=", 2 ) == 0 ) {
      startframe = strtol( tmp+2, NULL, 10 ) ;
    } else if( strncmp( tmp, "e=", 2 ) == 0 ) {  
      endframe = strtol( tmp+2, NULL, 10 ) ;
    } ;
  }

  FILE *reffp ;
  FILE *hypfp ;

  if( (reffp = fopen( argv[1], "r" )) == 0 ) {
    fprintf( stderr, "%s: can't open ref-file '%s'\n", argv[0], argv[1] ) ;
    exit(-1) ;
  } ;
  if( (hypfp = fopen( argv[2], "r" )) == 0 ) {
    fprintf( stderr, "%s: can't open hyp-file '%s'\n", argv[0], argv[2] ) ;
    exit(-1) ;
  } ;
  
  if( endframe > maxnrframes )
    maxnrframes = endframe ;

  refarray = new int[maxnrframes] ;
  hyparray = new int[maxnrframes] ;
  
  for( int frame=0 ; frame<maxnrframes ; frame++ ) {
    refarray[frame] = 0 ;
    hyparray[frame] = 0 ;
  }

  // fill refarray
  while( fgets( line, MaxLineLength, reffp ) != 0 ) {
    
    if( strcmp( argv[3], strtok( line, " \t" ) ) == 0 ) { 
      beg = strtol( strtok( NULL, " \t" ), NULL, 10 ) ;
      end = strtol( strtok( NULL, " \t\n" ), NULL, 10 ) ;
      for( frame=beg ; frame<=end ; frame++ )
	refarray[frame] = 1 ;
    }
  }
  lastframe = end ;


  // fill hyparray
  while( fgets( line, MaxLineLength, hypfp ) != 0 ) {
    
    if( strcmp( argv[4], strtok( line, " \t" ) ) == 0 ) { 
      beg = strtol( strtok( NULL, " \t" ), NULL, 10 )*100 ;
      end = strtol( strtok( NULL, " \t\n" ), NULL, 10 )*100 ;

      for( frame=beg ; frame<=end ; frame++ )
	hyparray[frame] = 1 ;
    }
  }
  if( end > lastframe )
    lastframe = end ;

  if( endframe == maxnrframes ) 
    endframe = lastframe ;

  // count matches:
  nrdet = nrref = nrdetcor = 0 ;

  for( frame=startframe ; frame<endframe ; frame++ ) {
    if( refarray[frame] == 1 ) {
      nrref++ ;
      if( hyparray[frame] == 1 ) 
	nrdetcor++ ;
    }
    if( hyparray[frame] == 1 ) 
      nrdet++ ;
  }


  if( nrdet != 0 )
    precision = (float)nrdetcor/(float)nrdet ;
 

  if( nrref != 0 )  
    recall = (float)nrdetcor/(float)nrref ;
  

  fprintf( stdout, "%4.2f\t%4.2f\t%d\n", precision*100.0, recall*100.0, nrref ) ;

  delete[] refarray ;
  delete[] hyparray ;
  
  fclose( reffp ) ;
  fclose( hypfp ) ;

}
