/******************************************************************************
 * file:        pref_drs.h  (for pref_drs.cc)
 * info:        class declarations for CPreFeature (TU-Dresden) 
 * written by:  M.Beham, 12.01.95 (org melfilt)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) C.M.Westendorf (TU-Dresden) und SIEMENS AG  15.06.1994
 *
 * Die Implementierung realisiert ein von der SIEMENS AG entwickeltes Signal-
 * analyseverfahren, da"s an der TU-Dresden neu implementiert wurde.
 *
 * Die Software ist ausschlie"slich f"ur die Verwendung innerhalb des Projektes
 * VERBMOBIL bestimmt. Eine Weitergabe der Programme an Partner au"serhalb des 
 * Projektes VERBMOBIL ist nur mit Zustimmung der SIEMENS AG und des Institutes
 * f"ur Technische Akustik der TU Dresden, sowie des Lehrstuhls MMK der TU
 * Muenchen zul"assig. 
 * 
 * Das Projekt VERBMOBIL wird vom Bundesminister f"ur Forschung und Technologie
 * unter dem F"orderkennzeichen 01IV 102K gef"ordert.
 * 
 ******************************************************************************/

#if !defined( _PREFEAT_H_ )
#define _PREFEAT_H_

#include "feat_drs.h"
#include "paramet.h"

// class CSpecConv:  (mel-scale cepstrally smoothed acc. to TU-Dresden)
class CSpecConv
{
private:
    const short    n_in;    /* input channels               */
    const short    n_out;   /* output channels              */
    short   *mid;           /* array of mid frequ.          */
    short   *width;         /* array of conv. core width    */
    float   *norm;          /* array of normal. coeff.      */
    float  **a;             /* matrix of core coeff.        */
public:
    CSpecConv( unsigned int powerspeclen, unsigned int convlen );
    ~CSpecConv();
    void   convSpec( float *powspecptr, float *convoutptr ) const;
    float *getRow( int i ) const { return a[i]; }
};

// class definition: CSpDetFeature, feature vec. for speech non-speech detect.
class CSpDetFeature
{
private:
   SSpDetFeatX          XVec;
   SSpDetFeature        YVec;
   const float          b;
   const float          a;
public:
   CSpDetFeature( double dt );
   ~CSpDetFeature( );
   unsigned int getFeatLen() const { return sizeof(SSpDetFeature)/sizeof(float); }
   float       *getFeature( const SPreFeature *pfptr );
};

// class definition: CPreFeature  (TU-Dresden)
class CPreFeature
{
private:
    const  CSpecConv       SpecConv;
    CSpDetFeature          SpDetFeature;
    const  unsigned int    PowerSpecLen;
    const  unsigned int    PreSpecLen;
    const  float           QuantLevel;
    float  * const         NoisePower;
public:
    CPreFeature( unsigned int powerspeclen, double frqsample, double noise, double quant );
    ~CPreFeature();
    void convPowSpec( float *powspecptr, float ndr, SPreFeature *pfptr ) const;
    void postPreSpec( SPreFeature *pfptr ) const;
    double getNormFactor() const;
    unsigned int getSpecLen() const { return PreSpecLen; }
    unsigned int getSpDetFeatureLen() const { return SpDetFeature.getFeatLen(); }
    float      * getSpDetFeature( const SPreFeature *pfptr ) { return SpDetFeature.getFeature(pfptr); }
    void printPreFeature( SPreFeature *pfptr ) const;
    void printDeltaFeature( SDeltaFeature *dfptr ) const;
    void printFeature( SFeature *fptr ) const;
    float *getNoiseSpec() { return  NoisePower; }
};

// class definition: CDeltaFeature  (TU-Dresden)
class CDeltaFeature
{
private:
    const unsigned int  DeltaSpecLen;
    const unsigned int  PreSpecLen;
    unsigned       int  SpecMerge;
public:
    CDeltaFeature( unsigned int prespeclen, unsigned int deltaspeclen );
    ~CDeltaFeature();
    void  calcDelta(  SFeatBufEntry *ep ) const;
    void  calcDDelta( SFeatBufEntry *ep ) const;
};

// class definition: CGainControl (TU-Dresden)
class CGainControl
{
private:
    CParFloat     TauRise;
    CParFloat     TauFall;
    CParEnum      UpdateUse;
    CParBool      UpdateFlag;
    CParBool      UpdateWrite;

    const float   bRise;
    const float   bFall;
    const float   aRise;
    const float   aFall;

    CParFloatVec  Max;
    FILE         *ParFile;

    CParFloat     EnergyDelta;
public:
    CGainControl( FILE *fp, double dt );
    ~CGainControl( );
    int   writePar( FILE *fp );
    void  norm( SPreFeature *prefp, SNormFeature *normfp, ESpeechFlag speechflag );
    float *getActMax() const { return Max.Vec; }
};



#endif

