/******************************************************************************
 * file:        pref_muc.h  (for pref_muc.cc)
 * info:        class declarations for CPreFeature (TU-Muenchen) 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _PREFEAT_H_ )
#define _PREFEAT_H_

#include "feat_muc.h"
#include "paramet.h"

// class CSpecConv:  (bark-scale convolution acc. to M.Beham)
class CSpecConv
{
private:
    const double        DeltaFrqSpec;
    const unsigned int  PowerSpecLen;
    const unsigned int  ConvLen;

    const double        FirstBarkGroup;
    const double        DeltaBarkGroup;
    const double        MinWeightLimit;

    struct SConvCoeffRow
    {   unsigned short  FirstFrq;
        unsigned short  LastFrq;
        float           MidFrqHz;
        float          *CoeffTab;
    };
    SConvCoeffRow      *ConvRowTab;

    double      bark( double f );
    double      erpeg( double z );
    double      exp10( double x );
    double      freq( double z );
    double      erreg( double f, double v );
    double      coeff( double fj, int i );
    int         coeffrow( float *coeffptr, double fj, unsigned short *von, unsigned short *bis );
public:
    CSpecConv( unsigned int powerspeclen, unsigned int convlen, double frqsample,
                  double firstBark, double deltaBark, double minLimit );
    ~CSpecConv();
    void   convSpec( float *powspecptr, float *convoutptr ) const;
    SConvCoeffRow &getRow( int i ) const { return ConvRowTab[i]; }
};

// class definition: CPreFeature (Bark scaled spectrum analysis acc. to M.Beham TU-Muenchen)
class CPreFeature
{
private:
    const  CSpecConv       SpecConv;
    const  unsigned int    PowerSpecLen;
    const  unsigned int    PreSpecLen;

    double totalenergy( float *powspec ) const;
    void   spezloudness( float *spec ) const;
    double sumloudness( float *spec ) const;
    double modloudness( float *spec ) const;
    double entropy( float *powspec ) const ;
    double energy_entropy( float *powspec ) const ;

    void   logspec( float *spec ) const;
    void   expspec( float *spec ) const;
    void   normalizeSpec( float *spec ) const;
public:
    CPreFeature( unsigned int powerspeclen, double frqsample );
    ~CPreFeature();
    void convPowSpec( float *powspecptr, float ndr, SPreFeature *pfptr ) const;
    void convPowSpec( float *powspecptr, float ndr, float entropy, SPreFeature *pfptr ) const;
    void postPreSpec( SPreFeature *prefeatptr ) const;
    double getNormFactor() const;
    unsigned int getSpecLen() const { return PreSpecLen; }
    unsigned int getSpDetFeatureLen() const { return sizeof(SPreFeature)/sizeof(float); }
    float      * getSpDetFeature( SPreFeature *pfptr ) const { return (float *)pfptr; }
    void printPreFeature( SPreFeature *pfptr ) const;
    void printDeltaFeature( SDeltaFeature *pfptr ) const;
    void printFeature( SFeature *fptr ) const;
};

// class definition: CDeltaFeature  (TU-Muenchen)
class CDeltaFeature
{
private:
    const unsigned int  DeltaSpecLen;
    const unsigned int  PreSpecLen;
public:
    CDeltaFeature( unsigned int prespeclen, unsigned int deltaspeclen );
    ~CDeltaFeature();
    void  calcDelta(  SFeatBufEntry *ep ) const;
    void  calcDDelta( SFeatBufEntry *ep ) const;
};

// class definition: CGainControl (TU-Muenchen)
class CGainControl
{
private:
    CParFloat     TauRise;
    CParFloat     TauFall;
    CParEnum      UpdateUse;
    CParBool      UpdateFlag;
    CParBool      UpdateWrite;

    const float   bRise;
    const float   bFall;
    const float   aRise;
    const float   aFall;

    CParFloatVec  Max;
    FILE         *ParFile;

    CParFloat     EnergyDelta;
public:
    CGainControl( FILE *fp, double dt );
    ~CGainControl( );
    int   writePar( FILE *fp );
    int   appendFixPar( FILE *fp );
    int   appendUpdPar( FILE *fp );
    void norm( SPreFeature *prefp, SNormFeature *normfp, ESpeechFlag speechflag );
    float *getActMax() const { return Max.Vec; }
};

#endif

