#include <stdio.h>
#include <string.h>
#include <math.h>

#include "dbg_mess.h"
#include "akupresimple.h"


float *energymin ;
float *sumloudmax ;
float *modloudmax ;
float *zcratemax ;

CAkuPreSimple **AkuPre=0 ; // array of pointers to acoustic preprocessing of the indiv channels
SPreFeature **prefp=0 ;               // array of pointers to feature vectors of indiv channels


void GetMinAndMax( int nrchanwav ) {

  // local variables
  int i, frame, running, chan ;

  // allocate size of global arrays
  energymin = new float[nrchanwav+1] ;
  sumloudmax = new float[nrchanwav+1] ;
  modloudmax = new float[nrchanwav+1] ;
  zcratemax = new float[nrchanwav+1] ;

  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( i=0 ; i<nrchanwav+1 ; i++ ) {
    energymin[i] = 1.0e30 ;
    sumloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
    zcratemax[i] = -1.0e30 ;
  } ;

#ifdef TESTLEVEL
  fprintf( stderr, "Determine minima and maxima of special features ...\n" ) ;
#endif

  // determine speech min resp max of special features (for offline normalization to get features for spdet)
  frame = 0 ;
  running = 1 ;
  while( running ) { 
    // check if any of the sigfiles faces EOF yet
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      if( (prefp[chan] = AkuPre[chan]->getNextPre()) == 0 ) {
	running = 0 ;
	break ;
      } ;
      
	
      //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( prefp[chan]->TotalEnergy < energymin[chan] )
	energymin[chan] =  prefp[chan]->TotalEnergy ;
      if( prefp[chan]->SumLoud > sumloudmax[chan] )
	sumloudmax[chan] = prefp[chan]->SumLoud ;
      if( prefp[chan]->ModLoud > modloudmax[chan] )
	modloudmax[chan] = prefp[chan]->ModLoud ;
      if( prefp[chan]->ZCRate > zcratemax[chan] )
	zcratemax[chan] = prefp[chan]->ZCRate ;
      
    } ;
    if( running == 0 )
      break ;
  } ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif

  // output of min energies. max sumloud, max modloud, max zcrate: 
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    fprintf( stderr, "channel %d: Min energy: %f\n", chan, energymin[chan] ) ;
    fprintf( stderr, "channel %d: Max sumloud: %f\n", chan, sumloudmax[chan] ) ;
    fprintf( stderr, "channel %d: Max modloud: %f\n", chan, modloudmax[chan] ) ;
    fprintf( stderr, "channel %d: Max zcrate: %f\n", chan, zcratemax[chan] ) ;
  } ;

} ;


int main( int argc, char *argv[] ) {

  // local variables
  int i ;
  FILE *cfgfp ;

  // check params
  if( argc < 4 ) {
    fprintf( stderr, "USAGE: %s <akupre-cfg-file> <wavfile1> <wavfile2> [wavfile3] ... [wavfileN]\n", argv[0] ) ;
    exit(-1) ;
  } ;

  cfgfp = fopen( argv[1], "r" ) ;

  AkuPre = new  CAkuPreSimple*[argc-2] ;
  for( i=0 ; i<argc-2 ; i++ ) {
    AkuPre[i] = new CAkuPreSimple( cfgfp ) ;
    AkuPre[i]->startAkuPre( argv[i+2] ) ;
  }
  prefp = new SPreFeature*[argc-2] ;

  


  GetMinAndMax( argc-2 ) ;
  
  // determine maximum of maxmodloud!
  float maxofmaxmodloud = -1.0e30 ;
  for( i=0 ; i<argc-2 ; i++ ) 
    if( modloudmax[i] > maxofmaxmodloud ) 
      maxofmaxmodloud = modloudmax[i] ;

  
  // rescale audiofiles
  for( i=0 ; i<argc-2 ; i++ ) {
    char command[1024] ;
    sprintf( command, "/u/janin/bin/jsndmix -o %s.amp %s %4.3f\n", argv[i+2], argv[i+2], (maxofmaxmodloud/modloudmax[i])*5.0 ) ;
    system( command ) ;

  } ;
  
  for( i=0 ; i<argc-2 ; i++ ) 
    delete AkuPre[i] ;
  delete[] AkuPre ;
  
  delete[] energymin ;
  delete[] sumloudmax ;
  delete[] modloudmax ;
  delete[] zcratemax ;

  delete[] prefp ;               // array of pointers to feature vectors of indiv channels



  fclose( cfgfp ) ;

  return(1) ;

} ;
