/******************************************************************************
 * file:        siganl.cc  
 * info:        implementation of CSigAnalyse  
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

// #define DEBUG_LEVEL 3
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "siganl.h"



//--------------------------- public 'CSigAnalyse' ----------------------------


// constructor: CSigAnalyse
CSigAnalyse::CSigAnalyse( unsigned int windelta, unsigned int winlen, unsigned int fftlen, char *devicestr, int sampleshift, float preemphase  )
 :  Sig( windelta, winlen, devicestr, sampleshift, preemphase  ),
    Spec( winlen, fftlen ),
    NextIndex( -1 )
{
    DEBUG_1_MSG("'CSigAnalyse' constructor called" );
}

// destructor: CSigAnalyse
CSigAnalyse::~CSigAnalyse( )
{
    DEBUG_1_MSG("'CSigAnalyse' destructor called" );
}

// init. buffer with open file or start ADC etc...
int CSigAnalyse::startAnl( const char *filename, int skipbytes )
{
    DEBUG_2_MSG_S("'CSigAnalyse' startAnl called with:", filename );

    // start read from stream (start adc and fill buffer first time)
    if( Sig.startBuf( filename, skipbytes ) == _ERR_ ) return _ERR_;
    NextIndex = 0;
    return _OK_;
}

// stop device, do not clear Buffer
void CSigAnalyse::stopAnl( )
{
    DEBUG_2_MSG("'CSigAnalyse' stopAnl called");

    // stop device, do not clear device buffers
    Sig.stopBuf( );
}
    
// stop device, do not clear Buffer
void CSigAnalyse::clearAnl( )
{
    DEBUG_2_MSG("'CSigAnalyse' clearAnl called");

    // stop device, and clear all device buffers 
    Sig.clearBuf( );
}

// calculate next spec
float *CSigAnalyse::getNextSpec( )
{
    DEBUG_4_MSG("'CSigAnalyse' getNextSpec called");
    if( NextIndex < 0 )
    {   ERROR_MSG("'CSigAnalyse' getNextSpec: analyse not started");
        return 0;
    }

    // check already calculated
    if( NextIndex == 0 )
    {   // read two window
        if( Sig.getWin( Spec.getBuffer(0) ) == _ERR_ ) return 0;
        if( Sig.getWin( Spec.getBuffer(1) ) == _ERR_ ) return 0;
        // calc two specs with one fft
        Spec.calcDoubleSpec( );
        NextIndex = 1;
        return Spec.getBuffer(0);
    } else
    {   // spec already calculated
        NextIndex = 0;
        return Spec.getBuffer(1);
    }
}

// calculate next spec and zc-rate
float *CSigAnalyse::getNextSpec( float *zcr )
{
    DEBUG_4_MSG("'CSigAnalyse' getNextSpec called");
    if( NextIndex < 0 )
    {   ERROR_MSG("'CSigAnalyse' getNextSpec: analyse not started");
        return 0;
    }

    // check already calculated
    if( NextIndex == 0 )
    {   // read two window
        if( Sig.getWin( Spec.getBuffer(0) ) == _ERR_ ) return 0;
        if( Sig.getWin( Spec.getBuffer(1) ) == _ERR_ ) return 0;
        // calc two specs with one fft
        Spec.calcDoubleSpec( );
        NextIndex = 1;
        *zcr = Spec.getZCRate(0);
        return Spec.getBuffer(0);
    } else
    {   // spec already calculated
        NextIndex = 0;
        *zcr = Spec.getZCRate(1);
        return Spec.getBuffer(1);
    }
}

//--------------------------- end of 'CSigWinBuffer' --------------------------

//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL

#include <stdlib.h>

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 5, "usage: testsigwin <filename> <WinLen> <WinDelta> <Device>" );  

    CSigAnalyse siganl( atoi(argv[3]), atoi(argv[2]), argv[4] );

    siganl.startAnl( argv[1] );

    while( siganl.getNextSpec( ) != 0 );

    printf( "* number of samples = %d, of windows = %d\n", siganl.getSampleAnz(), siganl.getWindowAnz() );

    siganl.clearAnl();

    return(0);
}


#endif  // TEST_LEVELcd source
