/******************************************************************************
 * file:        siginput.h  (for siginput.cc)
 * info:        class declarations for CSigInput (base class) and deri. classes
 * written by:  M.Beham, 12.01.95 (org. from readsig.c and voronl.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _SIGINPUT_H_ )
#define _SIGINPUT_H_

#include <string.h>

#include  "dbg_mess.h"

//--------------------------- base class 'CSigInput' ------------------------

class CSigInput
{
protected:
    const int      BlocksPerRequest;
    const unsigned long  BlockSize;
    int            FileDes;
    unsigned int   FrqSample;
    unsigned long  SampleAnz;
public:
    enum EStopMode { Stop, Clear };
    CSigInput( unsigned int blocksize, int blocksperrequest );
    virtual ~CSigInput();
    virtual int startRead( const char *filename, int NrSkip=0 ) = 0;
    virtual int stopRead( EStopMode mode ) = 0;
    virtual int sigReadBlock( short *buf ) = 0;
    unsigned long getSampleAnz() const { return SampleAnz; }
    int  getBlocksPerRequest() const { return BlocksPerRequest; }
    unsigned int getFrqSamp()  const { return FrqSample; } 
};

//------------------------ deri class 'CSigInputFileILS' --------------------

#include "ilshead.h"

class CSigInputFileILS : public CSigInput
{
private:
    SHeaderILS     Header;
    int            ByteSwap;
    void           header_swap();
    void           short_swap( short *sp );
    void           long_swap( long *lp );   
public:
    CSigInputFileILS( unsigned int blocksize );
    virtual ~CSigInputFileILS();
    int startRead( const char *filename, int NrSkip=0 );
    int stopRead( EStopMode mode );
    virtual int sigReadBlock( short *buf );
    int  writeHeader( char *filename );
    int  writeBlock( int fd, short *buf );
};

#include "nisthead.h"

class CSigInputFileNIST : public CSigInput
{
private:
    SHeaderNIST     Header;
    int            ByteSwap;
    void           header_swap();
    void           short_swap( short *sp );
    void           long_swap( long *lp );   
public:
    CSigInputFileNIST( unsigned int blocksize );
    virtual ~CSigInputFileNIST();
    int startRead( const char *filename, int NrSkip=0 );
    int stopRead( EStopMode mode );
    virtual int sigReadBlock( short *buf );
    int  writeHeader( char *filename );
    int  writeBlock( int fd, short *buf );
    int  readHeader( FILE *fp ) ;
};




//------------------------ deri class 'CSigInputSunADC' ---------------------
#ifdef RealTimeADC
#include <signal.h>
#include <task.h>

// interupt handler: SIGUSR1
class CSigSunADCReady : public Interrupt_handler
{
private:
    int    ADCFlag;
    void   interrupt() { ADCFlag = TRUE; }
public:
    CSigSunADCReady() : Interrupt_handler(SIGUSR1), ADCFlag(FALSE) { }
    int    testADC() const { return ADCFlag; }
    void   setADC( int flag ) { ADCFlag = flag; }
};

// interupt handler:  SIGUSR2
class CSigSunADCEnd : public Interrupt_handler   
{
private:
    int    ADCFlag;
    void   interrupt() { ADCFlag = TRUE; }
public:
    CSigSunADCEnd() : Interrupt_handler(SIGUSR2), ADCFlag(FALSE) { }
    int    testADC() const { return ADCFlag; }
    void   setADC( int flag ) { ADCFlag = flag; }
};


// signal input object for parent process
class CSigInputSunADC : public CSigInput
{
private:
    CSigSunADCReady Ready;
    CSigSunADCEnd   End;
    int           ChildPid;
    int           StartedFlag;
public:
    CSigInputSunADC( unsigned int blocksize, unsigned long frqsample = 16000 );
    ~CSigInputSunADC();
    int startRead( const char *filename, int NrSkip=0 );
    int stopRead( EStopMode mode = Stop );
    int sigReadBlock( short *buf );
};

#endif

//------------------------ deri class 'CSigInputAudioFile' --------------------

// the following struct and defines are from 'audio_filehdr.h' in
// '/usr/demo/SOUND/include/multimedia'
typedef  unsigned long u_32;

struct SAudioFileHdr {
	u_32		magic;		/* magic number */
	u_32		hdr_size;	/* size of this header */
	u_32		data_size;	/* length of data (optional) */
	u_32		encoding;	/* data encoding format */
	u_32		sample_rate;	/* samples per second */
	u_32		channels;	/* number of interleaved channels */
};


/* Define the magic number */
#define	AUDIO_FILE_MAGIC		((u_32)0x2e736e64)
#define	AUDIO_UNKNOWN_SIZE	        ((u_32)(~0L))


class CSigInputAudioFile : public CSigInput
{
private:
    SAudioFileHdr  Header; 
public:
    CSigInputAudioFile( unsigned int blocksize );
    ~CSigInputAudioFile();
    int startRead( const char *filename, int NrSkip=0 );
    int stopRead( EStopMode mode );
    int sigReadBlock( short *buf );
    int  writeHeader( char *filename );
    int  writeBlock( int fd, short *buf );
};


#endif

