/******************************************************************************
 * file:        sigwin.cc  
 * info:        implementation of CSigWinBuffer  
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <string.h> 

// #define DEBUG_LEVEL 3
// #define TEST_LEVEL 
// #define TEST_SWEEP
// #define RealTimeADC

// R.Faltlhauser: Modification for LINUX --> no online Version
#ifdef LINUX
#undef RealTimeADC
#endif


#include "dbg_mess.h"
#include "sigwin.h"
#include "math.h"


//--------------------------- public 'CSigWinBuffer' --------------------------


// constructor: CSigWinBuffer
CSigWinBuffer::CSigWinBuffer( unsigned int windelta, unsigned int winlen, char *devicestr, int samshift, float preem )
 :  WinLen( winlen ), 
    BufLen( windelta ),
    SampleShift( samshift ),
    Preemphase( preem ),
    MinBufFill( (winlen+windelta-1)/windelta ),
    WinNum( 0 )
{
    DEBUG_1_MSG("'CSigWinBuffer' constructor called" );

    // construct SigInput device
    if(      strcmp( devicestr, "FileILS" )   == 0 )
        SigInputPtr = new CSigInputFileILS( windelta );
    // construct SigInput device
    else if( strcmp( devicestr, "FileNIST" )  == 0 )
        SigInputPtr = new CSigInputFileNIST( windelta );
#ifdef RealTimeADC
    else if( strcmp( devicestr, "SunADC"  )   == 0 )
        SigInputPtr = new CSigInputSunADC( windelta );
#endif
    else if( strcmp( devicestr, "AudioFile" ) == 0 )
        SigInputPtr = new CSigInputAudioFile( windelta );
    else
        IF_FATAL_EXIT_S( TRUE, "'CSigWinBuffer' unknown audio device:", devicestr );

    IF_FATAL_EXIT( SigInputPtr == 0, "'CSigWinBuffer' out of memory" );

    // calloc and init float-sample-buffer
    BufAnz = MinBufFill + (SigInputPtr->getBlocksPerRequest() - 1);
    if( BufAnz % SigInputPtr->getBlocksPerRequest() > 0 )
        BufAnz = (BufAnz / SigInputPtr->getBlocksPerRequest() + 1) * SigInputPtr->getBlocksPerRequest();

    BasePtr = new float[BufAnz*BufLen];
    IF_FATAL_EXIT( BasePtr == 0, "'CSigWinBuffer' out of memory" );
    EndPtr  = BasePtr + BufAnz*BufLen;

    // Clear Buffer
    WPtr = RPtr = BasePtr;
    BufFill = 0;

    // calloc short-sample-buffer
    SampleBuf = new short[SigInputPtr->getBlocksPerRequest()*BufLen];
    IF_FATAL_EXIT( SampleBuf == 0, "'CSigWinBuffer' out of memory" );


#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 3 || DEBUG_LEVEL == -3 )
 fprintf( stderr, ">>> 'CSigWinBuffer': BufLen=%d, BufAnz=%d, MinBufFill=%d\n", BufLen, BufAnz, MinBufFill );
#endif
}

// destructor: CSigWinBuffer
CSigWinBuffer::~CSigWinBuffer( )
{
    DEBUG_1_MSG("'CSigWinBuffer' destructor called" );

    // delete buffers
    delete SigInputPtr;
    delete [] BasePtr;
}

// fill float buffer with samples of Blocks per Request; Bit-Shift, short->float, Preemphase
int CSigWinBuffer::fillBuf( float *dest )
{
    register int i;
    register int len = SigInputPtr->getBlocksPerRequest() * BufLen;
    register float x;
    static float x_1 = 0.0;

    // read n blocks
    if( SigInputPtr->sigReadBlock( SampleBuf ) == _ERR_ )
	return _ERR_;

    // Bit-Shifting
    if( SampleShift > 0 )
	for( i=0; i < len; i++ )
	    SampleBuf[i] >>= SampleShift;
    // old Siemens-Version of Shifting, with problematic characteristic !!
	// SampleBuf[i] = SampleBuf[i] >= 0 ?  
	// SampleBuf[i] >> SampleShift : 
	// -( -SampleBuf[i] >> SampleShift );

    // Preemphase
    if( Preemphase > 1E-32 )
    {
	for( i=0; i < len; i++ )
	{   x = (float)SampleBuf[i];
	    dest[i] = x - Preemphase * x_1;
	    x_1 = x;
	}
    } else
    {
	for( i=0; i < len; i++ )
	    dest[i] = (float)SampleBuf[i];
    }

    return _OK_;
}

// init. buffer with open file or start ADC etc...
int CSigWinBuffer::startBuf( const char *filename, int skipbytes )
{
    DEBUG_2_MSG_S("'CSigWinBuffer' startBuf called with:", filename );

    // start read from stream (start adc)
    if( SigInputPtr->startRead( filename, skipbytes ) == _ERR_ ) return _ERR_;

    // fill buffer first time
    RPtr = WPtr = BasePtr; 
    for( BufFill = 0; BufFill < BufAnz; BufFill+=SigInputPtr->getBlocksPerRequest() )
    {   if( fillBuf( WPtr ) == _ERR_ )
        {   ERROR_MSG_S( "'CSigWinBuffer' can't get first signal-window from:", filename );
            return _ERR_;
        }
        WPtr += SigInputPtr->getBlocksPerRequest() * BufLen;
        if( WPtr >= EndPtr ) WPtr = BasePtr;
    }
    WinNum = 0;
    return _OK_;
}

// stop device, do not clear Buffer
void CSigWinBuffer::stopBuf( )
{
    DEBUG_2_MSG("'CSigWinBuffer' stopBuf called");

    // stop device, do not clear device buffers
    SigInputPtr->stopRead( CSigInput::Stop );
}
    
// stop device, and clear Buffer
void CSigWinBuffer::clearBuf( )
{
    DEBUG_2_MSG("'CSigWinBuffer' clearBuf called");

    // stop device, and clear all device buffers 
    SigInputPtr->stopRead( CSigInput::Clear );
 
    WinNum = 0;
    BufFill = 0;
    RPtr = WPtr = BasePtr;
}

#ifndef TEST_SWEEP

// read one window from buffer, and subtract mean
int CSigWinBuffer::getWin( float *destptr )
{
    DEBUG_4_MSG("'CSigWinBuffer' getWin called");
    register int i;

    // check fill of buffer (one window min)
    if( BufFill < MinBufFill )
    {   if( fillBuf( WPtr ) == _ERR_ )
            return _ERR_;
        WPtr += SigInputPtr->getBlocksPerRequest() * BufLen;
        if( WPtr >= EndPtr ) WPtr = BasePtr;
        BufFill += SigInputPtr->getBlocksPerRequest();
        IF_FATAL_EXIT( BufFill > BufAnz, "'CSigWinBuffer' getWin buffer overflow" );
    }

    // copy samples to dest. and calc sum.
    register int rest = EndPtr - RPtr;
    register float sum = 0.0;
    if( rest >= WinLen )
    {   for( i = 0; i < WinLen; i++ )  
            sum += RPtr[i];
    } else
    {   for( i = 0; i < rest; i++ )
            sum += RPtr[i];
        for( i = 0; i < WinLen-rest; i++ )
            sum += BasePtr[i];
    }
    sum /= WinLen;

    // subtract mean value and copy
    if( rest >= WinLen )
    {   for( i = 0; i < WinLen; i++ )  
            *destptr++ = RPtr[i] - sum;
    } else
    {   for( i = 0; i < rest; i++ )
            *destptr++ = RPtr[i] - sum;
        for( i = 0; i < WinLen-rest; i++ )
            *destptr++ = BasePtr[i] - sum;
    }
    RPtr += BufLen;
    if( RPtr >= EndPtr ) RPtr = BasePtr;
    BufFill--;
    WinNum++;

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL >= 4 || DEBUG_LEVEL == -4 )
 fprintf( stderr, ">>> 'CSigWinBuffer': BufFill=%d\n", BufFill );
#endif

    return _OK_;
}

#else

// get sinus-window 
#define FRQ_0 50.0
#define FRQ_1 5000.0
#define FRQ_D 2.0
#define AMPL  10000.0

int CSigWinBuffer::getWin( float *destptr )
{
    register int i;
    double f = WinNum * FRQ_D + FRQ_0;
    double w = 2*M_PI*f / 16000.0;

    if( f > FRQ_1 )
	return _ERR_;

    for( i=0; i<WinLen; i++ )
	destptr[i] = AMPL * sin( w*i );

    printf( "WinNr: %d, f: %1.1f Hz\n", WinNum, f );

    WinNum++;
    return _OK_;
}

#endif
    



//--------------------------- end of 'CSigWinBuffer' --------------------------

//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL

#include <stdlib.h>

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 5, "usage: testsigwin <filename> <WinLen> <WinDelta> <Device>" );  

    CSigWinBuffer sigwin( atoi(argv[3]), atoi(argv[2]), argv[4] );
    float *win = new float[atoi(argv[2])];

    sigwin.startBuf( argv[1] );

    while( sigwin.getWin( win ) == _OK_ );

    printf( "* number of samples = %d, of windows = %d\n", sigwin.getSampleAnz(), sigwin.getWindowAnz() );

    delete [] win;

    return(0);
}


#endif  // TEST_LEVELcd source
