/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "genobj.h"

#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupresimple.h"
#include "spnonsp.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

#define MAXNRCHAN 20

// use first 120 minutes ONLY!
#define MAXFRAMESFORMAXDET 1080000  

#define TESTLEVEL
//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD, XML_MULT } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- global variables ----------------------------------------

float **Energy=0 ;                    // array of energies for each frame and channel
//float **EnergyNoSmooth=0 ;            // array of energies for each frame and channel, always contains unsmoothed energies
float **ModLoud=0 ;                   // array of modified loudness for each frame and channel
ESpeechFlag **SpeechFlag=0  ;         // array of decisions of spnonsp-detector for each frame and channel
ESpeechFlag **SpeechFlag2=0  ;   // array of decisions of spnonsp-detector for each frame and channel, use unnormalized features!
int *Channel=0 ;                     // array of active ChannelId for each frame, contains ChanId or -1 
//int **ChannelNoSmooth=0 ;             // array of active Channel-Ids for each frame, built on the basis of EnergyNoSmooth[][] 
CAkuPreSimple **AkuPre=0 ;            // array of pointers to acoustic preprocessing of the indiv channels
CAkuPreSimple  *AkuPreMixFile=0 ;     // pointer to acoustic preprocessing of mixfile 
CSpeechNonSpeech **SpNonSp=0  ;       // array of pointers to speechnonspeech-detectors for indiv channels
CSpeechNonSpeech **SpNonSp2=0  ;       // array of pointers to speechnonspeech-detectors for indiv channels
CSpeechNonSpeech *SpNonSpMixFile=0;   // pointer to speechnonspeech-detector of mixfile
CTool *Tool=0 ;                       // pointer to Tool-class

FILE **outfp=0 ;                      // filepointer for output of results for indiv channels
FILE *globoutfp=0 ;                   // filepointer for out of single resultfile
FILE *chanfp=0 ;                      // filepointer to basic cfg-file containing unchangeable params
FILE *spnonspchancfgfp=0;             // filepointer to cfg-file for spnonsp-detector of indiv channels (containing means and vars!)
FILE *spnonspchancfgfp2=0;            // fileptr to cfg-file for spnsp-det. of indiv channels (cont. means and vars!), more sensitive version
FILE *spnonspmixcfgfp=0;              // filepointer to cfg-file for spnonsp-detector of mixfile (containing means and vars!)
FILE *logfp ;

SPreFeature **prefp=0 ;               // array of pointers to feature vectors of indiv channels
SPreFeature *prefpmixfile=0 ;         // pointer to feature vector of mixfile

float meanenergy ;                  // meanenergy of current frame of all channels (including mixfile)
float meanenergyohnemix ;           // meanenergy of current frame of all channels (without mixfile)
float *energymin=0  ;                 // array of minimum energies found on each channel
float *sumloudmax=0 ;                 // array of max sumloudness found on each channel
float *modloudmax=0 ;                 // array of max modified loudness found on each channel
float *modloudmin=0 ;                 // array of max modified loudness found on each channel
//float *zcratemax=0 ;                  // array of max zerocrossingrate found on each channel
float *zcratemin=0 ;                  // array of max zerocrossingrate found on each channel

int nrchanwav ;                     // number of indiv channels (without mixfile)
int nrframe ;                       // number of frames of wavfiles (shortest wavfile should be counted!)

int wavfileoffset ;                 // number of first wavfile-argument
int totnrwav ;                      // total number of wavfiles (includes mixfile) 
int mixfileindex ;                  // number of mixfile-argument
int outfileindex ;                  // number of outfile-argument

int **ChanCount=0 ;                   // contains, how often a channels is the most probable within a smoothbuffer!
//int *Overlap=0 ;                      // info about overlap for each channel (while output is produced!
int **OverlapFlag=0 ;                 // info about overlap for each frame! 1: means this is a possible overlap channel (not the maxloudchannel!)

char **ChanId ;                     // array of strings which contain the ChannelId of each channel (Ids 0,1,2, ...,A,B, ..)

//------------------------------- functions ----------------------------------------

void PrintSpeechFlagArray( FILE *fp ) {

  // local variables 
  int chan, frame, count ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) {
    count = 0 ;
    for( chan=0 ;chan<nrchanwav+1 ; chan++ ) {
      fprintf( fp, "%d: %d\t", frame, SpeechFlag[frame][chan] ) ;
      if( SpeechFlag[frame][chan] >= PUS ) 
	count++ ;
    } ;
    fprintf( fp, "nract: %d\n", count ) ;
  } ;
} ;

void PrintChannelArray( FILE *fp ) {

  // local variables 
  int chan, frame ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) 
    fprintf( fp, "%d: %d\n", frame, Channel[frame] ) ;

} ;

/*
void PrintOverlapArray( FILE *fp ) {

  // local variables 
  int chan, frame ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) {
    fprintf( fp, "%d: %d\n", frame, Overlap[frame] ) ;   
  } ;
} ;
*/

void PrintEnergyArray( FILE *fp ) {

  // local variables 
  int chan, frame ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) {
    for( chan=0 ;chan<nrchanwav+1 ; chan++ ) {
      fprintf( fp, "%d: %f\t", frame, Energy[frame][chan] ) ;
    } ;
    fprintf( fp, "\n" ) ;
  } ;
} ;

void PrintOverlapFlagArray( FILE *fp ) {

  // local variables 
  int chan, frame ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) {
    for( chan=0 ;chan<nrchanwav ; chan++ ) {
      fprintf( fp, "%d: %d\t", frame, OverlapFlag[frame][chan] ) ;
    } ;
    fprintf( fp, "\n" ) ;
  } ;
} ;


void PrintModLoudArray( FILE *fp ) {

  // local variables 
  int chan, frame ;
  
  for( frame=0 ; frame<nrframe ; frame++ ) {
    for( chan=0 ;chan<nrchanwav+1 ; chan++ ) {
      fprintf( fp, "%d: %f\t", frame, ModLoud[frame][chan] ) ;
    } ;
    fprintf( fp, "\n" ) ;
  } ;
} ;




void SmoothEnergies( int delta ) {

  // local variables
  float sum ;
  int j, frame, chan, tmpframe ;  

  for( frame=0 ; frame<nrframe ; frame++ ) {
    
    for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
      sum = 0.0 ;
      for( j=-delta ; j<=delta ; j++ ) {
	if( (tmpframe = frame+j) < 0 )
	  tmpframe = 0 ;
	if( tmpframe >= nrframe )
	  tmpframe = nrframe-1 ;
	
	sum += Energy[tmpframe][chan] ;
      } ;
      Energy[frame][chan] = sum/(2*delta+1) ;
    } 
  } ;
} ;


void SmoothModLoud( int delta ) {

  // local variables
  float sum ;
  int j, frame, chan, tmpframe ;  

  for( frame=0 ; frame<nrframe ; frame++ ) {
    
    for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
      sum = 0.0 ;
      for( j=-delta ; j<=delta ; j++ ) {
	if( (tmpframe = frame+j) < 0 )
	  tmpframe = 0 ;
	if( tmpframe >= nrframe )
	  tmpframe = nrframe-1 ;
	
	sum += ModLoud[tmpframe][chan] ;
      } ;
      ModLoud[frame][chan] = sum/(2*delta+1) ;
    } 
  } ;
} ;



int CountFrames(  char **argument ) {

  // local variables
  int count ;
  
#ifdef TESTLEVEL
  fprintf( stderr, "Determine length of signal-files ...\n" ) ;
#endif
  // start akupre of mixfile
  AkuPreMixFile->clearAkuPre() ;
  AkuPreMixFile->startAkuPre( argument[mixfileindex] ) ;

  // count frames
  count = 0 ;
  while( AkuPreMixFile->getNextPre() != 0 ) {
    count++ ;
  } ;

  // restart MixFile
  AkuPreMixFile->clearAkuPre() ;
  AkuPreMixFile->startAkuPre( argument[mixfileindex] ) ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif
  
  return( count ) ;
  
} ;



void PrintHeader( EOutPutMode outputmode ) {
  // local variables
  int i ;


  // output header
  switch( outputmode ) {
    
  case FrameWise:
    for( i=0 ; i<nrchanwav ; i++ ) 
      fprintf( outfp[i], "FILENAME: %s\n", "dummy.wav" ) ;
    break ;
      
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      // dummy header: length of file is still unknown!
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    } ;
    // dummy header: length of file is still unknown!
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    break ;
  case XML_MULT:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      // dummy header: length of file is still unknown!
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13-icsi.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    } ;
    // dummy header: length of file is still unknown!
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13-icsi.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    break ;
  } ; 
} ;


void GetMinAndMax( float normchanperc=0.0 ) {

  // local variables
  int i, frame, running, chan, nrframesformaxdet, usedframes ;

  // allocate size of global arrays
  energymin = new float[nrchanwav+1] ;
  sumloudmax = new float[nrchanwav+1] ;
  modloudmax = new float[nrchanwav+1] ;
  modloudmin = new float[nrchanwav+1] ;
  zcratemin = new float[nrchanwav+1] ;

  CListOfObj ModLoudList ;
  CListOfObj SumLoudList ;
  //  CListOfObj ZCRateList ;

  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( i=0 ; i<nrchanwav+1 ; i++ ) {
    energymin[i] = 1.0e30 ;
    sumloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
    modloudmin[i] = 1.0e30 ;
    zcratemin[i] = 1.0e30 ;
  } ;

#ifdef TESTLEVEL
  fprintf( stderr, "Determine minima and maxima of special features ...\n" ) ;
#endif
  
  usedframes = MAXFRAMESFORMAXDET ;
  // determine speech min resp max of special features (for offline normalization to get features for spdet)
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    frame = 0 ;
    running = 1 ;

    // reset sortlists!
    if( !ModLoudList.isEmpty() )
      ModLoudList.removeAll() ;
    if( !SumLoudList.isEmpty() )
      SumLoudList.removeAll() ;
    //if( !ZCRateList.isEmpty() )
    //ZCRateList.removeAll() ;

    while( running ) {
      // check if any of the sigfiles faces EOF yet
      if( (prefp[chan] = AkuPre[chan]->getNextPre()) == 0 ) {
	running = 0 ;
	break ;
      } ;
      
      if( normchanperc != 0.0 ) {

	CFloatObj* modloud = new CFloatObj(prefp[chan]->ModLoud) ;
	CFloatObj* sumloud = new CFloatObj(prefp[chan]->SumLoud) ;
	//CFloatObj* zcrate  = new CFloatObj(prefp[chan]->ZCRate) ;

	ModLoudList+=( (CObj*)modloud ) ;
	SumLoudList+=( (CObj*)sumloud ) ;
	//ZCRateList+=( (CObj*)zcrate ) ;
      } ;
  
      // get min energy (just use the second minute of the meeting:
      if( frame > 6000 && frame < 36000 ) {
	if( (prefp[chan]->TotalEnergy < energymin[chan]) && (prefp[chan]->TotalEnergy >= 1.0) )
	  energymin[chan] =  prefp[chan]->TotalEnergy ;
	if( prefp[chan]->ZCRate < zcratemin[chan] )
	  zcratemin[chan] = prefp[chan]->ZCRate ;
	if( prefp[chan]->ModLoud < modloudmin[chan] )
	  modloudmin[chan] = prefp[chan]->ModLoud ;
      }
      // search for max of other features over whole meeting! (some speakers don't talk very often!)
      if( prefp[chan]->SumLoud > sumloudmax[chan] )
	sumloudmax[chan] = prefp[chan]->SumLoud ;
      if( prefp[chan]->ModLoud > modloudmax[chan] )
	modloudmax[chan] = prefp[chan]->ModLoud ;

      
      if( ++frame > MAXFRAMESFORMAXDET )
	break ;

    } ;

    // memorize length of shortest wavfile!
    if( frame < usedframes )
      usedframes = frame ;

    // overwrite max-values with percentiles, if requested!
    CListIterator modloudit ;
    CListIterator sumloudit;
    //CListIterator zcrateit; 
    if( normchanperc != 0.0 ) {
      ModLoudList.quickSort() ;
      SumLoudList.quickSort() ;
      //ZCRateList.quickSort() ;
      modloudit = ModLoudList.head() ;
      sumloudit = SumLoudList.head() ;
      //zcrateit = ZCRateList.head();
      for( i=0 ; i<usedframes-(int)(usedframes*normchanperc/100) ; i++ ) {
	++modloudit ;
	++sumloudit ;
	//++zcrateit ;
      }
      sumloudmax[chan] = ((CFloatObj*)(*sumloudit))->getData() ; 
      modloudmax[chan] = ((CFloatObj*)(*modloudit))->getData() ; 
      //zcratemax[chan] =  ((CFloatObj*)(*zcrateit))->getData(); 

      ModLoudList.removeAll() ;
      SumLoudList.removeAll() ;
      //ZCRateList.removeAll() ;
      
    } ;
  } ;


  for( frame=0 ; frame<usedframes ; frame++ ) {
 
    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 )
      break ;
    //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
    if( frame > 6000 && frame < 36000 ) {
      if( (prefpmixfile->TotalEnergy < energymin[nrchanwav]) && ( prefpmixfile->TotalEnergy >= 1.0 ) )
	energymin[nrchanwav] =  prefpmixfile->TotalEnergy ;
      if( prefpmixfile->ZCRate < zcratemin[nrchanwav] )
	zcratemin[nrchanwav] = prefpmixfile->ZCRate ;
      if( prefpmixfile->ModLoud < modloudmin[nrchanwav] )
	modloudmin[nrchanwav] = prefpmixfile->ModLoud ;
    } 
    if( prefpmixfile->SumLoud > sumloudmax[nrchanwav] )
      sumloudmax[nrchanwav] = prefpmixfile->SumLoud ;
    if( prefpmixfile->ModLoud > modloudmax[nrchanwav] )
      modloudmax[nrchanwav] = prefpmixfile->ModLoud ;
  } ;
#ifdef TESTLEVEL
  fprintf( stderr, "%d frames used for finding min/max\n", usedframes ) ;
  fprintf( stderr, "... DONE\n\n" ) ;
#endif

  // memorize length of mixfile if shorter than the other files!
  usedframes = frame ; 

  // output of min energies. max sumloud, max modloud, max zcrate: 
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    fprintf( logfp, "channel %d: Min energy: %f\n", chan, energymin[chan] ) ;
    fprintf( logfp, "channel %d: Max sumloud: %f\n", chan, sumloudmax[chan] ) ;
    fprintf( logfp, "channel %d: Max modloud: %f\n", chan, modloudmax[chan] ) ;
    fprintf( logfp, "channel %d: Min modloud: %f\n", chan, modloudmin[chan] ) ;
    fprintf( logfp, "channel %d: Min zcrate: %f\n", chan, zcratemin[chan] ) ;
  } ;
  fprintf( logfp, "mixfile: Min energy: %f\n", energymin[nrchanwav] ) ;
  fprintf( logfp, "mixfile: Max sumloud: %f\n",  sumloudmax[nrchanwav] ) ;
  fprintf( logfp, "mixfile: Max modloud: %f\n", modloudmax[nrchanwav] ) ;
  fprintf( logfp, "mixfile: Min modloud: %f\n", modloudmin[nrchanwav] ) ;
  fprintf( logfp, "mixfile: Min zcrate: %f\n",  zcratemin[nrchanwav] ) ;


} ;


void InitAkuPreAndSpNonSp( char **argument, int usemltimeszcr ) { 

  // local variables
  int i ;

#ifdef TESTLEVEL
  fprintf( stderr, "Initializing AkuPre and SpNonSp of all %d channels and of MixFile '%s' ... \n", nrchanwav, argument[mixfileindex] ) ;
#endif

  // construct Preprocessing
  Tool = new CTool( chanfp );
  AkuPre = new CAkuPreSimple*[nrchanwav] ;
  AkuPreMixFile = new CAkuPreSimple( chanfp ) ;

  // construct SpNonSp
  SpNonSp = new CSpeechNonSpeech*[nrchanwav] ;
  SpNonSp2 = new CSpeechNonSpeech*[nrchanwav] ;
  SpNonSpMixFile = new CSpeechNonSpeech( (sizeof(sPreFeature)/sizeof(float))-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspmixcfgfp ) ;  

  // init channel specific akupre and spnonsp
  for( i=0 ; i<nrchanwav; i++ ) {
    AkuPre[i] = new CAkuPreSimple( chanfp ) ;
    //SpNonSp[i] = new CSpeechNonSpeech( PRESPECLEN+4, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; // Achtung hard coded length of feature file
    if( usemltimeszcr ) {
      SpNonSp[i] = new CSpeechNonSpeech(sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ;
      //SpNonSp2[i] = new CSpeechNonSpeech( PRESPECLEN+4, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp2 ) ; 
      SpNonSp2[i] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp2 ) ; 
    } else {
      SpNonSp[i] = new CSpeechNonSpeech( (sizeof(sPreFeature)/sizeof(float))-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ;
      //SpNonSp2[i] = new CSpeechNonSpeech( PRESPECLEN+4, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp2 ) ; 
      SpNonSp2[i] = new CSpeechNonSpeech( (sizeof(sPreFeature)/sizeof(float))-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp2 ) ; 
    } ;
 
  } ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif

 
} ;

int StartAkuPreAll( char **argument ) { 
  
  // local variables
  int chan ;

#ifdef TESTLEVEL
  fprintf( stderr, "Start AkuPre of all %d channels ... \n", nrchanwav ) ;
#endif

  for( chan=0 ; chan<nrchanwav ; chan++) {
#ifdef TESTLEVEL
    fprintf( stderr, "... %d\n", chan  ) ;
#endif    // start AkuPre
    if( AkuPre[chan]->startAkuPre( argument[(chan)+wavfileoffset] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argument[(chan)+wavfileoffset] ) ;
      return(-1) ;
    } ;
#ifdef TESTLEVEL
  } ;
  fprintf( stderr, "... DONE\n\n" ) ;
#endif
 
#ifdef TESTLEVEL
  fprintf( stderr, "Start AkuPre of Mixfile ...\n" ) ;
#endif
  if(AkuPreMixFile->startAkuPre( argument[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argument[mixfileindex] ) ;
    return(-1) ;
  } ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif

  return(_OK_) ;

} ;


int GetSpNonSpAndEnergy( char **argument, int normchan=1, int normchansecondpass=0, int normmix=0, float normchanperc=0.0 ) {

  // local variables
  int frame, chan, tmpframe, running, count ;
  ESpeechFlag *lastflag = new ESpeechFlag[nrchanwav+1] ;
  ESpeechFlag *lastflag2 = new ESpeechFlag[nrchanwav+1] ;

  // initialize array to remember status of last identified spnonsp-state per channel
  for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
    lastflag[chan] = P ;
    lastflag2[chan] = P ;
  }

#ifdef TESTLEVEL
  fprintf( stderr, "Get SpNonSp-Info and Energy of all channels and mixfile ...\n" ) ;
#endif

  // prepare to store framewise spnonsp-info in global arrays
  SpeechFlag = new ESpeechFlag*[nrframe] ;
  SpeechFlag2 = new ESpeechFlag*[nrframe] ;
  Energy = new float*[nrframe] ;
  //EnergyNoSmooth = new float*[nrframe] ;
  ModLoud = new float*[nrframe] ;
  for( frame=0 ; frame<nrframe ; frame++ ) { 
    SpeechFlag[frame] = new ESpeechFlag[nrchanwav+1] ; // incl mixfile
    SpeechFlag2[frame] = new ESpeechFlag[nrchanwav+1] ; // incl mixfile
    Energy[frame] = new float[nrchanwav+1] ;
    //EnergyNoSmooth[frame] = new float[nrchanwav+1] ;
    ModLoud[frame] = new float[nrchanwav+1] ;
  }
  // reset acoustic preprocessing for channels
  for( chan=0 ; chan<nrchanwav ; chan++) {
    // clear AkuPre
    AkuPre[chan]->clearAkuPre() ;
    if( AkuPre[chan]->startAkuPre( argument[(chan)+wavfileoffset] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argument[(chan)+wavfileoffset] ) ;
      delete[] lastflag ;
      return(_ERR_) ;
    } ;
  } ;

  AkuPreMixFile->clearAkuPre() ;
  if( AkuPreMixFile->startAkuPre( argument[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argument[mixfileindex] ) ;
    delete[] lastflag ;
    return(_ERR_) ;
  } ;

  // do loop over all channels (including mixfile) and collect speechnonspeech-info
  for( frame=0 ; frame<nrframe ; frame++ ) {
    // get feature vecs, prepare feature vecs by normalizing
    meanenergy = 0.0 ;
    running = 1 ;
    for( chan=0 ; chan<nrchanwav ; chan++) {
      prefp[chan] = AkuPre[chan]->getNextPre() ;
      if( prefp[chan] == 0 ) {
	nrframe = frame ;
	fprintf( stderr, "Unexpected EOF reached for file '%s' after %d frames\n", argument[(chan)+wavfileoffset], frame ) ;
	running = 0 ; 
	break ;
      } ;
   
      // do spnonsp-detection on unnormalized features!
      if( !normchansecondpass )
	SpeechFlag2[frame][chan] = SpNonSp2[chan]->detectandupdate( (float*)prefp[chan] ) ;

      // process undef frames if act decision is 'S' or 'P'
      if( SpeechFlag2[frame][chan] == S ) {
	tmpframe = frame-1 ;
	while( tmpframe>=0 && SpeechFlag2[tmpframe][chan] == U ) {
	  if( lastflag2[chan] == S )
	    //SpeechFlag[tmpframe][chan] = SUS ;
	    SpeechFlag2[tmpframe][chan] = S ;
	  else if( lastflag2[chan] == P ) 
	    //SpeechFlag[tmpframe][chan] = PUS ;
	    SpeechFlag2[tmpframe][chan] = S ;
	  tmpframe-- ;
	} ;
	lastflag2[chan] = S ;
      } else if(  SpeechFlag2[frame][chan] == P ) {
	tmpframe = frame-1 ;
	while( tmpframe>=0 && SpeechFlag2[tmpframe][chan] == U ) {
	  if( lastflag2[chan] == S )
	    //SpeechFlag[tmpframe][chan] = SUP ;
	    SpeechFlag2[tmpframe][chan] = P ;
	  else if( lastflag2[chan] == P ) 
	    //SpeechFlag[tmpframe][chan] = PUP ;
	    SpeechFlag2[tmpframe][chan] = P ;
	  tmpframe-- ;
	} ;
	lastflag2[chan] = P ;
      } ;

      prefp[chan]->ModLoudTimesZCRate = ((prefp[chan]->ModLoud-modloudmin[chan])/(modloudmax[chan]-modloudmin[chan])) * (prefp[chan]->ZCRate-zcratemin[chan]) ;
      if( normchan || normchanperc != 0.0 ) {
	// normalize channel features: (maxarrays either conatin max or percentile of max at this time!!!)
	prefp[chan]->TotalEnergy -= energymin[chan] ;
	prefp[chan]->SumLoud /= sumloudmax[chan] ;
	//prefp[chan]->ZCRate /= zcratemax[chan] ;    
	prefp[chan]->ModLoud /= modloudmax[chan] ;
      } ; 

      // calc mean energy over all channel energies
      meanenergy += prefp[chan]->TotalEnergy ;
    } ;

    // check if any soundfile met EOF!
    if( running == 0 ) {
      // stop processing soundfiles!
      nrframe = frame ;
      break ;
    } ;

    // normalize mixfile-features
    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 ) {
	fprintf( stderr, "Unexpected EOF reached for file '%s'\n", argument[mixfileindex] ) ;
	nrframe = frame ;
	break ;
    } ;
    
    // set special feature: produkt aus modloud und zcrate
    prefpmixfile->ModLoudTimesZCRate = ((prefpmixfile->ModLoud-modloudmin[nrchanwav])/(modloudmax[nrchanwav]-modloudmin[nrchanwav])) * (prefpmixfile->ZCRate-zcratemin[nrchanwav]) ;
    if( normmix ) {
      // normalize mixfile features with max:
      prefpmixfile->TotalEnergy -= energymin[nrchanwav] ;
      prefpmixfile->SumLoud /= sumloudmax[nrchanwav] ;
      prefpmixfile->ModLoud /= modloudmax[nrchanwav] ;
      //prefpmixfile->ZCRate /= zcratemax[nrchanwav] ;    
    } ;

    meanenergyohnemix = (meanenergy/nrchanwav) ;
    meanenergy += prefpmixfile->TotalEnergy ;
    meanenergy /= (nrchanwav+1) ;

    // normalize channel energies with mean over channels    
    for( chan=0 ; chan<nrchanwav ; chan++) {
      // two different methods to normalize channel energies!
      //prefp[chan]->TotalEnergy -= meanenergy ;
      if( normchan )
	prefp[chan]->TotalEnergy -= meanenergyohnemix  ;
      Energy[frame][chan] = prefp[chan]->TotalEnergy ;
      //EnergyNoSmooth[frame][chan] = prefp[chan]->TotalEnergy ;
      ModLoud[frame][chan] = prefp[chan]->ModLoud ;

    } ;

    // normalize mixfile energy with mean over channels
    if( normmix ) 
      prefpmixfile->TotalEnergy -= meanenergy ;
    Energy[frame][nrchanwav] = prefpmixfile->TotalEnergy ;
    //EnergyNoSmooth[frame][nrchanwav] = prefpmixfile->TotalEnergy ;
    ModLoud[frame][nrchanwav] = prefpmixfile->ModLoud ;

    // do SpNonSp-detection
    for( chan=0 ; chan<nrchanwav ; chan++) {
      // detect and update (if flag is set in cfg-file!)
      SpeechFlag[frame][chan] = SpNonSp[chan]->detectandupdate( (float*)prefp[chan] ) ;
      // do spnonsp-detection for second pass (with hopefully more sensitive (normalized) Gaussians!)
      if( normchansecondpass )
	SpeechFlag2[frame][chan] = SpNonSp2[chan]->detectandupdate( (float*)prefp[chan] ) ;

      // process undef frames if act decision is 'S' or 'P'
      if( SpeechFlag[frame][chan] == S ) {
	tmpframe = frame-1 ;
	while( tmpframe>=0 && SpeechFlag[tmpframe][chan] == U ) {
	  if( lastflag[chan] == S )
	    //SpeechFlag[tmpframe][chan] = SUS ;
	    SpeechFlag[tmpframe][chan] = S ;
	  else if( lastflag[chan] == P ) 
	    //SpeechFlag[tmpframe][chan] = PUS ;
	    SpeechFlag[tmpframe][chan] = S ;
	  tmpframe-- ;
	} ;
	lastflag[chan] = S ;
      } else if(  SpeechFlag[frame][chan] == P ) {
	tmpframe = frame-1 ;
	while( tmpframe>=0 && SpeechFlag[tmpframe][chan] == U ) {
	  if( lastflag[chan] == S )
	    //SpeechFlag[tmpframe][chan] = SUP ;
	    SpeechFlag[tmpframe][chan] = P ;
	  else if( lastflag[chan] == P ) 
	    //SpeechFlag[tmpframe][chan] = PUP ;
	    SpeechFlag[tmpframe][chan] = P ;
	  tmpframe-- ;
	} ;
	lastflag[chan] = P ;
      } ;
    } ;

    // detect upon current frame and update, if wanted (cfg-file!)
    SpeechFlag[frame][nrchanwav] = SpNonSpMixFile->detectandupdate( (float*) prefpmixfile ) ;
    
    // process undef frames if act decision is 'S' or 'P'
    if( SpeechFlag[frame][nrchanwav] == S ) {
      tmpframe = frame-1 ;
      while( tmpframe>=0 && SpeechFlag[tmpframe][nrchanwav] == U ) {
	if( lastflag[nrchanwav] == S )
	  //SpeechFlag[tmpframe][nrchanwav] = SUS ;
	  SpeechFlag[tmpframe][nrchanwav] = S ;
	else if( lastflag[nrchanwav] == P ) 
	  //SpeechFlag[tmpframe][nrchanwav] = PUS ;
	  SpeechFlag[tmpframe][nrchanwav] = S ;
	tmpframe-- ;
      } ;
      lastflag[nrchanwav] = S ;
    } else if(  SpeechFlag[frame][nrchanwav] == P ) {
      tmpframe = frame-1 ;
      count = 1 ;
      while( tmpframe>=0 && SpeechFlag[tmpframe][nrchanwav] == U ) {
	if( lastflag[nrchanwav] == S )
	  // eigentlich SUP
	  SpeechFlag[tmpframe][nrchanwav] = P ;
	else if( lastflag[nrchanwav] == P ) 
	  //SpeechFlag[tmpframe][nrchanwav] = PUP ;
	  SpeechFlag[tmpframe][nrchanwav] = P ;
	tmpframe-- ;
	count++ ;
      } ;
      lastflag[nrchanwav] = P ;
    } ;
  } ;

  // check if status of lastframe is 'U'
  for( chan=0 ; chan<nrchanwav ; chan++) {
    if( SpeechFlag[nrframe-1][chan] == U ) {
      tmpframe = nrframe-1 ;
      while( tmpframe>=0 && SpeechFlag[tmpframe][chan] == U ) {
	if( lastflag[chan] == S )
	  //SpeechFlag[tmpframe][chan] = SUP ;
	  SpeechFlag[tmpframe][chan] = P ;
	else if( lastflag[chan] == P ) 
	  //SpeechFlag[tmpframe][chan] = PUP ;
	  SpeechFlag[tmpframe][chan] = P ;
	tmpframe-- ;
      } ;
    }
  } ;
  if( SpeechFlag[nrframe-1][nrchanwav] == U ) {
    tmpframe = nrframe-1 ;
    count=1 ;
    while( tmpframe>=0 && SpeechFlag[tmpframe][nrchanwav] == U ) {
      if( lastflag[nrchanwav] == S )
	SpeechFlag[tmpframe][nrchanwav] = S ;
      else if( lastflag[nrchanwav] == P ) 
	//SpeechFlag[tmpframe][nrchanwav] = PUP ;
	SpeechFlag[tmpframe][nrchanwav] = P ;
      tmpframe-- ;
      count++ ;
    } ;
  }

  // SpeechFlag2:
  // check if status of lastframe is 'U'
  for( chan=0 ; chan<nrchanwav ; chan++) {
    if( SpeechFlag2[nrframe-1][chan] == U ) {
      tmpframe = nrframe-1 ;
      while( tmpframe>=0 && SpeechFlag2[tmpframe][chan] == U ) {
	if( lastflag2[chan] == S )
	  //SpeechFlag[tmpframe][chan] = SUP ;
	  SpeechFlag2[tmpframe][chan] = P ;
	else if( lastflag2[chan] == P ) 
	  //SpeechFlag[tmpframe][chan] = PUP ;
	  SpeechFlag2[tmpframe][chan] = P ;
	tmpframe-- ;
      } ;
    }
  } ;
  delete[] lastflag2 ;
  delete[] lastflag ;

#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif

  return(_OK_) ;
} ;

// also look at channel energies to decide on activity, only, if both, the HMM-detector and the energy threshold
// decide for speech: set SpeechFlag to 'S'
int CorrectSpNonSpWithEnergyThres( int nrthres, float *thres ) {
  
  // local variables
  int chan, frame ;

  if( (nrthres != 0) && (nrthres!=nrchanwav) ) {
    // eihter one global threshold for all channels or indiv thresholds for ALL channels
    fprintf( stderr, "Sorry, can not perform comparison with energy threshold!\n" ) ;
    fprintf( stderr, "You must either use one global threshold or one threshold for each channel!\n" ) ;
    return(_ERR_) ;
  } ;
    
  for( frame=0 ; frame<nrframe ; frame++ ) {
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      if( SpeechFlag[frame][chan] == S && Energy[frame][chan] <= (nrthres==0 ? (*thres) : thres[chan]) )
	SpeechFlag[frame][chan] = P ;
    } ;
  } ;

  return(_OK_) ;
}

int getNrActiveChan( int frame ) {
  int nractive = 0, chan ;

  for( chan=0 ; chan<nrchanwav ; chan++ )
    if( OverlapFlag[frame][chan] == 1 )
      nractive++ ;

  return( nractive ) ;
}


void ReviseOverlapFlagsWithChanInfo( ) {

  int nractive, nractivelast, start, maxcount, maxindex, chan, frame, i ;
  int *count = new int[nrchanwav] ;

  // look into first frame
  if( SpeechFlag[0][nrchanwav] >= PUS && getNrActiveChan( 0 ) == 0 ) {
    start = 0 ;
    for( chan=0 ; chan<nrchanwav ; chan++ )
      count[chan] = 0 ;
    // increase count of active channel
    if( Channel[frame] != -1 )
      count[Channel[frame]]++ ;
  };

  // look for regions, where mixfile indicates speech. but no channel does so!
  for( frame=1 ;frame<nrframe ; frame++ ) {

    // look for speech regions:
    if( SpeechFlag[frame][nrchanwav] >= PUS && SpeechFlag[frame-1][nrchanwav] >= PUS) {
      // we are in the middle of a speech portion

      if( getNrActiveChan( frame ) == 0 && getNrActiveChan( frame-1 ) > 0 ) {
	// starting point of interesting reason
	start = frame ;
	for( chan=0 ; chan<nrchanwav ; chan++ )
	  count[chan] = 0 ;
	// increase count of active channel
	if( Channel[frame] != -1 )
	  count[Channel[frame]]++ ;
      } else if( getNrActiveChan( frame ) > 0 && getNrActiveChan( frame-1 ) == 0 ) {
	// end point of interesting reason
	// evaluate counts
	maxindex = -1 ;
	maxcount = 0 ;
	for( chan=0 ; chan<nrchanwav ; chan++ )
	  if( count[chan] > maxcount ) {
	    maxcount = count[chan] ;
	    maxindex = chan ;
	  } ;
	// set flags of channel, which was found most often!
	for( i=frame-1 ; i>=start ; i-- )
	  OverlapFlag[i][maxindex] = 1 ;
	    
      } ;
    } else if( SpeechFlag[frame][nrchanwav] >= PUS && SpeechFlag[frame-1][nrchanwav] < PUS) {
      // we are at the beginning of a speech portion
      if( getNrActiveChan( frame ) == 0 ) {
	start = frame ;

	for( chan=0 ; chan<nrchanwav ; chan++ )
	  count[chan] = 0 ;
	// increase count of active channel
	if( Channel[frame] != -1 )
	  count[Channel[frame]]++ ;
      } ;
    } else if( SpeechFlag[frame][nrchanwav] < PUS && SpeechFlag[frame-1][nrchanwav] >= PUS) {
      // we are at the end of a speech portion
      if( getNrActiveChan( frame-1 ) == 0 ) {
	// evaluate counts
	maxindex = -1 ;
	maxcount = 0 ;
	for( chan=0 ; chan<nrchanwav ; chan++ )
	  if( count[chan] > maxcount ) {
	    maxcount = count[chan] ;
	    maxindex = chan ;
	  } ;
	// set flags of channel, which was found most often!
	for( i=frame-1 ; i>=start ; i-- )
	  OverlapFlag[i][maxindex] = 1 ;
      } 
    }
  }
} ; 




// go through activty arrays (OverlapFlag[][]) and determine maxima of modloud in 1-regions
// only if max is higher than tres, this is a speech region!!!
void ReviseOverlapFlagsWithModLoudThres( float thres ) {

  int frame, chan, i ;

  int* start = new int[nrchanwav] ;
  float* mlmax = new float[nrchanwav] ;

  // check first frame:
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    if( OverlapFlag[0][chan] == 1 ) {
      start[chan] = 0 ;
      mlmax[chan] = ModLoud[0][chan] ;
    } ;
  } ;

  // check rest:
  for( frame=1 ; frame<nrframe ; frame++ ) {

    for( chan=0 ; chan<nrchanwav ; chan++ ) {

      if( OverlapFlag[frame][chan] == 1 ) {
	if( OverlapFlag[frame-1][chan] == 0 ) {
	  // start of speech-region!
	  start[chan] = frame ;
	  // reset max
	  mlmax[chan] = ModLoud[frame][chan] ;
	} else {
	  // within speech-region: search for max modloud
	  if( ModLoud[frame][chan] > mlmax[chan] ) 
	    mlmax[chan] = ModLoud[frame][chan] ;
	}
      } else if( OverlapFlag[frame][chan] == 0 ) {
	if( OverlapFlag[frame-1][chan] == 1 ) {
	  // end of speech region
	  // evaluate mlmax
	  if( mlmax[chan] < thres ) {
	    // this is  probably not a speech region: reset flags
	    for( i=start[chan] ; i<=frame ; i++ )
	      OverlapFlag[i][chan] = 0 ;
	  } ;
	}
      } ;
    } ;
  } ;
 
  // check last frame
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    if( OverlapFlag[nrframe-1][chan] == 1 ) {
      if( mlmax[chan] < thres ) {
	// this is  probably not a speech region: reset flags
	for( i=start[chan] ; i<nrframe ; i++ )
	  OverlapFlag[i][chan] = 0 ;
      } ;
    } ;
  } ;

} ;


void GetActiveChannel( float ModLoudThres=0.25 ) {

  // local variables
  int i, count, chan, frame, maxchanenergyindex, isactive, nractive, activeindex = -1 ;
  float maxchanenergy, distmean, distmax, distmin ;


#ifdef TESTLEVEL
  fprintf( stderr, "\nDecide which channel is active at each timeframe ...\n" ) ;
#endif

  // loop over all frames: use SpeechFlags and Energies to decide which channel is active (no overlaps allowed!)
  // use spnonsp-info of mixfile to decide, whether there is speech or not and compare channels energies to 
  // find channel with highest energy, which is assumed to be active
  // prepare to store framewise channel-info (no overlaps!)
  Channel = new int[nrframe] ;
  //ChannelNoSmooth = new int*[nrframe] ;
  OverlapFlag = new int*[nrframe] ;
  for( frame=0 ; frame<nrframe ; frame++ ) {

    // prepare to store more than only one active channel per frame
    //Channel[frame] = new int[nrchanwav] ;
    //ChannelNoSmooth[frame] = new int[nrchanwav] ;
    OverlapFlag[frame] = new int[nrchanwav] ;
    Channel[frame] = -1 ;
    //ChannelNoSmooth[frame][chan] = -1 ;
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      OverlapFlag[frame][chan] = 0 ;
    } 

    // speech or nonspeech?
    if( SpeechFlag[frame][nrchanwav] >= PUS ) {
      // how many channelas are active
      nractive = 0 ;
      for( chan=0 ; chan<nrchanwav ; chan++ ) {
	if( SpeechFlag[frame][chan] >= PUS ) {
	  nractive++ ;
	  activeindex = chan ;
	} ;
      } ;

      if( nractive == 1 ) {
	// only one channel is active, pick this one
	Channel[frame] = activeindex ;
	//ChannelNoSmooth[frame][0] = activeindex ;
	OverlapFlag[frame][activeindex] = 1 ;
	//lastnractive = 1 ;
      } else if( nractive == 0 ) {

	// no channel was found active, only mixfile!!!
	// version 1: however, assume, that the mixfile-detection is right!!!
	// Version 1a:
	// so, search for max energy (over N frames) and pick the corresponding channel
	//maxchanenergy = -1.0e30 ;
	//maxchanenergyindex = -1 ;
	//float *SumEnergy = new float[nrchanwav] ;
	//int N = 25 ;
	//for( chan=0 ;chan<nrchanwav ; chan++) {
	//  SumEnergy[chan] = 0.0 ;
	//  for( i=(frame-N>=0? frame-N:0), count=0 ; (i<frame+N && i<nrframe) ; i++, count++ )
	//    SumEnergy[chan] += ModLoud[i][chan] ;
	//} ;
	//for( chan=0 ;chan<nrchanwav ; chan++) {
	//  if( SumEnergy[chan] > maxchanenergy ) {
	//    maxchanenergy = SumEnergy[chan] ;
	//    maxchanenergyindex = chan ;
	//  } ;
	//} ;
	//if( SumEnergy[maxchanenergyindex]/count >= ModLoudThres ) 
	//  OverlapFlag[frame][maxchanenergyindex] = 1 ;

	//Channel[frame] = ChannelNoSmooth[frame][0] = maxchanenergyindex ;
	
	// version 1b: believe in mixfile: 
	// search channels, which show high modified loudness values for at least N frames
	//int N = 25 ;
	//for( chan=0 ; chan<nrchanwav ; chan++ ) {
	//  isactive = 1 ;
	//  for( i=frame ; (i<frame+N && i<nrframe) ; i++ )
	//    if( ModLoud[i][chan] < ModLoudThres )
	//      isactive = 0 ;
	//  if( isactive ) 
	//    OverlapFlag[frame][chan] = 1 ;
	//} ;

	// version 1c: believe in mixfile:
	// take the (hopefully) more sensitive SpeechFlag2-Array instead of the SpeechFlag-Array
	//for( chan=0 ; chan<nrchanwav ; chan++ ) 
	//  if( SpeechFlag2[frame][chan] >= PUS ) 
	//    OverlapFlag[frame][chan] = 1 ;

	// version 1d: believe in mixfile, like 1c, but only accept the channel with the max loudness!!!
	maxchanenergy = -1.0e30 ;
	maxchanenergyindex = -1 ;
	for( chan=0 ;chan<nrchanwav ; chan++) {
	  if( ModLoud[frame][chan] > maxchanenergy && SpeechFlag2[frame][chan] >= PUS ) {
	    maxchanenergy = ModLoud[frame][chan] ;
	    maxchanenergyindex = chan ;
	  } ;
	} ;
	if( maxchanenergyindex != -1 ) {
	  OverlapFlag[frame][maxchanenergyindex] = 1 ;
	  Channel[frame] = maxchanenergyindex ;
	  //ChannelNoSmooth[frame][0] = maxchanenergyindex ;
	} 

	// version 2: believe in channel specific detectors: set Channel to -1 (Pause!)
	//Channel[frame] =-1 ;
	//ChannelNoSmooth[frame][0] =-1 ;
	//SpeechFlag[frame][nrchanwav] = P ;

	//lastnractive = 0 ;
      } else {
	// speech in more than one channel: search for max modloud //TP 1.2.2001
	// also keep track of the distance between maximum and the other channels
	maxchanenergy = -1.0e30 ;
	maxchanenergyindex = -1 ;
	distmean = 0.0 ;
	distmax = 0.0 ;
	distmin = 1.0e30 ;
	for( chan=0 ;chan<nrchanwav ; chan++) {
	  // skip all channels in which no speech is detected!!!
	  if( SpeechFlag[frame][chan] < PUS )
	    continue ;
	  if( ModLoud[frame][chan] > maxchanenergy ) {
	  //if( Energy[frame][chan] > maxchanenergy ) {
	    maxchanenergy = ModLoud[frame][chan] ;
	    //maxchanenergy = Energy[frame][chan] ;
	    maxchanenergyindex = chan ;
	  } ;
	} ;
	Channel[frame] = maxchanenergyindex ;
	//ChannelNoSmooth[frame][0] = maxchanenergyindex ;

	// calculate distmean, distmax and distmin to the other active channels
	//for( chan=0 ;chan<nrchanwav ; chan++) {
	


	// search for further candidates for active channels by searching SpeechFlags
	int j;
	for( chan=0 ;chan<nrchanwav ; chan++) {
	  // is this the channel with the max?
	  if( chan == maxchanenergyindex ) {
	    OverlapFlag[frame][chan] = 1 ; // just mark this channel as possible overlap!!!
	    continue ;
	  } ;
	  // is loudness higher than 0.8 * max? Yes, new candidate found!!!
	  //if( SpeechFlag[frame][chan] >= PUS && ( ModLoud[frame][chan] >= ModLoud[frame][maxchanenergyindex] * 0.8 ) ) {
	  nractive = 1 ;
	  if( SpeechFlag[frame][chan] >= PUS ) {
	    OverlapFlag[frame][chan] = 1 ; // just mark this channel as possible overlap!!!
	    nractive++ ;
	  } ;
	}
	//lastnractive = nractive ;
      } ;

    } else {
      // nonspeech
      Channel[frame] = -1 ;
      //ChannelNoSmooth[frame][0] = -1 ;
      SpeechFlag[frame][nrchanwav] = P ;
    } 
  } ;

#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif
}


void SmoothOverlapInfo( int SmoothBufLen ) {

  // local variables
  int *SmoothBuf = new int[SmoothBufLen] ;
  int *SortBuf = new int[SmoothBufLen] ;
  int writepointer, i, chan, frame, outframe, j, k ;

  if( SmoothBufLen > 0 ) {
#ifdef TESTLEVEL
    fprintf( stderr, "Smoothing OverlapInfo... \n\n" ) ;
#endif

    // smmoth all channels
    for( chan=0 ;chan<nrchanwav ; chan++ ) {
      for( i=0 ; i<SmoothBufLen ; i++ ) {
	SmoothBuf[i] = 10 ;
	SortBuf[i] = 10 ;
      } ;
      writepointer = 0 ;

      // smooth all frames of each channel
      for( frame=0 ; frame<nrframe+SmoothBufLen/2 ; frame++ ) {
	// define which frame will be output (smoothing results in frame shift!)
	outframe = frame-SmoothBufLen/2 ;

	// enter current value into buufer
	SmoothBuf[writepointer] = (frame<nrframe ? OverlapFlag[frame][chan] : 0 ) ;

	// increase writepointer
	writepointer = (writepointer < SmoothBufLen-1 ? writepointer+1 : 0 ) ;

	// Sort all i=0...N buffer entries into SortBuf
	for( i=0 ; i<SmoothBufLen ; i++ ) 
	  SortBuf[i] = 10 ;

	for( i=0 ; i<SmoothBufLen ; i++ ) {
	  // compare current buffer entry with all sortbuf entries 
	  for( j=0 ; j<SmoothBufLen ; j++ ) {
	    if( SmoothBuf[i] < SortBuf[j] ) {
	      // enter this smmothbuf-entry at this position in SortBuf
	      for( k=SmoothBufLen-1 ; k>j ; k-- )
		SortBuf[k] = SortBuf[k-1] ;
	      SortBuf[j] = SmoothBuf[i] ;
	      break ;
	    } ;
	  } ;
	} ;
	
	// take median
	if( outframe >= 0 )
	  OverlapFlag[outframe][chan] = SortBuf[SmoothBufLen/2] ;
      } ;

    } ;
	
  } ;

} ;




// smooth Channelinfo: SmoothBufLen defines nr of frames over which the smoothing is done
// 'lowThres' defines how many switches in the most probable channel have to occur to be a possible candidate 
// for an overlapped region, only if later on more than 'upThres' switches occur, these frames are labelled
// as overlap-region!!! 
void SmoothChannelInfo( int SmoothBufLen, float lowThresPercent, float upThresPercent ) { 

  // local variables
  int *SmoothBuf = new int[SmoothBufLen] ;
  //int *SmoothBufNoSmooth = new int[SmoothBufLen] ;
  int i, j, k, sum, lastchan, frame, chan, outframe, last, next, writepointer = 0 ;
  int *ChanIdSort = new int[nrchanwav+1] ;
  int overlapmax = -1 ;
  int overlapmin = SmoothBufLen ;

  // smooth framewise channelinfo
  for( i=0 ; i<SmoothBufLen ; i++ ) {
    SmoothBuf[i] = -1 ;
    //SmoothBufNoSmooth[i] = -1 ;
  } ;

  // smooth info???
  //Overlap = new int[nrframe] ;
  ChanCount = new int*[nrframe];

  if( SmoothBufLen > 0 ) {
#ifdef TESTLEVEL
    fprintf( stderr, "Smoothing MaxChanInfo... \n\n" ) ;
#endif

    for( frame=0 ; frame<nrframe+SmoothBufLen/2 ; frame++ ) {
      // define which frame will be output (smoothing results in frame shift!)
      outframe = frame-SmoothBufLen/2 ;

      // allocate memory
      ChanCount[frame] = new int[nrchanwav+1] ; // 0-th channel means pause!

       // write current entry into buffer
      SmoothBuf[writepointer] = (frame<nrframe ? Channel[frame] : -1 ) ;
      //SmoothBufNoSmooth[writepointer] = (frame<nrframe ? ChannelNoSmooth[frame][0] : -1 ) ;

      // increase writepointer
      writepointer = (writepointer < SmoothBufLen-1 ? writepointer+1 : 0 ) ;

      /*
      // create overlap-array: count how often there is a switching between different channels in the smoothbuf
      //Overlap[frame] = 0 ;
      if( outframe >= 0 ) {
	Overlap[outframe] = 0 ;
	for( j=0, k=writepointer ; j<SmoothBufLen-1 ; j++, k=(k<SmoothBufLen-1 ? k+1 : 0) ) 
	  if( SmoothBufNoSmooth[k] != SmoothBufNoSmooth[(k<SmoothBufLen+1? k+1 : 0 )] )
	    Overlap[outframe]++ ;
	if( Overlap[outframe] > overlapmax ) 
	  overlapmax = Overlap[outframe] ;
	if( Overlap[outframe] < overlapmin ) 
	  overlapmin = Overlap[outframe] ;
	//fprintf( stderr, "[%d]: %d\t%f\n", outframe, Overlap[outframe], Energy[outframe][nrchanwav]  ) ;
      } ;
      */

      // evaluate buffer: only do that, when at least SmoothBufLen/2+1 entries are in the buffer!
      // and when the detector decided for 'speech'
      if( (frame >= SmoothBufLen/2) && (Channel[outframe] >= 0) ) {
	
	// reset channel counts:
	int j ;
	for( chan=0 ; chan<nrchanwav+1 ; chan++ )
	  ChanCount[outframe][chan] = 0 ;
	
	// count how often each channel is seen in buffer:
	for( j=0 ; j<SmoothBufLen ; j++ )
	  ChanCount[outframe][ SmoothBuf[j]+1 ] ++ ; // increase count for the observed channel (channel nr +1, since -1=pause)
	
	// sort channel-numbers according to the channelcount in the buffer!!!
	// reset array
	for( chan=0 ; chan<nrchanwav+1 ; chan++ ) 
	  ChanIdSort[chan] = -1 ;
	
	// go through ChanCounts and sort corresponding channel ids according to the count (highest first)
	//for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
	for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
	  for( j=0 ; j<nrchanwav+1 ; j++ ) {
	    // compare count of channel chan to count of j-th channelid in sorted array
	    if( ChanIdSort[j] == -1 ) {
	      ChanIdSort[j] = chan ;
	      break ;
	    } else if( ChanCount[outframe][chan] > ChanCount[outframe][ChanIdSort[j]] ) {
	      // enter chan at j-th place in sorted array
	      for( int k=nrchanwav ; k>j ; k-- )
		ChanIdSort[k] = ChanIdSort[k-1] ;
	      ChanIdSort[j] = chan ;
	      break ;
	    } ;
	  } ;
	} ;
	
	// enter all those Channels which together account for at least half of the buffer entries in Channel array,
	sum = 0 ;
	lastchan=0 ; /// default: take most frequent channel, even if not accounting for half the buffer entries
	for( chan=0 ; chan<nrchanwav+1 ; chan++ ) {
	  sum += ChanCount[outframe][ ChanIdSort[chan] ] ;
	  if( sum > (SmoothBufLen/2) ) {
	    lastchan = chan ;
	    break ;
	  } ;
	} ;
	
	// Version1: comment in Version1 and Version2 to activate Version1!!!!!!!!! 
	lastchan = 0 ; // just take most frequent channel!!---------------------------------------------------------------------------------

	// defunct: Channel[] does only contain ONE active channel at atime
	// Version2: just add those channels which are needed to get SmoothBufLen/2 'speech'-frames in the buffer
	//for( chan=0 ; chan<=lastchan ; chan++ ) {
	//  Channel[outframe][chan] = ChanIdSort[chan] -1 ; // subtract offset to get 'real' channel ids (+1 was added before)
	//} ;
	// set SpeechFlag of mixfile to P, if most frames are -1
	if( Channel[outframe] == -1 )
	  SpeechFlag[outframe][nrchanwav] = P ;


	//Version 3: just add those channels which account for more than SmoothBufLen/nrchanwav frames in  the buiffer
	//for( chan=0, j=0 ; chan<Overlap[outframe]+1 ; chan++ ) {
	//  if( ChanIdSort[chan] != 0 ) {
	//    Channel[outframe][j] = ChanIdSort[chan] -1 ; // subtract offset to get 'real' channel ids (+1 was added before)
	//    j++ ;
	//  } ;
	//} ;
      } ;
    } ;
      
    /*
    int lowThres = (overlapmax - overlapmin) * lowThresPercent ;
    int upThres  = (overlapmax - overlapmin) * upThresPercent ;
    if( lowThres >= upThres )
      upThres = lowThres +1 ;
    //fprintf( stderr, "ovlmax %d, ovlmin %d, lowthres %d, upthres %d\n", overlapmax, overlapmin, lowThres, upThres ) ;

    // process Overlap-Array: make binary entries: 1: overlapped region, 0: nonoverlapped region!
    //int lowThres = SmoothBufLen/10 ;
    // convert counts to: 0: if count below lower threshold, 1: if count between lower and upper thres, 2: if above upper thres 
    for( frame=0 ; frame<nrframe ; frame++ ) {
      if( Overlap[frame] < lowThres )
	Overlap[frame] = 0 ;
      else if( Overlap[frame] < upThres )
	Overlap[frame] = 1 ;
      else 
	Overlap[frame] = 2 ;

      //fprintf( stderr, "%d: %d\n", frame, Overlap[frame]);

    } ;

    last=0 ;
    for( frame=0 ; frame<nrframe ; frame++ ) {
      if( Overlap[frame] == 1 && last == 2 ) {
	// current frame is '1' and last frame was '2' -> current = '2'
	Overlap[frame] = last ;
	//fprintf( stderr, "%d: %d\n", frame, Overlap[frame]);
      } else if( Overlap[frame] == 1 ) {
	// current frame is '1' and last was not '2' (last was '0')
	// look into future until next non-'1' is found
	next = 0 ;
	if( frame>=nrframe-2 ) {
	  Overlap[frame] = 0 ;
	  //fprintf( stderr, "%d: %d\n", frame, Overlap[frame]);
	  continue ;
	} ;
	for( j=frame+1 ; j<nrframe ; j++ ) {
	  if( Overlap[j] != 1 ) {
	    // next non-'1' found, remember value and break loop;
	    next = Overlap[j] ;
	    break ;
	  } ;
	}
	// set all 
	for( k=frame ; k<=j ; k++ ) {
	  Overlap[k] = next ;
	  //fprintf( stderr, "%d: %d\n", k, Overlap[k]);
	} ;
	frame = j ;
      } else {
	// current frame is either '0' or '2'
	last = Overlap[frame] ;
	//fprintf( stderr, "%d: %d\n", frame, Overlap[frame]);
      } ;
    }
    */
    
  } ;
#ifdef TESTLEVEL
  fprintf( stderr, "... DONE\n\n" ) ;
#endif
} ;



void PrintResults( int overlapwanted, EOutPutMode outputmode, char *speechlabel, char *nonspeechlabel, char *overlaplabel="OVL", float OffSet=0.05 ) {

  // local variables 
  int nractive, nractivelast, lastspeaker, frame, chan ; 
  
  // OUTPUT:
  if( overlapwanted ) {
#ifdef TESTLEVEL
    fprintf( stderr, "Output WITH overlap\n\n" ) ;
#endif
    
    // do overlap-output:
    // Version 1, from overlap-info:
    switch(outputmode) {
    case XML_MULT:
      for( frame=0 ; frame<nrframe ; frame++ ) {
	
	// first frame???
	if( frame == 0 ) {
	  // first frame of a dialogue:
	  // global track:
	  fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"default\"/>\n", frame/100.0 ) ;
	  if( SpeechFlag[frame][nrchanwav] >= PUS ) {
	    fprintf( globoutfp, "%s\n", speechlabel ) ;
	  } else {
	    fprintf( globoutfp, "%s\n", nonspeechlabel ) ;
	  }
	  
	  // process each channel-track
	  for( chan=0 ; chan<nrchanwav ; chan++ ) {
	    if (OverlapFlag[frame][chan] == 1 ) {
	      fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"%s\"/>\n", frame/100.0, ChanId[chan] ) ;
	      fprintf( globoutfp, "%s\n", speechlabel ) ;
	      fprintf( outfp[chan], "<Sync time=\"%.3f\"/>\n", frame/100.0 ) ;
	      fprintf( outfp[chan], "%s\n", speechlabel ) ;
	    } else {
	      fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"%s\"/>\n", frame/100.0, ChanId[chan] ) ;
	      fprintf( globoutfp, "%s\n", nonspeechlabel ) ;
	      fprintf( outfp[chan], "<Sync time=\"%.3f\"/>\n", frame/100.0 ) ;
	      fprintf( outfp[chan], "%s\n", nonspeechlabel ) ;
	    } ;
	  }
	} else {
	  // all other frames:
	  // process global track
	  if( SpeechFlag[frame][nrchanwav] >= PUS && SpeechFlag[frame-1][nrchanwav] < PUS ) {
	    // mark global channel as active in the global track
	    fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"default\"/>\n", frame/100.0 ) ;
	    fprintf( globoutfp, "%s:\n", speechlabel ) ;
	  } else if( SpeechFlag[frame][nrchanwav] < PUS && SpeechFlag[frame-1][nrchanwav] >= PUS ) {
	    // mark global channel as inactive in the global track
	    fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"default\"/>\n", (frame/100.0 + OffSet) ) ;
	    fprintf( globoutfp, "%s\n", nonspeechlabel ) ;
	  } 
	  // process channels:
	  for( chan=0 ; chan<nrchanwav ; chan++ ) {
	    // look into channels:
	    if( OverlapFlag[frame][chan] == 1 && OverlapFlag[frame-1][chan] == 0 ) {
	      // this particular channel has switched from pause to speech		
	      fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"%s\"/>\n", frame/100.0, ChanId[chan] ) ;
	      fprintf( globoutfp, "%s\n", speechlabel ) ;
	      fprintf( outfp[chan], "<Sync time=\"%.3f\"/>\n", frame/100.0 ) ;
	      fprintf( outfp[chan], "%s\n", speechlabel ) ;
	    } else if( OverlapFlag[frame][chan] == 0 && OverlapFlag[frame-1][chan] == 1 ) {
	      // this particular channel has speech to pause
	      fprintf( globoutfp, "<Sync time=\"%.3f\" chan=\"%s\"/>\n", frame/100.0 + OffSet, ChanId[chan] ) ;
	      fprintf( globoutfp, "%s\n", nonspeechlabel ) ;
	      fprintf( outfp[chan], "<Sync time=\"%.3f\"/>\n", frame/100.0 + OffSet ) ;
	      fprintf( outfp[chan], "%s\n", nonspeechlabel ) ;  //// ????????? nonspeech?????????
	    }
	  } ;	 		
	  
	} ;	    
      }
      break ;
    } ;
  } else {
#ifdef TESTLEVEL
    fprintf( stderr, "Output WITHOUT overlap\n\n" ) ;
#endif

  
    switch(outputmode) {
    case FrameWise:
     
      break ;
    case XML:

         
       break ;

 
    case XML_MULT:
      
  
      break ;
    } ;
   
  } ;
  

}





void PrintFooter( EOutPutMode outputmode ) {
  // local variables
  int i ;

    
  // header and footer:
  switch( outputmode ) { 
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry
      fprintf( outfp[i], "</Turn>\n" ) ;
      fprintf( outfp[i], "</Section>\n" ) ;
      fprintf( outfp[i], "</Episode>\n" ) ;
      fprintf( outfp[i], "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( outfp[i] ) ;
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0  ) ;
    } ;
     // finish last entry
      fprintf( globoutfp, "</Turn>\n" ) ;
      fprintf( globoutfp, "</Section>\n" ) ;
      fprintf( globoutfp, "</Episode>\n" ) ;
      fprintf( globoutfp, "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( globoutfp ) ;
      fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( globoutfp, "<Episode>\n" ) ;
      fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0 ) ;
      fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0  ) ;
    break ;

  case XML_MULT:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry
      fprintf( outfp[i], "</Turn>\n" ) ;
      fprintf( outfp[i], "</Section>\n" ) ;
      fprintf( outfp[i], "</Episode>\n" ) ;
      fprintf( outfp[i], "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( outfp[i] ) ;
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13-icsi.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0  ) ;
    } ;
     // finish last entry
      fprintf( globoutfp, "</Turn>\n" ) ;
      fprintf( globoutfp, "</Section>\n" ) ;
      fprintf( globoutfp, "</Episode>\n" ) ;
      fprintf( globoutfp, "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( globoutfp ) ;
      fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13-icsi.dtd\">\n" ) ;
      fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( globoutfp, "<Episode>\n" ) ;
      fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0 ) ;
      fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (nrframe-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;
}




//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i, chan, frame, tmpframe ;
  // open files
  if( argc < 5) 
    {   fprintf(stderr,"usage: spdtest <akupre-cfg-file> <spnonsp-chan-cfg-file> <spnonsp-chan-cfg-file-secondpass> <spnonsp-mix-cfg-file> <wavfile1> <wavfile2> ... <wavfileN> <mixfile> <outfile>\n");
    return _ERR_;
    }
  
  // set filepointers
  chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open akupre-cfg-file:", argv[1]);
  spnonspchancfgfp = fopen( argv[2], "r+" );
  IF_ERROR_RETURN_S( spnonspchancfgfp == 0, "can't open first channel-specific spnonsp-cfg-file:", argv[2]);
  spnonspchancfgfp2 = fopen( argv[3], "r+" );
  IF_ERROR_RETURN_S( spnonspchancfgfp2 == 0, "can't open second channel-specific spnonsp-cfg-file:", argv[3]);
  spnonspmixcfgfp = fopen( argv[4], "r+" );
  IF_ERROR_RETURN_S( spnonspmixcfgfp == 0, "can't open spnonsp-cfg-file:", argv[4]);
  
  nrchanwav = argc-7 ;
  wavfileoffset = 5 ;
  totnrwav = nrchanwav+1 ; // includes mixfile 
  mixfileindex = argc-2 ;
  outfileindex = argc-1 ;

  // get ChanIds 
  int MaxChanIdLen = 2 ;
  ChanId = new char*[nrchanwav] ;
  char *tmpchar ;
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    // allocate memory
    ChanId[chan] = new char[MaxChanIdLen] ;
  
    // parse filenames of wav-files
    tmpchar = strrchr( argv[wavfileoffset+chan], '.' ) ;
    tmpchar--;
    strncpy( ChanId[chan], tmpchar, 1 ) ;
    ChanId[chan][1] = '\0'  ;

    //fprintf( stderr, "ChanId[%d]=%s\n", chan, ChanId[chan] ) ;
  } ;

    
  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD, XML_MULT", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;

  CParBool NormSpecialFeatChan( "Tools.NormSpecialFeatChan", chanfp, 1 ) ;
  CParBool NormSpecialFeatChanSecondPass( "Tools.NormSpecialFeatChanSecondPass", chanfp, 0 ) ;
  CParFloat NormSpecialFeatChanPerc( "Tools.NormSpecialFeatChanPerc", chanfp, 0.0 ) ; // use percentile of max instead of max!
  CParBool NormSpecialFeatMix( "Tools.NormSpecialFeatMix", chanfp, 0 ) ;
  CParBool UseModLoudTimesZCRate( "Tools.UseModLoudTimesZCRate", chanfp, 1 ) ;

  CParInt SmoothWindowLen( "Tools.Result.SmoothWindowLen", chanfp, 41 ) ;
  CParBool OverlapWanted( "Tools.Result.OverlapWanted", chanfp, 0 ) ;
  
  CParBool SmoothEnergyWanted( "Tools.Result.SmoothEnergyWanted", chanfp, 0 ) ;
  CParInt  SmoothEnergyDelta( "Tools.Result.SmoothEnergyDelta", chanfp, 0 ) ;

  CParBool UseEnergyToCorrectSpnonsp( "Tools.UseEnergyToCorrectSpnonsp", chanfp, 0 ) ;
  CParBool UseIndividualThres( "Tools.UseIndividualThres", chanfp, 0 ) ;

  CParFloat ModLoudMaxThres( "Tools.ModLoudMaxThres", chanfp, 0.0 ) ;

  CParFloat  ChanSwitchRateUpThresPercent( "Tools.ChanSwitchRateUpThresPercent", chanfp, 0.8 ) ;
  CParFloat  ChanSwitchRateLowThresPercent( "Tools.ChanSwitchRateLowThresPercent", chanfp, 0.5 ) ;

  CParBool CtrlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;


  // initialize acoustic preprocessing and speech non speech detection of all channels
  InitAkuPreAndSpNonSp( argv, UseModLoudTimesZCRate.Par ) ;

  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  

  // start preprocessing of all channels
  if ( StartAkuPreAll( argv ) == -1 ) {
    fprintf( stderr, "ERROR: Could not start acoustic preprocessing of all channels!\n" ) ;
    exit(-1);
  } ;
  
  // open outfiles
  char tmp[MaxLineLength] ;
  outfp = new FILE*[nrchanwav] ;
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    sprintf( tmp, "%s.%d.%s", argv[argc-1], chan, OutputMode == FrameWise ? "fwise" : "trs" ) ;
    outfp[chan] = fopen( tmp, "w" ) ;
    if( outfp[chan] == 0 ) {
      ERROR_MSG_S("can't open %d-th outfile %s!", chan, tmp );
      outfp[chan] = stderr ;
    }
    
  } ;
  globoutfp = fopen( argv[argc-1], "w" ) ;

  sprintf( tmp, "%s.log", argv[argc-1] ) ;
  logfp = fopen( tmp, "w" ) ;

  // output headers into all outfiles
  PrintHeader(OutputMode) ;
   
  // initialize array of pointers to feature vectors
  prefp = new SPreFeature*[nrchanwav];
  

  // determine minima resp maxima of special features!
  GetMinAndMax( NormSpecialFeatChanPerc.Par ) ;

  // count frames of mixfile to determine length
  nrframe = CountFrames( argv ) ;


  GetSpNonSpAndEnergy( argv, NormSpecialFeatChan.Par, NormSpecialFeatChanSecondPass.Par, NormSpecialFeatMix.Par, NormSpecialFeatChanPerc.Par ) ;

  if( CtrlOutput.Par ) {
    fprintf( stderr, "ENERGIES before smoothing\n" ) ;
    PrintEnergyArray( stderr ) ;
    PrintModLoudArray( stderr ) ;
  } ;

  if( SmoothEnergyWanted.Par ) {
    SmoothEnergies( SmoothEnergyDelta.Par) ;
    SmoothModLoud( SmoothEnergyDelta.Par ) ;
  } 

  if( CtrlOutput.Par ) {
    fprintf( stderr, "ENERGIES after smoothing\n" ) ;
    PrintEnergyArray( stderr ) ;
    PrintModLoudArray( stderr ) ;
  } ;

  // overwrite spnonsp-info for mixfile by 'or-ing' together channel specific spnonsp-info!!! -------------------------
  //for( frame=0 ; frame<nrframe ; frame++ ) {
  //SpeechFlag[frame][nrchanwav] = P ;
  //for( chan=0 ; chan<nrchanwav ; chan++ )
  //  if( SpeechFlag[frame][chan] >= PUS )
  //SpeechFlag[frame][nrchanwav] = S ;
  //} ;

  if( CtrlOutput.Par ) {
    fprintf( stderr, "SPEECHFLAGS before energy thresholding\n" ) ;
    PrintSpeechFlagArray( stderr ) ;
  } ;

  if( UseEnergyToCorrectSpnonsp.Par ) {
    if( UseIndividualThres.Par ) {
      // call matlab script to calcultae Thresholds: not yet implemented
      float *Thres = new float[nrchanwav] ;
      for( chan=0 ; chan<nrchanwav ; chan++ )
	Thres[chan] = -5.0 ;
      CorrectSpNonSpWithEnergyThres( nrchanwav, Thres ) ;
      delete[] Thres ;
    } else {
      // use global Threshold
      float globThres = -5.0 ;
      CorrectSpNonSpWithEnergyThres( 0, &globThres ) ;
    }
  } ;

  if( CtrlOutput.Par ) {
    fprintf( stderr, "SPEECHFLAGS after energy thresholding\n" ) ;
    PrintSpeechFlagArray( stderr ) ;
  } ;

  // serach for active channel for each frame: one channel at a time only!!!!
  GetActiveChannel() ;
  
  if( CtrlOutput.Par ) {
    fprintf( stderr, "Active CHANNEL before smoothing\n" ) ;
    PrintChannelArray( stderr ) ;
    fprintf( stderr, "Active OVERLAPFLAG before smoothing\n" ) ;
    PrintOverlapFlagArray( stderr ) ;
  } ;

  SmoothChannelInfo( SmoothWindowLen.Par, ChanSwitchRateLowThresPercent.Par, ChanSwitchRateUpThresPercent.Par ) ;
  //SmoothOverlapInfo( 15 ) ;


  // version 1: just use a threshold on the max modloud found on spurts to reject them.
  // hence: do more restrict speech-nonspeech-detection (when uncertain -> nonspeech
  ReviseOverlapFlagsWithModLoudThres( ModLoudMaxThres.Par ) ;

  // version 2: use ChanInfo: output of the channel with the most hits, 
  //            when mixfile indicates speech, but no channel indicates speech
  // hence, do more lax speech nonspeech detection (when uncertain -> speech)
  //ReviseOverlapFlagsWithChanInfo( ) ;

  if( CtrlOutput.Par ) {
    fprintf( stderr, "OVERLAPFLAG after revision (modloudthres for max)\n" ) ;
    PrintOverlapFlagArray( stderr ) ;
  } ;


  if( CtrlOutput.Par ) {
    fprintf( stderr, "Active CHANNEL after smoothing\n" ) ;
    PrintChannelArray( stderr ) ;
    fprintf( stderr, "Active OVERLAPFLAG after smoothing\n" ) ;
    PrintOverlapFlagArray( stderr ) ;
  } ;

  // use deafult overlaplabel!!!
  PrintResults( OverlapWanted.Par, OutputMode, SpeechLabel.Par, NonspeechLabel.Par) ;


  PrintFooter( OutputMode ) ;





#ifdef TESTLEVEL
  fprintf( stderr, "Free memory ...\n", nrchanwav, argv[argc-1] ) ;
#endif

    
  for( i=0 ; i<nrchanwav ; i++ ) {
    fclose( outfp[i] ) ;    
    if( AkuPre[i] != 0 ) {
      AkuPre[i]->clearAkuPre();
      delete AkuPre[i] ;
    } ;
  } ;
  delete[] AkuPre ;

  AkuPreMixFile->clearAkuPre() ;
  delete   AkuPreMixFile ;
  
  if( Channel != 0 )
    delete[] Channel ;
  if( energymin != 0 )
    delete[] energymin ;
  if( prefp != 0 )
    delete[] prefp ;
  
  //AkuPre.appendPar( stderr ) ;
#ifdef TESTLEVEL
  fprintf( stderr, "Done\n", nrchanwav, argv[argc-1] ) ;
#endif
  return _OK_;
}
