/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

// nr of frames which are used to find minima and maxima of energy, totloud, modloud and zcrate in first pass!!!
#define NrFramesMax 600

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i, j,sum ;
  // open files
  if( argc < 4) 
    {   fprintf(stderr,"usage: spdtest <cfg-file> <wavfile1> <thres1> <wavfile2> <thres2>... <wavfileN> <thresN> <mixfile> <outfile>\n");
    return _ERR_;
    }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open cfg-file:", argv[1]);
  
  int nrchanwav = (argc-4)/2 ; // only channel wavfiles are counted
  int totnrwav = nrchanwav+1 ; // includes mixfile 

  // construct Preprocessing
  CTool      Tool( chanfp );
  CAkuPre    **AkuPre = new CAkuPre*[nrchanwav] ;
  CAkuPre    *AkuPreMixFile = new CAkuPre( chanfp ) ;
  
  for( i=0 ; i<nrchanwav; i++ ) {
    AkuPre[i] = new CAkuPre( chanfp ) ;
  } ;
  

  // read Tool-related params

  // format output
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParInt  SmoothWinLen( "Tools.Result.SmoothWindowLen", chanfp, 41 ) ;
  CParBool OverlapOutput( "Tools.Result.OverlapWanted", chanfp, 0 ) ;

  // Label:
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  
  
  // start AkuPre of channels
  for( i=0 ; i<nrchanwav ; i++)
    if( AkuPre[i]->startAkuPre( argv[(i*2)+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open%d-th  sigfile '%s'\n", i, argv[(i*2)+2] ) ;
      exit(-1) ;
    } ;
  
  // start AkuPre of Mixfile
  if(AkuPreMixFile->startAkuPre( argv[argc-2] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argv[argc-2] ) ;
    exit(-1) ;
  } ;

  // set thresholds for channels from commandline args
  float *Thres = new float[nrchanwav] ;
  for( i=0 ; i<nrchanwav ; i++) {
    Thres[i] = strtod( argv[(i*2)+3], NULL ) ;
  } ;
  
  // open outfiles
  char tmp[MaxLineLength] ;
  FILE **outfp = new FILE*[argc-3] ;
  for( i=0 ; i<nrchanwav ; i++ ) {
    sprintf( tmp, "%s.%d", argv[argc-1], i ) ;
    outfp[i] = fopen( tmp, "w" ) ;
    if( outfp[i] == 0 ) {
      ERROR_MSG_S("can't open %d-th outfile %s!", i, tmp );
      outfp[i] = stderr ;
    }
    
  } ;
  FILE *globoutfp = fopen( argv[argc-1], "w" ) ;

  // output header
  switch( OutputMode ) {
    
  case FrameWise:
    //for( i=0 ; i<nrchanwav ; i++ ) 
      //fprintf( outfp[i], "FILENAME: %s\n", "dummy.wav" ) ;
    break ;
    
  case MUC:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      fprintf( outfp[i], "##filename %s\n", "dummy.wav" ) ;
      fprintf( outfp[i], "##nodelist\n" ) ;
      fprintf( outfp[i], "{\n0\t<\n" ) ;
    } ;
    break;
    
  case MUC_OLD:
    for( i=0 ; i<nrchanwav ; i++ ) 
      fprintf( outfp[i], "// filename=%s\n", "dummy.wav" ) ;
    break ;
    
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      // dummy header: length of file is still unknown!
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    } ;
    // dummy header: length of file is still unknown!
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    break ;
  } ; 
  
  // define variables for processing all channels and mixfile
  // status of speech nonspeech detection
  ESpeechFlag *spd = new ESpeechFlag[nrchanwav]; 
  SFeature **prefp = new SFeature*[nrchanwav];
  ESpeechFlag spdmixfile ;
  SFeature *prefpmixfile ;

  int lastflagmixfile = U ;
  int *lastflag = new int[nrchanwav] ; ;
  int *labelcount = new int[nrchanwav] ; // first label is "<"
  long framecount = 0 ;
  int lastchan = - 1;
  float *energymin = new float[nrchanwav] ;
  float *totloudmax = new float[nrchanwav] ;
  float *modloudmax = new float[nrchanwav] ;
  float *zerocrossmax = new float[nrchanwav] ;
  
  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( int i=0 ; i<nrchanwav ; i++ ) {
    lastflag[i] = U ;
    energymin[i] = 1.0e30 ;
    totloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
    zerocrossmax[i] = -1.0e30 ;
  } ;

  // determine speech energy min in each channel!!!
  while(1) { 
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 0 ;
    for( i=0 ; i<nrchanwav ; i++ )
      stillrunning += AkuPre[i]->testRdIsRun() ;
    if( stillrunning != nrchanwav )
      break ;

    // for each channel
    for( i=0 ; i<nrchanwav ; i++ ) {
      if( (prefp[i] = AkuPre[i]->getNext()) == 0 ) {
	stillrunning = 0 ;
	break ;
      } ;
      spd[i] = AkuPre[i]->getRdFlag();

      // find energy min
      if( prefp[i]->Pre.TotalEnergy < energymin[i] )
	energymin[i] =  prefp[i]->Pre.TotalEnergy ;
      
      // find totloudmax
      if( prefp[i]->Pre.SumLoud > totloudmax[i] )
	totloudmax[i] =  prefp[i]->Pre.SumLoud ;

      // find modloudmax
      if( prefp[i]->Pre.ModLoud > modloudmax[i] )
	modloudmax[i] =  prefp[i]->Pre.ModLoud ;

      // find zerocrossingmax
      if( prefp[i]->Pre.ZCRate > zerocrossmax[i] )
	zerocrossmax[i] =  prefp[i]->Pre.ZCRate ;

    } ;
    if( stillrunning < nrchanwav )
      break ;

    framecount++ ;

    if( framecount > NrFramesMax )
      break ;
  } ;

  // how long are files expected to be?
  framecount = 0 ;
  while( AkuPreMixFile->testRdIsRun() ) {
    AkuPreMixFile->getNext() ;
    framecount ++ ;
  } ;
  // restart MixFile
  AkuPreMixFile->clearAkuPre() ;
  AkuPreMixFile->startAkuPre( argv[argc-2] ) ;

  // prepare to store framewise channel-info (channel with max energy!)
  int *Channel = new int[framecount] ;
  // prepare to store info about, which channel energy is above the threshold!
  int **ActChanFlag = new int*[framecount] ;

  // initialize arrays
  for( j=0 ; j<framecount ; j++ ) {
    Channel[j] = -1 ;
    ActChanFlag[j] = new int[nrchanwav+1] ; // last entry will contain the spdet-info on the mixfile
    for( i=0 ; i<nrchanwav+1 ; i++ ) 
      ActChanFlag[j][i] = 0 ;
  } ;


  // output of min energies found: 
  for( i=0 ; i<nrchanwav ; i++ ) 
    fprintf( stderr, "channel %d: Min energy: %f\n", i, energymin[i] ) ;


  // reset AkuPre and start again (mixfile has already been started)
  for( i=0 ; i<nrchanwav ; i++) {
    AkuPre[i]->clearAkuPre() ;
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[(i*2)+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open%d-th  sigfile '%s'\n", i, argv[(i*2)+2] ) ;
      exit(-1) ;
    } ;
  }

  framecount = 0 ;
  // get spectra loop
  while( 1 ) {
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 1 ;
    for( i=0 ; i<nrchanwav ; i++ )
      stillrunning &= AkuPre[i]->testRdIsRun() ;
    stillrunning &= AkuPreMixFile->testRdIsRun() ;
    if( stillrunning == 0 )
      break ;

    // get next feature vectors
    for( i=0 ; i<nrchanwav ; i++ ) {
      prefp[i] = AkuPre[i]->getNext();
      // with normalized features!!!
      //prefp[i] = AkuPre[i]->getNext( energymin[i], totloudmax[i], modloudmax[i], zerocrossmax[i]);
      if( prefp[i] == 0 ) {
	stillrunning = 0 ;
	break;
      } ;
    } ;
    prefpmixfile = AkuPreMixFile->getNext() ;
    if( prefpmixfile == 0 ) 
      stillrunning = 0 ;
    
    if( stillrunning == 0 )
      break ;
    
    // get speech-nonspeech-info from probabilistic speech nonspeech detector
    for( i=0 ; i<nrchanwav ; i++ ) {
      spd[i] = AkuPre[i]->getRdFlag();
      if( spd[i] == U ) {   
	ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
	exit(-1) ;
      }
      spd[i] = ( spd[i] >= PUS ? S : P );
      spdmixfile = AkuPreMixFile->getRdFlag() ;
      if( spdmixfile == U ) {
	ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
	exit(-1) ;
      } ;
      spdmixfile = (spdmixfile >= PUS ? S : P );
    } ;
    
    // calculate mean energy (mean over all channels!)
    float maxenergy, meanenergy = 0.0 ;
    int maxindex, speechdet;
    for(  i=0 ; i<nrchanwav ; i++ ) {
      meanenergy += prefp[i]->Pre.TotalEnergy ;
    } ;
    meanenergy /= float(nrchanwav) ; 
    
    // decide upon the format of the output!
    switch( OutputMode ) {
    case FrameWise:
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      speechdet = 0 ;

      // only if speech detected in mixfile: search for channel with highest energy!
      if( spdmixfile >= PUS ) {
	ActChanFlag[framecount][nrchanwav] = 1 ; // memorize spdetinfo on mixfile!
	// find channels in which energy is higher than threshold	
	for( i=0 ; i<nrchanwav ; i++ ) {
	  //fprintf( outfp[i], "%d: ", AkuPre[i]->getRdNum() ) ;
	  if( spd[i] >= PUS ) {
	    fprintf( outfp[i], "%s\t%f\n", SpeechLabel.Par, (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    //fprintf( outfp[i], "%f\n", (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    speechdet++ ;
	    //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	    if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	      //maxenergy = prefp[i]->Pre.TotalEnergy ;
	      maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	      maxindex = i ;
	    } ;
	  } else {
	    fprintf( outfp[i], "%s\t%f\n", NonspeechLabel.Par,(prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    //fprintf( outfp[i], "%f\n", (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	  } ;
	  if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy >= Thres[i] )
	    ActChanFlag[framecount][i] = 1 ;
	} ;
	//fprintf( globoutfp, "%d: speech in channel %d\n", AkuPre[0]->getRdNum(), maxindex ) ;
      } else {
	//fprintf( globoutfp, "%d: nonspeech in all channels\n",  AkuPre[0]->getRdNum() ) ;
	;
      } ;
      
      Channel[framecount] = maxindex ;

      break ;
      
    case MUC:
      for( i=0 ; i<nrchanwav ; i++ ) {
	// check if spd-status has changed since last frame
	if( spd[i] != lastflag[i] ) {
	  if( framecount != 0 )
	    fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
	  fprintf( outfp[i], "%d\t%s\tBegF=%d\t", ++labelcount, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par, framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	} ;
      } ;
      break ;
      
    case MUC_OLD:
      for( i=0 ; i<nrchanwav ; i++ ) {
	// check if spd-status has changed since last frame
	if( lastflag[i] != spd[i] ) {
	  if( framecount != 0 ) 
	    fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
	  fprintf( outfp[i], "*###LA\t%d\t", framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	}
      } ;
      break ;
      
    case XML:
      
      // look for changes in the indiv channels!
      for( i=0 ; i<nrchanwav ; i++ ) {
	if( spd[i] != lastflag[i] ) {
	  fprintf( outfp[i], "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	  fprintf( outfp[i], "%s\n", spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	  lastflag[i] = spd[i] ;
	}
      } ;

      speechdet = 0 ;
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      // only if speech in mixfile detected: search for channel, else maxindex remains -1 ;
      if( spdmixfile >= PUS ) {
	ActChanFlag[framecount][nrchanwav] = 1 ; // memorize spdetstatus on mixfile!
	for( i=0 ; i<nrchanwav ; i++ ) {
	  if( spd[i] >= PUS ) {
	    speechdet++ ;
	    //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	    if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	      //maxenergy = prefp[i]->Pre.TotalEnergy ;
	      maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	      maxindex = i ;
	    } ;
	  } ;
	  if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy >= Thres[i])
	    ActChanFlag[framecount][i] = 1;
	} ;
      } ;

      // output in global file only, if new channel is active or switch between speech and silence
      if( maxindex != lastchan ) {
	// now a new channel is active!
	//fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	//fprintf( globoutfp, "%d: %s\n", maxindex, maxindex >= 0 ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	lastchan = maxindex ;
      } ;

      Channel[framecount] = maxindex ;
 
      break ;
    } ;
    
    framecount++ ;
    
  } // END loop over all frames
  
  switch( OutputMode ) {
  case MUC:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
      fprintf( outfp[i], "%d\t>\n", ++labelcount ) ;
      fprintf( outfp[i], "}\n" ) ;
    }	
    break ;
    
  case MUC_OLD:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
    } 
    break ;
    
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry
      fprintf( outfp[i], "</Turn>\n" ) ;
      fprintf( outfp[i], "</Section>\n" ) ;
      fprintf( outfp[i], "</Episode>\n" ) ;
      fprintf( outfp[i], "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( outfp[i] ) ;
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    } ;
     // finish last entry
      fprintf( globoutfp, "</Turn>\n" ) ;
      fprintf( globoutfp, "</Section>\n" ) ;
      fprintf( globoutfp, "</Episode>\n" ) ;
      fprintf( globoutfp, "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( globoutfp ) ;
      fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( globoutfp, "<Episode>\n" ) ;
      fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;

  // smooth info about channels with max energy
  int SmoothBufLen = SmoothWinLen.Par ;
  int *SmoothBuf = new int[SmoothBufLen] ;
  for( i=0 ; i<SmoothBufLen/2 ; i++ )
    SmoothBuf[i] = Channel[0] ; // pause!
  for( i=SmoothBufLen/2 ; i<SmoothBufLen ; i++ )
    SmoothBuf[i] = -1 ;
  int writepointer = SmoothBufLen/2 ;
  int MaxNrChan = 20 ;
  int *ChanCount = new int[MaxNrChan] ;

  for( i=0 ; i<framecount+SmoothBufLen/2 ; i++ ) {

    // write current entry into buffer
    SmoothBuf[writepointer] = (i<framecount ? Channel[i] : -1 ) ;

    // increase writepointer
    writepointer = (writepointer < SmoothBufLen-1 ? writepointer+1 : 0 ) ;

    // evaluate buffer
    // reset channel counts:
    int j ;
    for( j=0 ; j<MaxNrChan ; j++ )
      ChanCount[j] = 0 ;

    // count how often a channel is seen in buffer
    for( j=0 ; j<SmoothBufLen ; j++ )
      ChanCount[ SmoothBuf[j]+1 ] ++ ; // increase count for the observed channel (channel nr +1, since -1=pause)

    // search for maxcount
    int MaxChanCount=0 ;
    int MaxChanIndex=-1 ;
    for( j=0 ; j<MaxNrChan ; j++ ) 
      if( ChanCount[j] > MaxChanCount ) {
	MaxChanCount = ChanCount[j] ; 
	MaxChanIndex = j-1 ; // offset is -1 (=pause)
      } ;
    
    if( i >= SmoothBufLen/2 ) {
      Channel[i-SmoothBufLen/2] = MaxChanIndex ;
      //fprintf( stderr, "%d: %d\n", i-SmoothBufLen/2, Channel[i-SmoothBufLen/2] ) ;
    } ;

  } ; 

  // smooth array with active-channel info! (not for mixfile!!!!)
  // for each channel
  for( int chan=0 ; chan<nrchanwav ; chan++ ) {
    // reset smoothbuf
    for( i=0 ; i<SmoothBufLen ; i++ )
    SmoothBuf[i] = 0 ; // channel is inactive per default
    // init writepointer
    writepointer = 0 ;

    // for each frame
    for( i=0 ; i<framecount+SmoothBufLen/2 ; i++ ) {

      // write current entry into buffer
      SmoothBuf[writepointer] = (i<framecount ? ActChanFlag[i][chan] : 0 ) ;

      // increase writepointer
      writepointer = (writepointer < SmoothBufLen-1 ? writepointer+1 : 0 ) ;

      // evaluate buffer entries
      sum = 0 ;
      for( j=0 ; j<SmoothBufLen ; j++ ) {
	sum += SmoothBuf[j] ;
      } ;
      // first smoothing possible after SmoothBufLen/2+1 Frames !!! 
      if( i >= SmoothBufLen/2 ) {
	if( sum > SmoothBufLen/2 )
	  ActChanFlag[i-SmoothBufLen/2][chan] = 1 ;
	else
	  ActChanFlag[i-SmoothBufLen/2][chan] = 0 ;
      } ;
      
    } ;  // END loop over all frames   

  } ; // END loop over all channels


  // transform framewise info into format wanted:
  lastchan = -2 ;
  for( i=0 ; i<framecount ; i++ ) {
    switch( OutputMode ) {
    case FrameWise:
      // with overlaps??
      if( !OverlapOutput.Par ) {
	// only channel id of channel with max energy
	if( Channel[i] >= 0 )
	  fprintf( globoutfp, "%d: %d: %s\n", i, Channel[i], SpeechLabel.Par ) ;
	else
	  fprintf( globoutfp, "%d: %s\n", i, NonspeechLabel.Par ) ;
      } else {
	// all active channels (energy >= Thres[chan] )
	sum = 0 ;
	for( j=0 ; j<nrchanwav ; j++ ) {
	  sum += ActChanFlag[i][j] ;
	  if( ActChanFlag[i][j] == 1 )
	    fprintf( globoutfp, " %d", j ) ;
	} ;
	if( sum > 0 )
	  fprintf( globoutfp, ": %s\n", SpeechLabel.Par ) ;
	else if( ActChanFlag[i][nrchanwav] == 1 )
	  // speech detected on mixfile!
	  fprintf( globoutfp, "??? %s\n", SpeechLabel.Par ) ;
	else 
	  // neither on the mixfile nor on any of the channels speech was detected!
	  fprintf( globoutfp, "%s\n", NonspeechLabel.Par ) ;
      } ;
      break ;
      
    case MUC:
      // check if label switched!
      if( Channel[i] != lastchan ) {
	if( i != 0 )
	  fprintf( globoutfp, "EndF=%d\n", i - 1 ) ; // finish previous entry 
	if( Channel[i] >=0 ) 
	  fprintf( globoutfp, "%d\t%d:%s\tBegF=%d\t", ++labelcount, Channel[i], SpeechLabel.Par,  i ) ;// and begin next one
	else
	  fprintf( globoutfp, "%d\t%s\tBegF=%d\t", ++labelcount, NonspeechLabel.Par,  i ) ;// and begin next one

	lastchan = Channel[i] ;
      } ;
      break ;
      
    case MUC_OLD:
      // check if label switched!
      if( Channel[i] != lastchan ) {
	if( i != 0 )
	  if( lastchan >= 0 )
	    fprintf( globoutfp, "%d\t%d:%s\n",  i-1, Channel[i], SpeechLabel.Par ) ; // finish previous entry 
	  else
	    fprintf( globoutfp, "%d\t%s\n",  i-1, NonspeechLabel.Par ) ; // finish previous entry 

	fprintf( globoutfp, "*###LA\t%d\t", i ) ;// and begin next one
	lastchan = Channel[i] ;
      }     
      break ;
      
    case XML:

      if( !OverlapOutput.Par ) {
	//no overlaps allowed: check if label switched!  -> output of channel id with max energy
	if( Channel[i] != lastchan ) {
	  fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", i/100.0 ) ;
	  if( Channel[i] >= 0 )
	    fprintf( globoutfp, "%d:%s\n", Channel[i], SpeechLabel.Par ) ;
	  else
	    fprintf( globoutfp, "%s\n", NonspeechLabel.Par ) ;
	  lastchan = Channel[i] ;
	} ; 
      } else {
	// overlaps allowed: output of all active channels
	sum = 0 ; 
	if( i== 0 )
	  // first frame:
	  sum = 1 ;
	else 
	  // has there changed anything in terms of activity?
	  for( j=0 ; j<nrchanwav+1 ; j++ ) 
	    sum |= abs(ActChanFlag[i-1][j]-ActChanFlag[i][j]);
	
	// which channel switched from active to not active or vice versa (or was it the mixfile?):
	if( sum ) {
	  // something has changed, but what?
	  fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", i/100.0 ) ;
	  sum = 0 ; // 'sum' is now used to detect, if the status of one of the channels has switched!
	  // first look into the channels
 	  for( j=0 ; j<nrchanwav ; j++ ) {
 	    // which channels are active???
	    if( ActChanFlag[i][j] ) {
	      sum += 1 ;
	      fprintf( globoutfp, " %d", j ) ;
	    } ; 
	  } ;
	  // speech or nonspeech on mixfile
	  if( sum )
	    // one or more of the channels are active
	    fprintf( globoutfp, ": %s", SpeechLabel.Par ) ;
	  else if ( ActChanFlag[i][nrchanwav] == 1 )
	    // activity on the mixfile detected, but no activity on any of the channels
	    fprintf( globoutfp, "???" ) ;
	  else
	    // no activity at all
	    fprintf( globoutfp, "%s", NonspeechLabel.Par ) ; 
	} ;

      } ;
      break ;
    } ;

  } ;

  switch( OutputMode ) {
  case MUC:
 
    break ;
    
  case MUC_OLD:

    break ;
    
  case XML:

    // finish last entry
    fprintf( globoutfp, "</Turn>\n" ) ;
    fprintf( globoutfp, "</Section>\n" ) ;
    fprintf( globoutfp, "</Episode>\n" ) ;
    fprintf( globoutfp, "</Trans>\n" ) ;
    // correct header
    // correct header: now length of file is known!
    rewind( globoutfp ) ;
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;



    
  for( i=0 ; i<nrchanwav ; i++ ) {
    fclose( outfp[i] ) ;    
    AkuPre[i]->clearAkuPre();
    delete AkuPre[i] ;
  } ;
  delete[] AkuPre ;

  AkuPreMixFile->clearAkuPre() ;
  delete   AkuPreMixFile ;
  
  delete[] Channel ;
  delete[] energymin ;
  delete[] totloudmax ;
  delete[] modloudmax ;
  delete[] zerocrossmax ;
  delete[] labelcount ;
  delete[] lastflag ;
  delete[] spd ;
  delete[] prefp ;

  delete[] Thres ;
  delete[] ActChanFlag ;
  
  //AkuPre.appendPar( stderr ) ;
  return _OK_;
}
