/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i ;
  // open files
  if( argc < 4) 
    {   fprintf(stderr,"usage: spdtest <cfg-file> <wavfile1> <wavfile2> ... <wavfileN> <mixfile> <outfile>\n");
    return _ERR_;
    }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open cfg-file:", argv[1]);
  
  int nrchanwav = argc-4 ;
  int totnrwav = argc-3 ; // includes mixfile 

  // construct Preprocessing
  CTool      Tool( chanfp );
  CAkuPre    **AkuPre = new CAkuPre*[nrchanwav] ;
  CAkuPre    *AkuPreMixFile = new CAkuPre( chanfp ) ;
  
  for( i=0 ; i<nrchanwav; i++ ) {
    AkuPre[i] = new CAkuPre( chanfp ) ;
  } ;
  

  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  
  
  for( i=0 ; i<nrchanwav ; i++)
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[i+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open5d-th  sigfile '%s'\n", i, argv[i+2] ) ;
      exit(-1) ;
    } ;
  
  if(AkuPreMixFile->startAkuPre( argv[argc-2] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argv[argc-2] ) ;
    exit(-1) ;
  } ;

  
  // open outfiles
  char tmp[MaxLineLength] ;
  FILE **outfp = new FILE*[argc-3] ;
  for( i=0 ; i<nrchanwav ; i++ ) {
    sprintf( tmp, "%s.%d", argv[argc-1], i ) ;
    outfp[i] = fopen( tmp, "w" ) ;
    if( outfp[i] == 0 ) {
      ERROR_MSG_S("can't open %d-th outfile %s!", i, tmp );
      outfp[i] = stderr ;
    }
    
  } ;
  FILE *globoutfp = fopen( argv[argc-1], "w" ) ;

  
  // output header
  switch( OutputMode ) {
    
  case FrameWise:
    //for( i=0 ; i<nrchanwav ; i++ ) 
      //fprintf( outfp[i], "FILENAME: %s\n", "dummy.wav" ) ;
    break ;
    
  case MUC:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      fprintf( outfp[i], "##filename %s\n", "dummy.wav" ) ;
      fprintf( outfp[i], "##nodelist\n" ) ;
      fprintf( outfp[i], "{\n0\t<\n" ) ;
    } ;
    break;
    
  case MUC_OLD:
    for( i=0 ; i<nrchanwav ; i++ ) 
      fprintf( outfp[i], "// filename=%s\n", "dummy.wav" ) ;
    break ;
    
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) { 
      // dummy header: length of file is still unknown!
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    } ;
    // dummy header: length of file is still unknown!
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    break ;
  } ; 
  
  ESpeechFlag *spd = new ESpeechFlag[argc-3];
  SFeature **prefp = new SFeature*[argc-3];
  
  ESpeechFlag spdmixfile ;
  SFeature *prefpmixfile ;

  int lastflagmixfile = U ;

  int *lastflag = new int[argc-3] ; ;
  int *labelcount = new int[argc-3] ; // first label is "<"
  long framecount = 0 ;
  int lastchan = - 1;

  float *energymin = new float[argc-3] ;
  
  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( int i=0 ; i<nrchanwav ; i++ ) {
    lastflag[i] = U ;
    energymin[i] = 1.0e30 ;
  } ;

  // determine speech energy min in each channel!!!
  while(1) { 
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 0 ;
    for( i=0 ; i<nrchanwav ; i++ )
      stillrunning += AkuPre[i]->testRdIsRun() ;
    if( stillrunning != nrchanwav )
      break ;

    for( i=0 ; i<nrchanwav ; i++ ) {
      if( (prefp[i] = AkuPre[i]->getNext()) == 0 ) {
	stillrunning = 0 ;
	break ;
      } ;
      spd[i] = AkuPre[i]->getRdFlag();

      //if( prefp[i]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( prefp[i]->Pre.TotalEnergy < energymin[i] )
	energymin[i] =  prefp[i]->Pre.TotalEnergy ;
    } ;
    if( stillrunning < nrchanwav )
      break ;

    framecount++ ;

    if( framecount > 300 )
      break ;
  } ;

  // how long are files expected to be?
  framecount = 0 ;
  while( AkuPreMixFile->testRdIsRun() ) {
    AkuPreMixFile->getNext() ;
    framecount ++ ;
  } ;
  // restart MixFile
  AkuPreMixFile->clearAkuPre() ;
  AkuPreMixFile->startAkuPre( argv[argc-2] ) ;

  // prepare to store framewise channel-info
  int *Channel = new int[framecount] ;

  // output of min energies found: 
  for( i=0 ; i<nrchanwav ; i++ )
    fprintf( stderr, "channel %d: Min energy: %f\n", i, energymin[i] ) ;


  // reset AkuPre and start again (mixfile has already been started)
  for( i=0 ; i<nrchanwav ; i++) {
    AkuPre[i]->clearAkuPre() ;
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[i+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open5d-th  sigfile '%s'\n", i, argv[i+2] ) ;
      exit(-1) ;
    } ;
  }

  framecount = 0 ;
  // get spectra loop
  while( 1 ) {
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 1 ;
    for( i=0 ; i<nrchanwav ; i++ )
      stillrunning &= AkuPre[i]->testRdIsRun() ;
    stillrunning &= AkuPreMixFile->testRdIsRun() ;
    if( stillrunning == 0 )
      break ;

    // get next feature vectors
    for( i=0 ; i<nrchanwav ; i++ ) {
      prefp[i] = AkuPre[i]->getNext();
      if( prefp[i] == 0 ) {
	stillrunning = 0 ;
	break;
      } ;
    } ;
    prefpmixfile = AkuPreMixFile->getNext() ;
    if( prefpmixfile == 0 ) 
      stillrunning = 0 ;
    
    if( stillrunning == 0 )
      break ;
    
    // get speech-nonspeech-info
    for( i=0 ; i<nrchanwav ; i++ ) {
      spd[i] = AkuPre[i]->getRdFlag();
      if( spd[i] == U ) {   
	ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
	exit(-1) ;
      }
      spd[i] = ( spd[i] >= PUS ? S : P );
      spdmixfile = AkuPreMixFile->getRdFlag() ;
      if( spdmixfile == U ) {
	ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
	exit(-1) ;
      } ;
      spdmixfile = (spdmixfile >= PUS ? S : P );
    } ;
    
    float maxenergy, meanenergy = 0.0 ;
    int maxindex, speechdet;

    // calculate mean energy (mean over all channels!)
    for(  i=0 ; i<nrchanwav ; i++ ) {
      meanenergy += prefp[i]->Pre.TotalEnergy ;
    } ;
    meanenergy /= float(nrchanwav) ; 
    
    
    // decide upon the format of the output!
    switch( OutputMode ) {
    case FrameWise:
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      speechdet = 0 ;

      // only if speech detected in mixfile: search for channel with highest energy!
      if( spdmixfile >= PUS ) {
	
	// V1: output for frames, in which speech in mixfile was detected!
	// fprintf( outfp[i], "%s\t%f\n", SpeechLabel.Par, (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	
	for( i=0 ; i<nrchanwav ; i++ ) {
	  //fprintf( outfp[i], "%d: ", AkuPre[i]->getRdNum() ) ;
	  if( spd[i] >= PUS ) {
	    fprintf( outfp[i], "%s\t%f\n", SpeechLabel.Par, (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    //fprintf( outfp[i], "%f\n", (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    speechdet++ ;
	    //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	    if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	      //maxenergy = prefp[i]->Pre.TotalEnergy ;
	      maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	      maxindex = i ;
	    } ;
	  } else {
	    fprintf( outfp[i], "%s\t%f\n", NonspeechLabel.Par,(prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	    //fprintf( outfp[i], "%f\n", (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	  } ;
	} ;
	//fprintf( globoutfp, "%d: speech in channel %d\n", AkuPre[0]->getRdNum(), maxindex ) ;
      } else {
	//fprintf( globoutfp, "%d: nonspeech in all channels\n",  AkuPre[0]->getRdNum() ) ;
	;
      } ;
      
      Channel[framecount] = maxindex ;

      break ;
      
    case MUC:
      for( i=0 ; i<nrchanwav ; i++ ) {
	// check if spd-status has changed since last frame
	if( spd[i] != lastflag[i] ) {
	  if( framecount != 0 )
	    fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
	  fprintf( outfp[i], "%d\t%s\tBegF=%d\t", ++labelcount, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par, framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	} ;
      } ;
      break ;
      
    case MUC_OLD:
      for( i=0 ; i<nrchanwav ; i++ ) {
	// check if spd-status has changed since last frame
	if( lastflag[i] != spd[i] ) {
	  if( framecount != 0 ) 
	    fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
	  fprintf( outfp[i], "*###LA\t%d\t", framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	}
      } ;
      break ;
      
    case XML:
      
      // look for changes in the indiv channels!
      for( i=0 ; i<nrchanwav ; i++ ) {
	if( spd[i] != lastflag[i] ) {
	  fprintf( outfp[i], "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	  fprintf( outfp[i], "%s\n", spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	  lastflag[i] = spd[i] ;
	}
      } ;

      speechdet = 0 ;
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      // only if speech in mixfile detected: search for channel, else maxindex remains -1 ;
      if( spdmixfile >= PUS ) {
	for( i=0 ; i<nrchanwav ; i++ ) {
	  if( spd[i] >= PUS ) {
	    speechdet++ ;
	    //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	    if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	      //maxenergy = prefp[i]->Pre.TotalEnergy ;
	      maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	      maxindex = i ;
	    } ;
	  } ;
	} ;
      } ;

      // output in global file only, if new channel is active or switch between speech and silence
      if( maxindex != lastchan ) {
	// now a new channel is active!
	//fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	//fprintf( globoutfp, "%d: %s\n", maxindex, maxindex >= 0 ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	lastchan = maxindex ;
      } ;

      Channel[framecount] = maxindex ;
 
      break ;
    } ;
    
    framecount++ ;
    
  } // END loop over all frames
  
  switch( OutputMode ) {
  case MUC:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
      fprintf( outfp[i], "%d\t>\n", ++labelcount ) ;
      fprintf( outfp[i], "}\n" ) ;
    }	
    break ;
    
  case MUC_OLD:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
    } 
    break ;
    
  case XML:
    for( i=0 ; i<nrchanwav ; i++ ) {
      // finish last entry
      fprintf( outfp[i], "</Turn>\n" ) ;
      fprintf( outfp[i], "</Section>\n" ) ;
      fprintf( outfp[i], "</Episode>\n" ) ;
      fprintf( outfp[i], "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( outfp[i] ) ;
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    } ;
     // finish last entry
      fprintf( globoutfp, "</Turn>\n" ) ;
      fprintf( globoutfp, "</Section>\n" ) ;
      fprintf( globoutfp, "</Episode>\n" ) ;
      fprintf( globoutfp, "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( globoutfp ) ;
      fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( globoutfp, "<Episode>\n" ) ;
      fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;

  // smooth framewise labelinfo (fixed bufferlen!)
  int SmoothBufLen = 41 ;
  int *SmoothBuf = new int[SmoothBufLen] ;
  for( i=0 ; i<SmoothBufLen/2 ; i++ )
    SmoothBuf[i] = Channel[0] ; // pause!
  for( i=SmoothBufLen/2 ; i<SmoothBufLen ; i++ )
    SmoothBuf[i] = -1 ;
  int writepointer = SmoothBufLen/2 ;
  int MaxNrChan = 20 ;
  int *ChanCount = new int[MaxNrChan] ;

  for( i=0 ; i<framecount+SmoothBufLen/2 ; i++ ) {

    // write current entry into buffer
    SmoothBuf[writepointer] = (i<framecount ? Channel[i] : -1 ) ;

    // increase writepointer
    writepointer = (writepointer < SmoothBufLen-1 ? writepointer+1 : 0 ) ;

    // evaluate buffer
    // reset channel counts:
    int j ;
    for( j=0 ; j<MaxNrChan ; j++ )
      ChanCount[j] = 0 ;

    // count how often a channel is seen in buffer
    for( j=0 ; j<SmoothBufLen ; j++ )
      ChanCount[ SmoothBuf[j]+1 ] ++ ; // increase count for the observed channel (channel nr +1, since -1=pause)

    // search for maxcount
    int MaxChanCount=0 ;
    int MaxChanIndex=-1 ;
    for( j=0 ; j<MaxNrChan ; j++ ) 
      if( ChanCount[j] > MaxChanCount ) {
	MaxChanCount = ChanCount[j] ; 
	MaxChanIndex = j-1 ; // offset is -1 (=pause)
      } ;
    
    if( i >= SmoothBufLen/2 ) {
      Channel[i-SmoothBufLen/2] = MaxChanIndex ;
      //fprintf( stderr, "%d: %d\n", i-SmoothBufLen/2, Channel[i-SmoothBufLen/2] ) ;
    } ;

  } ; 

  // transform framewise info into format wanted:
  lastchan = -2 ;
  for( i=0 ; i<framecount ; i++ ) {
    switch( OutputMode ) {
    case FrameWise:
      // check if label switched!
      if( Channel[i] >= 0 )
	fprintf( globoutfp, "%d: %d: %s\n", i, Channel[i], SpeechLabel.Par ) ;
      else
	fprintf( globoutfp, "%d: %s\n", i, NonspeechLabel.Par ) ;
	
      break ;
      
    case MUC:
      // check if label switched!
      if( Channel[i] != lastchan ) {
	if( i != 0 )
	  fprintf( globoutfp, "EndF=%d\n", i - 1 ) ; // finish previous entry 
	if( Channel[i] >=0 ) 
	  fprintf( globoutfp, "%d\t%d:%s\tBegF=%d\t", ++labelcount, Channel[i], SpeechLabel.Par,  i ) ;// and begin next one
	else
	  fprintf( globoutfp, "%d\t%s\tBegF=%d\t", ++labelcount, NonspeechLabel.Par,  i ) ;// and begin next one

	lastchan = Channel[i] ;
      } ;
      break ;
      
    case MUC_OLD:
      // check if label switched!
      if( Channel[i] != lastchan ) {
	if( i != 0 )
	  if( lastchan >= 0 )
	    fprintf( globoutfp, "%d\t%d:%s\n",  i-1, Channel[i], SpeechLabel.Par ) ; // finish previous entry 
	  else
	    fprintf( globoutfp, "%d\t%s\n",  i-1, NonspeechLabel.Par ) ; // finish previous entry 

	fprintf( globoutfp, "*###LA\t%d\t", i ) ;// and begin next one
	lastchan = Channel[i] ;
      }     
      break ;
      
    case XML:

      // check if label switched!
      if( Channel[i] != lastchan ) {
	fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", i/100.0 ) ;
	if( Channel[i] >= 0 )
	  fprintf( globoutfp, "%d:%s\n", Channel[i], SpeechLabel.Par ) ;
	else
	  fprintf( globoutfp, "%s\n", NonspeechLabel.Par ) ;
	lastchan = Channel[i] ;
      } ; 
      break ;
    } ;

  } ;

  switch( OutputMode ) {
  case MUC:
 
    break ;
    
  case MUC_OLD:

    break ;
    
  case XML:

    // finish last entry
    fprintf( globoutfp, "</Turn>\n" ) ;
    fprintf( globoutfp, "</Section>\n" ) ;
    fprintf( globoutfp, "</Episode>\n" ) ;
    fprintf( globoutfp, "</Trans>\n" ) ;
    // correct header
    // correct header: now length of file is known!
    rewind( globoutfp ) ;
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;



    
  for( i=0 ; i<nrchanwav ; i++ ) {
    fclose( outfp[i] ) ;    
    AkuPre[i]->clearAkuPre();
    delete AkuPre[i] ;
  } ;
  delete[] AkuPre ;

  AkuPreMixFile->clearAkuPre() ;
  delete   AkuPreMixFile ;
  
  delete[] Channel ;
  delete[] energymin ;
  delete[] labelcount ;
  delete[] lastflag ;
  delete[] spd ;
  delete[] prefp ;
  
  //AkuPre.appendPar( stderr ) ;
  return _OK_;
}
