/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

#define MAXFRAMESUSEDFORENERGY 6000

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i ;
  // open files
  if( argc < 4) 
    {   fprintf(stderr,"usage: spdtest <cfg-file> <wavfile1> <wavfile2> ... <wavfileN> <outfile>\n");
    return _ERR_;
    }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open cfg-file:", argv[1]);
  
  // construct Preprocessing
  CTool      Tool( chanfp );
  CAkuPre    **AkuPre = new CAkuPre*[argc-3] ;
  
  for( i=0 ; i<argc-3; i++ ) {
    AkuPre[i] = new CAkuPre( chanfp ) ;
  } ;
  
  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  
  
  for( i=0 ; i<argc-3 ; i++)
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[i+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open5d-th  sigfile '%s'\n", i, argv[i+2] ) ;
      exit(-1) ;
    } ;
  
  // open outfile
  char tmp[MaxLineLength] ;
  FILE **outfp = new FILE*[argc-3] ;
  for( i=0 ; i<argc-3 ; i++ ) {
    sprintf( tmp, "%s.%d", argv[argc-1], i ) ;
    outfp[i] = fopen( tmp, "w" ) ;
    if( outfp[i] == 0 ) {
      ERROR_MSG_S("can't open %d-th outfile %s!", i, tmp );
      outfp[i] = stderr ;
    }
    
  } ;
  FILE *globoutfp = fopen( argv[argc-1], "w" ) ;

  
  // output header
  switch( OutputMode ) {
    
  case FrameWise:
    for( i=0 ; i<argc-3 ; i++ ) 
      fprintf( outfp[i], "FILENAME: %s\n", "dummy.wav" ) ;
    break ;
    
  case MUC:
    for( i=0 ; i<argc-3 ; i++ ) { 
      fprintf( outfp[i], "##filename %s\n", "dummy.wav" ) ;
      fprintf( outfp[i], "##nodelist\n" ) ;
      fprintf( outfp[i], "{\n0\t<\n" ) ;
    } ;
    break;
    
  case MUC_OLD:
    for( i=0 ; i<argc-3 ; i++ ) 
      fprintf( outfp[i], "// filename=%s\n", "dummy.wav" ) ;
    break ;
    
  case XML:
    for( i=0 ; i<argc-3 ; i++ ) { 
      // dummy header: length of file is still unknown!
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    } ;
    // dummy header: length of file is still unknown!
    fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
    fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
    fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
    fprintf( globoutfp, "<Episode>\n" ) ;
    fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
    break ;
  } ; 
  
  ESpeechFlag *spd = new ESpeechFlag[argc-3];
  SFeature **prefp = new SFeature*[argc-3];
  
  int *lastflag = new int[argc-3] ; ;
  int *labelcount = new int[argc-3] ; // first label is "<"
  long framecount = 0 ;
  int lastchan = - 1;

  float *energymin = new float[argc-3] ;
  
  for( int i=0 ; i<argc-3 ; i++ ) {
    lastflag[i] = U ;
    energymin[i] = 1.0e30 ;
  } ;


  // determine speech energy min in each channel!!!
  while(1) { 
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 0 ;
    for( i=0 ; i<argc-3 ; i++ )
      stillrunning += AkuPre[i]->testRdIsRun() ;
    if( stillrunning != argc-3 )
      break ;

    for( i=0 ; i<argc-3 ; i++ ) {
      if( (prefp[i] = AkuPre[i]->getNext()) == 0 ) {
	stillrunning = 0 ;
	break ;
      } ;
      spd[i] = AkuPre[i]->getRdFlag();

      //if( prefp[i]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( prefp[i]->Pre.TotalEnergy < energymin[i] )
	energymin[i] =  prefp[i]->Pre.TotalEnergy ;
    } ;
    if( stillrunning < argc-3 )
      break ;

    framecount++ ;
    if( framecount > MAXFRAMESUSEDFORENERGY )
      break ;
  } ;

  for( i=0 ; i<argc-3 ; i++ )
    fprintf( stderr, "channel %d: Min energy: %f\n", i, energymin[i] ) ;


  for( i=0 ; i<argc-3 ; i++) {
    AkuPre[i]->clearAkuPre() ;
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[i+2] ) == _ERR_ ) {
      fprintf( stderr, "Can't open5d-th  sigfile '%s'\n", i, argv[i+2] ) ;
      exit(-1) ;
    } ;
  }

  framecount = 0 ;
  // get spectra loop
  while( 1 ) {
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 1 ;
    for( i=0 ; i<argc-3 ; i++ )
      stillrunning &= AkuPre[i]->testRdIsRun() ;
    if( stillrunning == 0 )
      break ;
    
    for( i=0 ; i<argc-3 ; i++ ) {
      prefp[i] = AkuPre[i]->getNext();
      if( prefp[i] == 0 ) {
	stillrunning = 0 ;
	break;
      } ;
    } ;
    if( stillrunning == 0 )
      break ;
    
    for( i=0 ; i<argc-3 ; i++ ) {
      spd[i] = AkuPre[i]->getRdFlag();
      if( spd[i] == U ) {   
	ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
	continue;
      }
      spd[i] = ( spd[i] >= PUS ? S : P );
    } ;
    
    float maxenergy, meanenergy = 0.0 ;
    int maxindex, speechdet;

    // calculate mean energy (mean over all channels!)
    for(  i=0 ; i<argc-3 ; i++ ) {
      meanenergy += prefp[i]->Pre.TotalEnergy ;
    } ;
    meanenergy /= float(argc-3) ; 
    
    
    // how to format output!
    switch( OutputMode ) {
    case FrameWise:
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      speechdet = 0 ;
      for( i=0 ; i<argc-3 ; i++ ) {
	fprintf( outfp[i], "%d: ", AkuPre[i]->getRdNum() ) ;
	if( spd[i] >= PUS ) {
	  fprintf( outfp[i], "%s(%f)\n", SpeechLabel.Par, (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	  speechdet++ ;
	  //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	  if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	    //maxenergy = prefp[i]->Pre.TotalEnergy ;
	    maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	    maxindex = i ;
	  } ;
	} else {
	  fprintf( outfp[i], "%s(%f)\n", NonspeechLabel.Par,(prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy );
	} ;
      } ;
      if( speechdet > 0 ) {
	fprintf( globoutfp, "%d: speech in channel %d\n", AkuPre[0]->getRdNum(), maxindex ) ;
      } else {
	fprintf( globoutfp, "%d: nonspeech in all channels\n",  AkuPre[0]->getRdNum() ) ;
      } ;
      
      break ;
      
    case MUC:
      for( i=0 ; i<argc-3 ; i++ ) {
	// check if spd-status has changed since last frame
	if( spd[i] != lastflag[i] ) {
	  if( framecount != 0 )
	    fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
	  fprintf( outfp[i], "%d\t%s\tBegF=%d\t", ++labelcount, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par, framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	} ;
      } ;
      break ;
      
    case MUC_OLD:
      for( i=0 ; i<argc-3 ; i++ ) {
	// check if spd-status has changed since last frame
	if( lastflag[i] != spd[i] ) {
	  if( framecount != 0 ) 
	    fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i]>=PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
	  fprintf( outfp[i], "*###LA\t%d\t", framecount ) ;// and begin next one
	  lastflag[i] = spd[i] ;
	}
      } ;
      break ;
      
    case XML:
      speechdet = 0 ;
      maxenergy = -1.0e30 ;
      maxindex = -1 ;
      for( i=0 ; i<argc-3 ; i++ ) {
	if( spd[i] != lastflag[i] ) {
	  fprintf( outfp[i], "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	  fprintf( outfp[i], "%s\n", spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	  lastflag[i] = spd[i] ;
	} 
	if( spd[i] >= PUS ) {
	  speechdet++ ;
	  //if( prefp[i]->Pre.TotalEnergy > maxenergy ) {
	  if( (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy > maxenergy ) {
	    //maxenergy = prefp[i]->Pre.TotalEnergy ;
	    maxenergy = (prefp[i]->Pre.TotalEnergy)-energymin[i]-meanenergy ;
	    maxindex = i ;
	  } ;
	} ;
      } ;

      // output in global file only, if new channel is active or switch between speech and silence
      if( maxindex != lastchan ) {
	// now a new channel is active!
	fprintf( globoutfp, "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
	fprintf( globoutfp, "%d: %s\n", maxindex, maxindex >= 0 ? SpeechLabel.Par : NonspeechLabel.Par ) ;
	lastchan = maxindex ;
      } ;

      break ;
    } ;
    
    framecount++ ;
    
  } // END loop over all frames
  
  switch( OutputMode ) {
  case MUC:
    for( i=0 ; i<argc-3 ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "EndF=%d\n", framecount - 1 ) ; // finish previous entry 
      fprintf( outfp[i], "%d\t>\n", ++labelcount ) ;
      fprintf( outfp[i], "}\n" ) ;
    }	
    break ;
    
  case MUC_OLD:
    for( i=0 ; i<argc-3 ; i++ ) {
      // finish last entry 
      fprintf( outfp[i], "%d\t%s\n", framecount - 1, spd[i] >= PUS ? SpeechLabel.Par : NonspeechLabel.Par ) ; // finish previous entry 
    } 
    break ;
    
  case XML:
    for( i=0 ; i<argc-3 ; i++ ) {
      // finish last entry
      fprintf( outfp[i], "</Turn>\n" ) ;
      fprintf( outfp[i], "</Section>\n" ) ;
      fprintf( outfp[i], "</Episode>\n" ) ;
      fprintf( outfp[i], "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( outfp[i] ) ;
      fprintf( outfp[i], "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( outfp[i], "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( outfp[i], "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( outfp[i], "<Episode>\n" ) ;
      fprintf( outfp[i], "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( outfp[i], "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    } ;
     // finish last entry
      fprintf( globoutfp, "</Turn>\n" ) ;
      fprintf( globoutfp, "</Section>\n" ) ;
      fprintf( globoutfp, "</Episode>\n" ) ;
      fprintf( globoutfp, "</Trans>\n" ) ;
      // correct header
      // correct header: now length of file is known!
      rewind( globoutfp ) ;
      fprintf( globoutfp, "<?xml version=\"1.0\"?>\n" ) ;
      fprintf( globoutfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
      fprintf( globoutfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", "dummy.wav" ) ;
      fprintf( globoutfp, "<Episode>\n" ) ;
      fprintf( globoutfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
      fprintf( globoutfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
    break ;
    
  case FrameWise:
    // do nothing
    break ;
  } ;
  
  for( i=0 ; i<argc-3 ; i++ ) {
    fclose( outfp[i] ) ;    
    AkuPre[i]->clearAkuPre();
  } ;
  
  //AkuPre.appendPar( stderr ) ;
  return _OK_;
}
