/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
    int i ;
    // open files
    if( argc != 3) 
    {   fprintf(stderr,"usage: spdtest <cfg-file> <filelist>\n");
        return _ERR_;
    }
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open cfg-file:", argv[1]);
    FILE *catfp = fopen( argv[2], "r" );
    IF_ERROR_RETURN_S( catfp == 0, "can't open filelist:", argv[2] );

    FILE *labfp, *outfp ;

    // construct
    CTool      Tool( chanfp );
    CAkuPre    AkuPre( chanfp );
    CListOfLabel Lab( Tool.LabelKeyword.Par );

    // read Tool-related params
    CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
    EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
    CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
    CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
    CParString OverlapLabel( "Tools.Result.OverlapLabel", chanfp, "-" ) ;
    // set Labels to empty string if requested!
    if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
      strcpy( SpeechLabel.Par, "" ) ;
    if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
      strcpy( NonspeechLabel.Par, "" ) ;
    if( strcmp( OverlapLabel.Par, "NONE" ) == 0 ) 
      strcpy( OverlapLabel.Par, "" ) ;
  

    // setup histogramm
    long   HistoDstart[2*DMAX+1], HistoDstop[2*DMAX+1];
    long  *dstarthisto = &HistoDstart[DMAX], *dstophisto = &HistoDstop[DMAX];
    long   nhisto = 0;
    for( i = -DMAX; i <= DMAX; i++ )
        dstarthisto[i] = dstophisto[i] = 0;

    // loop over catalog
    char  str[MAXCATLINELEN], *sigfilename, *labfilename, *outfilename;
    while( fgets( str, MAXCATLINELEN, catfp) != 0 )
    {   if( (sigfilename = strtok(str, " ,\t\n")) == 0) break;
    //if( (labfilename = strtok(  0, " ,\t\n")) == 0) break;
	if( (outfilename = strtok(  0, " ,\t\n")) == 0) break; 

        // start AkuPre
        if( AkuPre.startAkuPre( sigfilename ) == _ERR_ ) continue;
 
	// open outfile
	outfp = fopen( outfilename, "w" ) ;
	if( outfp == 0 ) {
	  ERROR_MSG_S("can't open outfile %s!", outfilename );
	  outfp = stderr ;
        }
	 
	// output header
	switch( OutputMode ) {

	case FrameWise:
	  fprintf( outfp, "FILENAME: %s\n", sigfilename ) ;
	  break ;

	case MUC:
	  fprintf( outfp, "##filename %s\n", sigfilename ) ;
	  fprintf( outfp, "##nodelist\n" ) ;
	  fprintf( outfp, "{\n0\t<\n" ) ;
	  break;

	case MUC_OLD:
	  fprintf( outfp, "// filename=%s\n", sigfilename ) ;
	  break ;

	case XML:
	  // dummy header: length of file is still unknown!
	  fprintf( outfp, "<?xml version=\"1.0\"?>\n" ) ;
	  fprintf( outfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
	  fprintf( outfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", sigfilename ) ;
	  fprintf( outfp, "<Episode>\n" ) ;
	  fprintf( outfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;
	  fprintf( outfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", 0.0 ) ;

	  break ;
	} ; 

        long spdstart, spdstop;
        ESpeechFlag spd;
        SFeature *prefp;
        spdstop = spdstart = -1;

	ESpeechFlag lastflag = U ;
	int labelcount = 1 ; // first label is "<", only for MUC-version!
	long framecount = 0 ;

        // get spectra loop
        while( AkuPre.testRdIsRun() )
        {   prefp = AkuPre.getNext();
            if( prefp == 0 )
                break;
            spd = AkuPre.getRdFlag();
            if( spd == U )
            {   ERROR_MSG("wrong cfg-file use: AkuPre.ReadMode = P");
                continue;
            }
	    spd = ( spd >= PUS ? S : spd <= SUP ? P : O ) ;

	    // auskommentiert: tpfau, 12/21/2000
            //if( spdstart < 0 && spd >= PUS ) {
	    //  fprintf( stderr, "%d: SPEECH (%lf)\n", AkuPre.getRdNum(), AkuPre.getRdProbSp() );
	    //  spdstart = AkuPre.getRdNum();
	    //} ;
	    
	    // how to format output!
	    switch( OutputMode ) {
	    case FrameWise:
	      if( spd == S ) {
		fprintf( stderr, "%d: %s (prob: %lf, Energy: %f, ModLoud: %f)\n", AkuPre.getRdNum(), SpeechLabel.Par, AkuPre.getRdProbSp(), prefp->Pre.TotalEnergy,  prefp->Pre.ModLoud );
		spdstop = AkuPre.getRdNum();
	      } else if( spd == P ) {
		fprintf( stderr, "%d: %s (prob: %lf, Energy: %f, ModLoud: %f)\n", AkuPre.getRdNum(), NonspeechLabel.Par, AkuPre.getRdProbSi(), prefp->Pre.TotalEnergy,  prefp->Pre.ModLoud );
	      } else if( spd == O ) {
		fprintf( stderr, "%d: %s (prob: %lf, Energy: %f, ModLoud: %f)\n", AkuPre.getRdNum(), OverlapLabel.Par, AkuPre.getRdProbSi(), prefp->Pre.TotalEnergy,  prefp->Pre.ModLoud );
	      } ;
	      break ;

	    case MUC:
	      // check if spd-status has changed since last frame
	      if( lastflag != spd ) {
		// finish previous entry (not for first entry!)
		if( framecount != 0 )
		  fprintf( outfp, "EndF=%d\n", framecount - 1 ) ; // finish previous entry
		// begin this entry: (endf still unknown!)
		fprintf( outfp, "%d\t%s\tBegF=%d\t", ++labelcount, spd == S ? SpeechLabel.Par : spd == P ? NonspeechLabel.Par : OverlapLabel.Par, framecount ) ;// and begin next one
		lastflag = ( spd ) ;
	      }
	      break ;

	    case MUC_OLD:
	      // check if spd-status has changed since last frame
	      if( lastflag != spd ) {
		if( framecount != 0 ) 
		  fprintf( outfp, "%d\t%s\n", framecount - 1, lastflag == S ? SpeechLabel.Par : lastflag == P ? NonspeechLabel.Par : OverlapLabel.Par ) ; // finish previous entry 
		fprintf( outfp, "*###LA\t%d\t", framecount ) ;// and begin next one
		lastflag = ( spd ) ;
	      } ;
	      break ;

	    case XML:
	      if( lastflag != spd ) {
		fprintf( outfp, "<Sync time=\"%.3f\"/>\n", framecount/100.0 ) ;
		fprintf( outfp, "%s\n", spd == S ? SpeechLabel.Par : spd == P ? NonspeechLabel.Par : OverlapLabel.Par ) ;
		lastflag = ( spd ) ;
	      } 
	      break ;
	    } ;
	    
	    framecount++ ;
	    
        } // END loop over all frames

	switch( OutputMode ) {
	case MUC:
	  // finish last entry 
	  fprintf( outfp, "EndF=%d\n", framecount - 1 ) ; // finish previous entry
	  fprintf( outfp, "%d\t>\n", ++labelcount ) ;
	  fprintf( outfp, "}\n" ) ;
	  break ;
	case MUC_OLD:
	  // finish last entry 
	  fprintf( outfp, "%d\t%s\n", framecount - 1, lastflag == S ? SpeechLabel.Par : lastflag == P ? NonspeechLabel.Par : OverlapLabel.Par ) ; // finish previous entry 
	  break ;
	case XML:
	  // finish turn
	  fprintf( outfp, "</Turn>\n" ) ;
	  fprintf( outfp, "</Section>\n" ) ;
	  fprintf( outfp, "</Episode>\n" ) ;
	  fprintf( outfp, "</Trans>\n" ) ;
	  // correct header
	  // correct header: now length of file is known!
	  rewind( outfp ) ;
	  fprintf( outfp, "<?xml version=\"1.0\"?>\n" ) ;
	  fprintf( outfp, "<!DOCTYPE Trans SYSTEM \"trans-13.dtd\">\n" ) ;
	  fprintf( outfp, "<Trans scribe=\"dummy\" audio_filename=\"%s\" version=\"4\" version_date=\"010102\">\n", sigfilename ) ;
	  fprintf( outfp, "<Episode>\n" ) ;
	  fprintf( outfp, "<Section type=\"report\" startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0 ) ;
	  fprintf( outfp, "<Turn startTime=\"0\" endTime=\"%.8e\">\n", (framecount-1)/100.0  ) ;
	  
	  break ;
	case FrameWise:
	  // do nothing
	  break ;
	} ;

	fclose( outfp ) ;
  
        // clear AkuPre
        AkuPre.clearAkuPre();


    }



    return _OK_;
}
