/******************************************************************************
 * file:        spdtrain.cc  (main)
 * info:        training of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>

#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  256

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
    CParString         LabelSilentBeg;
    CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" ),
    LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),
    LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }



//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
    // open files
    if( argc != 3) 
    {   fprintf(stderr,"usage: spdtrain <channel-file> <catalog-file>\n");
        return _ERR_;
    }
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);
    FILE *catfp = fopen( argv[2], "r" );
    IF_ERROR_RETURN_S( catfp == 0, "can't open catalog-file:", argv[2] );

    // construct
    CTool      Tool( chanfp );
    CAkuPre    AkuPre( chanfp );
    CListOfLabel Lab( Tool.LabelKeyword.Par );

    // loop over catalog
    char  str[MAXCATLINELEN], *sigfilename, *labfilename;
    while( fgets( str, MAXCATLINELEN, catfp) != 0 )
    {   if( (sigfilename = strtok(str, " ,\t\n")) == 0) break;
        if( (labfilename = strtok(  0, " ,\t\n")) == 0) break;

        // start AkuPre
        if( AkuPre.startAkuPre( sigfilename ) == _ERR_ ) {
		fprintf( stderr, "Can't start akupre with signalfile %s\n", sigfilename ) ;	
		continue;
	} ;

        // open lab-file
        FILE *labfp = fopen( labfilename, "r");
        if( labfp == 0 )
        {   ERROR_MSG_S("can't open label-file:", labfilename );
            continue; 
        }

        // read start and stop from label-file
        long  start = 0, stop = 0;
        if( Lab.readLabels( labfp ) == _ERR_ || 
            Lab.getStartStop(&start,&stop, Tool.LabelSilentBeg.Par, Tool.LabelSilentEnd.Par) == _ERR_ )
            // Lab.getStartStop(&start,&stop ) == _ERR_ )
        {   ERROR_MSG_S("can't read start and stop in label-file:", labfilename );
            fclose( labfp );
            continue; 
        }
        fclose( labfp );

        // update proc. loop
        SPreFeature *prefp;
        while( (prefp=AkuPre.getNextPre()) != 0 )
        {   if( AkuPre.getRdNum() < start || AkuPre.getRdNum() > stop )
                AkuPre.updateSpNonSp( prefp, P );
            else
                AkuPre.updateSpNonSp( prefp, S );
        }

        // clear AkuPre
        AkuPre.clearAkuPre();
    }

    FILE *fp = fopen( "test.cfg", "w" ) ;
    AkuPre.appendPar( fp ) ;
    fclose(fp) ;

    return _OK_;
}
