/******************************************************************************
 * file:        spdtrain-multi-norm.cc  (main)
 * info:        training of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95, T.Pfau, 12/22/2000
 * last change: 12/22/2000, im ersten Durchgang werden die Maxima der ersten vier
 *                          Feature festgestellt und damit wird normiert!
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>

#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  1024
#define   MaxRuns 20000

//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
    CParString         LabelSilentBeg;
    CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" ),
    LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "<p:>" ),
    LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "<nib>" )
{ }



//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
    // open files
  int i ;
    if( argc != 4) 
    {   fprintf(stderr,"usage: spdtrain <cfg-file> <filelist> <new-cfgfile>\n");
        return _ERR_;
    }
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);
    FILE *catfp = fopen( argv[2], "r" );
    IF_ERROR_RETURN_S( catfp == 0, "can't open catalog-file:", argv[2] );

    // construct
    CTool      Tool( chanfp );
    CAkuPre    AkuPre( chanfp );
    CListOfLabel Lab( Tool.LabelKeyword.Par );

    //fprintf( stderr, "Initialisiert!\n" ) ;

    // loop over catalog
    char  str[MAXCATLINELEN], *sigfilename, *labfilename;
    while( fgets( str, MAXCATLINELEN, catfp) != 0 )
    {   if( (sigfilename = strtok(str, " ,\t\n")) == 0) break;
        if( (labfilename = strtok(  0, " ,\t\n")) == 0) break;

        // start AkuPre: first loop over file: get maxima of  Energy, loudness ...
        if( AkuPre.startAkuPre( sigfilename ) == _ERR_ ) {
		fprintf( stderr, "Can't start akupre with signalfile %s\n", sigfilename ) ;	
		continue;
	} ;
	//fprintf( stderr, "AkuPre gestartet mit file '%s'\n", sigfilename ) ;
	float MaxEnergy = -1.0 ;
	float MaxZCRate = -1.0 ;
	float MaxSumLoud = -1.0 ;
	float MaxModLoud = -1.0 ;
	long count = 0 ;
	long argMaxEnergy = 0 ;
	long argMaxZCRate = 0 ;
	long argMaxSumLoud = 0 ;
	long argMaxModLoud = 0 ;
	while( (prefp=AkuPre.getNextPre()) != 0 ) { 
	  if( prefp->TotalEnergy > MaxEnergy ) {
	    MaxEnergy = prefp->TotalEnergy ; argMaxEnergy = count ;
	  } ;
	  if( prefp->ZCRate > MaxZCRate ) {
	    MaxZCRate = prefp->ZCRate ; argMaxZCRate = count ;
	  } ;
	  if( prefp->SumLoud > MaxSumLoud ) {
	    MaxSumLoud = prefp->SumLoud ; argMaxSumLoud = count ;
	  } ;
	  if( prefp->ModLoud > MaxModLoud ) {
	    MaxModLoud = prefp->ModLoud ; argMaxModLoud = count ;
	  } ;
	  count++ ;
	} ;
	AkuPre.stopAkuPre() ;

	// start AkuPre: second loop over file: use maxima of  Energy, loudness ... for feature normalization!
        if( AkuPre.startAkuPre( sigfilename ) == _ERR_ ) {
	  fprintf( stderr, "Can't start akupre with signalfile %s\n", sigfilename ) ;	
	  continue;
	} ;

        // open lab-file
        FILE *labfp = fopen( labfilename, "r");
        if( labfp == 0 )
        {   ERROR_MSG_S("can't open label-file:", labfilename );
            continue; 
        }

	// define array for starts and stops of runs
	fprintf( stderr, "maxRuns: %d\n", MaxRuns ) ;
	long *start = new long[MaxRuns] ;
	long *stop = new long[MaxRuns] ;
	// default values
	for( i=0 ; i<MaxRuns ; i++ ) 
	  start[i] = stop[i] = -1 ;
	//fprintf( stderr, "Arrays initialisiert!\n" ) ;
	

	long count = 0 ;
        // read multiple starts and stops from label-file
        if( Lab.readLabels( labfp ) == _ERR_ ) { 
	  ERROR_MSG_S("can't read label-file:", labfilename );
	  fclose( labfp );
	  continue; 
	}
	//fprintf( stderr, "Labelfile gelesen!\n" ) ;
	while( count < MaxRuns && Lab.getNextStartStop( (count>0 ? stop[count-1] : 0), start[count], stop[count], Tool.LabelSilentBeg.Par, Tool.LabelSilentEnd.Par) != _ERR_ ) {
	  fprintf( stdout, "%ld: start: %ld stop: %ld\n", count, start[count], stop[count] );
	  count++ ;
	} ;
	// Lab.getFirstStartStop(&start,&stop, Tool.LabelSilentBeg.Par, Tool.LabelSilentEnd.Par) == _ERR_ )
	fclose( labfp );
	long nr_runs = count ;

	
	// loop over all frames, update spnonsp-info
        SPreFeature *prefp;	  
	count = 0 ;  // counts frames!
        // update proc. loop
        while( (prefp=AkuPre.getNextPre()) != 0 )
        { 
	  // check, if this frame is part of a speech portion
	  int speech = 0 ;
	  for( i=0 ; i<nr_runs ; i++ ) {
	    if( count >= start[i] && count <= stop[i] ) {
	      // speech!
	      speech = 1 ;
	      AkuPre.updateSpNonSp( prefp, S );
	      fprintf( stderr, "%d.%03d: SPEECH\n", count/100, count%100*6 ) ;
	      break ;
	    } ;
	  } ;
	  // frame does not belong to any run
	  if( speech == 0 ) {
	    AkuPre.updateSpNonSp( prefp, P );
	    fprintf( stderr, "%d.%03d: SILENCE\n", count/100, count%100*6 ) ;
	  } ;
	  count++ ;
        }

        // clear AkuPre
	AkuPre.clearAkuPre();
	
    }

    FILE *outfp = fopen( argv[3], "w" );
    if( outfp == 0 ) {
      fprintf( stderr, "Can't open new-cfgfile: %s\n", argv[3]);
      return(-1) ;
    } ;
    AkuPre.appendPar( outfp ) ;

    fclose(outfp) ;
    //fclose(chanfp) ;
    fclose(catfp ) ;

    return _OK_;
}
