/******************************************************************************
 * file:        spdtrain.cc  (main)
 * info:        training of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>

#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  1024
#define   MaxRuns 20000

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
{ ; } ;



//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] ) {
  // open files
  int i ;
  if( argc != 4) {   
    fprintf(stderr,"usage: spdtrain <cfg-file> <filelist> <new-cfgfile>\n");
    return _ERR_;
  }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);
  FILE *catfp = fopen( argv[2], "r" );
  IF_ERROR_RETURN_S( catfp == 0, "can't open catalog-file:", argv[2] );
  
  // construct
  CTool      Tool( chanfp );
  rewind( chanfp ) ;
  CAkuPre    AkuPre( chanfp );

  // read tool-related things
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParEnum InputPar( "Tools.Label.InputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode InputMode = (EOutPutMode)InputPar.Par ;
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, "SPEECH" ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  CParString OverlapLabel( "Tools.Result.OverlapLabel", chanfp, "OVERLAP" ) ;
  CParBool  UpdateSpNonSpAutoSil( "Tools.SpdAutoUpdate.Silent", chanfp, 1 ) ;
  CParBool  UpdateSpNonSpAutoSp( "Tools.SpdAutoUpdate.Speech", chanfp, 1 ) ;
  CParBool  UpdateSpNonSpAutoOv( "Tools.SpdAutoUpdate.Overlap", chanfp, 1 ) ;
  CParBool  ControlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;

  CListOfLabel *Lab ;
  // just quick and dirty, class CListOfLabel should be redesigned!
  if( InputMode == XML )
    // XML Input, use XML as KeyWord
    Lab = new CListOfLabel( InputPar.getValue( XML ) );
  else 
    // MUC or Kiel Input! use Keyword as Keyword
    Lab = new CListOfLabel( Tool.LabelKeyword.Par );

  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  if( strcmp( OverlapLabel.Par, "NONE" ) == 0 ) 
    strcpy( OverlapLabel.Par, "" ) ;
  
  // create array of Silence-Labels
  int NrSiLabels = AkuPre.getSpNonSpNrOfMixSi() ; 
  char **SiLabel = new char*[NrSiLabels] ;
  char tmp[MAXKEYWORDLEN] ;
  for( i=0 ; i<NrSiLabels ; i++ ) {
    SiLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SilentName.%d", i ) ;
    strcpy( SiLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SiLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SilentLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;

  // create array of Speech-Labels
  int NrSpLabels = AkuPre.getSpNonSpNrOfMixSp() ; 
  char **SpLabel = new char*[NrSpLabels] ;
  for( i=0 ; i<NrSpLabels ; i++ ) {
    SpLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SpeechName.%d", i ) ;
    strcpy( SpLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SpLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SpeechLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;
 
  int NrOvLabels ;
  char **OvLabel ;
  if( AkuPre.getSpNonSpUseOverlap() ) {
    // create array of overlap-Labels
    NrOvLabels = AkuPre.getSpNonSpNrOfMixOv() ; 
    OvLabel = new char*[NrOvLabels] ;
    for( i=0 ; i<NrOvLabels ; i++ ) {
      OvLabel[i] = new char[MAXKEYWORDLEN] ;
      sprintf( tmp, "Tools.Label.OverlapName.%d", i ) ;
      strcpy( OvLabel[i], CParString( tmp, chanfp, "???").Par ) ;
      if( !strcmp( OvLabel[i], "???" ) ) {
	fprintf( stderr, "ERROR: can't find %d-th OverlapLabel in cfg-file '%s'\n", i, argv[1] ) ;
	exit(-1) ;
      } ;
    } ;
  } ;
 
  // loop over catalog
  char  str[MAXCATLINELEN], *sigfilename, *labfilename;
  while( fgets( str, MAXCATLINELEN, catfp) != 0 ) {   
    if( (sigfilename = strtok(str, " ,\t\n")) == 0) break;
    if( (labfilename = strtok(  0, " ,\t\n")) == 0) break;
    
    // start AkuPre
    if( AkuPre.startAkuPre( sigfilename ) == _ERR_ ) {
      fprintf( stderr, "Can't start akupre with signalfile %s\n", sigfilename ) ;	
      continue;
    } ;
    
    // open lab-file
    FILE *labfp = fopen( labfilename, "r");
    if( labfp == 0 ) {   
      ERROR_MSG_S("can't open label-file:", labfilename );
      continue; 
    }
       
    // read label-file
    long count = 0 ;
    if( Lab->readLabels( labfp ) == _ERR_ ) { 
      ERROR_MSG_S("can't read label-file:", labfilename );
      fclose( labfp );
      continue; 
    }
    fclose( labfp );
	
    // loop over all frames, update spnonsp-info, stop, when label file is at end (must NOT be end of wav-file!!!)
    SPreFeature *prefp;	  
    count = 0 ;  // counts frames

    if( ControlOutput.Par ) 
      fprintf( stderr, "Loop until eof in Wav-file or endframe (%d) of labelfile!\n", Lab->getEndFrameOfLastEntry() ) ;

    // update proc. loop
    while( (prefp=AkuPre.getNextPre()) != 0 && count < Lab->getEndFrameOfLastEntry() ) { 
 
      // is it speech, nonspeech or overlapped speech?
      char *labtmp = Lab->getLabelName( count ) ;
      
      // is it speech?
      int found = 0 ;
      for( i=0 ; i<NrSpLabels ; i++ ) {
	//if( strcmp( labtmp, SpLabel[i] ) == 0 ) {
	if( strstr( labtmp, SpLabel[i] ) != 0 ) {
	  // yes, it is speech:
	  if( UpdateSpNonSpAutoSp.Par ) {
	    // mixture component with highest prob is picked automatically!
	    AkuPre.updateSpNonSpAuto( prefp, S ) ;
	    if( ControlOutput.Par ) 
	      fprintf( stderr, "%d.%03d: SPEECH (AUTOMATIC UPDATE)\n", count/100, count%100*6 ) ;
	  } else {
	    if( !AkuPre.updateSpNonSpOneMix( prefp, S, i ) ) 
	      fprintf( stderr, "ERROR: updating, %d-th speech mixture\n", i ) ;
	    if( ControlOutput.Par ) 
	      fprintf( stderr, "%d.%03d: SPEECH (Label %d: %s)\n", count/100, count%100*6, i, SpLabel[i] ) ;
	  } ;
	  found = 1 ;
	  break ;
	}
      } ;

      // proceed to next frame if it was speech
      if( found ) 
	goto CONTINUE ;	

      // is it nonspeech
      for( i=0 ; i<NrSiLabels ; i++ ) {
	//if( strcmp( labtmp, SiLabel[i] ) == 0 ) {
	if( strstr( labtmp, SiLabel[i] ) != 0 ) {
	  // yes, it is speech:
	  if( UpdateSpNonSpAutoSil.Par ) {
	    // mixture component with highest prob is picked automatically!
	    AkuPre.updateSpNonSpAuto( prefp, P ) ;
	    if( ControlOutput.Par ) 
	      fprintf( stderr, "%d.%03d: NONSPEECH (AUTOMATIC UPDATE)\n", count/100, count%100*6 ) ;
	  } else {
	    if( !AkuPre.updateSpNonSpOneMix( prefp, P, i ) ) 
	      fprintf( stderr, "ERROR: updating, %d-th nonspeech mixture\n", i ) ;
	    if( ControlOutput.Par ) 
	      fprintf( stderr, "%d.%03d: NONSPEECH (Label %d: %s)\n", count/100, count%100*6, i, SiLabel[i] ) ;
	  } ;
	  found = 1 ;
	  break ;
	}
      } ;

      // proceed to next frame if it was speech
      if( found ) 
	goto CONTINUE ;	  

      if( AkuPre.getSpNonSpUseOverlap() ) {
	// is it overlapped speech?
	for( i=0 ; i<NrOvLabels ; i++ ) {
	  //if( strcmp( labtmp, OvLabel[i] ) == 0 ) {
	  if( strstr( labtmp, OvLabel[i] ) != 0 ) {
	    // yes, it is overlapped speech:
	    if( UpdateSpNonSpAutoOv.Par ) {
	      // mixture component with highest prob is picked automatically!
	      AkuPre.updateSpNonSpAuto( prefp, O ) ;
	      if( ControlOutput.Par ) 
		fprintf( stderr, "%d.%03d: OVERLAPPED SPEECH (AUTOMATIC UPDATE)\n", count/100, count%100*6 ) ;
	    } else {
	      if( !AkuPre.updateSpNonSpOneMix( prefp, O, i ) ) 
		fprintf( stderr, "ERROR: updating, %d-th overlapped speech mixture\n", i ) ;
	      if( ControlOutput.Par ) 
		fprintf( stderr, "%d.%03d: OVERLAPPED SPEECH (Label %d: %s)\n", count/100, count%100*6, i, OvLabel[i] ) ;
	    } ;
	    found = 1 ;
	    break ;
	  }
	} ;
	// proceed to next frame if it was speech
	if( found ) 
	  goto CONTINUE ;	  
      } ;

    CONTINUE:
      count++ ;

    } ; // END loop over all frames of a speechfile
    
    // clear AkuPre
    AkuPre.clearAkuPre();
    
  } ; // END loop over all files
  
  FILE *outfp = fopen( argv[3], "w" );
  if( outfp == 0 ) {
    fprintf( stderr, "Can't open new-cfgfile: %s\n", argv[3]);
    return(-1) ;
  } ;


  // output Tools-parameter (default, if not used here!)
  fprintf( outfp, "// some old label used to define labels in labelfiles!\n" ) ;  
  Tool.LabelKeyword.appendPar(outfp) ;

  fprintf( outfp, "\n// specify Input and Output format\n" ) ;  
  OutputPar.appendPar( outfp ) ;
  InputPar.appendPar( outfp ) ;
  UpdateSpNonSpAutoSil.appendPar( outfp ) ; 
  UpdateSpNonSpAutoSp.appendPar( outfp ) ; 
  if( AkuPre.getSpNonSpUseOverlap() ) 
    UpdateSpNonSpAutoOv.appendPar( outfp ) ; 
  

  fprintf( outfp, "\n// which labels are used for the different speech resp. nonspeech classes?\n" ) ;  
  fprintf( outfp, "// labels must match the labelfile, e.g. if only \"speech\" and \"nonspeech\"\n" ) ;  
  fprintf( outfp, "// is contained in labelfile, then no explicit update is possible. In this case,\n" ) ;  
  fprintf( outfp, "// set all Names to \"speech\" resp. \"nonspeech\"!\n" ) ;  
  for( i=0 ; i<NrSiLabels ; i++ )
    fprintf( outfp, "Tools.Label.SilentName.%d = %s\n", i, SiLabel[i] ) ;
  for( i=0 ; i<NrSpLabels ; i++ )
    fprintf( outfp, "Tools.Label.SpeechName.%d = %s\n", i, SpLabel[i] ) ;
  if( AkuPre.getSpNonSpUseOverlap() ) 
    for( i=0 ; i<NrOvLabels ; i++ )
      fprintf( outfp, "Tools.Label.OverlapName.%d = %s\n", i, OvLabel[i] ) ;

  // set Labels back to NONE if required!
  if( strcmp( SpeechLabel.Par, "" ) == 0 ) 
    strcpy( SpeechLabel.Par, "NONE" ) ;
  if( strcmp( NonspeechLabel.Par, "" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "NONE" ) ;
  if( AkuPre.getSpNonSpUseOverlap() ) 
    if( strcmp( OverlapLabel.Par, "" ) == 0 ) 
      strcpy( OverlapLabel.Par, "NONE" ) ;

  fprintf( outfp, "\n// which labels do you want to be used for classification results?\n" ) ;  
  SpeechLabel.appendPar(outfp) ;
  NonspeechLabel.appendPar(outfp) ;
  if( AkuPre.getSpNonSpUseOverlap() ) 
    OverlapLabel.appendPar(outfp) ;
  
  fprintf( outfp, "\n// do you want some additional output?\n" ) ;
  ControlOutput.appendPar(outfp) ;

  fprintf( outfp, "\n// Params for akupre, spnonsp, chancomp and gainctrl!\n" ) ;
  AkuPre.appendPar( outfp ) ;

  
  fclose(outfp) ;
  //fclose(chanfp) ;
  fclose(catfp ) ;
  
  // free memory for silence labels
  for( i=0 ; i<NrSiLabels ; i++ ) 
    delete[] SiLabel[i] ;
  delete[] SiLabel ;
  // free memory for speech labels
  for( i=0 ; i<NrSpLabels ; i++ ) 
    delete[] SpLabel[i] ;
  delete[] SpLabel ;

  delete Lab ;

  return _OK_;
}
