/******************************************************************************
 * file:        spdtest.cc  (main)
 * info:        testing of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <stdio.h>
#include <string.h>
#include <math.h>


#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupresimple.h"
#include "spnonsp.h"

#define   MAXCATLINELEN  256
#define   DMAX  50
#define   OUTPUT_LOG

#define MAXNRCHAN 20
#define MaxRuns 20000
#define MAXNRFRAMEFORMAXDET 30000


#define TESTLEVEL
//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
  //CParString         LabelSilentBeg;
  //CParString         LabelSilentEnd;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
    //LabelSilentBeg(  "Tools.Label.SilentNameBeg", fp, "sw_/" ),    
    //LabelSilentEnd(  "Tools.Label.SilentNameEnd", fp, "sw_\\" )
{ }

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] )
{
  int i, run, chan, frame, tmpframe ;
  // open files
  if( argc < 9) 
    {   fprintf(stderr,"usage: spdtrain <akupre-cfg-file> <spnonsp-chan-cfg-file> <spnonsp-mix-cfg-file> <wavfile1> <wavfile2> ... <wavfileN> <labfile1> <labfile2> ... <labfileN> <mixfile>\n");
    return _ERR_;
    }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open akupre-cfg-file:", argv[1]);
  FILE *spnonspchancfgfp = fopen( argv[2], "r+" );
  IF_ERROR_RETURN_S( spnonspchancfgfp == 0, "can't open spnonsp-cfg-file:", argv[2]);
  FILE *spnonspmixcfgfp = fopen( argv[3], "r+" );
  IF_ERROR_RETURN_S( spnonspmixcfgfp == 0, "can't open spnonsp-cfg-file:", argv[3]);
  
  int nrchanwav = (argc-5)/2 ;
  int wavfileoffset = 4 ;
  int labfileoffset = wavfileoffset + nrchanwav ;
  int totnrwav = nrchanwav+1 ; // includes mixfile 
  int mixfileindex = argc-1 ;

  // read Tool-related params
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParEnum InputPar( "Tools.Label.InputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode InputMode = (EOutPutMode)InputPar.Par ;  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, " " ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  CParBool  SpdAutoUpdateSilent( "Tools.SpdAutoUpdate.Silent", chanfp, 1 ) ;
  CParBool  SpdAutoUpdateSpeech( "Tools.SpdAutoUpdate.Speech", chanfp, 1 ) ;
  CParBool  SpDetUseCommonForAllChan( "Tools.SpdUseCommonForAllChan", chanfp, 1 ) ;

  CParBool NormSpecialFeatChan( "Tools.NormSpecialFeatChan", chanfp, 1 ) ;
  CParFloat NormSpecialFeatChanPerc( "Tools.NormSpecialFeatChanPerc", chanfp, 0.0 ) ; // use percentile of max instead of max!
  CParBool NormSpecialFeatMix( "Tools.NormSpecialFeatMix", chanfp, 0 ) ;
  
  CParBool  ControlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;

  CParInt SmoothWindowLen( "Tools.Result.SmoothWindowLen", chanfp, 41 ) ;
  CParBool OverlapWanted( "Tools.Result.OverlapWanted", chanfp, 0 ) ;


#ifdef TESTLEVEL
  fprintf( stdout, "Initializing AkuPre and SpNonSp of all %d channels and of MixFile '%s' ... \n", nrchanwav, argv[mixfileindex] ) ;
#endif
  // construct Preprocessing
  CTool      Tool( chanfp );
  CAkuPreSimple    **AkuPre = new CAkuPreSimple*[nrchanwav] ;
  CAkuPreSimple    *AkuPreMixFile = new CAkuPreSimple( chanfp ) ;

  // construct SpNonSp
  CSpeechNonSpeech **SpNonSp = new CSpeechNonSpeech*[nrchanwav] ;
  //CSpeechNonSpeech *SpNonSpMixFile = new CSpeechNonSpeech( PRESPECLEN+4, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspmixcfgfp ) ;  // Achtung hard coded length of feature file
  CSpeechNonSpeech *SpNonSpMixFile = new CSpeechNonSpeech( (sizeof(sPreFeature)/sizeof(float))-1, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspmixcfgfp ) ;  // Achtung hard coded length of feature file
  
  // init channel specific akupre and spnonsp
  // use only one SpeechNonSpeech-Detector for all channels!!!
  if( SpDetUseCommonForAllChan.Par )
    SpNonSp[0] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; 
  for( i=0 ; i<nrchanwav; i++ ) {
    AkuPre[i] = new CAkuPreSimple( chanfp ) ;
    if( SpDetUseCommonForAllChan.Par ) {
      // use one detector for all channels:
      if( i!= 0 )
	SpNonSp[i] = SpNonSp[0] ;
    } else {
      // use indiv Detector for each channel: 
      //SpNonSp[i] = new CSpeechNonSpeech( PRESPECLEN+4, (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; // Achtung hard coded length of feature file
      SpNonSp[i] = new CSpeechNonSpeech( sizeof(sPreFeature)/sizeof(float), (double)WINDOW_DELTA/(double)SAMPLE_FRQ , spnonspchancfgfp ) ; // Achtung hard coded length of feature file
   } ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif



  // initialize label-info for all channels
  CListOfLabel **Lab = new CListOfLabel*[nrchanwav] ; ;
  // just quick and dirty, class CListOfLabel should be redesigned!
  if( InputMode == XML )
    // XML Input, use XML as KeyWord
    for( chan=0 ; chan<nrchanwav ; chan++ )
      Lab[chan] = new CListOfLabel( InputPar.getValue( XML ) );
  else 
    for( chan=0 ; chan<nrchanwav ; chan++ )
      // MUC or Kiel Input! use Keyword as Keyword
      Lab[chan] = new CListOfLabel( Tool.LabelKeyword.Par );


  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;

  // create array of Silence-Labels
  int NrSiLabels = SpNonSp[0]->getNrOfMixSi() ; 
  char **SiLabel = new char*[NrSiLabels] ;
  char tmp[MAXKEYWORDLEN] ;
  for( i=0 ; i<NrSiLabels ; i++ ) {
    SiLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SilentName.%d", i ) ;
    strcpy( SiLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SiLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SilentLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;

  // create array of Speech-Labels
  int NrSpLabels = SpNonSp[0]->getNrOfMixSp() ; 
  char **SpLabel = new char*[NrSpLabels] ;
  for( i=0 ; i<NrSpLabels ; i++ ) {
    SpLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SpeechName.%d", i ) ;
    strcpy( SpLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SpLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SpeechLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;


#ifdef TESTLEVEL
  fprintf( stdout, "Open labfiles for all channels ...\n" ) ;
#endif
  // open lab-files
  FILE **labfp = new FILE*[nrchanwav] ;
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    labfp[chan]= fopen( argv[labfileoffset+chan], "r");
    if( labfp[chan] == 0 ) {   
      ERROR_MSG_S("can't open label-file:", argv[labfileoffset+chan] );
      exit(-1) ; 
    }
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
  

#ifdef TESTLEVEL
  fprintf( stdout, "Define 'runs' for all channels ...\n" ) ;
#endif
  long *nr_runs = new long[nrchanwav] ; 
  
  // define array for starts and stops of runs
  //fprintf( stderr, "maxRuns: %d\n", MaxRuns ) ;
  long **start = new long*[nrchanwav] ;
  long **stop = new long*[nrchanwav] ;
  
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    start[chan] = new long[MaxRuns] ;
    stop[chan] = new long[MaxRuns] ;
    
    // default values for start and stop of each run
    for( run=0 ; run<MaxRuns ; run++ ) 
      start[chan][run] = stop[chan][run] = -1 ;
    
    // read multiple starts and stops from label-file
    long count = 0 ;
    if( Lab[chan]->readLabels( labfp[chan] ) == _ERR_ ) { 
      ERROR_MSG_S("can't read label-file:", argv[labfileoffset+chan] );
      fclose( labfp[chan] );
      continue; 
    }
  
    // loop over all runs to determine start and stop-frame
    //	while( count < MaxRuns && Lab.getNextStartStop( (count>0 ? stop[count-1] : 0), start[count], stop[count], Tool.LabelSilentBeg.Par, Tool.LabelSilentEnd.Par) != _ERR_ ) {
    while( count < MaxRuns && Lab[chan]->getNextStartStop( (count>0 ? stop[chan][count-1] : 0 ), start[chan][count], stop[chan][count], NrSiLabels, SiLabel ) != _ERR_ ) { 
      if( ControlOutput.Par ) 
	fprintf( stdout, "chan %d: %ld: start: %ld stop: %ld\n", chan, count, start[chan][count], stop[chan][count] );
      count++ ;
    } ;
    fclose( labfp[chan] );
    nr_runs[chan] = count ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif




#ifdef TESTLEVEL
  fprintf( stdout, "Start AkuPre of all %d channels ... \n", nrchanwav ) ;
#endif
  for( chan=0 ; chan<nrchanwav ; chan++) {
#ifdef TESTLEVEL
    fprintf( stdout, "... %d\n", chan  ) ;
#endif    // start AkuPre
    if( AkuPre[chan]->startAkuPre( argv[(chan)+wavfileoffset] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argv[(chan)+wavfileoffset] ) ;
      exit(-1) ;
    } ;
#ifdef TESTLEVEL
  } ;
  fprintf( stdout, "... DONE\n\n" ) ;
#endif
 


#ifdef TESTLEVEL
  fprintf( stdout, "Start AkuPre of Mixfile ...\n" ) ;
#endif
  if(AkuPreMixFile->startAkuPre( argv[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argv[mixfileindex] ) ;
    exit(-1) ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  

  SPreFeature **prefp = new SPreFeature*[nrchanwav];
  SPreFeature *prefpmixfile ;

  float meanenergy ;
  float meanenergyohnemix ;
  float *energymin = new float[nrchanwav+1] ;
  // float *entropymin = new float[nrchanwav+1] ;
  float *sumloudmax = new float[nrchanwav+1] ;
  float *modloudmax = new float[nrchanwav+1] ;
  float *modloudmin = new float[nrchanwav+1] ;
  float *zcratemin = new float[nrchanwav+1] ;

  // initialize flags for detecting change between speech and nonspeech and variables to detect min energy!
  for( int i=0 ; i<nrchanwav+1 ; i++ ) {
    energymin[i] = 1.0e30 ;
    // entropymin[i] = 1.0e30 ;
    sumloudmax[i] = -1.0e30 ;
    modloudmax[i] = -1.0e30 ;
    modloudmin[i] = 1.0e30 ;
    zcratemin[i] = 1.0e30 ;
  } ;

#ifdef TESTLEVEL
  fprintf( stdout, "Determine minima and maxima of special features ...\n" ) ;
#endif

  float **ModLoudSort, **SumLoudSort, **ZCRateSort ;
  
  // determine speech min resp max of special features (for offline normalization to get features for spdet)
  if( NormSpecialFeatChanPerc.Par != 0.0 ) {
    ModLoudSort = new float*[MAXNRFRAMEFORMAXDET] ; // sort modloudnesses of all channels to get 10% off max value
    SumLoudSort = new float*[MAXNRFRAMEFORMAXDET] ; // sort modloudnesses of all channels to get 10% off max value
    //ZCRateSort = new float*[MAXNRFRAMEFORMAXDET] ; // sort modloudnesses of all channels to get 10% off max value
  } ;

  frame = 0 ;
  int running = 1 ;
  while( running ) { 

    if( NormSpecialFeatChanPerc.Par != 0.0 ) {
      // allocate memory
      ModLoudSort[frame] = new float[nrchanwav] ;
      SumLoudSort[frame] = new float[nrchanwav] ;
      //ZCRateSort[frame] = new float[nrchanwav] ;
    } ;
    
    // check if any of the sigfiles faces EOF yet
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      if( (prefp[chan] = AkuPre[chan]->getNextPre()) == 0 ) {
	running = 0 ;
	break ;
      } ;
      
      if( NormSpecialFeatChanPerc.Par != 0.0 ) {
	ModLoudSort[frame][chan] = 1.0e30 ;
	SumLoudSort[frame][chan] = 1.0e30 ;
	//ZCRateSort[frame][chan] = 1.0e30 ;
      } ;
      
      //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( frame > 6000 && frame < 36000 ) {
	if( prefp[chan]->TotalEnergy < energymin[chan] )
	  energymin[chan] =  prefp[chan]->TotalEnergy ;
	//if( prefp[chan]->ModLoudTimesZCRate < entropymin[chan] )
	//entropymin[chan] = prefp[chan]->ModLoudTimesZCRate ;
	if( prefp[chan]->ModLoud < modloudmin[chan] )
	  modloudmin[chan] = prefp[chan]->ModLoud ;
	if( prefp[chan]->ZCRate < zcratemin[chan] )
	  zcratemin[chan] = prefp[chan]->ZCRate ;
       }
      if( prefp[chan]->SumLoud > sumloudmax[chan] )
	sumloudmax[chan] = prefp[chan]->SumLoud ;
      if( prefp[chan]->ModLoud > modloudmax[chan] )
	modloudmax[chan] = prefp[chan]->ModLoud ;
      
      if( NormSpecialFeatChanPerc.Par != 0.0 ) {
	// add modloud to sorted array
	for( i=0 ; i<=frame ; i++ ) {
	  if( prefp[chan]->ModLoud < ModLoudSort[i][chan] ) {
	    // schiebe all eins nach hinten!
	    for( int k=frame ; k>i ; k-- )
	      ModLoudSort[k][chan] = ModLoudSort[k-1][chan] ;
	    ModLoudSort[i][chan] = prefp[chan]->ModLoud ;
	    break ;
	  } ;
	} ;
	
	// add sumloud to sorted array
	for( i=0 ; i<=frame ; i++ ) {
	  if( prefp[chan]->SumLoud < SumLoudSort[i][chan] ) {
	    // schiebe alle eins nach hinten!
	    for( int k=frame ; k>i ; k-- )
	      SumLoudSort[k][chan] = SumLoudSort[k-1][chan] ;
	    SumLoudSort[i][chan] = prefp[chan]->SumLoud ;
	    break ;
	  } ;
	} ;

	// add zcrate to sorted array
	//for( i=0 ; i<=frame ; i++ ) {
	//if( prefp[chan]->ZCRate < ZCRateSort[i][chan] ) {
	    // schiebe alle eins nach hinten!
	//  for( int k=frame ; k>i ; k-- )
	//ZCRateSort[k][chan] = ZCRateSort[k-1][chan] ;
	//  ZCRateSort[i][chan] = prefp[chan]->ZCRate ;
	// break ;
	//} ;
	//} ;
      } ;
	
    } ;
      
    if( running == 0 )
      break ;

    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 )
      break ;
    //if( prefp[chan]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
    if( frame > 6000 && frame < 36000 ) {
      if( prefpmixfile->TotalEnergy < energymin[nrchanwav] )
	energymin[nrchanwav] =  prefpmixfile->TotalEnergy ;
      //if( prefpmixfile->ModLoudTimesZCRate < entropymin[nrchanwav] )
      //entropymin[nrchanwav] = prefpmixfile->ModLoudTimesZCRate ;
      if( prefpmixfile->ModLoud < modloudmin[nrchanwav] )
	modloudmin[nrchanwav] = prefpmixfile->ModLoud ;
      if( prefpmixfile->ZCRate < zcratemin[nrchanwav] )
	zcratemin[nrchanwav] = prefpmixfile->ZCRate ;
    }    
    if( prefpmixfile->SumLoud > sumloudmax[nrchanwav] )
      sumloudmax[nrchanwav] = prefpmixfile->SumLoud ;
    if( prefpmixfile->ModLoud > modloudmax[nrchanwav] )
      modloudmax[nrchanwav] = prefpmixfile->ModLoud ;

    if( ++frame > MAXNRFRAMEFORMAXDET )
      break ;
  } ;
#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  int usedframes = frame ; // memorize how many frames were used for max-detection!!! (use for percentiles!)

  // overwrite max-values with percentiles, if requested!
  if( NormSpecialFeatChanPerc.Par != 0.0 ) {
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      sumloudmax[chan] = SumLoudSort[usedframes-((int)(usedframes*NormSpecialFeatChanPerc.Par/100))][chan] ; 
      modloudmax[chan] = ModLoudSort[usedframes-((int)(usedframes*NormSpecialFeatChanPerc.Par/100))][chan] ; 
      //zcratemax[chan] = ZCRateSort[usedframes-((int)(usedframes*NormSpecialFeatChanPerc.Par/100))][chan] ; 
    } ;
  } ;
 



#ifdef TESTLEVEL
  // output of min energies. max sumloud, max modloud, max zcrate: 
  for( chan=0 ; chan<nrchanwav ; chan++ ) {
    fprintf( stdout, "channel %d: Min energy: %f\n", chan, energymin[chan] ) ;
    fprintf( stdout, "channel %d: Max sumloud: %f\n", chan, sumloudmax[chan] ) ;
    fprintf( stdout, "channel %d: Max modloud: %f\n", chan, modloudmax[chan] ) ;
    fprintf( stdout, "channel %d: Min modloud: %f\n", chan, modloudmin[chan] ) ;
    fprintf( stdout, "channel %d: Min zcrate: %f\n", chan, zcratemin[chan] ) ;
  } ;
  fprintf( stdout, "mixfile: Min energy: %f\n", energymin[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Max sumloud: %f\n",  sumloudmax[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Max modloud: %f\n", modloudmax[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Min modloud: %f\n", modloudmin[nrchanwav] ) ;
  fprintf( stdout, "mixfile: Min zcrate: %f\n",  zcratemin[nrchanwav] ) ;
#endif

  //#ifdef TESTLEVEL
  //fprintf(stderr, "ModLoudSort-Array:\n" ) ;
  //for( i=0 ;i<frame ; i++ )
  //for( chan=0 ; chan<nrchanwav; chan++)
  //fprintf(stderr, "%f\n", ModLoudSort[frame][chan] ) ;
  //#endif


#ifdef TESTLEVEL
  fprintf( stdout, "\nDetermine length of lab-files ...\n" ) ;
#endif
  // how long are files expected to be?
  int minlabfilelen = Lab[0]->getEndFrameOfLastEntry() ;
  for( chan=1 ; chan<nrchanwav ; chan++ ) {
    if( Lab[chan]->getEndFrameOfLastEntry() < minlabfilelen )
      minlabfilelen = Lab[chan]->getEndFrameOfLastEntry() ;
  } ;
  long nrframe = minlabfilelen ;
#ifdef TESTLEVEL
  fprintf( stdout, "... length of shortest labfile is %d frames\n", minlabfilelen ) ;
  fprintf( stdout, "... DONE\n\n" ) ;
#endif



#ifdef TESTLEVEL
  if( NormSpecialFeatChan.Par || NormSpecialFeatChanPerc.Par != 0.0 )
    fprintf( stdout, "Normalize Special features of individual channels and update SpNonSp-models ...\n" ) ;
  if( NormSpecialFeatMix.Par )
    fprintf( stdout, "Normalize Special features of mixfile and update SpNonSp-models ...\n" ) ;
#endif


#ifdef TESTLEVEL
  fprintf( stdout, "... Clear and restart AkuPre of all %d channels ... \n", nrchanwav ) ;
#endif
  // prepare to store framewise spnonsp-info
  // reset acoustic preprocessing for channels
  for( chan=0 ; chan<nrchanwav ; chan++) {
#ifdef TESTLEVEL
    fprintf( stdout, "... %d\n", chan  ) ;
#endif
    // clear AkuPre
    AkuPre[chan]->clearAkuPre() ;
    if( AkuPre[chan]->startAkuPre( argv[(chan)+wavfileoffset] ) == _ERR_ ) {
      fprintf( stderr, "Can't open %d-th  sigfile '%s'\n", chan, argv[(chan)+wavfileoffset] ) ;
      exit(-1) ;
    } ;
  } ;
  // clear AkuPre
  AkuPreMixFile->clearAkuPre() ;
  if( AkuPreMixFile->startAkuPre( argv[mixfileindex] ) == _ERR_ ) {
    fprintf( stderr, "Can't open mixfile '%s'\n", argv[mixfileindex] ) ;
    exit(-1) ;
  } ;


  // do loop over all channels (including mixfile) and normalize special features
  for( frame=0 ; frame<nrframe ; frame++ ) {
    // get feature vecs, prepare feature vecs by normalizing
    meanenergy = 0.0 ;
    for( chan=0 ; chan<nrchanwav ; chan++) {
      prefp[chan] = AkuPre[chan]->getNextPre() ;
      if( prefp[chan] == 0 ) {
	fprintf( stderr, "Unexpected EOF reached for file '%s'\n", argv[(chan)+wavfileoffset] ) ;
	break ;
      } ;

      // anyway: overwrite entropy, which did not seem to be useful at all, use prduct of modloud and zcrate instead
      prefp[chan]->ModLoudTimesZCRate = (prefp[chan]->ZCRate - zcratemin[chan]) * ((prefp[chan]->ModLoud - modloudmin[chan])/(modloudmax[chan]-modloudmin[chan])) ;
      if( NormSpecialFeatChan.Par || NormSpecialFeatChanPerc.Par != 0.0 ) {
	// normalize channel features: 
	prefp[chan]->TotalEnergy -= energymin[chan] ;
	prefp[chan]->SumLoud /= sumloudmax[chan] ;
	//prefp[chan]->ZCRate /= zcratemax[chan] ;    
	prefp[chan]->ModLoud /= modloudmax[chan] ;
      } ;

  
      // calc mean energy over all channel energies
      meanenergy += prefp[chan]->TotalEnergy ;
    } ;
    // normalize mixfile features:
    prefpmixfile = AkuPreMixFile->getNextPre() ;
    if( prefpmixfile == 0 ) {
	fprintf( stderr, "Unexpected EOF reached for file '%s'\n", argv[mixfileindex] ) ;
	break ;
    } ;
    // anyway: overwrite entropy, which did not seem to be useful at all, use prduct of modloud and zcrate instead
    prefpmixfile->ModLoudTimesZCRate = (prefpmixfile->ZCRate - zcratemin[nrchanwav]) * ((prefpmixfile->ModLoud - modloudmin[nrchanwav])/(modloudmax[nrchanwav]-modloudmin[nrchanwav])) ;
    if( NormSpecialFeatMix.Par ) {
      prefpmixfile->TotalEnergy -= energymin[nrchanwav] ;
      prefpmixfile->SumLoud /= sumloudmax[nrchanwav] ;
      //prefpmixfile->ZCRate /= zcratemax[nrchanwav] ; 
      prefpmixfile->ModLoud /= modloudmax[nrchanwav] ;
    } ;


    // calc mean energy:
    meanenergyohnemix = meanenergy ;
    meanenergyohnemix /= nrchanwav ;
    meanenergy += prefpmixfile->TotalEnergy ;
    meanenergy /= (nrchanwav+1) ;

    // normalize energies with mean over channels
    if( NormSpecialFeatChan.Par || NormSpecialFeatChanPerc.Par != 0.0 ) {
      for( chan=0 ; chan<nrchanwav ; chan++) {
	prefp[chan]->TotalEnergy -= meanenergyohnemix ;
	//(prefp[chan]->ModLoudTimesZCRate -= entropymin[chan]) *= pow( 10.0,(prefp[chan]->TotalEnergy)/10) ;
      	//prefp[chan]->ModLoudTimesZCRate *= (prefp[chan]->ModLoud) ;
      } ;
    } ;
    if( NormSpecialFeatMix.Par ) {
      // normalize with mean over channels
      prefpmixfile->TotalEnergy -= meanenergyohnemix ;
      //(prefpmixfile->ModLoudTimesZCRate -= entropymin[nrchanwav]) *= pow(10.0,(prefpmixfile->TotalEnergy/10)) ;
      //prefpmixfile->ModLoudTimesZCRate *= (prefpmixfile->ModLoud) ;
    } 
    // do training on current frame:
    // first decide if there is speech or nonspeech on indiv channels -> if there is speech in any channel,
    // it is assumed that there is also speech in the mixfile!!!
    int speechmix = 0 ;
    for( chan=0 ; chan<nrchanwav ; chan++ ) {
      int speechchan = 0 ;
      // is it speech in this channel?
      for( run=0 ; run<nr_runs[chan] ; run++ ) {
	if( frame >= start[chan][run] && frame <= stop[chan][run] ) {
	  // speech in this chan found
	  speechchan = 1 ;
	  if( SpdAutoUpdateSpeech.Par ) {
#ifdef TESTLEVEL
	    fprintf( stderr, "[%d]: chan%d: AUTO update SPEECH (%f)\n", frame, chan, prefp[chan]->ModLoud ) ;
#endif
	    // automatic update
	    SpNonSp[chan]->update( (float*)prefp[chan], S ) ;
	  } else {
	    // find matching mixture number:
	    char *labtmp = Lab[chan]->getLabelName( frame ) ;
	    if( labtmp != 0 ) { 
	      // valid label found
	      for( i=0 ; i<NrSpLabels ; i++ ) {
		if( strstr( labtmp, SpLabel[i] ) != 0 ) {
#ifdef TESTLEVEL
		  fprintf( stderr, "[%d]: chan%d: update SPEECH, mix %d (%f)\n", frame, chan, i, prefp[chan]->ModLoud ) ;
#endif
		  SpNonSp[chan]->updateOneMix( (float*)prefp[chan], S, i ) ;
		} ;
	      } ;
	    } ;
	  } ;
	  break ; // matching run found, leave loop over all runs!
	} ;
      } ; 
      speechmix |= speechchan ;
      // don't do nonspeech-update of lapel-channel (chan0)
      if( !speechchan && chan != 0) {
	// no speech in this channel:
	if( SpdAutoUpdateSilent.Par ) {
	  // automatic update
	  SpNonSp[chan]->update( (float*)prefp[chan], P ) ;
#ifdef TESTLEVEL
	  fprintf( stderr, "[%d]: chan%d: AUTO update NONSPEECH (%f) \n", frame, chan, prefp[chan]->ModLoud ) ;
#endif
	} else {
	  // find matching mixture number:
	  char *labtmp = Lab[chan]->getLabelName( frame ) ;
	  if( labtmp != 0 ) { 
	    // valid label found
	    for( i=0 ; i<NrSiLabels ; i++ ) {
	      if( strstr( labtmp, SiLabel[i] ) != 0 ) {
		SpNonSp[chan]->updateOneMix( (float*)prefp[chan], P, i ) ; 
#ifdef TESTLEVEL
		fprintf( stderr, "[%d]: chan%d: update NONSPEECH, mix %d (%f)\n", frame, chan, i, prefp[chan]->ModLoud) ;
#endif
	      } ;
	    } ;
	  } ;
	} ; // END manual update
      } ; // END if no speech in this channel
      
    } ; // END loop over all channels to decide whether speech or nonspeech!
    if( speechmix ) {
      // only automatic update possible: speech
#ifdef TESTLEVEL
      fprintf( stderr, "[%d]: mixfile: AUTO update SPEECH (%f)\n", frame, prefpmixfile->ModLoud) ;
#endif
      SpNonSpMixFile->update( (float*)prefpmixfile, S ) ;

    } else {
      // only automatic update possible: nonspeech
#ifdef TESTLEVEL
      fprintf( stderr, "[%d]: mixfile: AUTO update NONSPEECH (%f)\n", frame, prefpmixfile->ModLoud) ;
#endif
      SpNonSpMixFile->update( (float*)prefpmixfile, P ) ;
    } ;
  } ; // END loop over all frames

#ifdef TESTLEVEL
  fprintf( stdout, "... DONE\n\n" ) ;
#endif

  FILE *outfp ;
  // output of new chan-cfg-files:
  for( chan=0 ;chan<nrchanwav ; chan++ ) {
    sprintf( tmp, "spnonsp-chan%d.cfg", chan ) ;
    outfp = fopen( tmp, "w" ) ;
    SpNonSp[chan]->appendPar( outfp ) ;
    fclose( outfp ) ;
  } ;
  
  // output new mix-cfg-file
  sprintf( tmp, "spnonsp-mix.cfg" ) ;
  outfp = fopen( tmp, "w" ) ;
  SpNonSpMixFile->appendPar( outfp ) ;
  fclose( outfp ) ;
      
#ifdef TESTLEVEL
  fprintf( stdout, "Free memory ...\n", nrchanwav, argv[argc-1] ) ;
#endif

    
  for( i=0 ; i<nrchanwav ; i++ ) {
    AkuPre[i]->clearAkuPre();
    delete AkuPre[i] ;
    if( !SpDetUseCommonForAllChan.Par ) 
      delete SpNonSp[i] ;
  } ;
  if( !SpDetUseCommonForAllChan.Par ) 
    delete SpNonSp[0] ;

  delete[] SpNonSp ;
  delete[] AkuPre ;

  AkuPreMixFile->clearAkuPre() ;
  delete   AkuPreMixFile ;
  
  delete[] energymin ;
  delete[] prefp ;
  
  //AkuPre.appendPar( stderr ) ;
#ifdef TESTLEVEL
  fprintf( stdout, "Done\n", nrchanwav, argv[argc-1] ) ;
#endif
  return _OK_;
}
