/******************************************************************************
 * file:        specdoub.h  (for specdoub.cc)
 * info:        class declarations for CSpecDoubleBuf 
 * written by:  M.Beham, 12.01.95 (org. from fft.c, sts.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _SPECDOUB_H_ )
#define _SPECDOUB_H_

// object members included in CSpecDoubleBuf
#include "hamming.h"
#include "fft.h"

// class definition for CSpecDoubleBuf

class CSpecDoubleBuf
{
private:
    const unsigned int WinLen;         // length of signal-window for one fft
    const unsigned int FFTLen;         // length of fft-vector (FFTLen >= WinLen)
                                    // signal-window and spectral-buffer (double FFT)
    float            *DoubleBuf[2];   // 2 vectors of length WinLen
    float             ZCRateBuf[2];   // 2 zero-crossing-rates for 2 windows
                                    // member objects
    CHammingWindow    HamWin;         // hamming window and dc supress
    CFastFourier      FFT;            // FFT butterfly tabele 
                                   // private functions
    void  fillZero();
    void  unfoldPowerSpec();
    void  calcDC( float *s0, float *s1 );
public:
    CSpecDoubleBuf( unsigned int winlen, unsigned int fftlen );
    ~CSpecDoubleBuf();
    float *getBuffer( unsigned int index ) const { return DoubleBuf[index]; }
    float  getZCRate( unsigned int index ) const { return ZCRateBuf[index]; }
    void   setNorm( double norm ) { HamWin.setNorm(norm); }
    void   calcDoubleSpec();
};    


#endif

