/******************************************************************************
 * file:        spnonsp.cc  
 * info:        implementation of CSpeechNonSpeech, speech pause detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <math.h>
#include <string.h>

// #define DEBUG_LEVEL 2
// #define TEST_LEVEL 
#include "fileconst.h"

#include "dbg_mess.h"
#include "spnonsp.h"

#define TEST
//#define TEST_HIGHLEVEL


#define  MIN(a,b)  ( a < b ? a : b )
#define  MAX(a,b)  ( a > b ? a : b )

//--------------------------- public 'CProbGaussVar' --------------------------

// constructor
CProbGaussVar::CProbGaussVar( unsigned int len, double dt, FILE *fp, char *key, int shorttimeflag, int nr ) 
  :   Len(len),
    FlagShortTime(shorttimeflag),
    ParFile(fp),
    TauUpdate( "", 1.0  ),
    Mean( "", len, (float)0.0 ),
    Var(  "", len, (float)0.0 )
{
    char keyword[MAXKEYWORDLEN+1], *ext;
    strncpy(keyword, key, MAXKEYWORDLEN );
    ext = keyword + strlen(keyword);

    // read gauss parameter
    strncpy( ext, "UpdateTau", MAXKEYWORDLEN-strlen(keyword) );
    TauUpdate.setKeyWord(keyword);
    //strncpy( ext, "Mean", MAXKEYWORDLEN-strlen(keyword) );
    sprintf( ext, "%d.Mean", nr ) ;
    Mean.setKeyWord(keyword);
    //fprintf( stderr, "%s\n", keyword ) ;
    sprintf( ext, "%d.Var", nr ) ;
    //strncpy( ext, "Var", MAXKEYWORDLEN-strlen(keyword) );
    Var.setKeyWord(keyword);
    //fprintf( stderr, "%s\n", keyword ) ;
    TauUpdate.readPar(fp);  
    Mean.readPar(fp);  
    Var.readPar(fp);  

    // calculate factors for update constants
    if( FlagShortTime )
        Tf = 1.0 - exp( -dt / TauUpdate.Par );
    else
        Tf = 1.0;
    // calc. determinants
    TwoPiPowerN = pow( 2.0*M_PI, Len );
    double det=1.0;
    for( unsigned int i=0; i<Len; i++ ) 
        det *= Var.Vec[i];
    LogDet = -0.5 * log( TwoPiPowerN * det );
}

// calc log of prop density
double CProbGaussVar::getLogProb( float *xvec )
{
    register unsigned int i;
    // calc distance
    double  dis=0.0;
    for( i=0; i<Len; i++ )
        dis += (0.5 * (xvec[i]-Mean.Vec[i]) * (xvec[i]-Mean.Vec[i])) / Var.Vec[i];   
    // calc log. of prob. density
    return  LogDet - dis;
}


// update gauss function
void   CProbGaussVar::updateProb( float *xvec )
{
    register unsigned int i;
    // update mean and var
    double  af = 1.0 - Tf, bf = af * Tf;
    for( i=0; i<Len; i++ )
    {   Var.Vec[i]  = af *  Var.Vec[i]  + bf * (xvec[i]-Mean.Vec[i]) * (xvec[i]-Mean.Vec[i]);
        Mean.Vec[i] = af * Mean.Vec[i]  + Tf * xvec[i];
    }
    af=1.0;
    for( i=0; i<Len; i++ ) 
        af *= Var.Vec[i];
    LogDet = -0.5 * log( TwoPiPowerN * af );

    // for recursiv mode update constants (=> M = M+1)
    if( !FlagShortTime )
       Tf = Tf / (1.0+Tf);
}

// write gauss function parameter: mean, var
int    CProbGaussVar::writePar( FILE *fp )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' write parameter called" );

    if( Mean.writePar( fp ) == _ERR_ ||
         Var.writePar( fp ) == _ERR_ )
        return _ERR_;
    else
        return _OK_;
}

// append fix gauss function parameter: TauUpdate
int    CProbGaussVar::appendFixPar( FILE *fp )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' appendFIXparameter called" );
    
    int ret = _OK_ ;

    if( TauUpdate.appendPar( fp ) == _ERR_ )
      ret = _ERR_;
    
    return ret ;
}

// append updatable gauss function parameter: mean, var
int    CProbGaussVar::appendUpdPar( FILE *fp )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' appendUPDparameter called" );

fprintf(stderr, "'CSpeechNonSpeech' appendUPDparameter called" );

    if( Mean.appendPar( fp ) == _ERR_ ||
         Var.appendPar( fp ) == _ERR_ )
        return _ERR_;
    else
        return _OK_;
}

// write gauss function parameter: mean, var
int    CProbGaussVar::writePar( )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' write parameter called" );

    if( Mean.writePar( ParFile ) == _ERR_ ||
         Var.writePar( ParFile ) == _ERR_ )
        return _ERR_;
    else
        return _OK_;
}

//--------------------------- public 'CObjStochasticState' --------------------

// constructor
CObjStochasticState::CObjStochasticState( float *pp,  CObjStochasticState *state, ESpeechFlag type )
:   PropDens( pp )
{
    if( state != 0 )
        PreStateList.addTail(state);
    Alpha = TmpAlpha = -HUGE;
    SpnonspType = type ; // tpfau, 12/21/2000
}

// constructor
CObjStochasticState::CObjStochasticState( float *pp, CObjStochasticState *state1, CObjStochasticState *state2, ESpeechFlag type )
:   PropDens( pp )
{
    if( state1 != 0 )
        PreStateList.addTail(state1);
    if( state2 != 0 )
        PreStateList.addTail(state2);
    Alpha = TmpAlpha = -HUGE;
    SpnonspType = type ; // tpfau, 12/21/2000
}

CObjStochasticState::~CObjStochasticState() {
    ;
}

// calc new alpha recursion and return score
float CObjStochasticState::calcAlpha( )
{
    register PListEntry pos = 0;
    register CObjStochasticState *preState = (CObjStochasticState *)PreStateList.getHeadEntry(pos);
    TmpAlpha = preState->getAlpha();     
    while( (preState = (CObjStochasticState *)PreStateList.getNextEntry(pos)) != 0 )
        if( preState->getAlpha() > TmpAlpha )
            TmpAlpha = preState->getAlpha();
    return TmpAlpha += *PropDens;
}

//--------------------------- public 'CSpeechNonSpeech' -----------------------

// constructor: CSpeechNonSpeech
CSpeechNonSpeech::CSpeechNonSpeech( unsigned int len, double dt, FILE *fp )
:   UpdateFlag( "SpDet.Update.MeansAndVars", fp,  FALSE ),
    UpdateShortTime( "SpDet.Update.ShortTime", fp, FALSE ),
    UpdateWrite( "SpDet.Update.Write", fp, FALSE ),
    UpdateUseSp( "SpDet.Update.Speech.Use", ESPEECHFLAG, fp, S ),
    UpdateUseSi( "SpDet.Update.Silent.Use", ESPEECHFLAG, fp, P ),
    ThresSp( "SpDet.Speech.Threshold", fp, 0.5 ),
    ThresSi( "SpDet.Silent.Threshold", fp, 0.5 ),
    MinTimeSp( "SpDet.Speech.MinTime", fp, 0.2 ),
    MinTimeSi( "SpDet.Silent.MinTime", fp, 0.2 ),
    //PropLogSp( len, dt, fp, "SpDet.Speech.", UpdateShortTime.Par),
    //PropLogSi( len, dt, fp, "SpDet.Silent.", UpdateShortTime.Par)
    NrGaussSp( "SpDet.Speech.NrGauss", fp, 1 ),
    NrGaussSi( "SpDet.Silent.NrGauss", fp, 1 ) 

{
    DEBUG_1_MSG("'CSpeechNonSpeech' constructor called" );

    int i ;
    char keyword[MAXKEYWORDLEN] ;

    // allocate array of gaussian densities for speech and silence
    PropLogSp = new CProbGaussVar*[NrGaussSp.Par] ;
    PropLogSi = new CProbGaussVar*[NrGaussSi.Par] ;
    for( i=0 ;i<NrGaussSp.Par ; i++ ) {
      sprintf( keyword, "SpDet.Speech." ) ;
      PropLogSp[i] = new CProbGaussVar( len, dt, fp, keyword, UpdateShortTime.Par, i ) ;
    } ;
    for( i=0 ;i<NrGaussSi.Par ; i++ ) {
      sprintf( keyword, "SpDet.Silent." ) ;
      PropLogSi[i] = new CProbGaussVar( len, dt, fp, keyword, UpdateShortTime.Par,i ) ;
    } ;
    
    // build special stochastic model constants
    CountSiToSp = MAX((int) (MinTimeSp.Par/dt), 1 );
    CountSpToSi = MAX((int) (MinTimeSi.Par/dt), 1 );
    CountSpToSp = MIN( CountSpToSi, CountSiToSp );
    CountSiToSi = MIN( CountSpToSi, CountSiToSp );
    TotalStates = 2 + CountSiToSp + CountSiToSi + CountSpToSi + CountSpToSp;
    CountThres = 0;
    LastFlag = P;

    // model state array
    typedef CObjStochasticState * PState;
    StateArray = new PState[TotalStates];
    IF_FATAL_EXIT( StateArray==0, "'CSpeechNonSpeech' constructor out of memory" );

    // build model structure (only here)
  
    // Sp main state, prob. for Sp
    unsigned int j, n=0;
    SpMain = StateArray[n] = new CObjStochasticState( &LogSp, NULL, (ESpeechFlag)P );
    SpMain->addPreState( SpMain );
    n++;
 
    // Si' states from Sp to Si, prob. for Si    
    for( i=0; i<CountSpToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], (ESpeechFlag)P );
    n += CountSpToSi;

    // Sp' states from Sp to Sp, prob. for Sp
    j = n - ((CountSpToSi - CountSpToSp) + 1);
    StateArray[n] = new CObjStochasticState( &LogSp, StateArray[j], (ESpeechFlag)S );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSpToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)S );
    n += CountSpToSp;
    SpMain->addPreState( StateArray[n-1] );

    // Si main state, prob for Si
    SiMain = StateArray[n] = new CObjStochasticState( &LogSi, StateArray[n-(CountSpToSp+1)], (ESpeechFlag)P );
    SiMain->addPreState( SiMain );
    n++;

    // Sp' states from Si to Sp, prob. for Sp
    for( i=0; i<CountSiToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], (ESpeechFlag)S );
    n += CountSiToSp;

    // Si' states from Si to Si, prob. for Si
    j = n - ((CountSiToSp - CountSiToSi) + 1);
    StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)P );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSiToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)P );
    n += CountSiToSi;
    SiMain->addPreState( StateArray[n-1] );

    // close ergodic model - loop
    SpMain->addPreState( StateArray[n-(CountSiToSi+1)] );
    IF_FATAL_EXIT( n != TotalStates, "'CSpeechNonSpeech' constructor internal model-structure error" );
    
    // test memory alloc
    for( i=0; i<TotalStates; i++ )
        IF_FATAL_EXIT( StateArray[i] == 0, "'CSpeechNonSpeech' constructor out of memory")
}


// special constructor of CSpeechNonSpeech, no statearray is made when buildarray is set to zero
CSpeechNonSpeech::CSpeechNonSpeech( unsigned int len, double dt, FILE *fp, int buildarray )
:   UpdateFlag( "SpDet.Update.MeansAndVars", fp,  FALSE ),
    UpdateShortTime( "SpDet.Update.ShortTime", fp, FALSE ),
    UpdateWrite( "SpDet.Update.Write", fp, FALSE ),
    UpdateUseSp( "SpDet.Update.Speech.Use", ESPEECHFLAG, fp, S ),
    UpdateUseSi( "SpDet.Update.Silent.Use", ESPEECHFLAG, fp, P ),
    ThresSp( "SpDet.Speech.Threshold", fp, 0.5 ),
    ThresSi( "SpDet.Silent.Threshold", fp, 0.5 ),
    MinTimeSp( "SpDet.Speech.MinTime", fp, 0.2 ),
    MinTimeSi( "SpDet.Silent.MinTime", fp, 0.2 ),
    //PropLogSp( len, dt, fp, "SpDet.Speech.", UpdateShortTime.Par),
    //PropLogSi( len, dt, fp, "SpDet.Silent.", UpdateShortTime.Par)
    NrGaussSp( "SpDet.Speech.NrGauss", fp, 1 ),
    NrGaussSi( "SpDet.Silent.NrGauss", fp, 1 ) 

{
    DEBUG_1_MSG("'CSpeechNonSpeech' constructor called" );

    int i ;
    char keyword[MAXKEYWORDLEN] ;

    // allocate array of gaussian densities for speech and silence
    PropLogSp = new CProbGaussVar*[NrGaussSp.Par] ;
    PropLogSi = new CProbGaussVar*[NrGaussSi.Par] ;
    for( i=0 ;i<NrGaussSp.Par ; i++ ) {
      sprintf( keyword, "SpDet.Speech." ) ;
      PropLogSp[i] = new CProbGaussVar( len, dt, fp, keyword, UpdateShortTime.Par, i ) ;
    } ;
    for( i=0 ;i<NrGaussSi.Par ; i++ ) {
      sprintf( keyword, "SpDet.Silent." ) ;
      PropLogSi[i] = new CProbGaussVar( len, dt, fp, keyword, UpdateShortTime.Par,i ) ;
    } ;
    
    // build special stochastic model constants
    CountSiToSp = MAX((int) (MinTimeSp.Par/dt), 1 );
    CountSpToSi = MAX((int) (MinTimeSi.Par/dt), 1 );
    CountSpToSp = MIN( CountSpToSi, CountSiToSp );
    CountSiToSi = MIN( CountSpToSi, CountSiToSp );
    TotalStates = 2 + CountSiToSp + CountSiToSi + CountSpToSi + CountSpToSp;
    CountThres = 0;
    LastFlag = P;

    // model state array
    typedef CObjStochasticState * PState;
    if( buildarray == 1 ) {
      StateArray = new PState[TotalStates];
      IF_FATAL_EXIT( StateArray==0, "'CSpeechNonSpeech' constructor out of memory" );
      
      // build model structure (only here)
      
      // Sp main state, prob. for Sp
      unsigned int j, n=0;
      SpMain = StateArray[n] = new CObjStochasticState( &LogSp, NULL, (ESpeechFlag)P );
      SpMain->addPreState( SpMain );
      n++;
      
      // Si' states from Sp to Si, prob. for Si    
      for( i=0; i<CountSpToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], (ESpeechFlag)P );
      n += CountSpToSi;
      
      // Sp' states from Sp to Sp, prob. for Sp
      j = n - ((CountSpToSi - CountSpToSp) + 1);
      StateArray[n] = new CObjStochasticState( &LogSp, StateArray[j], (ESpeechFlag)S );
      for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
      for( i=1; i<CountSpToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)S );
      n += CountSpToSp;
      SpMain->addPreState( StateArray[n-1] );
      
      // Si main state, prob for Si
      SiMain = StateArray[n] = new CObjStochasticState( &LogSi, StateArray[n-(CountSpToSp+1)], (ESpeechFlag)P );
      SiMain->addPreState( SiMain );
      n++;
      
      // Sp' states from Si to Sp, prob. for Sp
      for( i=0; i<CountSiToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], (ESpeechFlag)S );
      n += CountSiToSp;
      
      // Si' states from Si to Si, prob. for Si
      j = n - ((CountSiToSp - CountSiToSi) + 1);
      StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)P );
      for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
      for( i=1; i<CountSiToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)P );
      n += CountSiToSi;
      SiMain->addPreState( StateArray[n-1] );
      
      // close ergodic model - loop
      SpMain->addPreState( StateArray[n-(CountSiToSi+1)] );
      IF_FATAL_EXIT( n != TotalStates, "'CSpeechNonSpeech' constructor internal model-structure error" );
      
      // test memory alloc
      for( i=0; i<TotalStates; i++ )
        IF_FATAL_EXIT( StateArray[i] == 0, "'CSpeechNonSpeech' constructor out of memory") ;
    } else if( buildarray == 0 ) {
      StateArray = 0 ;
      SpMain = 0 ;
      SiMain = 0 ;
    } ;
}


// destructor: CSpeechNonSpeech
CSpeechNonSpeech::~CSpeechNonSpeech( )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' destructor called" );
    int i ;
    if( UpdateWrite.Par )
    { 
      //NrGaussSp.writePar() ;
      //NrGaussSi.writePar() ;
      for( i=0 ; i<NrGaussSp.Par ; i++ )
	PropLogSp[i]->writePar();
      for( i=0 ; i<NrGaussSi.Par ; i++ )
	PropLogSi[i]->writePar();
    }

    // delete all states
    for( unsigned int i=0; i<TotalStates; i++ )
    {   StateArray[i]->clearPreState();
        delete StateArray[i];
    }
    if( StateArray != 0 )
      delete [] StateArray;

    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogSp[i] != 0 ) 
	delete PropLogSp[i] ;
    for( i=0 ; i<NrGaussSi.Par ; i++ )
      if( PropLogSi[i] != 0 )
	delete PropLogSi[i] ;
    if( PropLogSp != 0 )
	delete[] PropLogSp ;
    if( PropLogSi != 0 )
      delete[] PropLogSi ;
}

// write parameter to file
int  CSpeechNonSpeech::writePar( FILE *fp )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' write parameter called" );

    NrGaussSp.writePar(fp) ;
    NrGaussSi.writePar(fp) ;
  
    int i ;
    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogSp[i]->writePar( fp ) == _ERR_ )
        return _ERR_;
        
    for( i=0 ; i<NrGaussSi.Par ; i++ ) 
      if( PropLogSi[i]->writePar( fp ) == _ERR_ )
	return _ERR_ ;

    
    return _OK_;

}

/////////////////////////////////////////////////////////////////////////
//
// append fix params
// 
/////////////////////////////////////////////////////////////////////////
int  CSpeechNonSpeech::appendFixPar( FILE *fp ) {
    
    int ret = 0 ; 
    ret |= UpdateFlag.appendPar( fp ) ;
    ret |= UpdateShortTime.appendPar( fp ) ;
    ret |= UpdateWrite.appendPar( fp ) ;
    ret |= UpdateUseSp.appendPar( fp ) ;
    ret |= UpdateUseSi.appendPar( fp ) ;
    ret |= NrGaussSp.appendPar(fp) ;
    ret |= NrGaussSi.appendPar(fp) ;
    ret |= PropLogSp[0]->appendFixPar( fp ) ;
    ret |= PropLogSi[0]->appendFixPar( fp ) ;
    ret |= ThresSp.appendPar( fp ) ;
    ret |= ThresSi.appendPar( fp ) ;
    ret |= MinTimeSp.appendPar( fp ) ;
    ret |= MinTimeSi.appendPar( fp ) ;

    return( ret == 0 ? ret : -1 ) ;
} ; // END appendFixPar

/////////////////////////////////////////////////////////////////////////
//
// append updatable params
// 
/////////////////////////////////////////////////////////////////////////
int  CSpeechNonSpeech::appendUpdPar( FILE *fp ) {

    DEBUG_1_MSG("'CSpeechNonSpeech' append updatable parameter called" );

    int i ;
    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogSp[i]->appendUpdPar( fp ) == _ERR_ )
	return( _ERR_ ) ;
   
    for( i=0 ; i<NrGaussSi.Par ; i++ )
      if( PropLogSi[i]->appendUpdPar( fp ) == _ERR_ )
        return _ERR_;
    
    return _OK_;

}




// start alpha recursion
void  CSpeechNonSpeech::start()
{
    for( unsigned int i=0; i<TotalStates; i++ )
        StateArray[i]->setAlpha( -HUGE );
    SiMain->setAlpha( 1.0 );
    CountThres = 0;
    LastFlag = P;
}

// speech silent detector
ESpeechFlag  CSpeechNonSpeech::detect( float *xvec )
{
    register unsigned int i, argmax = 0, argspeechmax=0, argsilentmax=0;
    register float tmp, max, speechmax, silentmax ;
    double pSp, pSi ;
    ESpeechFlag Decision = U;

    ArgMaxSp = ArgMaxSi = 0 ;
    // calc. log of propabilities (in advance)
    // use max over all mixtures as log probability!
    LogSp = PropLogSp[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSp.Par ; i++ ) 
      if( (tmp = PropLogSp[i]->getLogProb( xvec )) > LogSp )
	{ LogSp = tmp ; ArgMaxSp = i ; }
    LogSi = PropLogSi[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSi.Par ; i++ ) 
      if( (tmp = PropLogSi[i]->getLogProb( xvec )) > LogSi )
	{ LogSi = tmp ; ArgMaxSi = i ; }

    // alpha recursion for all states in TmpAlpha 
    // look for speechmax and silentmax for calculating aposterioris
    max = StateArray[0]->calcAlpha();
    if( StateArray[0]->getSpnonspType() == P ) { 
      silentmax = max ; speechmax = -1.0e30 ;
    } else {
      speechmax = max ; silentmax = -1.0e30 ;
    } ;
    for( i=1; i<TotalStates; i++ ) {
        if( (tmp=StateArray[i]->calcAlpha()) > max ) 
	  {   max = tmp; argmax = i; }
	if( tmp > silentmax && (StateArray[i]->getSpnonspType() == P)) {
	  silentmax = tmp; argsilentmax = i ;
	} else if( tmp > speechmax && (StateArray[i]->getSpnonspType() == S)) {
	  speechmax = tmp ; argspeechmax = i ; 
	} ;
    } ;

    // normalize to max Alpha and transfer tmpAlpha -> Alpha
    for( i=0; i<TotalStates; i++ )
        StateArray[i]->normAlpha(max); 

    //fprintf( stderr, "Speechalpha: %e, SilenceAlpha: %e\n", (double)SpMain->getAlpha(), (double)SiMain->getAlpha() ) ;
 
    // compute normalized aposterioris for main states 
    pSp = exp( (double)SpMain->getAlpha() ) ;
    pSi = exp( (double)SiMain->getAlpha() ) ;
    //fprintf( stderr, "Prob(Speechalpha): %e, Prob(SilenceAlpha): %e\n", pSp, pSi ) ;
    pSp = pSp / (pSi +pSp);
    pSi = 1.0 - pSp ;
    
    // use those aposts as features for output 
    //probSp = pSp ; 
    //probSi = pSi ; 
    
    // compute aposterioris of P and S using argspeechmax and argsilentmax
    //probSp = exp( (double)((StateArray[argspeechmax])->getAlpha())) ;
    //probSi = exp( (double)((StateArray[argsilentmax])->getAlpha())) ;
    //fprintf( stderr, "probSp: %e, probSi: %e\n", probSp, probSi ) ;
    //probSp = probSp / (probSi +probSp);
    //probSi = 1.0-probSp ;
 
    // compute local aposterioris of P and S using argspeechmax and argsilentmax
    probSp = exp( (double)LogSp ) ;
    probSi = exp( (double)LogSi ) ;
    probSp = probSp / (probSi + probSp);
    probSi = 1 - probSp;
 

    // use argmax for decision
    if( StateArray[argmax] == SpMain ) 
      Decision = pSp > ThresSp.Par ? S : U;
    else if( StateArray[argmax] == SiMain )
      Decision = pSi > ThresSi.Par ? P : U;

    // otherwise undef.
    if( Decision == U )
    {   if( ++CountThres > MAXTHRESCOUNT )
            Decision = LastFlag;
        else
            return Decision;
    }
    CountThres = 0;
    LastFlag   = Decision;
    return Decision;
}


// speech silent detector
ESpeechFlag  CSpeechNonSpeech::detectandupdate( float *xvec )
{
    register unsigned int i, argmax = 0, argspeechmax=0, argsilentmax=0;
    register float tmp, max, speechmax, silentmax ;
    double pSp, pSi ;
    ESpeechFlag Decision = U;

    ArgMaxSp = ArgMaxSi = 0 ;
    // calc. log of propabilities (in advance)
    // use max over all mixtures as log probability!
    LogSp = PropLogSp[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSp.Par ; i++ ) 
      if( (tmp = PropLogSp[i]->getLogProb( xvec )) > LogSp )
	{ LogSp = tmp ; ArgMaxSp = i ; }
    LogSi = PropLogSi[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSi.Par ; i++ ) 
      if( (tmp = PropLogSi[i]->getLogProb( xvec )) > LogSi )
	{ LogSi = tmp ; ArgMaxSi = i ; }

    // alpha recursion for all states in TmpAlpha 
    // look for speechmax and silentmax for calculating aposterioris
    max = StateArray[0]->calcAlpha();
    if( StateArray[0]->getSpnonspType() == P ) { 
      silentmax = max ; speechmax = -1.0e30 ;
    } else {
      speechmax = max ; silentmax = -1.0e30 ;
    } ;
    for( i=1; i<TotalStates; i++ ) {
        if( (tmp=StateArray[i]->calcAlpha()) > max ) 
	  {   max = tmp; argmax = i; }
	if( tmp > silentmax && (StateArray[i]->getSpnonspType() == P)) {
	  silentmax = tmp; argsilentmax = i ;
	} else if( tmp > speechmax && (StateArray[i]->getSpnonspType() == S)) {
	  speechmax = tmp ; argspeechmax = i ; 
	} ;
    } ;

    // normalize to max Alpha and transfer tmpAlpha -> Alpha
    for( i=0; i<TotalStates; i++ )
        StateArray[i]->normAlpha(max); 

    //fprintf( stderr, "Speechalpha: %e, SilenceAlpha: %e\n", (double)SpMain->getAlpha(), (double)SiMain->getAlpha() ) ;
 
    // compute normalized aposterioris for main states 
    pSp = exp( (double)SpMain->getAlpha() ) ;
    pSi = exp( (double)SiMain->getAlpha() ) ;
    //fprintf( stderr, "Prob(Speechalpha): %e, Prob(SilenceAlpha): %e\n", pSp, pSi ) ;
    pSp = pSp / (pSi +pSp);
    pSi = 1.0 - pSp ;
    
    // use those aposts as features for output 
    //probSp = pSp ; 
    //probSi = pSi ; 
    
    // compute aposterioris of P and S using argspeechmax and argsilentmax
    //probSp = exp( (double)((StateArray[argspeechmax])->getAlpha())) ;
    //probSi = exp( (double)((StateArray[argsilentmax])->getAlpha())) ;
    //fprintf( stderr, "probSp: %e, probSi: %e\n", probSp, probSi ) ;
    //probSp = probSp / (probSi +probSp);
    //probSi = 1.0-probSp ;
 
    // compute local aposterioris of P and S using argspeechmax and argsilentmax
    probSp = exp( (double)LogSp ) ;
    probSi = exp( (double)LogSi ) ;
    probSp = probSp / (probSi + probSp);
    probSi = 1 - probSp;
 

    // use argmax for decision
    if( StateArray[argmax] == SpMain ) {
      if( pSp > ThresSp.Par ) {
	Decision = S ;
	if( UpdateFlag.Par ) 
	  update( xvec, S ) ;
      } else
	Decision = U ;
    } else if( StateArray[argmax] == SiMain ) {
      if( pSi > ThresSi.Par ) {
	Decision = P ;
	if( UpdateFlag.Par )
	  update( xvec, P ) ;
      } else 
	Decision = U;
    } 

    // otherwise undef.
    if( Decision == U )
    {   if( ++CountThres > MAXTHRESCOUNT )
            Decision = LastFlag;
        else
            return Decision;
    }
    CountThres = 0;
    LastFlag   = Decision;
    return Decision;
}


// update mean and variance using the mixture component which yields the highest probability
void  CSpeechNonSpeech::update( float *xvec, ESpeechFlag ind, short NrUndef, short InUndef )
{
    NrUndef = InUndef = 0;
    // update with 1/T
    if( ind == U || !UpdateFlag.Par )
        return;
 
    // calc. log of probabilities to search for
    // max over all mixtures!
    float tmp, logprob ;
    int i ;
    ArgMaxSp = ArgMaxSi = 0 ;
    tmp = PropLogSp[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSp.Par ; i++ ) 
      if( (logprob = PropLogSp[i]->getLogProb( xvec )) > tmp )
	{ tmp = logprob ; ArgMaxSp = i ; }
    tmp = PropLogSi[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSi.Par ; i++ ) 
      if( (logprob = PropLogSi[i]->getLogProb( xvec )) > tmp )
	{ tmp = logprob ; ArgMaxSi = i ; }


    if( ind >= UpdateUseSp.Par )
        PropLogSp[ArgMaxSp]->updateProb( xvec );
    else if( ind <= UpdateUseSi.Par )
        PropLogSi[ArgMaxSi]->updateProb( xvec );
}
 

// update mean and variance using the given mixture mixnr
int  CSpeechNonSpeech::updateOneMix( float *xvec, ESpeechFlag ind, int mixnr, short NrUndef, short InUndef )
{
    NrUndef = InUndef = 0;
    // update with 1/T
    if( ind == U || !UpdateFlag.Par )
        return 0;

    if( ind >= UpdateUseSp.Par && mixnr < NrGaussSp.Par ) {
      //fprintf( stderr, "SPeech update\n" ) ;
      PropLogSp[mixnr]->updateProb( xvec );
    } 
    else if( ind <= UpdateUseSi.Par && mixnr < NrGaussSi.Par ) {
      //fprintf( stderr, "Silence update\n" ) ;
      PropLogSi[mixnr]->updateProb( xvec );
    }
    return 1 ;
}
 






//--------------------------- public 'CSpeechNonSpeechOverlap' -----------------------

// constructor: CSpeechNonSpeech
CSpeechNonSpeechOverlap::CSpeechNonSpeechOverlap( unsigned int len, double dt, FILE *fp ) 
:  CSpeechNonSpeech( len, dt, fp, 0 ) ,
    UpdateUseOv( "SpDet.Update.Overlap.Use", ESPEECHFLAG, fp, O ),
    ThresOv( "SpDet.Overlap.Threshold", fp, 0.5 ),
    MinTimeOv( "SpDet.Overlap.MinTime", fp, 0.2 ),
    NrGaussOv( "SpDet.Overlap.NrGauss", fp, 1 ) 

{
    DEBUG_1_MSG("'CSpeechNonSpeechOverlap' constructor called" );

    int i ;
    char keyword[MAXKEYWORDLEN] ;

    // allocate array of gaussian densities for speech and silence
    PropLogOv = new CProbGaussVar*[NrGaussOv.Par] ;
    for( i=0 ;i<NrGaussOv.Par ; i++ ) {
      sprintf( keyword, "SpDet.Overlap." ) ;
      PropLogOv[i] = new CProbGaussVar( len, dt, fp, keyword, UpdateShortTime.Par, i ) ;
    } ;
    
    // build special stochastic model constants
    CountSpToOv = MAX((int) (MinTimeOv.Par/dt), 1 );
    CountSiToOv = MAX((int) (MinTimeOv.Par/dt), 1 );
    CountOvToSp = MAX((int) (MinTimeSp.Par/dt), 1 );
    CountOvToSi = MAX((int) (MinTimeSi.Par/dt), 1 );
   
    CountOvToOv_Si = MIN( CountOvToSi, CountSiToOv );
    CountOvToOv_Sp = MIN( CountOvToSp, CountSpToOv );
    CountSiToSi_Ov = MIN( CountSiToOv, CountOvToSi );
    CountSpToSp_Ov = MIN( CountSpToOv, CountOvToSp );

    TotalStates = 3 + CountSiToSp + CountSiToSi + CountSpToSi + CountSpToSp
      + CountSiToOv + CountSiToSi_Ov + CountSpToOv + CountSpToSp_Ov 
      + CountOvToSi + CountOvToOv_Si + CountOvToSp + CountOvToOv_Sp ;

    CountThres = 0;
    LastFlag = P;

    // model state array
    typedef CObjStochasticState * PState;
    StateArray = new PState[TotalStates];
    IF_FATAL_EXIT( StateArray==0, "'CSpeechNonSpeech' constructor out of memory" );

    // build model structure (only here)
  
    // Sp main state, prob. for Sp
    unsigned int j, n=0;
    SpMain = StateArray[n] = new CObjStochasticState( &LogSp, NULL, (ESpeechFlag)P );
    SpMain->addPreState( SpMain );
    n++;
 
    // Si' states from Sp to Si, prob. for Si    
    for( i=0; i<CountSpToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], (ESpeechFlag)P );
    n += CountSpToSi;

    // Sp' states from Sp to Sp, prob. for Sp
    j = n - ((CountSpToSi - CountSpToSp) + 1);
    int SpToSiLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogSp, StateArray[j], (ESpeechFlag)S );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSpToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)S );
    n += CountSpToSp;
    SpMain->addPreState( StateArray[n-1] );

    // Ov' states from Sp to Ov, prob for Ov 
    for( i=0; i<CountSpToOv; i++ )
      StateArray[n+i] = new CObjStochasticState( &LogOv, StateArray[n+i-1], (ESpeechFlag)O );
    n += CountSpToOv;

    // Sp' states from Sp to Sp, prob. for Sp
    j = n - ((CountSpToOv - CountSpToSp_Ov) + 1);
    int SpToOvLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogSp, StateArray[j], (ESpeechFlag)S );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSpToSp_Ov; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)S );
    n += CountSpToSp;
    SpMain->addPreState( StateArray[n-1] );

    // Si main state, prob for Si
    SiMain = StateArray[n] = new CObjStochasticState( &LogSi, StateArray[SpToSiLast], (ESpeechFlag)P );
    SiMain->addPreState( SiMain );
    n++;

    // Sp' states from Si to Sp, prob. for Sp
    for( i=0; i<CountSiToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], (ESpeechFlag)S );
    n += CountSiToSp;

    // Si' states from Si to Si, prob. for Si
    j = n - ((CountSiToSp - CountSiToSi) + 1);
    int SiToSpLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)P );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSiToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)P );
    n += CountSiToSi;
    SiMain->addPreState( StateArray[n-1] );

    // close ergodic model - loop
    SpMain->addPreState( StateArray[SiToSpLast] );

    // Ov' states from Si to Ov, prob. for Ov
    for( i=0; i<CountSiToOv; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogOv, StateArray[n+i-1], (ESpeechFlag)O );
    n += CountSiToOv;

    // Si' states from Si to Si, prob. for Si
    j = n - ((CountSiToSp - CountSiToSi) + 1);
    int SiToOvLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)P );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSiToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)P );
    n += CountSiToSi;
    SiMain->addPreState( StateArray[n-1] );




    // Ov main state, prob for Ov
    OvMain = StateArray[n] = new CObjStochasticState( &LogOv, StateArray[SpToOvLast], (ESpeechFlag)O );
    OvMain->addPreState( StateArray[SiToOvLast] ) ;
    OvMain->addPreState( OvMain ); // self loop
    n++;

    // Sp' states from Ov to Sp, prob. for Sp
    for( i=0; i<CountOvToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], (ESpeechFlag)S );
    n += CountOvToSp;

    // Ov' states from Ov to Ov, prob. for Ov
    j = n - ((CountOvToSp - CountOvToOv_Sp) + 1);
    int OvToSpLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)O );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountOvToOv_Sp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogOv, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)O );
    n += CountOvToOv_Sp;
    OvMain->addPreState( StateArray[n-1] );

    // close ergodic model - loop
    SpMain->addPreState( StateArray[OvToSpLast] );

    // Si' states from Ov to Si, prob. for Si
    for( i=0; i<CountOvToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], (ESpeechFlag)P );
    n += CountOvToSi;

    // Ov' states from Ov to Ov, prob. for Ov
    j = n - ((CountOvToSi - CountOvToOv_Si) + 1);
    int OvToSiLast = n-1 ;
    StateArray[n] = new CObjStochasticState( &LogOv, StateArray[j], (ESpeechFlag)O );
    for( i=j+1; i<n; i++ )
        StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountOvToOv_Si; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogOv, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)O );
    n += CountOvToOv_Si;
    OvMain->addPreState( StateArray[n-1] );

    // close ergodic model - loop
    SiMain->addPreState( StateArray[OvToSiLast] );
    
    IF_FATAL_EXIT( n != TotalStates, "'CSpeechNonSpeech' constructor internal model-structure error" );
    
    // test memory alloc
    for( i=0; i<TotalStates; i++ )
        IF_FATAL_EXIT( StateArray[i] == 0, "'CSpeechNonSpeech' constructor out of memory")
}

// destructor: CSpeechNonSpeech
CSpeechNonSpeechOverlap::~CSpeechNonSpeechOverlap( )
{
    DEBUG_1_MSG("'CSpeechNonSpeechOverlap' destructor called" );
    int i ;
    if( UpdateWrite.Par )
    { 
      for( i=0 ; i<NrGaussOv.Par ; i++ )
	PropLogSi[i]->writePar();
    }

    // delete all states
    for( unsigned int i=0; i<TotalStates; i++ )
      {   
	StateArray[i]->clearPreState();
	if( StateArray[i] != 0 )
	  delete StateArray[i];
    }
    if( StateArray != 0 )
      delete [] StateArray;

    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogOv[i] != 0 )
	delete PropLogOv[i] ;

    if( PropLogOv != 0 )
      delete[] PropLogOv ;
}

// write parameter to file
int  CSpeechNonSpeechOverlap::writePar( FILE *fp )
{
    DEBUG_1_MSG("'CSpeechNonSpeech' write parameter called" );

    NrGaussSp.writePar(fp) ;
    NrGaussSi.writePar(fp) ;
    NrGaussOv.writePar(fp) ;

    int i ;
    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogSp[i]->writePar( fp ) == _ERR_ )
        return _ERR_;
        
    for( i=0 ; i<NrGaussSi.Par ; i++ ) 
      if( PropLogSi[i]->writePar( fp ) == _ERR_ )
	return _ERR_ ;

    for( i=0 ; i<NrGaussOv.Par ; i++ ) 
      if( PropLogOv[i]->writePar( fp ) == _ERR_ )
	return _ERR_ ;

    
    return _OK_;

}

/////////////////////////////////////////////////////////////////////////
//
// append fix params
// 
/////////////////////////////////////////////////////////////////////////
int  CSpeechNonSpeechOverlap::appendFixPar( FILE *fp ) {
    
    int ret = 0 ; 
    ret |= UpdateFlag.appendPar( fp ) ;
    ret |= UpdateShortTime.appendPar( fp ) ;
    ret |= UpdateWrite.appendPar( fp ) ;
    ret |= UpdateUseSp.appendPar( fp ) ;
    ret |= UpdateUseSi.appendPar( fp ) ;
    ret |= UpdateUseOv.appendPar( fp ) ;
    ret |= NrGaussSp.appendPar(fp) ;
    ret |= NrGaussSi.appendPar(fp) ;
    ret |= NrGaussOv.appendPar(fp) ;
    ret |= PropLogSp[0]->appendFixPar( fp ) ;
    ret |= PropLogSi[0]->appendFixPar( fp ) ;
    ret |= PropLogOv[0]->appendFixPar( fp ) ;
    ret |= ThresSp.appendPar( fp ) ;
    ret |= ThresSi.appendPar( fp ) ;
    ret |= ThresOv.appendPar( fp ) ;
    ret |= MinTimeSp.appendPar( fp ) ;
    ret |= MinTimeSi.appendPar( fp ) ;
    ret |= MinTimeOv.appendPar( fp ) ;

    return( ret == 0 ? ret : -1 ) ;
} ; // END appendFixPar

/////////////////////////////////////////////////////////////////////////
//
// append updatable params
// 
/////////////////////////////////////////////////////////////////////////
int  CSpeechNonSpeechOverlap::appendUpdPar( FILE *fp ) {

    DEBUG_1_MSG("'CSpeechNonSpeechOverlap' append updatable parameter called" );

    int i ;
    for( i=0 ; i<NrGaussSp.Par ; i++ )
      if( PropLogSp[i]->appendUpdPar( fp ) == _ERR_ )
	return( _ERR_ ) ;
   
    for( i=0 ; i<NrGaussSi.Par ; i++ )
      if( PropLogSi[i]->appendUpdPar( fp ) == _ERR_ )
        return _ERR_;
 
    for( i=0 ; i<NrGaussOv.Par ; i++ )
      if( PropLogOv[i]->appendUpdPar( fp ) == _ERR_ )
        return _ERR_;
    
    return _OK_;

}



// speech silent detector
ESpeechFlag  CSpeechNonSpeechOverlap::detect( float *xvec )
{
    register unsigned int i, argmax = 0, argspeechmax=0, argsilentmax=0, argoverlapmax=0;
    register float tmp, max, speechmax, silentmax, overlapmax ;
    double pSp, pSi, pOv ;
    ESpeechFlag Decision = U;

    ArgMaxSp = ArgMaxSi = ArgMaxOv = 0 ;
    // calc. log of propabilities (in advance)
    // use max over all mixtures as log probability!
    LogSp = PropLogSp[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSp.Par ; i++ ) 
      if( (tmp = PropLogSp[i]->getLogProb( xvec )) > LogSp )
	{ LogSp = tmp ; ArgMaxSp = i ; }
    LogSi = PropLogSi[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSi.Par ; i++ ) 
      if( (tmp = PropLogSi[i]->getLogProb( xvec )) > LogSi )
	{ LogSi = tmp ; ArgMaxSi = i ; }
    LogOv = PropLogOv[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussOv.Par ; i++ ) 
      if( (tmp = PropLogOv[i]->getLogProb( xvec )) > LogOv )
	{ LogOv = tmp ; ArgMaxOv = i ; }

    // alpha recursion for all states in TmpAlpha 
    // look for speechmax and silentmax for calculating aposterioris
    max = StateArray[0]->calcAlpha();
    if( StateArray[0]->getSpnonspType() == P ) { 
      silentmax = max ; speechmax = -1.0e30 ; overlapmax = -1.0e30 ;
    } else if( StateArray[0]->getSpnonspType() == S ){
      speechmax = max ; silentmax = -1.0e30 ; overlapmax = -1.0e30 ;
    } else {
      speechmax = -1.0e30 ; silentmax = -1.0e30 ; overlapmax = max ;
    } ;

    for( i=1; i<TotalStates; i++ ) {
        if( (tmp=StateArray[i]->calcAlpha()) > max ) 
	  {   max = tmp; argmax = i; }
	if( tmp > silentmax && (StateArray[i]->getSpnonspType() == P)) {
	  silentmax = tmp; argsilentmax = i ;
	} else if( tmp > speechmax && (StateArray[i]->getSpnonspType() == S)) {
	  speechmax = tmp ; argspeechmax = i ; 
	} else if( tmp > overlapmax &&  (StateArray[i]->getSpnonspType() == O)) {
	  overlapmax = tmp ; argoverlapmax = i ;
	}
    } ;

    // normalize to max Alpha and transfer tmpAlpha -> Alpha
    for( i=0; i<TotalStates; i++ )
        StateArray[i]->normAlpha(max); 

    //fprintf( stderr, "Speechalpha: %e, SilenceAlpha: %e\n", (double)SpMain->getAlpha(), (double)SiMain->getAlpha() ) ;
 
    // compute normalized aposterioris for main states 
    pSp = exp( (double)SpMain->getAlpha() ) ;
    pSi = exp( (double)SiMain->getAlpha() ) ;
    pOv = exp( (double)OvMain->getAlpha() ) ;
    double sum = pSi + pSp + pOv ;
    //fprintf( stderr, "Prob(Speechalpha): %e, Prob(SilenceAlpha): %e\n", pSp, pSi ) ;
    pSp = pSp / sum ;
    pSi = pSi / sum ;
    pOv = 1.0 - pSi - pSp ;
    
    // compute local aposterioris of P and S using argspeechmax and argsilentmax
    probSp = exp( (double)LogSp ) ;
    probSi = exp( (double)LogSi ) ;
    probOv = exp( (double)LogOv ) ;
    sum =  probSp +  probSi +  probOv ;
    probSp = probSp / sum ;
    probSi = probSi / sum;
    probOv = 1.0 - probSp - probSi ;

    // use argmax for decision
    if( StateArray[argmax] == SpMain ) 
      Decision = pSp > ThresSp.Par ? S : U;
    else if( StateArray[argmax] == SiMain )
      Decision = pSi > ThresSi.Par ? P : U;
    else if( StateArray[argmax] == OvMain )
      Decision = pOv > ThresOv.Par ? O : U;

    // otherwise undef.
    if( Decision == U )
    {   if( ++CountThres > MAXTHRESCOUNT )
            Decision = LastFlag;
        else
            return Decision;
    }
    CountThres = 0;
    LastFlag   = Decision;
    return Decision;
} ;


// update mean and variance using the mixture component which yields the highest probability
void  CSpeechNonSpeechOverlap::update( float *xvec, ESpeechFlag ind, short NrUndef, short InUndef )
{
    NrUndef = InUndef = 0;
    // update with 1/T
    if( ind == U || !UpdateFlag.Par )
        return;
 
    // calc. log of probabilities to search for
    // max over all mixtures!
    float tmp, logprob ;
    int i ;
    ArgMaxSp = ArgMaxSi = ArgMaxOv = 0 ;
    tmp = PropLogSp[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSp.Par ; i++ ) 
      if( (logprob = PropLogSp[i]->getLogProb( xvec )) > tmp )
	{ tmp = logprob ; ArgMaxSp = i ; }
    tmp = PropLogSi[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussSi.Par ; i++ ) 
      if( (logprob = PropLogSi[i]->getLogProb( xvec )) > tmp )
	{ tmp = logprob ; ArgMaxSi = i ; }

    tmp = PropLogOv[0]->getLogProb( xvec );
    for( i=1 ; i<NrGaussOv.Par ; i++ ) 
      if( (logprob = PropLogOv[i]->getLogProb( xvec )) > tmp )
	{ tmp = logprob ; ArgMaxOv = i ; }


    if( ind >= UpdateUseSp.Par )
      PropLogSp[ArgMaxSp]->updateProb( xvec );
    else if( ind <= UpdateUseSi.Par )
      PropLogSi[ArgMaxSi]->updateProb( xvec );
    else if( ind >= UpdateUseOv.Par && ind <= O )
      PropLogOv[ArgMaxOv]->updateProb( xvec );
}
 

// update mean and variance using the given mixture mixnr
int  CSpeechNonSpeechOverlap::updateOneMix( float *xvec, ESpeechFlag ind, int mixnr, short NrUndef, short InUndef )
{
    NrUndef = InUndef = 0;
    // update with 1/T
    if( ind == U || !UpdateFlag.Par )
        return 0;

    if( ind >= UpdateUseSp.Par && mixnr < NrGaussSp.Par ) {
      //fprintf( stderr, "SPeech update\n" ) ;
      PropLogSp[mixnr]->updateProb( xvec );
    } 
    else if( ind <= UpdateUseSi.Par && mixnr < NrGaussSi.Par ) {
      //fprintf( stderr, "Silence update\n" ) ;
      PropLogSi[mixnr]->updateProb( xvec );
    } 
    else if( ind >= UpdateUseOv.Par && ind <= O && mixnr < NrGaussOv.Par ) {
      //fprintf( stderr, "Overlap update\n" ) ;
      PropLogOv[mixnr]->updateProb( xvec );
    } ;

    return 1 ;
}
 




// constructor: CSpeechNonSpeech
CSpeechNonSpeechANN::CSpeechNonSpeechANN( double dt, FILE *cfgfp )
:   ThresSp( "SpDet.Speech.Threshold", cfgfp, 0.5 ),
    ThresSi( "SpDet.Silent.Threshold", cfgfp, 0.5 ),
    MinTimeSp( "SpDet.Speech.MinTime", cfgfp, 0.2 ),
    MinTimeSi( "SpDet.Silent.MinTime", cfgfp, 0.2 ),
    PriorSp( "SpDet.Speech.Prior", cfgfp, 0.5 ),
    PriorSi( "SpDet.Silent.Prior", cfgfp, 0.5 ) 
{
    DEBUG_1_MSG("'CSpeechNonSpeechANN' constructor called" );

    int i ;
    char keyword[MAXKEYWORDLEN] ;
    
    // build special stochastic model constants
    CountSiToSp = MAX((int) (MinTimeSp.Par/dt), 1 );
    CountSpToSi = MAX((int) (MinTimeSi.Par/dt), 1 );
    CountSpToSp = MIN( CountSpToSi, CountSiToSp );
    //CountSpToSp = CountSpToSi ;
    CountSiToSi = MIN( CountSpToSi, CountSiToSp );
    //CountSiToSi = CountSiToSp ;
    TotalStates = 2 + CountSiToSp + CountSiToSi + CountSpToSi + CountSpToSp;
    CountThres = 0;
    LastFlag = P;

    MaxNrUndef = 3000 ; // 30 seconds!!!
    //MaxNrUndef = 2*MAX(CountSiToSp, CountSpToSi) ; // only use this setting for online use -> can introduce errors!!!!

#ifdef TEST
    fprintf( stderr, "SiToSp %d, SpToSi %d, SpToSp %d, SiToSi %d, Total: %d\n", CountSiToSp, CountSpToSi, CountSpToSp, CountSiToSi, TotalStates ) ;
    fprintf( stderr, "MaxNrUndef: %d\n", MaxNrUndef ) ;
#endif

    // model state array
    typedef CObjStochasticState * PState;
    StateArray = new PState[TotalStates];
    IF_FATAL_EXIT( StateArray==0, "'CSpeechNonSpeechANN' constructor out of memory" );

    // build model structure (only here)
  
    // Sp main state, prob. for Sp
    unsigned int j, n=0;
    SpMain = StateArray[n] = new CObjStochasticState( &LogSp, NULL, (ESpeechFlag)S );
    SpMain->addPreState( SpMain );
    n++;
 
    // Si' states from Sp to Si, prob. for Si    
    for( i=0; i<CountSpToSi; i++ )
      StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], (ESpeechFlag)P );
    n += CountSpToSi;

    // Sp' states from Sp to Sp, prob. for Sp
    j = n - ((CountSpToSi - CountSpToSp) + 1);
    StateArray[n] = new CObjStochasticState( &LogSp, StateArray[j], (ESpeechFlag)S );
    for( i=j+1; i<n; i++ )
      StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSpToSp; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)S );
    n += CountSpToSp;
    SpMain->addPreState( StateArray[n-1] );

    // Si main state, prob for Si
    SiMain = StateArray[n] = new CObjStochasticState( &LogSi, StateArray[n-(CountSpToSp+1)], (ESpeechFlag)P );
    SiMain->addPreState( SiMain );
    n++;

    // Sp' states from Si to Sp, prob. for Sp
    for( i=0; i<CountSiToSp; i++ )
      StateArray[n+i] = new CObjStochasticState( &LogSp, StateArray[n+i-1], (ESpeechFlag)S );
    n += CountSiToSp;

    // Si' states from Si to Si, prob. for Si
    j = n - ((CountSiToSp - CountSiToSi) + 1);
    StateArray[n] = new CObjStochasticState( &LogSi, StateArray[j], (ESpeechFlag)P );
    for( i=j+1; i<n; i++ )
      StateArray[n]->addPreState( StateArray[i] ); 
    for( i=1; i<CountSiToSi; i++ )
        StateArray[n+i] = new CObjStochasticState( &LogSi, StateArray[n+i-1], StateArray[j-i], (ESpeechFlag)P );
    n += CountSiToSi;
    SiMain->addPreState( StateArray[n-1] );

    // close ergodic model - loop
    SpMain->addPreState( StateArray[n-(CountSiToSi+1)] );
    IF_FATAL_EXIT( n != TotalStates, "'CSpeechNonSpeechANN' constructor internal model-structure error" );
    
    // test memory alloc
    for( i=0; i<TotalStates; i++ )
      IF_FATAL_EXIT( StateArray[i] == 0, "'CSpeechNonSpeechANN' constructor out of memory")
}

// destructor: CSpeechNonSpeech
CSpeechNonSpeechANN::~CSpeechNonSpeechANN( )
{
  DEBUG_1_MSG("'CSpeechNonSpeechANN' destructor called" );
  int i ;
  
  // delete all states
  for( i=0; i<TotalStates; i++ )
    {   StateArray[i]->clearPreState();
    delete StateArray[i];
    }
  if( StateArray != 0 )
    delete [] StateArray;
}


// start alpha recursion
int  CSpeechNonSpeechANN::start( char * filename, char *filename2 )
{
  char line[MaxLineLength] ;
  char tmp[MaxLineLength] ;
  // open file with probs: 
  ProbFp = fopen( filename, "r" ) ;
  if( ProbFp == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Can't open file '%s'\n", filename ) ;
    return(-1) ;
  } ;

  if( ReadPfileHeader( ProbFp ) != 1 ) {
    fprintf( stderr, "ERROR: reading header of file '%s' \n", filename ) ;
    return(-1) ;
  } 

  if( filename2 != 0 ) {
    ProbFp2 = fopen( filename2, "r" ) ;
    if( ProbFp2 == 0 ) {
      fprintf( stderr, "CSpeechNonSpeechANN::start: Can't open file '%s' with posteriors\n", filename2 ) ;
      return(-1) ;
    } ;
    if( ReadPfileHeader( ProbFp2 ) != 1 ) {
      fprintf( stderr, "ERROR: reading header of file '%s' \n", filename2 ) ;
      return(-1) ;
    } 
  } else {
    ProbFp2 = 0 ;
  } ;


  // init search space
  for( unsigned int i=0; i<TotalStates; i++ )
    StateArray[i]->setAlpha( -HUGE );
  SiMain->setAlpha( 1.0 );
  CountThres = 0;
  LastFlag = P;
  FrameCount = 0 ;

  return(1) ;
}


int CSpeechNonSpeechANN::ReadPfileHeader( FILE *fp ) {

  char line[MaxLineLength] ;
  char tmp[MaxLineLength] ;

  // check header:
  fgets( line, MaxLineLength, fp ) ;
  if( strstr( line, "pfile_header version 0 size 32768" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, wrong header format '%s' found!", line ) ;
    return(-1) ;
  }
  fgets( line, MaxLineLength, fp ) ;
  if( strstr( line, "num_sentences" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find number of sentences!\n" ) ;
    return(-1) ;
  } 
  fgets( line, MaxLineLength, fp ) ; 
  strcpy( tmp, line ) ;
  if( strstr( tmp, "num_frames" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find number of frames!\n" ) ;
    return(-1) ;
  } else {
    // read number of frames
    strtok( line, " \t" ) ;
    if( (NumFrames = strtol( strtok( NULL, " \t\n" ), NULL, 10 )) == 0 ) {
      fprintf( stderr, "CSpeechNonSpeechANN::start: File should contain data for at least one frame\n" ) ;
      return(-1) ;
    }
  }
  fgets( line, MaxLineLength, fp ) ; 
  if( strstr( line, "first_feature_column" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find number first feature column!\n" ) ;
    return(-1) ;
  } 
  fgets( line, MaxLineLength, fp ) ; 
  strcpy( tmp, line ) ;
  if( strstr( line, "num_features" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find number of features!\n" ) ;
    return(-1) ;
  } else {
    // read number of feature
    strtok( tmp, " \t" ) ;
    if( strtol( strtok( NULL, " \t\n" ), NULL, 10 ) != 2 ) {
      fprintf( stderr, "CSpeechNonSpeechANN::start: File should contain two 'features' (prob for 'sil' and prob for 'speech')\n" ) ;
      return(-1) ;
    }
  }
  fgets( line, MaxLineLength, fp ) ; 
  if( strstr( line, "first_label_column" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find first label column!\n" ) ;
    return(-1) ;
  } 
  fgets( line, MaxLineLength, fp ) ; 
  if( strstr( line, "num_labels" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find number of labels!\n" ) ;
    return(-1) ;
  } 
  fgets( line, MaxLineLength, fp ) ; 
  if( strstr( line, "format" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find format!\n" ) ;
    return(-1) ;
  } 
  fgets( line, MaxLineLength, fp ) ; 
  if( strstr( line, "data size" ) == 0 ) {
    fprintf( stderr, "CSpeechNonSpeechANN::start: Sorry, could not find data entries!\n" ) ;
    return(-1) ;
  } 
  
  // move to first data row:
  fseek( fp, 32768, 0 ) ; 
  
  return(1) ;
} 



// start alpha recursion
int  CSpeechNonSpeechANN::restartat( int startframe )
{
  char line[MaxLineLength] ;
  char tmp[MaxLineLength] ;

  rewind( ProbFp ) ;
  if( ProbFp2 != 0 )
    rewind( ProbFp2 ) ;

  // init search space
  for( unsigned int i=0; i<TotalStates; i++ )
    StateArray[i]->setAlpha( -HUGE );
  SiMain->setAlpha( 1.0 );
  CountThres = 0;
  LastFlag = P;
  FrameCount = 0 ;

  // goto startframe
  fseek( ProbFp, (startframe*16+32768), 0 ) ;

  if( ProbFp2 != 0 )
    fseek( ProbFp2, (startframe*16+32768), 0 ) ;

  return(1) ;
}


// speech silence detector
ESpeechFlag  CSpeechNonSpeechANN::detect( int checkframe )
{
    register unsigned int i, argmax ;
    register float tmp, max ;
    double pSp, pSi, separation1, separation2 ;
    ESpeechFlag Decision = U;

    // read probs from file
    if( FrameCount < NumFrames ) {
      Read(checkframe) ; // count is increased in function Read()
      //FrameCount++ ;
    } else { 
      // eof reached
      return( INVALID );
    }


    //fprintf( stderr, "%d %d %f %f\n", 0, FrameCount, ApostSp, ApostSi ) ;


    // check if TWO posteriors for each class given:
    if( ProbFp2 != 0 ) {
#ifdef MaxApost
      // method one: go for max apost for each class 
      LogSp = log( ApostSp/PriorSp.Par ) ;
      LogSi = log( ApostSi/PriorSi.Par ) ;
      if( ApostSp2 > ApostSp )
	LogSp = log( ApostSp2/PriorSp.Par ) ;
      if( ApostSi2 > ApostSi )
	LogSi = log( ApostSi/PriorSi.Par ) ;
#endif

      //fprintf( stderr, "detect maxapost: %f %f\n", LogSi, LogSp ) ;

#ifdef MaxSeparation
      // method 2: go for the net which gives the better separation for this frame
      separation1 = ( ApostSi > ApostSp ? ApostSi/ApostSp : ApostSp/ApostSi ) ;
      separation2 = ( ApostSi2 > ApostSp2 ? ApostSi2/ApostSp2 : ApostSp2/ApostSi2 ) ;
      if( separation1 > separation2 ) {
	// use first net output
	LogSp = log( ApostSp/PriorSp.Par ) ;
	LogSi = log( ApostSi/PriorSi.Par ) ;
      } else {
        // use second net output
	LogSp = log( ApostSp2/PriorSp.Par ) ;
	LogSi = log( ApostSi2/PriorSi.Par ) ;
      }
#endif 

#ifdef SumOfLog
      LogSp = log( ApostSp/PriorSp.Par ) + log( ApostSp2/PriorSp.Par ) ;
      LogSi = log( ApostSi/PriorSi.Par ) + log( ApostSi2/PriorSi.Par ) ;
#endif

#ifdef WeightedSumOfLog
      LogSp = log( (0.5*ApostSp+0.5*ApostSp2)/PriorSp.Par ) ;
      LogSi = log( (0.5*ApostSi+0.5*ApostSi2)/PriorSi.Par ) ;
#endif

    } else { 
      LogSp = log( ApostSp/PriorSp.Par ) ;
      LogSi = log( ApostSi/PriorSi.Par ) ;
    }

    //fprintf( stderr, "%d: Log. Likelihoods: %f %f\n", FrameCount, LogSp, LogSi ) ;


    // alpha recursion for all states in TmpAlpha 
    // look for speechmax and silentmax for calculating aposterioris
    //max = StateArray[0]->calcAlpha();
    max = -HUGE ;
    argmax = 0 ;
    for( i=0; i<TotalStates; i++ ) {
      if( (tmp=StateArray[i]->calcAlpha()) > max ) {   
	max = tmp; 
	argmax = i; 
      }
    } ;

    // normalize to max Alpha and transfer tmpAlpha -> Alpha
    for( i=0; i<TotalStates; i++ ) {
      StateArray[i]->normAlpha(max); 
    }  

    // compute normalized aposterioris for main states 
    pSp = exp( (double)SpMain->getAlpha() ) ;
    pSi = exp( (double)SiMain->getAlpha() ) ;
    pSp = pSp / (pSi +pSp);
    pSi = 1.0 - pSp ;


    //fprintf( stderr, "%d: Alpha: SiMain: %f\tSpMain: %f\n", FrameCount, (double)SiMain->getAlpha(), (double)SpMain->getAlpha()) ;

    
    //fprintf( stderr, "MAX found in state %d (%f)\n", argmax, pSp ) ;


    // use argmax for decision
    if( StateArray[argmax] == SpMain ) 
      //Decision = pSp > ThresSp.Par ? S : U;
      Decision = S ;
    else if( StateArray[argmax] == SiMain )
      //Decision = pSi > ThresSi.Par ? P : U;
      Decision = P ;

    // otherwise undef.
    if( Decision == U ) {  
      //if( ++CountThres > MAXTHRESCOUNT )
      if( ++CountThres >= MaxNrUndef )
	Decision = LastFlag;
      else
	return Decision;
    }
    CountThres = 0;
    LastFlag   = Decision;
    return(Decision);
}





int CSpeechNonSpeechANN::Read(int checkframe) {

  int dummy ;

  // read sent number:
  fread( &dummy, sizeof(int), 1, ProbFp ) ;
  // read frame number:
  fread( &dummy, sizeof(int), 1, ProbFp ) ;
  
  if( checkframe && dummy != FrameCount ) {
    fprintf( stderr, "CSpeechNonSpeechANN::Read: ProbFp: %d-th frame expected, but %d-th frame read\n", FrameCount, dummy ) ;
    return(-1) ;
  } ;
  // read next pair of probs
  fread( &ApostSi, sizeof(float), 1, ProbFp ) ; 
  fread( &ApostSp, sizeof(float), 1, ProbFp ) ; 

#ifdef TEST_HIGHLEVEL
  fprintf( stderr, "ProbFp: frame %d: read 1: %f %f\n", FrameCount, ApostSi, ApostSp ) ;
#endif

  if( ProbFp2 != 0 ) {
    // read sent number:
    fread( &dummy, sizeof(int), 1, ProbFp2 ) ;
    // read frame number:
    fread( &dummy, sizeof(int), 1, ProbFp2 ) ;
    if( checkframe && dummy != FrameCount ) {
      fprintf( stderr, "CSpeechNonSpeechANN::Read: ProbFp2: %d-th frame expected, but %d-th frame read\n", FrameCount, dummy ) ;
      return(-1) ;
    } ;
    // read next pair of probs
    fread( &ApostSi2, sizeof(float), 1, ProbFp2 ) ; 
    fread( &ApostSp2, sizeof(float), 1, ProbFp2 ) ; 

#ifdef TEST_HIGHLEVEL
    fprintf( stderr, "ProbFp2: frame %d: read 2: %f %f\n", FrameCount, ApostSi2, ApostSp2 ) ;
#endif 
  }

  FrameCount++ ;
  
  return(1) ;
} ;





//--------------------------- end of 'CSpeechNonSpeechOverlap' -----------------------

//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL


#endif  // TEST_LEVEL
