/******************************************************************************
 * file:        spnonsp.h  (for spnonsp.cc)
 * info:        class declarations for CSpeechNonSpeech, speech pause detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/

#if !defined( _SPNONSP_H_ )
#define _SPNONSP_H_

#define  MAXUNDEFCOUNT  90

#include  "paramet.h"
#include  "obj_list.h"
#include  "feature.h"


// gauss density function; deri. from CProbDensity 
class CProbGaussVar
{
protected:
    // parameter of x-vector:
    const unsigned int  Len;
    // parameter of prop-density:  gauss with diagonal covar.mat 
    CParFloatVec        Mean;
    CParFloatVec        Var;
    double              LogDet;
    double              TwoPiPowerN;
    // update parameter
    CParFloat           TauUpdate;
    double              Tf;
    FILE  *const        ParFile;
    int                 FlagShortTime;

public:
    CProbGaussVar( unsigned int len, double dt, FILE *fp, char *keyext, int shorttimeflag, int nr=0 );
    double getLogProb( float *xvec );
    void   updateProb( float *xvec );
    int    writePar( FILE *fp );
    int    appendFixPar( FILE *fp ) ;
    int    appendUpdPar( FILE *fp ) ;
    int    writePar();
};




// state of (HMM) stochastic automat
class CObjStochasticState : public CObj
{
private:
  float          Alpha;
  float          TmpAlpha;
  float         *PropDens;
  CListOfObj     PreStateList;

  ESpeechFlag SpnonspType ; // is it a Speech or a Silence State?
public:
  CObjStochasticState( float *proptr,  CObjStochasticState *state, ESpeechFlag type );
  CObjStochasticState( float *proptr,  CObjStochasticState *st1, CObjStochasticState *st2, ESpeechFlag type );
  ~CObjStochasticState();
  void   addPreState( CObjStochasticState *state ) { PreStateList.addTail(state); }
  float  calcAlpha();
  void   normAlpha( float norm ) { Alpha = TmpAlpha - norm; }
  float  getAlpha() const { return  Alpha; }
  float  getTmpAlpha() const { return  TmpAlpha; }
  void   setAlpha( float alpha ) { Alpha = alpha; }
  void   clearPreState()  { PreStateList.clearAll(); }
  // virtual function for abstract class CObj
  int    compare( const CObj *obj ) const { return (int)(Alpha - ((CObjStochasticState*)obj)->Alpha); }
  ESpeechFlag getSpnonspType() { return( SpnonspType ) ; } ; // tpfau, 12/21/2000
};

#define MAXTHRESCOUNT (FEATURE_BUFLEN-10)

// class CSpeechNonSpeech:  speech (Sp) silent (Si) detector
class CSpeechNonSpeech
{
protected:
    // parameter of Mean and Var update  
    CParBool            UpdateFlag;
    CParBool            UpdateShortTime;
    CParBool            UpdateWrite;
    CParEnum            UpdateUseSp;
    CParEnum            UpdateUseSi;
    // threshold in prob.
    CParFloat           ThresSp;
    CParFloat           ThresSi;
    // model constants
    CParFloat           MinTimeSp;
    CParFloat           MinTimeSi;
    int                 CountSpToSi;
    int                 CountSpToSp;
    int                 CountSiToSp;
    int                 CountSiToSi;
    int                 TotalStates;
    // model parameter
    CObjStochasticState   *SpMain;
    CObjStochasticState   *SiMain;                    
    CObjStochasticState  **StateArray;


  CParInt                NrGaussSp ;    // nr of mixtures of Guassian dens. for speech
  CParInt                NrGaussSi ;    // nr of mixtures of Gaussian density for silence
  CProbGaussVar          **PropLogSp;   // parameter der Gaussdichte fuer Sprache
  CProbGaussVar          **PropLogSi;   // parameter der Gaussdichte fuer Silence

  int                    ArgMaxSp ;     // remember which Gaussian had max value (for update!)
  int                    ArgMaxSi ;     // remember which Gaussian had max value (for update!)
  float                  LogSp;         // aktuelle Emission fuer "Sprache" (max aller Mixtures)
  float                  LogSi;         // aktuelle Emission fuer "Silence" (max aller Mixtures)

  // start count, status
  long                CountThres;
  ESpeechFlag         LastFlag;
  
  // TPfau 12/21/2000
  double probSp ; // probability of Speech for current frame
  double probSi ; // probability of Silence for current frame

public:
  
  CSpeechNonSpeech( unsigned int LenOfX, double dt, FILE *fp ); 
  CSpeechNonSpeech( unsigned int LenOfX, double dt, FILE *fp, int buildarray ); 
  ~CSpeechNonSpeech();
  virtual int          writePar( FILE *fp );
  virtual int          appendFixPar( FILE *fp ) ;
  virtual int          appendUpdPar( FILE *fp ) ;
  virtual int          appendPar( FILE *fp ) { return( appendFixPar(fp) || appendUpdPar(fp) ) ; } ;
  void                 start();
  virtual ESpeechFlag  detect( float *xvec );
  virtual ESpeechFlag  detectandupdate( float *xvec );
  virtual void         update( float *xvec, ESpeechFlag ind, short NrUndef = 0, short InUndef = 0 );
  virtual int          updateOneMix( float *xvec, ESpeechFlag ind, int mixnr, short NrUndef = 0, short InUndef = 0 );
  int                 &getUpdateFlag()      { return UpdateFlag.Par; }
  int                 &getUpdateShortTime() { return UpdateShortTime.Par; }
  int                 &getUpdateWrite()     { return UpdateWrite.Par; }
  double               getProbSp()          { return( probSp ) ; } ; // tpfau, 12/21/2000
  double               getProbSi()          { return( probSi ) ; } ; // tpfau, 12/21/2000
  
  int                  getNrOfMixSi()       { return( NrGaussSi.Par ) ; } ;
  int                  getNrOfMixSp()       { return( NrGaussSp.Par ) ; } ;

  int getSpToSi() { return( CountSpToSi ) ; } ;
  int getSiToSp() { return( CountSiToSp ) ; } ;
  int getSpToSp() { return( CountSpToSp ) ; } ;
  int getSiToSi() { return( CountSiToSi ) ; } ;
  
  float GetAlphaSiMain() { return SiMain->getAlpha() ; } ;
  float GetAlphaSpMain() { return SpMain->getAlpha() ; } ;


};


class CSpeechNonSpeechANN {

protected:
  // threshold in prob.
  CParFloat           ThresSp;
  CParFloat           ThresSi;
  
  // model constants
  CParFloat           MinTimeSp;
  CParFloat           MinTimeSi;
  int                 CountSpToSi;
  int                 CountSpToSp;
  int                 CountSiToSp;
  int                 CountSiToSi;
  int                 TotalStates;
  int                 MaxNrUndef ;   // TP, 31/10/2001: force decision after this many frames

  CParFloat           PriorSp ;      // prior for speech
  CParFloat           PriorSi ;      // prior for silence
  float               ApostSp ;      // posterior for speech (output of NN)
  float               ApostSi ;      // posterior for silence (output of NN)
  float               ApostSp2 ;      // posterior for speech (output of NN)
  float               ApostSi2 ;      // posterior for silence (output of NN)

  float               LogSp;         // log prob for speech (posterior divided by prior)
  float               LogSi;         // log prob for silence (posterior divided by prior)
  
  FILE               *ProbFp ;       // pointer to file with posteriors from ANN
  FILE               *ProbFp2 ;      // pointer to second file with posteriors (model combination!)

  // states of HMM
  CObjStochasticState   *SpMain;
  CObjStochasticState   *SiMain;                    
  CObjStochasticState  **StateArray;
  
  // start count, status
  long                CountThres;
  ESpeechFlag         LastFlag;
  long                FrameCount ;
  long                NumFrames ;
  
  int                 Read( int checkframe=1) ;  // read next pair of aposterioris

 int ReadPfileHeader( FILE *fp ) ;
public:
  CSpeechNonSpeechANN( double dt, FILE *cfgfp ); 
  ~CSpeechNonSpeechANN();


  int start( char *filename, char *filename2=0 ) ;  // open file, check header and goto first probvec, init search   
  int restartat( int startframe ) ; // restart detection at given frame!
  ESpeechFlag detect( int checkframe=1 ) ;                 // get next probvec and do Viterbi search

  long getNumFrames() { return( NumFrames ) ; } ;

  // einfache Zugriffe:
  int getSpToSi() { return( CountSpToSi ) ; } ;
  int getSiToSp() { return( CountSiToSp ) ; } ;
  int getSpToSp() { return( CountSpToSp ) ; } ;
  int getSiToSi() { return( CountSiToSi ) ; } ;

  void setSpPrior( float spprior ) { PriorSp.Par = spprior ; } ;
  void setSiPrior( float siprior ) { PriorSi.Par = siprior ; } ;

  float getSpPrior() { return( PriorSp.Par) ; } ;
  float getSiPrior() { return( PriorSi.Par) ; } ;

  float getApostSi() { return( ApostSi ) ; } ;
  float getApostSp() { return( ApostSp ) ; } ;
  float getApostSi2() { return( ApostSi2 ) ; } ;
  float getApostSp2() { return( ApostSp2 ) ; } ;

  float GetAlphaSiMain() { return SiMain->getAlpha() ; } ;
  float GetAlphaSpMain() { return SpMain->getAlpha() ; } ;

} ;




class CSpeechNonSpeechOverlap : public CSpeechNonSpeech 
{
protected:
  CParEnum            UpdateUseOv;
  CParFloat           ThresOv;
  CParFloat           MinTimeOv;
  int                 CountSpToOv;
  int                 CountSiToOv;
  int                 CountOvToOv_Si;
  int                 CountOvToOv_Sp;
  int                 CountSpToSp_Ov;
  int                 CountSiToSi_Ov;
  int                 CountOvToSi;
  int                 CountOvToSp;

  CObjStochasticState *OvMain;       

  CParInt             NrGaussOv ; // nr of mixtures of Gaussian density for overlap
  CProbGaussVar     **PropLogOv;  // parameter der Gaussdichte fuer Overlap         

  int                 ArgMaxOv ;  // remember which Gaussian had max value (for update!)
  float               LogOv;      // aktuelle Emission fuer "Sprache" (max aller Mixtures)    
  double              probOv;     // probability of Silence for current frame

public:
    CSpeechNonSpeechOverlap( unsigned int LenOfX, double dt, FILE *fp ); 
    ~CSpeechNonSpeechOverlap();
    int          writePar( FILE *fp );
    int          appendFixPar( FILE *fp ) ;
    int          appendUpdPar( FILE *fp ) ;
    ESpeechFlag  detect( float *xvec );
    void         update( float *xvec, ESpeechFlag ind, short NrUndef = 0, short InUndef = 0 );
    int         updateOneMix( float *xvec, ESpeechFlag ind, int mixnr, short NrUndef = 0, short InUndef = 0 );
  double getProbOv() { return( probOv ) ; } ; // tpfau, 12/21/2000
  
  int getNrOfMixOv() { return( NrGaussOv.Par ) ; } ;
} ;
#endif

