/******************************************************************************
 * file:        spdtrain.cc  (main)
 * info:        training of speech / non-speech detector 
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

// if output to produce histograms for energy are wanted!!!!

#include <stdio.h>
#include <string.h>

#include "paramet.h"
#include "lab_list.h"

#include "dbg_mess.h"
#include "akupre.h"

#define   MAXCATLINELEN  1024
#define   MaxRuns 20000

#ifndef OutPutModeDef
typedef enum { FrameWise, MUC, XML, MUC_OLD } EOutPutMode ; 
#define OutPutModeDef
#endif


//--------------------------------- CTool object ------------------------------

class CTool
{
public:
    CParString         LabelKeyword;
    CTool( FILE *fp );
};

CTool::CTool( FILE *fp )
:   LabelKeyword( "Tools.Label.Keyword", fp, "???" )
{ ; } ;



//------------------------------- main ----------------------------------------

// main prg.
int main( int argc, char *argv[] ) {

  // open files
  int i, j, count, wavindex, labindex ;
  if( argc < 6) {   
    fprintf(stderr,"usage: %s <cfg-file> <wavfile1> <labfile1> <wavfile2> <labfile2>... <wavfileN> <labfileN>\n");
    return _ERR_;
  }
  FILE *chanfp = fopen( argv[1], "r+" );
  IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);
 
  int nrchan = (argc-2)/2 ;
  CAkuPre    **AkuPre = new CAkuPre*[nrchan] ;

  // construct
  CTool      Tool( chanfp );
  rewind( chanfp ) ;

  // initialize all channels
  for( i=0 ; i<nrchan; i++ ) {
    AkuPre[i] = new CAkuPre( chanfp ) ;
    rewind( chanfp ) ;
  } ;
 
  // create filepointer for energyoutput:
  char tmp[MaxLineLength] ;
  FILE **speechfp = new FILE*[nrchan] ;
  FILE **nonspeechfp = new FILE*[nrchan] ;
  for( i=0, labindex=3 ; i<nrchan ; i++, labindex+=2 ) {
    sprintf( tmp, "%s.speech.energy", argv[labindex] ) ;
    speechfp[i] = fopen( tmp, "w" ) ;
    sprintf( tmp, "%s.nonspeech.energy", argv[labindex] ) ;
    nonspeechfp[i] = fopen( tmp, "w" ) ;
  } ;    

  // read tool-related things
  CParEnum OutputPar( "Tools.Result.OutputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode OutputMode = (EOutPutMode)OutputPar.Par ;
  CParEnum InputPar( "Tools.Label.InputMode", "FrameWise, MUC, XML, MUC_OLD", chanfp ) ;
  EOutPutMode InputMode = (EOutPutMode)InputPar.Par ;
  CParString SpeechLabel( "Tools.Result.SpeechLabel", chanfp, "NONE" ) ;
  CParString NonspeechLabel( "Tools.Result.SilentLabel", chanfp, "-" ) ;
  CParBool  UpdateSpNonSpAutoSil( "Tools.SpdAutoUpdate.Silent", chanfp, 1 ) ;
  CParBool  UpdateSpNonSpAutoSp( "Tools.SpdAutoUpdate.Speech", chanfp, 1 ) ;
  CParBool  ControlOutput( "Tools.Ctrl.Output", chanfp, 0 ) ;

  // start AkuPre of each channel
  for( i=0, wavindex=2 ; i<nrchan ; i++, wavindex+=2 )
    // start AkuPre
    if( AkuPre[i]->startAkuPre( argv[wavindex] ) == _ERR_ ) {
      fprintf( stderr, "Can't open5d-th  sigfile '%s'\n", i, argv[wavindex] ) ;
      exit(-1) ;
    } ;

  // initialize label-info for all channels
  CListOfLabel **Lab = new CListOfLabel*[nrchan] ; ;
  // just quick and dirty, class CListOfLabel should be redesigned!
  if( InputMode == XML )
    // XML Input, use XML as KeyWord
    for( i=0 ; i<nrchan ; i++ )
      Lab[i] = new CListOfLabel( InputPar.getValue( XML ) );
  else 
    for( i=0 ; i<nrchan ; i++ )
      // MUC or Kiel Input! use Keyword as Keyword
      Lab[i] = new CListOfLabel( Tool.LabelKeyword.Par );

  // set Labels to empty string if requested!
  if( strcmp( SpeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( SpeechLabel.Par, "" ) ;
  if( strcmp( NonspeechLabel.Par, "NONE" ) == 0 ) 
    strcpy( NonspeechLabel.Par, "" ) ;
  
  // create array of Silence-Labels
  int NrSiLabels = AkuPre[0]->getSpNonSpNrOfMixSi() ; 
  char **SiLabel = new char*[NrSiLabels] ;
  //  char tmp[MAXKEYWORDLEN] ;
  for( i=0 ; i<NrSiLabels ; i++ ) {
    SiLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SilentName.%d", i ) ;
    strcpy( SiLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SiLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SilentLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;

  // create array of Speech-Labels
  int NrSpLabels = AkuPre[0]->getSpNonSpNrOfMixSp() ; 
  char **SpLabel = new char*[NrSpLabels] ;
  for( i=0 ; i<NrSpLabels ; i++ ) {
    SpLabel[i] = new char[MAXKEYWORDLEN] ;
    sprintf( tmp, "Tools.Label.SpeechName.%d", i ) ;
    strcpy( SpLabel[i], CParString( tmp, chanfp, "???").Par ) ;
    if( !strcmp( SpLabel[i], "???" ) ) {
      fprintf( stderr, "ERROR: can't find %d-th SpeechLabel in cfg-file '%s'\n", i, argv[1] ) ;
      exit(-1) ;
    } ;
  } ;
     
  // loop over catalog
  char  str[MAXCATLINELEN], *sigfilename, *labfilename;
  
  // start AkuPre
  for( i=0, wavindex=2 ; i< nrchan ; i++, wavindex+=2 ) {
    if( AkuPre[i]->startAkuPre( argv[wavindex] ) == _ERR_ ) {
      fprintf( stderr, "Can't start akupre with %d-th signalfile %s\n", i, argv[wavindex] ) ;	
      continue;
    } ;
  } ;

  // open lab-files
  FILE **labfp = new FILE*[nrchan] ;
  for( i=0, labindex=3 ; i<nrchan ; i++, labindex += 2 ) {
    labfp[i]= fopen( argv[labindex], "r");
    if( labfp[i] == 0 ) {   
      ERROR_MSG_S("can't open label-file:", argv[labindex] );
      exit(-1) ; 
    }
  } ;
  
  long *nr_runs = new long[nrchan] ; 
  
  // define array for starts and stops of runs
  //fprintf( stderr, "maxRuns: %d\n", MaxRuns ) ;
  long **start = new long*[nrchan] ;
  long **stop = new long*[nrchan] ;
  
  for( i=0, labindex=3 ; i<nrchan ; i++, labindex+=2 ) {
    start[i] = new long[MaxRuns] ;
    stop[i] = new long[MaxRuns] ;
    
    // default values for start and stop of each run
    for( j=0 ; j<MaxRuns ; j++ ) 
      start[i][j] = stop[i][j] = -1 ;
    
    // read multiple starts and stops from label-file
    long count = 0 ;
    if( Lab[i]->readLabels( labfp[i] ) == _ERR_ ) { 
      ERROR_MSG_S("can't read label-file:", argv[labindex] );
      fclose( labfp[i] );
      continue; 
    }
  
    // loop over all runs to determine start and stop-frame
    //	while( count < MaxRuns && Lab.getNextStartStop( (count>0 ? stop[count-1] : 0), start[count], stop[count], Tool.LabelSilentBeg.Par, Tool.LabelSilentEnd.Par) != _ERR_ ) {
    while( count < MaxRuns && Lab[i]->getNextStartStop( (count>0 ? stop[i][count-1] : 0 ), start[i][count], stop[i][count], NrSiLabels, SiLabel ) != _ERR_ ) { 
      if( ControlOutput.Par ) 
	fprintf( stdout, "chan %d: %ld: start: %ld stop: %ld\n", i, count, start[i][count], stop[i][count] );
      count++ ;
    } ;
    fclose( labfp[i] );
    nr_runs[i] = count ;
  } ;

  
  // loop over all frames, update spnonsp-info, stop, when label file is at end (must NOT be end of wav-file!!!)
  SPreFeature **prefp = new SPreFeature*[nrchan] ;	  
  count = 0 ;  // counts frames now!
  
 
  float *energymin = new float[nrchan] ;
  for( int i=0 ; i<nrchan ; i++ ) 
    energymin[i] = 1.0e30 ;
  
  // determine min energy per channel (use first 300 frames only)
  // determine speech energy min in each channel!!!
  while(1) { 
    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 0 ;
    for( i=0 ; i<nrchan ; i++ )
      stillrunning += AkuPre[i]->testRdIsRun() ;
    if( stillrunning != nrchan )
      break ;
    
    for( i=0 ; i<nrchan ; i++ ) {
      if( (prefp[i] = AkuPre[i]->getNextPre()) == 0 ) {
	stillrunning = 0 ;
	break ;
      } ;
      
      //if( prefp[i]->Pre.TotalEnergy < energymin[i] && spd[i] >= PUS )
      if( prefp[i]->TotalEnergy < energymin[i] )
	energymin[i] =  prefp[i]->TotalEnergy ;
    } ;
    if( stillrunning < nrchan )
      break ;
    
    count++ ;
    
    if( count > 300 )
      break ;
  } ;
  
  // reset channels and restart them
  for( i=0, wavindex=2 ; i<nrchan ; i++, wavindex+=2 ) {
    AkuPre[i]->clearAkuPre() ;
    AkuPre[i]->startAkuPre( argv[wavindex] ) ;
  } ;
  
  
  // output loop
  count = 0 ;
  while( 1 ) {
    // check if any of the labelfiles has reqached EOF yet
    if( Lab[i]->getEndFrameOfLastEntry() < count ) 
      break ;

    // check if any of the sigfiles has seen EOF yet
    int stillrunning = 1 ;
    for( i=0 ; i<nrchan ; i++ )
      stillrunning &= AkuPre[i]->testRdIsRun() ;
    if( stillrunning == 0 )
      break ;
    
    // get next feature vectors
    for( i=0 ; i<nrchan ; i++ ) {
      prefp[i] = AkuPre[i]->getNextPre();
      if( prefp[i] == 0 ) {
	stillrunning = 0 ;
	break;
      } ;
    } ;
    
    // calculate mean energy of this frame!
    float meanenergy = 0.0 ;
    for( i=0 ; i<nrchan ; i++ ) 
      meanenergy += prefp[i]->TotalEnergy ;
    meanenergy /= nrchan ;
    
    // for all channels
    for( i=0 ; i<nrchan ; i++ ) {
      int speech = 0 ;
      // is it speech in this channel?
      for( j=0 ; j<nr_runs[i] ; j++ ) {
	if( count >= start[i][j] && count <= stop[i][j] ) {
	  // speech
	  fprintf( speechfp[i], "%f\n", (prefp[i]->TotalEnergy) - energymin[i] - meanenergy ) ;
	  speech = 1 ;
	  break ;
	}
      }
      if( speech == 0 )
	// nonspeech
	fprintf( nonspeechfp[i], "%f\n", (prefp[i]->TotalEnergy) - energymin[i] - meanenergy ) ;
    } ;
    
    count++ ;
  }

 
  // free memory:
  for( i=0 ; i<nrchan ; i++ ) {
    delete[] start[i] ;
    delete[] stop[i] ;
    
    fclose( speechfp[i] ) ;
    fclose( nonspeechfp[i] ) ;

    AkuPre[i]->stopAkuPre() ;
    delete AkuPre[i] ;

    delete Lab[i] ;
  } ;

  delete[] speechfp ;
  delete[] nonspeechfp ;
  
  delete[] nr_runs ;
  delete[] start ;
  delete[] stop ;

  delete[] prefp ;
  delete[] AkuPre ;

  // free memory for silence labels
  for( i=0 ; i<NrSiLabels ; i++ ) 
    delete[] SiLabel[i] ;
  delete[] SiLabel ;
  // free memory for speech labels
  for( i=0 ; i<NrSpLabels ; i++ ) 
    delete[] SpLabel[i] ;
  delete[] SpLabel ;

  delete[] Lab ;

  return _OK_;
}
